/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-07
**
***************************************************************************/

#ifndef QTB_Angle_H
#define QTB_Angle_H

#include <math.h>

#include "qtbpoint2d.h"
#include "qtbtoolsdllexport.h"

class QTBTOOLS_EXPORT QtbAngle
{
public:
  inline QtbAngle();
  QtbAngle(double dx, double dy) { set(dx, dy); }
  inline QtbAngle(double dx, double dy, double r) { set(dx,dy,r); }
  inline QtbAngle(double degrees) { setDegrees(degrees); }
  ~QtbAngle() {}

  inline QtbAngle& operator=(const QtbAngle& o);
  inline bool operator<(const QtbAngle& o) const {return _rad<o._rad; }
  inline bool operator==(const QtbAngle& o) const {return _rad==o._rad; }

  inline void set(double dx, double dy);
  inline void set(double dx, double dy, double r);
  void setDegreeAzimuth(double azimuth) {setDegrees(90.0-azimuth);}
  void setRadianAzimuth(double azimuth) {setRadians(0.5*M_PI-azimuth);}
  inline void setDegrees(double degrees);
  inline void setRadians(double radians);

  inline void initDegrees();
  inline void initRadians();

  double degrees() const {return _deg;}
  double radians() const {return _rad;}
  double cos() const {return _cos;}
  double sin() const {return _sin;}
  inline void rotate(QtbPoint2D& p) const;
  inline void chSign();
  inline void mirror();
private:
  double _deg;
  double _rad;
  double _cos;
  double _sin;
};

inline QtbAngle::QtbAngle()
{
  QTB_TRACE;
  _deg=0.0;
  _rad=0.0;
  _cos=1.0;
  _sin=0.0;
}

inline void QtbAngle::set(double dx, double dy)
{
  QTB_TRACE;
  _deg=0.0;
  _rad=0.0;
  double r=sqrt(dx*dx+dy*dy);
  if (r>0.0) {
    _cos=dx/r;
    _sin=dy/r;
  } else {
    _cos=1.0;
    _sin=0.0;
  }
}

inline void QtbAngle::set(double dx, double dy, double r)
{
  QTB_TRACE;
  _deg=0.0;
  _rad=0.0;
  if (r>0.0) {
    _cos=dx/r;
    _sin=dy/r;
  } else {
    _cos=1.0;
    _sin=0.0;
  }
}

inline QtbAngle& QtbAngle::operator=(const QtbAngle& o)
{
  QTB_TRACE;
  _deg=o._deg;
  _rad=o._rad;
  _cos=o._cos;
  _sin=o._sin;
  return *this;
}

inline void QtbAngle::initDegrees()
{
  QTB_TRACE;
  _deg=_rad*180.0/M_PI;
}

inline void QtbAngle::initRadians()
{
  QTB_TRACE;
  _rad=::atan2(_sin, _cos);
}

inline void QtbAngle::setDegrees(double degrees)
{
  QTB_TRACE;
  _deg=degrees;
  _rad=_deg/180.0*M_PI;
  _cos=::cos(_rad);
  _sin=::sin(_rad);
}

inline void QtbAngle::setRadians(double radians)
{
  QTB_TRACE;
  _rad=radians;
  _cos=::cos(_rad);
  _sin=::sin(_rad);
}

inline void QtbAngle::chSign()
{
  QTB_TRACE;
  _deg=-_deg;
  _rad=-_rad;
  _sin=-_sin;
}

inline void QtbAngle::mirror()
{
  QTB_TRACE;
  _cos = -_cos;
  _sin = -_sin;
  if (_deg>180.0) _deg-=180; else _deg+=180.0;
  if (_rad>M_PI) _rad-=M_PI; else _rad+=M_PI;
}

inline void QtbAngle::rotate(QtbPoint2D& p) const
{
  QTB_TRACE;
  double tmp=p.x()*_cos+p.y()*_sin;
  p.setY( p.y()*_cos-p.x()*_sin );
  p.setX( tmp );
}

#endif
