/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-07-06
**
***************************************************************************/

#include <QDir>
#include <QFileInfo>
#include <QMutex>

#include "qtbcache.h"
#include "qtbcacheitem.h"
#include "qtbcoreapplication.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbCache qtbcache.h
  \brief Cache space with "inteligent" swap mechanism

  QtbCacheItem can allocate data vectors into a QtbCache. QtbCache controls that the total memory does not exceed
  a maximum number of bytes. If so, already allocated blocks are automatically deallocated if they are not locked.

  Contrary to most of the system swaps, this swap mechanism knows what the item usage will be in the next operations
  avoiding unefficient deallocation of blocks. This "intelligent" deallocation is performed with a priority level
  assigned to each item. Items with a lower priority are deallocated first. If not enough space can be found with lower
  priority, the highest priority items are deallocated. Without appeal, the items with the same priority as the current
  one will be deallocated. Hence before starting a process, it is wize to set the priorities according to the order
  the items will be used as in the following example:

  \code
  QtbCache mainCache;

  class QtbItem : public QtbCacheItem
  {
  public:
    inine virtual bool lock();
  protected:
    virtual bool isAllocated() { return _data; }
    virtual int dataSize() { return 100*1024*1024; } // blocks of 100 Mb
    virtual bool allocate() { _data = new char [ 180*1024*1024 ]; return _data; }
    virtual void free() { delete [] _data; }
    virtual void save( QDir& d );
    virtual void load( QDir& d );
  private:
    char * _data;
  };

  bool QtbItem::lock()
  {
    if ( isAllocated() || mainCache->makeAvailable(this) ) return QtbCacheItem::lock();
    else return false;
  }

  void QtbItem::save( QDir& d )
  {
    QFile f( d.absoluteFilePath( swapFileName() ) );
    if ( f.open( QIODevice::WriteOnly ) ) {
      f.write( _data, dataSize() );
    } else qtbApp << QObject::tr( "Cannot save temporary file, check disk space!" ) << endl;
  }

  void QtbItem::load( QDir& d )
  {
    QFile f( d.absoluteFilePath( swapFileName() ) );
    if ( f.open( QIODevice::ReadOnly ) ) {
      f.read( _data, dataSize() );
    } else qtbApp << QObject::tr( "Cannot load temporary file" ) << endl;
  }

  int main ( int argc, char ** argv )
  {
    c.setSize(350);  // A cache of 350 Mb

    QtbItem i1;
    QtbItem i2;
    QtbItem i3;
    QtbItem i4;

    i3.setAllocationPriority(1);

    i4.setAllocationPriority(&i3); // Set the same priority for i3 and i4
    if ( i3.lock() ) {
      if ( i4.lock() ) {
        // processing i3 and i4
        i4.unlock();
      }
      i3.unlock();
    }

    // i3 and i4 still allocated

    i1.setAllocationPriority(1);
    i2.setAllocationPriority(2);
    i3.setAllocationPriority(3);
    i4.setAllocationPriority(0);

    if ( i1.lock() ) {  // i1 allocated
      if ( i2.lock ) {  // i2 allocated and i4 deallocated
        // processing i1 and i2
        i2.unlock();
      }
      if ( i3.lock ) {  // i3 allocated and i2 deallocated
        // processing i1 and i3
        i3.unlock();
      }
      i1.unlock();
    }
  }
  \endcode
*/

/*!
  Initialize a cache that manage allocation and deallocation of memory.

  The two parameters of a cache are its size in byte and the path of a temporary directory
  used for swapping if the requested memory for data exceed its size.
*/
QtbCache::QtbCache() : _mutex(QMutex::Recursive)
{
  QTB_TRACE;
  _loadedBytes = 0;
  _maxLoadedBytes = 0;
}

/*!
  \fn QtbCache::~QtbCache()
  It does nothing because the removal of data is the responsability of the QtbCacheItem
*/

/*!
  \fn QtbCache::setSize(double nMegaBytes)
  Set the maximum number of mega bytes that can be allocated simultaneously. The default is 0.
*/

/*!
  Set the path where to swap. The default is the home directory.
*/
void QtbCache::setSwapDir(const QDir& d)
{
  QTB_TRACE;
  _swapDir=d;
  if (_swapDir.exists()) { // clean file that might be left after a crash
    QStringList filters;
    filters << "cache_data_*";
    QStringList cacheFiles = _swapDir.entryList(filters);
    for (QStringList::iterator it=cacheFiles.begin();it!=cacheFiles.end();it++) {
      _swapDir.remove(*it);
    }
  } else {
    _swapDir.mkpath(_swapDir.absolutePath ());
  }
}

/*!
  Free items with priority lest than \a priority in order to obtain at least \a nBytes.
*/
bool QtbCache::free( double nBytes, int priority )
{
  QTB_TRACE;
  double obtainedBytes = _maxLoadedBytes - _loadedBytes;
  do {
    const QtbCacheItem * item = _allocatedList.bestToFree( priority );
    if ( item ) {
      //printf("Saving item %s\n",item->debugName().toAscii().data());
      item->save( _swapDir );
      item->_saved = true;
      item->free();
      double dataSize = ( double ) item->dataSize();
      obtainedBytes += dataSize;
      _loadedBytes -= dataSize;
      _allocatedList.remove( item );
      item->_dataLock.unlock();
      /*printf("Freeing %lf Kb from item %s, priority=%i/%i\t\t",dataSize/1024,
             item->debugName().toAscii().data(),item->_priority,priority);
      printf("Free space=%lf Kb\n",freeBytes()/1024);*/
    } else return false;
  } while ( obtainedBytes < nBytes );
  return true;
}

/*!
  Free data used by item. If \a saveTemp is false and if data is swaped, the corresponding file is removed.
  It \a saveTemp is true and if data is allocated, data is saved to file.
*/
void QtbCache::free( QtbCacheItem * item, bool saveTemp )
{
  QTB_TRACE;
  if ( !saveTemp && item->_saved ) {
    _swapDir.remove( item->swapFileName() );
    item->_saved = false;
  }
  if ( item->isAllocated() ) {
    if ( saveTemp ) {
      //printf("Saving item %s\n",item->debugName().toAscii().data());
      item->save( _swapDir );
      item->_saved = true;
    }
    item->free();
    _mutex.lock();
    _loadedBytes -= ( double ) item->dataSize();
    _allocatedList.remove( item );
    _mutex.unlock();
    //printf("Freeing item, %i are still allocated\n",_allocatedList.count());
    //printf("Freeing %lf Kb from item %s\t\t",(double)item->dataSize()/1024, item->debugName().toAscii().data());
    //printf("Free space=%lf Kb\n",freeBytes()/1024);
  }
}

/*!
  In all cases the data must not be allocated. This is not double checked.
  In a multi-threaded context, this function must be used by only one thread at a time and lock() of item
  must be called with this object locked.
*/
bool QtbCache::makeAvailable( const QtbCacheItem * item )
{
  QTB_TRACE;
  // Process low level allocation of data
  double neededSize = ( double ) item->dataSize();
  QMutexLocker ml(&_mutex);
  if ( neededSize > freeBytes() ) {
    // Not enough space in cache
    if ( !free( neededSize, item->priority() ) ) {
      QString msg = tr( "Impossible to allocate new data\n"
                        "   current used space: %1 Kb\n"
                        "   current free space: %2 Kb\n"
                        "   current buffer size: %3 Kb\n"
                        "   required space: %4 Kb\n"
                        "Increase buffer size, a good compromize is 80% of you physical memory\n"
                        "This buffer is allocated dynamically upon requests, so it will not overload\n"
                        "your memory usage.\n\n")
                        .arg( _loadedBytes / 1024.0 ).arg( freeBytes() / 1024.0 )
                        .arg( _maxLoadedBytes / 1024.0 ).arg( neededSize / 1024.0)+_allocatedList.humanInfo();
      qtbApp << msg << endl;
      qtbApp.reportBug( QtbMessage::Warning, msg.toAscii().data() );
      _allocatedList.humanInfo();
      return false; // really not lucky, no more space
    }
    // Perfect got memory free
  }
  if ( item->allocate() ) {
    _allocatedList.add( item );
    _loadedBytes += neededSize;
    ml.unlock();
    if ( item->_saved ) {
      //printf("Loading item %s\n",item->debugName().toAscii().data());
      item->load( _swapDir );
      _swapDir.remove( item->swapFileName() );
      item->_saved = false;
    }
    //printf("Allocating %lf Kbytes for item %s\t\t",neededSize/1024, item->debugName().toAscii().data());
    //printf("Free space=%lf Kb\n",freeBytes()/1024);
    return true;
  }
  // Impossible to allocate new memory, problem with your computer?
  QString msg = tr( "Impossible to allocate new data of size %1: current used space: %2 Kb.\n\n")
                   .arg( neededSize )
                   .arg( _loadedBytes * sizeof( double ) / 1024.0 )+_allocatedList.humanInfo();
  qtbApp << msg << endl;
  qtbApp.reportBug( QtbMessage::Warning, msg.toAscii().data() );
  return false;
}

/*!
  In all cases the data must be allocated with its original size. This is not double checked.
*/
bool QtbCache::enlarge( QtbCacheItem * item, double finalSize )
{
  QTB_TRACE;
  // Process low level allocation of data
  double neededSize = finalSize - ( double ) item->dataSize();
  QMutexLocker ml(&_mutex);
  if ( neededSize > freeBytes() ) {
    // Not enough space in cache
    if ( !free( neededSize, item->priority() ) ) {
      QString msg = tr( "Impossible to allocate new data\n"
                        "   current used space: %1 Kb\n"
                        "   current free space: %2 Kb\n"
                        "   current buffer size: %3 Kb\n"
                        "   required space: %4 Kb\n"
                        "Increase buffer size, a good compromize is 80% of you physical memory\n"
                        "This buffer is allocated dynamically upon requests, so it will not overload\n"
                        "your memory usage.\n\n")
                        .arg( _loadedBytes / 1024.0 ).arg( freeBytes() / 1024.0 )
                        .arg( _maxLoadedBytes / 1024.0 ).arg( neededSize / 1024.0)+_allocatedList.humanInfo();
      qtbApp << msg << endl;
      qtbApp.reportBug( QtbMessage::Warning, msg.toAscii().data() );
      return false; // really not lucky, no more space
    }
    // Perfect got memory free
  }
  _loadedBytes += neededSize;
  return true;
}

/*!
  According to current \a priority, returns the best item to free.
  The item with the maximum priority and less than \a priority is returned.
  If no item can be found, the item with the maximum priority is returned.
  In case of equal priorities, the block with the oldest access time is returned.
*/
const QtbCacheItem * QtbAllocatedCacheItem::bestToFree( int priority )
{
  QTB_TRACE;
  int maxPriority = priority;
  const QtbCacheItem * itemMax = 0;
  const QtbCacheItem * itemLess = 0;
  clock_t timeMax = clock()+1; // Last touched item might have the same clock as now
  clock_t timeLess = timeMax;
  for (iterator it=begin();it!=end();++it) {
    const QtbCacheItem& item=**it;
    if ( item._dataLock.tryLock() ) {
      int p = item.priority();
      if ( p >= maxPriority ) {
        if ( p > maxPriority ) {
          maxPriority = p;
          if (itemMax) itemMax->_dataLock.unlock();
          itemMax = &item;
          timeMax = item._lastAccess;
        } else if ( item._lastAccess < timeMax ){
          if (itemMax) itemMax->_dataLock.unlock();
          itemMax = &item;
          timeMax = item._lastAccess;
        } else {
          item._dataLock.unlock();
        }
      } else if ( p < priority ) {
        if ( item._lastAccess < timeLess ) {
          if (itemLess) itemLess->_dataLock.unlock();
          itemLess = &item;
          timeLess = item._lastAccess;
        } else {
          item._dataLock.unlock();
        }
      } else {
        item._dataLock.unlock();
      }
    }
  }
  if ( itemLess ) {
    if (itemMax) itemMax->_dataLock.unlock();
    return itemLess;
  }
  return itemMax;
}

/*!
  Print information about allocated blocks (may be useful for checking memory usage of processes)
*/
QString QtbAllocatedCacheItem::humanInfo() const
{
  QTB_TRACE;
  int sizeBlocks = 0, sizeLockedBlocks = 0, countLockedBlocks = 0;
  for (const_iterator it=begin();it!=end();++it) {
    const QtbCacheItem& item=**it;
    sizeBlocks+=item.dataSize();
    if (item._dataLock.tryLock() ) {
      sizeLockedBlocks+=item.dataSize();
      item._dataLock.unlock();
      countLockedBlocks++;
    }
  }
  return translate("QtbAllocatedCacheItem","%1 allocated blocks (%2 Mb), %3 locked blocks (%4 Mb)\n")
               .arg(count()).arg(sizeBlocks >> 20).arg(countLockedBlocks).arg(sizeLockedBlocks >> 20);
}
