/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2003-05-12
**
***************************************************************************/

#include <math.h>

#include "qtbcircle.h"
#include "qtbmemorychecker.h"

QtbCircle::QtbCircle(double x0,double y0, double r)
{
  QTB_TRACE;
  _x0=x0;
  _y0=y0;
  _r=fabs(r);
}

double QtbCircle::distanceTo( const QtbPoint2D& p ) const
{
  QTB_TRACE;
  double d=p.distanceTo(QtbPoint2D(_x0,_y0));
  return fabs(d-_r);
}

/*!
  Calculate intersections between this and \a o.
*/
QVector<QtbPoint2D> QtbCircle::intersect( const QtbCircle& o )
{
  QTB_TRACE;
  QVector<QtbPoint2D> inter;
  double r02 = _r*_r;
  double r12 = o._r*o._r;
  double x02=_x0*_x0;
  double y02=_y0*_y0;
  double x12=o._x0*o._x0;
  double y12=o._y0*o._y0;
  double a,b,c,rho,srho,A,B,C,D;
  A = 2.0 * ( o._x0-_x0 );
  B = 2.0 * ( o._y0-_y0 );
  C = r02 - r12 - ( x02 + y02 - x12 - y12 );
  // The solution must satisfy
  //  A * xs + B * ys = C, which is simply the subtraction of circle equations
  if (fabs(A)>fabs(B)) {
    // Prefer working with xs=(C-B*ys)/A transformed into xs = B-A*ys, with B=C/A and A=B/A
    D = 1.0 / A;
    A = B*D;
    B = C*D;
    a=1.0+A*A;
    b=2*A*_x0-2*A*B-2*_y0;
    c=x02+B*B-2*_x0*B+y02-r02;
    rho=b*b-4*a*c;
    if (rho>0) {
      srho=sqrt(rho);
      inter.resize(2);
      inter[0].setY( (-b+srho) / (2.0*a) );
      inter[1].setY( (-b-srho) / (2.0*a) );
      inter[0].setX( B-A*inter[0].y() );
      inter[1].setX( B-A*inter[1].y() );
      // Check:
      //printf("Check c1 %lf==%lf\n",_r,sqrt((inter[0].x()-_x0)*(inter[0].x()-_x0)+(inter[0].y()-_y0)*(inter[0].y()-_y0)));
      //printf("Check c2 %lf==%lf\n",o._r,sqrt((inter[0].x()-o._x0)*(inter[0].x()-o._x0)+(inter[0].y()-o._y0)*(inter[0].y()-o._y0)));
    } else if (rho==0) {
      inter.resize(1);
      inter[0].setY( -b / a );
      inter[0].setX( B-A*inter[0].y() );
    }
  } else {
    // Prefer working with ys=(C-A*xs)/B transformed into ys = B-A*xs, with B=C/A and A=B/A
    D = 1.0 / B;
    A = A*D;
    B = C*D;
    a=1.0+A*A;
    b=2*A*_y0-2*A*B-2*_x0;
    c=x02+B*B-2*_y0*B+y02-r02;
    rho=b*b-4*a*c;
    if (rho>0) {
      srho=sqrt(rho);
      inter.resize(2);
      inter[0].setX( (-b+srho) / (2.0*a) );
      inter[1].setX( (-b-srho) / (2.0*a) );
      inter[0].setY( B-A*inter[0].x() );
      inter[1].setY( B-A*inter[1].x() );
    } else if (rho==0) {
      inter.resize(1);
      inter[0].setX( -b / a );
      inter[0].setY( B-A*inter[0].x() );
    }
  }
  return inter;
}

/*!
  For convenience return only one solution if \a ok is true. If two solution are found only the closest to \a disc is returned.
*/
QtbPoint2D QtbCircle::intersect( const QtbCircle& o, bool& ok, const QtbCircle& disc )
{
  QTB_TRACE;
  QVector<QtbPoint2D> solutions = intersect(o);
  switch(solutions.count()) {
  case 1:
    ok = true;
    return solutions[0];
  case 2:
    ok = true;
    if (disc.distanceTo(solutions[0])<disc.distanceTo(solutions[1]))
      return solutions[0];
    else
      return solutions[1];
  default:
    ok = false;
    return QtbPoint2D();
  }
}
