/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2008-07-11
**
***************************************************************************/

#include <QSet>

#include "qtbcolumntextparser.h"
#include "qtbstringsection.h"
#include "qtbxmlmap.h"
#include "qtbcoreapplication.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbColumnTextParser qtbcolumntextparser.h
  \brief Brief description of class still missing

  Full description of class still missing
*/

const QString QtbColumnTextParser::xmlColumnTextParserTag = "ColumnTextParser";

/*!
  Description of constructor still missing
*/
QtbColumnTextParser::QtbColumnTextParser( QObject * parent )
    : QtbThread( parent )
{
  QTB_TRACE;
  _text = 0;
  _terminate = false;
  _currentRow = 0;

  // Init at least one standard type
  setStandardTypes( QStringList() );

  _startLine = 0;
  _endLine = 0;
  _acceptPattern = "";
  _rejectPattern = "^#";

  _delimiters = " \t";
  _columnCount = -1;
  _acceptNullColumns = false;
  _fixedWidth = false;
  _columnDescriptions.resize(1); // At least one default column is mandatory

  _isSectionLinePattern = true;
  _sectionLinePattern = "^#";
  _isSectionLineCount = false;
  _sectionLineCount = 0;
  _isSectionRowCount = false;
  _sectionRowCount = 0;
  _isSectionCellPattern = false;
  _sectionCellColumn = 0;
  _sectionCellPattern = "";
}

/*!
  Copy constructor, constructed without parent.
  Keep only properties to run on another file.
*/
QtbColumnTextParser::QtbColumnTextParser( const QtbColumnTextParser& o )
    : QtbThread(), QtbXMLClass()
{
  _text = 0;
  _terminate = false;
  _currentRow = 0;

  _standardTypes = o._standardTypes;
  _columnDescriptions = o._columnDescriptions;

  _startLine = o._startLine;
  _endLine = o._endLine;
  _acceptPattern = o._acceptPattern;
  _rejectPattern = o._rejectPattern;

  _delimiters = o._delimiters;
  _columnWidths = o._columnWidths;
  _fixedWidth = o._fixedWidth;
  _acceptNullColumns = o._acceptNullColumns;

  _isSectionLinePattern = o._isSectionLinePattern;
  _sectionLinePattern = o._sectionLinePattern;
  _isSectionLineCount = o._isSectionLineCount;
  _sectionLineCount = o._sectionLineCount;
  _isSectionRowCount = o._isSectionRowCount;
  _sectionRowCount = o._sectionRowCount;
  _isSectionCellPattern = o._isSectionCellPattern;
  _sectionCellColumn = o._sectionCellColumn;
  _sectionCellPattern = o._sectionCellPattern;
}

/*!
  Description of destructor still missing
*/
QtbColumnTextParser::~QtbColumnTextParser()
{
  QTB_TRACE;
  delete _text;
}

void QtbColumnTextParser::stopUpdates()
{
  QTB_TRACE;
  if (isRunning()) {
    _terminate.fetchAndStoreOrdered(true);
    wait();
  }
}

void QtbColumnTextParser::startUpdates()
{
  QTB_TRACE;
  _terminate = false;
  start();
}

/*!
  Set the list of standard types. Used when restoring from an xml stream and by setType().
  The first item of \a standardTypes has type index 1. A default "Skipped" type is always
  prepended, hence there is no need to include one in \a standardTypes.
*/
void QtbColumnTextParser::setStandardTypes( const QStringList& standardTypes )
{
  QTB_TRACE;
  _standardTypes = standardTypes;
  _standardTypes.prepend(tr("Skipped"));
}

/*!
  Returns column description of column \a column. Size of internal vector is eventually
  increased if \a column does not exist yet.
*/
QtbColumnTextDescription& QtbColumnTextParser::columnDescription( int column )
{
  if (column>=_columnDescriptions.count()) {
    _columnDescriptions.resize(column+1);
  }
  return _columnDescriptions[column];
}

/*!
  Returns column description of column \a column. If \a column does not exist yet, the
  description of the last column is returned.
*/
const QtbColumnTextDescription& QtbColumnTextParser::columnDescription( int column ) const
{
  if (column>=_columnDescriptions.count()) {
    return _columnDescriptions.last();
  } else {
    return _columnDescriptions[column];
  }
}

/*!
  Sets type of column index \a column to \a type.

  setTypeName() is called only from QtbColumnTextItem, in response to user action.
  The list of standard type is never huge, hence we search in a QStringList.
*/
void QtbColumnTextParser::setTypeName( int column, QString type )
{
  int i = _standardTypes.indexOf( type );
  if (i>-1) {
    columnDescription(column).setType(i);
  }
}

/*!
  \fn QString QtbColumnTextParser::type( int column ) const
  Returns the type of column \a column.
*/

/*!
  \fn QString QtbColumnTextParser::typeName( int column ) const
  Returns the type of column \a column.
*/

/*!
  \fn double QtbColumnTextParser::factor( int column ) const
  Returns the factor of column \a column.
*/

/*!
  \fn QString QtbColumnTextParser::replaceRx( int column ) const
  Returns the regular expression of column \a column.
*/

/*!
  \fn QString QtbColumnTextParser::replaceAfter( int column ) const
  Returns the replacement string of column \a column.
*/


/*!
  \fn void QtbColumnTextParser::setStartLine( int i )
  Lines start at 1, If \a i is 0, no limit on start line is set.
*/

/*!
  \fn void QtbColumnTextParser::setEndLine( int i )
  Lines start at 1, If \a i is 0, no limit on end line is set.
*/

QVector<int> QtbColumnTextParser::columnWidths( QString ws )
{
  QVector<int> wi;
  QStringList wStrList = ws.split(",");
  for(QStringList::iterator it=wStrList.begin();it!=wStrList.end();it++) {
    wi.append( it->trimmed().toInt() );
  }
  return wi;
}

QString QtbColumnTextParser::columnWidths( QVector<int> wi )
{
  QString ws;
  QVector<int>::iterator it=wi.begin();
  if (it!=wi.end()) {
    ws = QString::number(*it);
    for(it++;it!=wi.end();it++) {
      ws += ", " + QString::number(*it);
    }
  }
  return ws;
}

int QtbColumnTextParser::rowCount() const
{
  QTB_TRACE;
  QMutexLocker ml( &_mutex );
  return _rows.count();
}

int QtbColumnTextParser::columnCount() const
{
  QTB_TRACE;
  QMutexLocker ml( &_mutex );
  return _columnCount;
}

int QtbColumnTextParser::lineNumber( int row ) const
{
  QMutexLocker ml( &_mutex );
  if ( row<_rows.count() ) {
    return _lines2rows[row];
  } else {
    return 0;
  }
}

QString QtbColumnTextParser::text( int row, int column ) const
{
  QTB_TRACE;
  QMutexLocker ml( &_mutex );
  QString t;
  if ( row<_rows.count() ) {
    QMap<int,QStringList>::const_iterator it = _cols.find(row);
    if (it!=_cols.end()) {
      const QStringList& cols = it.value();
      if ( column < cols.count() ) {
        t = cols.at( column );
      }
    }
  }
  return t;
}

void QtbColumnTextParser::run()
{
  QTB_TRACE;
  _columnCount = -1;
  _cols.clear();
  _sectionMarkers.clear();
  _lines2rows.clear();
  _rows.clear();
  parseRows();
  emit dataAvailable();
}

void QtbColumnTextParser::parseRows()
{
  _lineNumber = 0;
  _lastSectionMarkerLine = 0;
  _lastSectionMarkerRow = 0;
  int nextReportLine = 2;
  if (!_text->seek( 0 )) return;
  QRegExp * acceptRegExp, * rejectRegExp, * sectionLineRegExp, * sectionCellRegExp;
  if (_acceptPattern.isEmpty()) {
    acceptRegExp = 0;
  } else {
    acceptRegExp = new QRegExp(_acceptPattern);
  }
  if (_rejectPattern.isEmpty()) {
    rejectRegExp = 0;
  } else {
    rejectRegExp = new QRegExp(_rejectPattern);
  }
  if (_isSectionLinePattern) {
    sectionLineRegExp = new QRegExp(_sectionLinePattern);
  } else {
    sectionLineRegExp = 0;
  }
  if (_isSectionCellPattern) {
    sectionCellRegExp = new QRegExp(_sectionCellPattern);
  } else {
    sectionCellRegExp = 0;
  }
  while ( !_text->atEnd() ) {
    QString rowStr = _text->readLine();
    _mutex.lock();
    _lineNumber++;
    if(_isSectionLinePattern && rowStr.contains(*sectionLineRegExp)) {
      _sectionMarkers.insert(_rows.count());
      _lastSectionMarkerLine = _lineNumber;
      _lastSectionMarkerRow = _rows.count();
    }
    if (_isSectionLineCount && _lineNumber >= (_lastSectionMarkerLine+_sectionLineCount ) ) {
      _sectionMarkers.insert(_rows.count());
      _lastSectionMarkerLine = _lineNumber;
      _lastSectionMarkerRow = _rows.count();
    }
   _mutex.unlock();
    if ( (_startLine==0 || _lineNumber>=_startLine) &&
          (_endLine==0 || _lineNumber<=_endLine) &&
          (!acceptRegExp || rowStr.contains(*acceptRegExp)) &&
          (!rejectRegExp || !rowStr.contains(*rejectRegExp))) {
      _mutex.lock();
      if (_isSectionRowCount && _rows.count() >= (_lastSectionMarkerRow+_sectionRowCount ) ) {
        _sectionMarkers.insert(_rows.count());
        _lastSectionMarkerLine = _lineNumber;
        _lastSectionMarkerRow = _rows.count();
      }
      _lines2rows.insert(_rows.count(), _lineNumber);
      _rows.append(rowStr);
      parseColumns( _rows.count()-1, sectionCellRegExp );
      _mutex.unlock();
    }
     if(_endLine>0 && _lineNumber>_endLine) break;
    if(_lineNumber==nextReportLine) {
      nextReportLine = nextReportLine << 1;
      emit dataAvailable();
    }
    if(_terminate.testAndSetOrdered(true, true)) break;
  }
  delete sectionLineRegExp;
  delete sectionCellRegExp;
  delete acceptRegExp;
  delete rejectRegExp;
}

void QtbColumnTextParser::parseColumns( int row, QRegExp * sectionCellRegExp )
{
  QStringList cols;
  if (_fixedWidth) {
    const QString& rowStr = _rows.at( row );
    int pos = 0;
    for(QVector<int>::const_iterator it = _columnWidths.begin(); it!=_columnWidths.end();it++) {
      if(_terminate.testAndSetOrdered(true, true)) return;
      if (pos>=rowStr.count()) break;
      cols.append( cellContent( cols.count(), rowStr.mid( pos, *it ) ) );
      pos+=*it;
    }
    if (pos<rowStr.count()) {
      cols.append( cellContent( cols.count(), rowStr.mid( pos ) ) );
    }
  } else {
    QtbStringSection rowStr = _rows.at( row );
    QtbStringSection colStr;
    const QChar * ptr = 0;
    while( true ) {
      if(_terminate.testAndSetOrdered(true, true)) return;
      colStr = rowStr.nextField(ptr, _delimiters, !_acceptNullColumns);
      if (!colStr.isValid()) break;
      cols.append( cellContent( cols.count(), colStr.toString() ) );
    }
  }
  if (_isSectionCellPattern && cols.count()>_sectionCellColumn) {
    if (cols[_sectionCellColumn].contains(*sectionCellRegExp)) {
      _sectionMarkers.insert(_rows.count()-1); // include this row as well
      _lastSectionMarkerLine = _lineNumber;
      _lastSectionMarkerRow = _rows.count()-1;
    }
  }
  _cols.insert(row,cols);
  if (cols.count()>_columnCount) {
    _columnCount = cols.count();
  }
}

QString QtbColumnTextParser::cellContent( int column, QString s ) const
{
  QString rxStr = replaceRx(column);
  if (!rxStr.isEmpty()) {
    QRegExp rx(rxStr);
    s.replace( rx, replaceAfter(column) );
  }
  double fac = factor(column);
  if (fac!=1.0) {
    return QString::number(s.toDouble()*fac);
  }
  return s;
}

void QtbColumnTextParser::xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const
{
  QTB_TRACE;
  Q_UNUSED(context);
  writeProperty( s, "startLine", _startLine );
  writeProperty( s, "endLine", _endLine );
  writeProperty( s, "acceptPattern", _acceptPattern );
  writeProperty( s, "rejectPattern", _rejectPattern );

  writeProperty( s, "fixedWidth", (bool)_fixedWidth );
  writeProperty( s, "acceptNullColumns", (bool)_acceptNullColumns );
  writeProperty( s, "delimiters", _delimiters );
  writeProperty( s, "columnWidths", columnWidths(_columnWidths) );

  writeProperty( s, "isSectionLinePattern", (bool)_isSectionLinePattern );
  writeProperty( s, "sectionLinePattern", _sectionLinePattern );
  writeProperty( s, "isSectionLineCount", (bool)_isSectionLineCount );
  writeProperty( s, "sectionLineCount", _sectionLineCount );
  writeProperty( s, "isSectionRowCount", (bool)_isSectionRowCount );
  writeProperty( s, "sectionRowCount", _sectionRowCount );
  writeProperty( s, "isSectionCellPattern", (bool)_isSectionCellPattern );
  writeProperty( s, "sectionCellColumn", _sectionCellColumn );
  writeProperty( s, "sectionCellPattern", _sectionCellPattern );
}

void QtbColumnTextParser::xml_writeChildren(XML_WRITECHILDREN_ARGS) const
{
  QTB_TRACE;
  Q_UNUSED(context);
  static const QString key("column");
  QtbXMLSaveAttributes att;
  att.add(key);
  QString& value = att.value(0);
  int n = _columnDescriptions.count();
  for (int i = 0;i<n;i++) {
    value = QString::number(i);
    _columnDescriptions[i].xml_save(s, context, att);
  }
}

QtbXMLMember QtbColumnTextParser::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(context);
  if ( tag == "ColumnTextDescription" ) {
    static const QString tmp("column");
    QtbXMLRestoreAttributeIterator it=attributes.find(tmp);
    if (it!=attributes.end()) {
      return QtbXMLMember( &columnDescription( it.value().toInt() ) );
    } else {
      qtbApp << tr("No column index defined for description") << endl;
      return QtbXMLMember(QtbXMLMember::Unknown);
    }
  } else if ( tag == "startLine" ) return QtbXMLMember( 0 );
  else if ( tag == "endLine" ) return QtbXMLMember( 1 );
  else if ( tag == "acceptPattern" ) return QtbXMLMember( 2 );
  else if ( tag == "rejectPattern" ) return QtbXMLMember( 3 );

  else if ( tag == "fixedWidth" ) return QtbXMLMember( 4 );
  else if ( tag == "acceptNullColumns" ) return QtbXMLMember( 5 );
  else if ( tag == "delimiters" ) return QtbXMLMember( 6 );
  else if ( tag == "columnWidths" ) return QtbXMLMember( 7 );

  else if ( tag == "isSectionLinePattern" ) return QtbXMLMember( 8 );
  else if ( tag == "sectionLinePattern" ) return QtbXMLMember( 9 );
  else if ( tag == "isSectionLineCount" ) return QtbXMLMember( 10 );
  else if ( tag == "sectionLineCount" ) return QtbXMLMember( 11 );
  else if ( tag == "isSectionRowCount" ) return QtbXMLMember( 12 );
  else if ( tag == "sectionRowCount" ) return QtbXMLMember( 13 );
  else if ( tag == "isSectionCellPattern" ) return QtbXMLMember( 14 );
  else if ( tag == "sectionCellColumn" ) return QtbXMLMember( 15 );
  else if ( tag == "sectionCellPattern" ) return QtbXMLMember( 16 );
  return QtbXMLMember(QtbXMLMember::Unknown);
}

bool QtbColumnTextParser::xml_setProperty(XML_SETPROPERTY_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  switch ( memberID ) {
  case 0: _startLine = content.toInt(); return true;
  case 1: _endLine = content.toInt(); return true;
  case 2: _acceptPattern = content.toString(); return true;
  case 3: _rejectPattern = content.toString(); return true;

  case 4: _fixedWidth = content.toBool(); return true;
  case 5: _acceptNullColumns = content.toBool(); return true;
  case 6: _delimiters = content.toString(); return true;
  case 7: _columnWidths = columnWidths( content.toString() ); return true;

  case 8: _isSectionLinePattern = content.toBool(); return true;
  case 9: _sectionLinePattern = content.toString(); return true;
  case 10: _isSectionLineCount = content.toBool(); return true;
  case 11: _sectionLineCount = content.toInt(); return true;
  case 12: _isSectionRowCount = content.toBool(); return true;
  case 13: _sectionRowCount = content.toInt(); return true;
  case 14: _isSectionCellPattern = content.toBool(); return true;
  case 15: _sectionCellColumn = content.toInt(); return true;
  case 16: _sectionCellPattern = content.toString(); return true;
  default: return false;
  }
}

/*!
  After restoring from an xml stream there might be some non-standard column types, remove them.
*/
void QtbColumnTextParser::xml_polish(XML_POLISH_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(context);
  int standardTypeCount = _standardTypes.count();
  int nColumns = _columnDescriptions.count();
  for (int i = 0;i<nColumns;i++) {
    if (_columnDescriptions[i].type()>=standardTypeCount ) {
      _columnDescriptions[i].setType(0);
    }
  }
}
