/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2008-07-11
**
***************************************************************************/

#ifndef QTB_COLUMNTEXTPARSER_H
#define QTB_COLUMNTEXTPARSER_H

#include <QMap>
#include <QSet>
#include <QTextStream>
#include <QStringList>

#include "qtbthread.h"
#include "qtbcompatqt43.h"
#include "qtbxmlclass.h"
#include "qtbcolumntextdescription.h"
#include "qtbtoolsdllexport.h"

class QTBTOOLS_EXPORT QtbColumnTextParser : public QtbThread, public QtbXMLClass
{
  Q_OBJECT
public:
  QtbColumnTextParser( QObject * parent = 0 );
  QtbColumnTextParser( const QtbColumnTextParser& o );
  ~QtbColumnTextParser();

  virtual const QString& xml_tagName() const {return xmlColumnTextParserTag;}
  static const QString xmlColumnTextParserTag;

  // Initialization
  void setText( QTextStream * s ) { _text = s; }
  void setStandardTypes( const QStringList& standardTypes );
  const QStringList& standardTypes() const { return _standardTypes; }

  // Line properties
  int startLine() const { return _startLine; }
  void setStartLine( int s ) { _startLine = s<0 ? 0 : s;  }

  int endLine() const { return _endLine; }
  void setEndLine( int e ) { _endLine = e<0 ? 0 : e; }

  QString acceptPattern() const { return _acceptPattern; }
  void setAcceptPattern( QString a ) { _acceptPattern = a; }

  QString rejectPattern() const { return _rejectPattern; }
  void setRejectPattern( QString r ) { _rejectPattern = r; }

  // Column properties
  void setFixedWidth( bool f) { _fixedWidth = f; }
  bool isFixedWidth() const { return _fixedWidth; }

  QString delimiters() const { return _delimiters; }
  void setDelimiters( QString d ) { _delimiters = d; }

  bool acceptNullColumns() const { return _acceptNullColumns; }
  void setAcceptNullColumns( bool a ) { _acceptNullColumns = a; }

  QString columnWidths() const { return columnWidths(_columnWidths); }
  void setColumnWidths( QString w ) { _columnWidths = columnWidths(w); }

  // Section properties
  bool isSectionLinePattern() const { return _isSectionLinePattern; }
  void setSectionLinePattern( bool b ) { _isSectionLinePattern = b; }

  QString sectionLinePattern() const { return _sectionLinePattern; }
  void setSectionLinePattern( QString s ) { _sectionLinePattern = s; }

  bool isSectionLineCount() const { return _isSectionLineCount; }
  void setSectionLineCount( bool b ) { _isSectionLineCount = b; }

  int sectionLineCount() const { return _sectionLineCount; }
  void setSectionLineCount( int s ) { _sectionLineCount = s; }

  bool isSectionRowCount() const { return _isSectionRowCount; }
  void setSectionRowCount( bool b ) { _isSectionRowCount = b; }

  int sectionRowCount() const { return _sectionRowCount; }
  void setSectionRowCount( int s ) { _sectionRowCount = s; }

  bool isSectionCellPattern() const { return _isSectionCellPattern; }
  void setSectionCellPattern( bool b ) { _isSectionCellPattern = b; }

  int sectionCellColumn() const { return _sectionCellColumn; }
  void setSectionCellColumn( int s ) { _sectionCellColumn = s; }

  QString sectionCellPattern() const { return _sectionCellPattern; }
  void setSectionCellPattern( QString s ) { _sectionCellPattern = s; }

  // Column type and operations
  int type( int column ) const { return columnDescription(column).type(); }
  void setType( int column, int type ) { return columnDescription(column).setType( type ); }
  QString typeName( int column ) const { return _standardTypes[columnDescription(column).type()]; }
  void setTypeName( int column, QString type );

  double factor( int column ) const { return columnDescription(column).factor(); }
  void setFactor( int column, double f ) { columnDescription(column).setFactor(f); }

  QString replaceRx( int column ) const { return columnDescription(column).replaceRx(); }
  void setReplaceRx( int column, QString rx ) { columnDescription(column).setReplaceRx(rx); }

  QString replaceAfter( int column ) const { return columnDescription(column).replaceAfter(); }
  void setReplaceAfter( int column, QString after ) { columnDescription(column).setReplaceAfter(after); }

  // Start/stop parser engine
  void startUpdates();
  void stopUpdates();

  // Final queries
  QString text( int row, int column ) const;
  bool isSectionFirstRow( int row ) const { return _sectionMarkers.contains( row ); }
  int rowCount() const;
  int columnCount() const;
  int lineNumber( int row ) const;
  int currentRow() const { return _currentRow; }
  void setCurrentRow( int r ) const { _currentRow = r; }
protected:
  virtual void xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const;
  virtual void xml_writeChildren(XML_WRITECHILDREN_ARGS) const;
  virtual bool xml_setProperty(XML_SETPROPERTY_ARGS);
  virtual QtbXMLMember xml_member(XML_MEMBER_ARGS);
  virtual void xml_polish(XML_POLISH_ARGS);
signals:
  void dataAvailable();
private:
  virtual void run();
  void parseRows();
  void parseColumns( int row, QRegExp * sectionCellRegExp );
  QString cellContent( int column, QString s ) const;
  static QVector<int> columnWidths( QString ws );
  static QString columnWidths( QVector<int> wi );

  QtbColumnTextDescription& columnDescription( int column );
  const QtbColumnTextDescription& columnDescription( int column ) const;

  QTextStream * _text;
  QStringList _standardTypes;

  QVector<QtbColumnTextDescription> _columnDescriptions;

  int _startLine, _endLine;
  QString _acceptPattern, _rejectPattern;
  QString _sectionLinePattern, _sectionCellPattern;
  int _sectionLineCount, _sectionRowCount, _sectionCellColumn;
  uint _fixedWidth:1;
  uint _acceptNullColumns:1;
  uint _isSectionLinePattern:1;
  uint _isSectionLineCount:1;
  uint _isSectionRowCount:1;
  uint _isSectionCellPattern:1;
  QString _delimiters;
  QVector<int> _columnWidths;

  QAtomicInt _terminate;
  mutable QMutex _mutex;
  QMap<int,QStringList> _cols;
  int _columnCount;
  QMap<int,int> _lines2rows;
  QStringList _rows;
  int _lineNumber, _lastSectionMarkerLine, _lastSectionMarkerRow;
  QSet<int> _sectionMarkers;

  mutable int _currentRow;
};

#endif // QTB_COLUMNTEXTPARSER_H
