/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-01-21
**
***************************************************************************/

#ifndef QTB_Complex_H
#define QTB_Complex_H

#include <math.h>
#include <stdio.h>

#include "qtbtoolsdllexport.h"

// Trying to use SSE extension
typedef double v2df __attribute__ ((vector_size (16)));

class QTBTOOLS_EXPORT QtbComplex
{
public:
  QtbComplex( const QtbComplex& c ) { _data.vector = c._data.vector;}
  QtbComplex( double re = 0, double im = 0 ) { _data.values.re = re;_data.values.im = im; }
  ~QtbComplex() {}

  inline bool operator==( const QtbComplex& c ) const;

  inline QtbComplex& operator=( const QtbComplex& c );
  inline QtbComplex& operator+=( const QtbComplex& c );
  inline QtbComplex& operator-=( const QtbComplex& c );
  inline QtbComplex& operator*=( const QtbComplex& c );
  inline QtbComplex& operator/=( const QtbComplex& c );

  inline QtbComplex& operator=( const double& d );
  inline QtbComplex& operator+=( const double& d );
  inline QtbComplex& operator-=( const double& d );
  inline QtbComplex& operator*=( const double& d );
  inline QtbComplex& operator/=( const double& d );

  inline QtbComplex operator+( const QtbComplex& c ) const;
  inline QtbComplex operator-( const QtbComplex& c ) const;
  inline QtbComplex operator*( const QtbComplex& c ) const;
  inline QtbComplex operator/( const QtbComplex& c ) const;

  inline QtbComplex operator+( const double& d ) const;
  inline QtbComplex operator-( const double& d ) const;
  inline QtbComplex operator*( const double& d ) const;
  inline QtbComplex operator/( const double& d ) const;

  bool isNull() const { return _data.values.re==0 && _data.values.im==0; }
  double re() const { return _data.values.re; }
  double im() const { return _data.values.im; }
  void set( double re, double im ) { _data.values.re = re; _data.values.im = im; }
  void setRe( double re ) {_data.values.re = re; }
  void setIm( double im ) { _data.values.im = im; }
  void setExp( double fac, double arg )
  {
    _data.values.re = fac*::cos( arg );
    _data.values.im = fac*::sin( arg );
  }
  void setUnitExp( double arg )
  {
    _data.values.re = ::cos( arg );
    _data.values.im = ::sin( arg );
  }
  double abs() const { return ::sqrt( abs2() ); }
  double abs2() const { return _data.values.re * _data.values.re + _data.values.im * _data.values.im; }
  double phase() const { return atan2( _data.values.im, _data.values.re ); }
protected:
  union {
    v2df vector;
    struct {
      double re, im;
    } values;
  } _data;
};

inline QtbComplex sqrt( const QtbComplex& c );
inline QtbComplex cos( const QtbComplex& c );
inline QtbComplex sin( const QtbComplex& c );
inline QtbComplex tan( const QtbComplex& c );
inline QtbComplex asin( const QtbComplex& c );
inline QtbComplex log( const QtbComplex& c );
inline QtbComplex exp( const QtbComplex& c );
inline QtbComplex conjugate( const QtbComplex& c );
inline QtbComplex inverse( const QtbComplex& c );

inline bool QtbComplex::operator==( const QtbComplex& c ) const
{
  return _data.values.re == c._data.values.re && _data.values.im == c._data.values.im;
}

inline QtbComplex& QtbComplex::operator=( const double& d )
{
  _data.values.re = d;
  _data.values.im = 0;
  return *this;
}

inline QtbComplex& QtbComplex::operator=( const QtbComplex& c )
{
  _data.vector=c._data.vector;
  return *this;
}

inline QtbComplex& QtbComplex::operator+=( const double& d )
{
  _data.values.re += d;
  return *this;
}

inline QtbComplex& QtbComplex::operator+=( const QtbComplex& c )
{
  _data.vector+=c._data.vector;
  return *this;
}

inline QtbComplex& QtbComplex::operator-=( const double& d )
{
  _data.values.re -= d;
  return *this;
}

inline QtbComplex& QtbComplex::operator-=( const QtbComplex& c )
{
  _data.vector -= c._data.vector;
  return *this;
}

inline QtbComplex& QtbComplex::operator*=( const double& d )
{
#if defined(QT_HAVE_3DNOW) && defined(QT_HAVE_SSE)
  v2df factor = __builtin_ia32_movddup( d );
  __builtin_ia32_mulpd( _data.vector, factor );
#else
  _data.values.re *= d;
  _data.values.im *= d;
#endif
  return *this;
}

inline QtbComplex& QtbComplex::operator*=( const QtbComplex& c )
{
#if defined(QT_HAVE_3DNOW) && defined(QT_HAVE_SSE)
  v2df thisIm = __builtin_ia32_shufpd( _data.vector, _data.vector,3 );       // Im part of this in both
  v2df thisRe = __builtin_ia32_shufpd( _data.vector, _data.vector,0 );       // Re part of this in  both
  // Next two lines not accepted by gcc ?? even with -msse3 option
  //v2df thisIm = __builtin_ia32_movddup( _data.values.im );                    // Im part of this in both
  //v2df thisRe = __builtin_ia32_movddup( _data.values.re );                    // Re part of this in  both
  v2df c_flipped = __builtin_ia32_shufpd( c._data.vector,c._data.vector,1 ); // Swap re and im parts of c
  _data.vector = __builtin_ia32_addsubpd( __builtin_ia32_mulpd( thisRe, c._data.vector ),
                                          __builtin_ia32_mulpd( thisIm, c_flipped ) );
#elif defined( QT_HAVE_SSE2 )
  v2df thisIm = __builtin_ia32_shufpd( _data.vector, _data.vector,3 );       // Im part of this in both
  v2df thisRe = __builtin_ia32_shufpd( _data.vector, _data.vector,0 );       // Re part of this in  both
  v2df c_flipped = __builtin_ia32_shufpd( c._data.vector,c._data.vector,1 ); // Swap re and im parts of c
  static const union {                                                  // (signbit,0)
    int i[4]; v2df v;
  } signbitlow = {{0,0x80000000,0,0}};
  thisIm = __builtin_ia32_xorpd(thisIm, signbitlow.v);                 // Change sign of low
  // Multiply and add:
  _data.vector = __builtin_ia32_mulpd( thisRe, c._data.vector ) +
                 __builtin_ia32_mulpd( thisIm, c_flipped );
#else
  double tmpre = _data.values.re * c._data.values.re - _data.values.im * c._data.values.im;
  _data.values.im = _data.values.re * c._data.values.im + _data.values.im * c._data.values.re;
  _data.values.re = tmpre;
#endif
  return *this;
}

inline QtbComplex& QtbComplex::operator/=( const double& d )
{
  _data.values.re /= d;
  _data.values.im /= d;
  return *this;
}

inline QtbComplex& QtbComplex::operator/=( const QtbComplex& c )
{
  double m2 = c.abs2();
  double tmpre = ( _data.values.re * c._data.values.re + _data.values.im * c._data.values.im ) / m2;
  _data.values.im = ( _data.values.im * c._data.values.re - _data.values.re * c._data.values.im ) / m2;
  _data.values.re = tmpre;
  return *this;
}

inline QtbComplex QtbComplex::operator+( const double& d ) const
{
  return QtbComplex( _data.values.re + d, _data.values.im );
}

inline QtbComplex QtbComplex::operator+( const QtbComplex& c ) const
{
  QtbComplex r;
  r._data.vector = _data.vector + c._data.vector;
  return r;
}

inline QtbComplex QtbComplex::operator-( const double& d ) const
{
  return QtbComplex( _data.values.re -d, _data.values.im );
}

inline QtbComplex QtbComplex::operator-( const QtbComplex& c ) const
{
  QtbComplex r;
  r._data.vector = _data.vector - c._data.vector;
  return r;
}

inline QtbComplex QtbComplex::operator*( const double& d ) const
{
  return QtbComplex( _data.values.re * d, _data.values.im * d );
}

inline QtbComplex QtbComplex::operator*( const QtbComplex& c ) const
{
  return QtbComplex( _data.values.re * c._data.values.re - _data.values.im * c._data.values.im,
                     _data.values.re * c._data.values.im + _data.values.im * c._data.values.re );
}

inline QtbComplex QtbComplex::operator/( const double& d ) const
{
  return QtbComplex( _data.values.re / d, _data.values.im / d );
}

inline QtbComplex QtbComplex::operator/( const QtbComplex& c ) const
{
  double m2 = c.abs2();
  return QtbComplex( ( _data.values.re * c._data.values.re + _data.values.im * c._data.values.im ) / m2,
                     ( _data.values.im * c._data.values.re - _data.values.re * c._data.values.im ) / m2 );
}

inline QtbComplex sqrt( const QtbComplex& c )
{
  double r2 = ::sqrt( c.abs() );
  double phi2 = c.phase() / 2;
  return QtbComplex(r2*::cos( phi2 ), r2*::sin( phi2 ));
}

inline QtbComplex inverse( const QtbComplex& c )
{
  double m2 = 1.0 / c.abs2();
  return QtbComplex( c.re() * m2, -m2 * c.im() );
}

inline QtbComplex exp( const QtbComplex& c )
{
  double r = ::exp( c.re() );
  return QtbComplex(r*::cos( c.im() ), r*::sin( c.im() ));
}

inline QtbComplex log( const QtbComplex& c )
{
  return QtbComplex(::log(c.abs()), c.phase());
}

/*!
  asin(C)=-i log (sqrt(1-C^2)+i*C), where C is this
*/
inline QtbComplex asin( const QtbComplex& c )
{
  return QtbComplex(0.0,-1.0)*log(sqrt(QtbComplex(1.0)-c*c)+QtbComplex(0.0,1.0)*c);
}

inline QtbComplex cos( const QtbComplex& c )
{
  return QtbComplex( ::cos( c.re() ) *::cosh( c.im() ), -::sin( c.re() ) * sinh( c.im() ) );
}

inline QtbComplex sin( const QtbComplex& c )
{
  return QtbComplex( ::sin( c.re() ) *::cosh( c.im() ), ::cos( c.re() ) * sinh( c.im() ) );
}

inline QtbComplex tan( const QtbComplex& c )
{
  return sin(c)/cos(c);
}

inline QtbComplex conjugate( const QtbComplex& c )
{
  return QtbComplex( c.re(),-c.im() );
}

#endif
