/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-01-09
**
***************************************************************************/

#include <math.h>

#include "qtbglobal.h"
#include "qtbconcentricsearch.h"
#include "qtbabstractfunction2.h"
#include "qtbmemorychecker.h"

/*!
*/
QtbConcentricSearch::QtbConcentricSearch()
{
  QTB_TRACE;
  _segment=Top;
  _r0 = 1e99;
  _rMax = 0.0;
  _dr = 0.0;
}

QtbConcentricSearch::~QtbConcentricSearch()
{
  QTB_TRACE;
}

/*!
  \fn setCenter( double x, double y ) { _x=x; _y=y; }
  x,y are the coordinates of the center.
*/

/*!
  \fn void setNoiseLevel( double nl ) {  _noiseLevel=nl; }
  noiseLevel is the minimum value necessary to consider a refinement of the peak.
*/

/*!
  \a r0 is the starting radius.
  \a dr is the general step (either in radius or along the circle).
  \a rMax is the maximum radius to cover.
*/
void QtbConcentricSearch::setGrid( double minR, double maxR, double dR)
{
  QTB_TRACE;
  _r0=minR;
  _rMax=maxR;
  _pos=_x-_r0;
  _dr=dR;
  QTB_ASSERT(_r0>2.0*_dr);
  _precision=_dr*0.001;
}

double QtbConcentricSearch::maximum()
{
  QTB_TRACE;
  while (_r0<_rMax) {
    switch (_segment) {
    case Top:
      if (scanX(_x-_r0,_x+_r0,_y-_r0)) return _maxValue;
      _segment=Bottom;
      _pos=_x-_r0;
      break;
    case Bottom:
      if (scanX(_x-_r0,_x+_r0,_y+_r0)) return _maxValue;
      _segment=Left;
      _pos=_y-_r0;
      break;
    case Left:
      if (scanY(_y-_r0,_y+_r0,_x-_r0)) return _maxValue;
      _segment=Right;
      _pos=_y-_r0;
      break;
    case Right:
      if (scanY(_y-_r0,_y+_r0,_x+_r0)) return _maxValue;
      _segment=Top;
      _pos=_x-_r0;
      break;
    }
    _r0+=_dr;
  }
  return 0.0;
}

double QtbConcentricSearch::maxRadius() const
{
  QTB_TRACE;
  double dx=_maxX-_x;
  double dy=_maxY-_y;
  return sqrt(dx*dx+dy*dy);
}

bool QtbConcentricSearch::scanX(double min, double max, double y)
{
  QTB_TRACE;
  if (fabs(_pos-min)<1e-15) {
    _v1=_function->value(_pos,y);
    _pos+=_dr;
  }
  if (fabs(_pos-(min+_dr))<1e-15) {
    _v2=_function->value(_pos,y);
    _pos+=_dr;
    if (_v2<_v1 && _v1>_noiseLevel) {
      _maxValue=_v1;
      crossRefineX(_pos-2*_dr, y, _dr);
      return true;
    }
  }
  for (;_pos<max;_pos+=_dr) {
    _v3=_function->value(_pos,y);
    if (_v2>_v1 && _v2>_v3 && _v2>_noiseLevel) {
      _maxValue=_v2;
      crossRefineY(_pos-_dr, y, 0.5*_dr);
      _v1=_v2;
      _v2=_v3;
      _pos+=_dr;
      return true;
    }
    _v1=_v2;
    _v2=_v3;
  }
  return false;
}

bool QtbConcentricSearch::scanY(double min, double max, double x)
{
  QTB_TRACE;
  if (fabs(_pos-min)<1e-15) {
    _v1=_function->value(x,_pos);
    _pos+=_dr;
  }
  if (fabs(_pos-(min+_dr))<1e-15) {
    _v2=_function->value(x,_pos);
    _pos+=_dr;
    if (_v2<_v1 && _v1>_noiseLevel) {
      _maxValue=_v1;
      crossRefineY(x, _pos-2*_dr, _dr);
      return true;
    }
  }
  for (;_pos<max;_pos+=_dr) {
    _v3=_function->value(x,_pos);
    if (_v2>_v1 && _v2>_v3 && _v2>_noiseLevel) {
      _maxValue=_v2;
      crossRefineX(x,_pos-_dr, 0.5*_dr);
      _v1=_v2;
      _v2=_v3;
      _pos+=_dr;
      return true;
    }
    _v1=_v2;
    _v2=_v3;
  }
  return false;
}

void QtbConcentricSearch::crossRefineX(double x, double y, double dr)
{
  QTB_TRACE;
  if (dr<_precision) {
    _maxX=x;
    _maxY=y;
    return;
  }
  double v=_function->value(x+dr, y);
  if (v>_maxValue) {
    x+=dr;
    while (v>_maxValue) {
      x+=dr;
      _maxValue=v;
      v=_function->value(x, y);
    }
    crossRefineY(x-dr, y, 0.5*dr);
  } else {
    x-=dr;
    v=_function->value(x, y);
    while (v>_maxValue) {
      x-=dr;
      _maxValue=v;
      v=_function->value(x, y);
    }
    crossRefineY(x+dr, y, 0.5*dr);
  }
}

void QtbConcentricSearch::crossRefineY(double x, double y, double dr)
{
  QTB_TRACE;
  if (dr<_precision) {
    _maxX=x;
    _maxY=y;
    return;
  }
  double v=_function->value(x, y+dr);
  if (v>_maxValue) {
    y+=dr;
    while (v>_maxValue) {
      y+=dr;
      _maxValue=v;
      v=_function->value(x, y);
    }
    crossRefineX(x, y-dr, 0.5*dr);
  } else {
    y-=dr;
    v=_function->value(x, y);
    while (v>_maxValue) {
      y-=dr;
      _maxValue=v;
      v=_function->value(x, y);
    }
    crossRefineX(x, y+dr, 0.5*dr);
  }
}
