/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-02-08
**
***************************************************************************/

#include <time.h>

#include <QLocale>
#include <QTranslator>
#include <QProcess>
#include <QDateTime>
#include <QSettings>

#if defined(Q_WS_WIN)
#  include <unistd.h>
#  include <windows.h>
#endif

#if defined(Q_WS_X11) && !defined(Q_CC_SUN) && !defined(QTB_NO_BACKTRACE)
#  include <execinfo.h>
#endif
#include <signal.h>
#include <cxxabi.h>
#ifndef Q_WS_WIN
#  include <sys/ioctl.h>
#  include <fcntl.h>
#endif

#include "qtbcoreapplicationprivate.h"
#include "qtbthread.h"
#include "qtbtracebug.h"
#include "qtbglobal.h"
#include "qtbstringsection.h"
#include "qtbpackageinfo.h"
#include "qtbmessage.h"
#include "qtbfile.h"
#include "qtbcompatqt43.h"
#include "qtbpathtranslator.h"
#include "qtbxmlclass.h"
#include "qtbxmlstringstream.h"
#include "qtbabstractnumericalcache.h"
#include "qtbmemorycheckerprivate.h"

QtbCoreApplicationPrivate * QtbCoreApplicationPrivate::_self = 0;
QString QtbCoreApplicationPrivate::_crashingThread;

/*!
  \class QtbCoreApplicationPrivate qtbcoreapplicationprivate.h
  \brief Brief description of class still missing

  Full description of class still missing
*/

/*!
  Initialize the application and analyse arguments. If there's a request for help, call back \a help is activated.
*/
void QtbCoreApplicationPrivate::init ( int & argc, char ** argv )
{
  srand ( time( NULL ) ); // init random generator

  if (!_self) {
    _self = this;
  } else {
    qDebug("Only one instance of QtbApplication or QtbApplicationCore is allowed.");
    abort();
  }

#ifdef DO_MEMORY_CHECK
  // Activate recording for undeleted blocks
  QtbMemoryCheckerPrivate::setRecord(true);
#endif
  // Record application main stream
  _thread=QThread::currentThread();
  // Initialize log stream system for the main thread
  initStream();
  // Message handler is initialized by QtbCoreApplication and QtbApplication
  _messageHandler = 0;
  // Facility to translate path if absolute file path stored in files are changed
  _pathTranslator = 0;
  // Size of terminal
  _terminalCols = 0;
  _terminalRows = 0;

  // Check arguments
  int i, j;
  j = 1;
  for (i=1; i<argc; i++) {
    QByteArray arg = argv[i];
    if (arg=="--version" || arg=="-version") {
      printf( "Versions:\n%s", version().toAscii().data() );
      ::exit(0);
    } else if (arg=="--app-version" || arg=="-app-version") {
      printf( "%s %s\n", constructorApplicationName().toAscii().data(),
                         version(constructorApplicationName()).toAscii().data() );
      ::exit(0);
#ifdef Q_WS_MACX
    } else if (arg.left(5)=="-psn_") { // does nothing: just get rid of that argument
#endif // Q_WS_MACX
    } else {
      argv[j++] = argv[i];
    }
  }
  if(j < argc) {
    argv[j] = 0;
    argc = j;
  }
}

/*!
  Description of destructor still missing
*/
QtbCoreApplicationPrivate::~QtbCoreApplicationPrivate()
{
  delete _pathTranslator;
}

/*!
  In normal destructor, QCoreApplication is no longer available, hence strange behavior.
  This function must be called by QtbCoreApplication and other related objects
*/
void QtbCoreApplicationPrivate::destructorCleanUp()
{
  delete _messageHandler;
  // Cleanup global pointers
  qDeleteAll(_globalPointers);
  clearStreams();
  checkBlocks(false, true);
  if (!_errors.isEmpty()) {
    reportBugNow( QtbMessage::Warning, "Quitting application with remaining error messages" );
  }
  QtbAbstractNumericalCache::clear();
}

void QtbCoreApplicationPrivate::initInternalDebugger(bool reportBugs)
{
  if (reportBugs) {
    // Redirect exception signals
    signal (SIGABRT, osSignal);
    signal (SIGSEGV, osSignal);
#ifndef Q_WS_WIN
    signal (SIGBUS, osSignal);
#endif
    // User interrupts
    signal (SIGINT, osSignal);
    signal (SIGTERM, osSignal);
    // Redirect QTB_ASSERT messages
    qInstallMsgHandler(messageOutput);
  } else {
    // Redirect QTB_ASSERT messages
    qInstallMsgHandler(messageOutputAbort);
  }
}

/*!
  Check whether atomic operations are available for current processor and issue
  warnings if not satisfied. Implemented for Qt>=4.4.0. Streams must be initialized
*/
void QtbCoreApplicationPrivate::checkAtomicOperations()
{
#if (QT_VERSION >= QT_VERSION_CHECK(4, 4, 0))
  /* Currently not used in sesarray package so left commented
  if ( !QAtomicInt::isReferenceCountingNative() )
    qtbApp << "WARNING: Atomic 'reference counting' is not available for your processor" << endl;
  if (!QAtomicInt::isFetchAndAddNative())
    qtbApp << "WARNING: Atomic 'fetch and add' is not available for your processor" << endl;
  */
  if ( !QAtomicInt::isTestAndSetNative() )
    qtbApp << "WARNING: Atomic 'test and set' is not available for your processor" << endl;
  if ( !QAtomicInt::isFetchAndStoreNative() )
    qtbApp << "WARNING: Atomic 'fetch and store' is not available for your processor" << endl;
#endif
}

/*!
  Internal use only: needed in QtbCoreApplication and QtbApplication constructors.
  There QCoreApplication::applicationName() is not yet available. As progam is executed
  this function no longer returns the application name if plugins are loaded.
*/
QString QtbCoreApplicationPrivate::constructorApplicationName()
{
  return QtbPackageInfo::list()->last().package();
}

/*!
  Returns a complete description of all versions
*/
QString QtbCoreApplicationPrivate::version()
{
  QString str="  Qt                  ";
  str+=qVersion();
  str+="\n";
  for (QList<QtbPackageInfo>::iterator it=QtbPackageInfo::list()->begin();it!=QtbPackageInfo::list()->end(); it++ ) {
    str+="  "+it->package().leftJustified(20,' ')+it->version();
    if (it->distribution().isEmpty()) {
      str+="\n";
    } else {
      str+=" from "+it->distribution()+"\n";
    }
    str+="      (built on "+it->versionTime()+")\n";
  }
  return str;
}

/*!
  Returns the version of \a item
*/
QString QtbCoreApplicationPrivate::version( QString item )
{
  for (QList<QtbPackageInfo>::iterator it=QtbPackageInfo::list()->begin();it!=QtbPackageInfo::list()->end(); it++ ) {
    if (it->package()==item) {
      return it->version();
    }
  }
  return QString::null;
}

/*!
  Returns the copyright message
*/
QString QtbCoreApplicationPrivate::copyright()
{
  QString appName = QCoreApplication::applicationName();
  for (QList<QtbPackageInfo>::iterator it=QtbPackageInfo::list()->begin();it!=QtbPackageInfo::list()->end(); it++ ) {
    if (it->package()==appName) {
      return it->copyright()+"\n"+it->contact();
    }
  }
  return QString::null;
}

void QtbCoreApplicationPrivate::initTranslations()
{
  for (QList<QtbPackageInfo>::iterator it=QtbPackageInfo::list()->begin();it!=QtbPackageInfo::list()->end(); it++ ) {
    QString langFile = it->package() + "." + QLocale::system().name();
    QTranslator * t = new QTranslator( QCoreApplication::instance() );
    bool isLoaded;
#ifdef Q_WS_MACX
    QDir d( "/Library/Geopsy/lang" );
    isLoaded = t->load( langFile, d.absolutePath() );
    if (!isLoaded) {
      d.setPath( it->shareDir() );
      d.cd( "lang" );
      isLoaded = t->load( langFile, d.absolutePath() );
    }
#else
    QDir d( it->shareDir() );
    d.cd( "lang" );
    isLoaded = t->load( langFile, d.absolutePath() );
#endif
    if ( isLoaded ) {
      QCoreApplication::instance() ->installTranslator( t );
    } else {
      delete t;
      switch ( QLocale::system().language() ) {
      case QLocale::C:
      case QLocale::English:
        break;
      default: {
        // In any case the graphical log viewer is still not started. Hence this message will be lost
        // to system terminal. For command line tools (e.g. gpdcreport) these kind of message must be
        // sent through stderr and the redirection of the main application stream is done after the
        // application initialization. Therefore we use the normal stderr directly.
        QTextStream s(stderr);
        s << QString("No translation found for language %1" ).arg( QLocale::system().name() ) << endl;
        }
        break;
      }
    }
  }
}

void QtbCoreApplicationPrivate::setMemoryVerbose( bool v )
{
#ifdef DO_MEMORY_CHECK
  QtbMemoryCheckerPrivate::setVerbose( v );
#else
  Q_UNUSED(v)
#endif
}

/*!
  Check for undeleted blocks
*/
void QtbCoreApplicationPrivate::checkBlocks(bool showLib, bool showUndeleted, char* file, int line)
{
#ifdef DO_MEMORY_CHECK
  QtbMemoryCheckerPrivate::checkBlocks(showLib, showUndeleted, file, line);
#else
  Q_UNUSED(showLib)
  Q_UNUSED(showUndeleted)
  Q_UNUSED(file)
  Q_UNUSED(line)
#endif
}

/*!
  Check for damaged blocks
*/
bool QtbCoreApplicationPrivate::isBlockDamaged( void* v, char* file, int line )
{
#ifdef DO_MEMORY_CHECK
  return QtbMemoryCheckerPrivate::isBlockDamaged( v, file, line );
#else
  Q_UNUSED(v)
  Q_UNUSED(file)
  Q_UNUSED(line)
  return false;
#endif
}

/*!
  Print block information
*/
void QtbCoreApplicationPrivate::printBlockInfo(void * ptr, char* file, int line )
{
#ifdef DO_MEMORY_CHECK
  QtbMemoryCheckerPrivate::printBlockInfo( ptr, file, line );
#else
  Q_UNUSED(ptr)
  Q_UNUSED(file)
  Q_UNUSED(line)
#endif
}

void QtbCoreApplicationPrivate::setMessageHandler( QtbMessage * m )
{
  delete _messageHandler;
  _messageHandler = m;
}

void QtbCoreApplicationPrivate::freezeStream( bool b ) const
{
  currentStream().freeze( b );
}

/*!
  Warning after this call, no stream can be used before initStream()
*/
void QtbCoreApplicationPrivate::clearStreams()
{
  QList<QtbLogStream *> s = _streams.values();
  for ( QList<QtbLogStream *>::iterator it = s.begin();it != s.end();++it ) {
    if ( (*it)->detach() ) delete *it;
  }
}

/*!
  Initialize the stream of \a thread to stdout. Does nothing if stream already exist for \a thread.
*/
void QtbCoreApplicationPrivate::initStream( QThread * thread )
{
  if ( !thread ) thread = QThread::currentThread();
  if ( !_streams.contains( thread ) ) {
    // Create at least one stream to stdout
    QtbLogStream * s = new QtbLogStream( stdout );
    //printf("Init log dispatch for thread %x with stream %x\n", (long)QThread::currentThread(), (long) s);
    setStream( s, thread );
  }
}

void QtbCoreApplicationPrivate::setStream( QtbLogStream * stream, QThread * thread )
{
  if ( !thread ) thread = QThread::currentThread();
  QMap<QThread *, QtbLogStream *>::iterator it;
  it=_streams.find( thread );
  if ( it != _streams.end() ) {
    QtbLogStream * old_stream = it.value();
    if (old_stream->detach()) delete old_stream;
  }
  _streams.insert( thread, stream );
}

QtbLogStream& QtbCoreApplicationPrivate::stream( QThread * thread ) const
{
  QMap<QThread *, QtbLogStream *>::const_iterator it;
  it=_streams.find( thread );
  if ( it != _streams.end() ) {
    return * it.value();
  } else return currentStream();
}

void QtbCoreApplicationPrivate::streamDestroyed(QtbLogStream * stream)
{
  if (!stream) return;
  QList<QThread *> threads=_streams.keys(stream);
  if (!threads.isEmpty()) {
    QTB_ASSERT (threads.count()==1);
    QThread * thread=threads.first();
    _streams.remove( thread );
    _originalStreams.remove( thread );
  }
}

/*!
  Redirect all output of current thread to stream \a stream.

  \sa restoreStream(), QtbStreamRedirection
*/
void QtbCoreApplicationPrivate::redirectStream(QtbLogStream * stream)
{
  QMap<QThread *, QtbLogStream *>::iterator it;
  QThread * t = QThread::currentThread();
  it=_originalStreams.find( t );
  if (it==_originalStreams.end()) {
    it = _streams.find( t );
    if (it!=_streams.end()) {
      _originalStreams.insert( t,  it.value() );
    }
  }
  _streams[ QThread::currentThread() ]=stream;
}

/*!
  Stop redirection and restore original stream for the current thread.

  \sa redirectStream(), QtbStreamRedirection
*/
void QtbCoreApplicationPrivate::restoreStream()
{
  QMap<QThread *, QtbLogStream *>::iterator it;
  QThread * t = QThread::currentThread();
  it=_originalStreams.find( t );
  if (it!=_originalStreams.end()) {
    _streams[t]=it.value();
  }
}

/*!
  Receive all message from Qt library, stack warning or output bugs
*/
void QtbCoreApplicationPrivate::messageOutput(QtMsgType type, const char *msg)
{
  qInstallMsgHandler(0); // avoid recursion
  switch (type) {
  case QtDebugMsg:
#ifdef QT_NO_DEBUG
    qtbApp.reportBug( QtbMessage::Information, msg );
#else
    printf("[qDebug] Information: %s\n", msg);
#endif
    break;
  case QtWarningMsg:
    qtbApp.reportBug( QtbMessage::Warning, msg );
    break;
  case QtCriticalMsg:
    qtbApp.reportBug( QtbMessage::Critical, msg );
    break;
  case QtFatalMsg:
    qtbApp.reportBugNow( QtbMessage::Fatal, msg );
    ::exit(2);
  }
  qInstallMsgHandler(messageOutput);
}

/*!
  Receive all message from Qt library, stack warning or output bugs
*/
void QtbCoreApplicationPrivate::messageOutputAbort(QtMsgType type, const char *msg)
{
  qInstallMsgHandler(0); // avoid recursion
  switch (type) {
  case QtDebugMsg:
    printf("[qDebug] Information: %s\n", msg);
    break;
  case QtWarningMsg:
    printf("[qDebug] Warning: %s\n", msg);
    break;
  case QtCriticalMsg:
    printf("[qDebug] Critical: %s\n", msg);
    break;
  case QtFatalMsg:
    printf("[qDebug] Fatal: %s\n", msg);
    abort();
  }
  qInstallMsgHandler(messageOutputAbort);
}

/*!
  Cross-platform sleep main thread for \a ms milliseconds
*/
void QtbCoreApplicationPrivate::sleep( int ms )
{
#ifdef Q_WS_WIN
  ::Sleep( ms );
#else
  if (ms>10000)
    ::sleep(ms/1000);
  else
    ::usleep(ms*1000);
#endif
}

/*!
  In case an exception occurs during normal execution
*/
void QtbCoreApplicationPrivate::osSignal( int sig )
{
  static QAtomicInt lock = false;
  if (lock.fetchAndStoreOrdered(true)) {
    signal( sig, SIG_DFL );
    sleep(10000);
    fprintf( stderr, "Exception occured while processing exceptions.\n");
    fprintf( stderr, "No backtrace information can be provided\n");
    fprintf( stderr, "Eventually kill remaining instances:\n");
    fprintf( stderr, "  ps\n" );
    fprintf( stderr, "  pkill <application name>\n" );
    ::exit(2);
  }
  _crashingThread = currentThreadName();
  switch (sig) {
  case SIGABRT:
    qtbApp.reportBugNow( QtbMessage::Fatal, "Abort Signal" );
    break;
  case SIGINT:
    qtbApp.reportBugNow( QtbMessage::Fatal, "Interrupted Signal" );
    break;
  case SIGTERM:
    qtbApp.reportBugNow( QtbMessage::Fatal, "Terminated Signal" );
    break;
  case SIGSEGV:
    qtbApp.reportBugNow( QtbMessage::Fatal, "Segmentation fault Signal" );
    break;
#ifndef Q_WS_WIN
  case SIGBUS:
    qtbApp.reportBugNow( QtbMessage::Fatal, "Bus error Signal" );
    break;
#endif
  default:
    break;
  }
  ::exit(2);
}

/*!
  Delayed bug report to user (wait for end of program).
  Not fatal error message. Stack it and report only when quitting or on fatal error
*/
void QtbCoreApplicationPrivate::reportBug( QtbMessage::Severity sev, const char * msg )
{
  QString str=bugInfo( sev, msg);
  printf("[bug report] %s: %s\n", QtbMessage::severityString(sev).toAscii().data(), msg);
  // Bug submitted to trolltech internal to QWorkspace on closing application
  QByteArray msgBA(msg);
  if (msgBA.startsWith("QCoreApplication::postEvent: Unexpected null receiver")) return;
  // Message output under Mac by graphical applications
  //if (msgBA.startsWith("QObject::moveToThread") ) return;
  // Message output given when restoring workspaces after an upgrade
  if (msgBA.startsWith("QMainWindow::restoreState()") ) return;
  // Message output given by Qt especially under Windows
  if (msgBA.startsWith("QObject::startTimer") ) return;
  if (msgBA.startsWith("X Error")) {
    /*static bool userNotWarned = true;
    if (userNotWarned) {
      userNotWarned = false;
      QtbMessage::warning( MSG_ID, translate("QtbCoreApplicationPrivate", "X11 server error"),
                           translate("QtbCoreApplicationPrivate", "If you are running this application through a ssh connection, "
                           "you can avoid X11 errors and improve the graphical quality by a proper configuration of your X11 server. "
                           "See http://www.geopsy.org/faq.html for details."));
    }*/
    //return;
  }
  _errors << str;
}

void QtbCoreApplicationPrivate::askForUserInterrupt()
{
  if (!hasGui()) printf("\n"); // Avoid uggly output if interrupt occured while printing something.
  if (QtbMessage::question( MSG_ID, translate("QtbCoreApplicationPrivate", "User interrupt"),
                          translate("QtbCoreApplicationPrivate", "Is this process blocked? Answer 'Yes' to "
                                    "generate a bug report."),
                          translate("QtbCoreApplicationPrivate", "Yes"),
                          translate("QtbCoreApplicationPrivate", "No"))==QtbMessage::Answer1) {
    ::exit(0);
  }
}

/*!
  Immediate bug report to user (fatal or quitting application)
*/
void QtbCoreApplicationPrivate::reportBugNow( QtbMessage::Severity sev, const char * msg )
{
  QtbXMLStringStream sError(QIODevice::WriteOnly);
  QtbXMLClass::writeChildren( sError, "Error", bugInfo( sev, msg ) );
  if (!_errors.isEmpty()) {
    for (int i=_errors.count()-1;i>=0;i--) {
      QtbXMLClass::writeChildren( sError, "Error", _errors.at(i) );
    }
    _errors.clear();
  }
  QtbXMLStringStream sReport(QIODevice::WriteOnly);
  QtbXMLClass::writeChildren( sReport, "CrashReport", sError.toString() + appInfo() );
  QString str = sReport.toString();
  QProcess p;
  if (strcmp(msg,"Terminated Signal")==0 || strcmp(msg,"Interrupted Signal")==0) {
    if (!hasGui()) qtbApp.askForUserInterrupt();
    p.start("\""+QCoreApplication::applicationFilePath()+"\" -reportint");
  } else {
    p.start("\""+QCoreApplication::applicationFilePath()+"\" -reportbug");
  }
  p.write(str.toAscii().data());
  p.write("END\n");
  p.setProcessChannelMode(QProcess::MergedChannels);
  if ( p.waitForStarted(10000) ) {
    while(!p.waitForFinished ( 10000 )) {}
    QTextStream(stderr) << p.readAll().data() << flush;
  } else {
    QString bugFileName = QDir::home().absoluteFilePath("sendmeto_bug_at_geopsy.org");
    FILE * f = fopen(bugFileName.toAscii().data(),"a");
    fprintf(f,"%s\n",str.toAscii().data());
    fclose(f);
    printf("Cannot restart \"%s\" in bug report mode (timed out)\n",QCoreApplication::applicationFilePath().toAscii().data());
    printf("Bug report save in %s.\nPlease send this bug report to bug@geopsy.org\n",bugFileName.toAscii().data());
  }
}

/*!
  Analyse current stack and return it as a string
*/
QString QtbCoreApplicationPrivate::backTrace()
{
  QString str;
#ifdef QTB_TRACE_ENABLED
  QStringList tList = QtbThread::threadNames();
  QtbThread::lockThreads();
  QList< QStack<const QtbTraceStamp *> * > sList = QtbThread::stacks();
  for( int i = 0; i<sList.count(); i++ ) {
    str += "---- thread ";
    str += tList.at(i);
    str += "\n";
    str += QtbTrace::backTrace(sList.at(i));
  }
  QtbThread::unlockThreads();
#elif defined(Q_WS_X11) && !defined(Q_CC_SUN) && !defined(QTB_NO_BACKTRACE)
  void *frames[50];
  size_t size;
  char ** strings;
  size_t i;
  size = backtrace (frames, 50);
  strings = backtrace_symbols (frames, size);

  size_t functionLen = 0;
  int status;
  // Demangle, see doc at http://gcc.gnu.org/onlinedocs/libstdc++/latest-doxygen/namespaceabi.html
  for (i = 0; i < size; i++) {
    // get value between (...+
    char * mangledName = strchr( strings[i], '(' );
    if (mangledName) {
      char * mangledNameEnd = strchr( mangledName, '+' );
      if (mangledNameEnd) {
        char * mangledString = new char[mangledNameEnd-mangledName];
        strncpy(mangledString, mangledName+1, mangledNameEnd-mangledName-1);
        mangledString[mangledNameEnd-mangledName-1]='\0';
        char * functionName = abi::__cxa_demangle(mangledString, 0, &functionLen, &status);
        str+=functionName;
        delete [] mangledString;
      } else {
        str+=strings[i];
      }
    } else {
      str+=strings[i];
    }
    str+="\n";
  }
  /*str+="---- raw function names\n";
  for (i = 0; i < size; i++) {
    str+=strings[i];
    str+="\n";
  }*/

  free (strings);
#endif
  return str;
}

/*!
  Analyse current stack for instrumented bugs and return it as a string
*/
QString QtbCoreApplicationPrivate::backTraceBug()
{
  QString str;
  QStringList tList = QtbThread::threadNames();
  QtbThread::lockThreads();
  QList< QStack<QtbTraceBug *> * > sList = QtbThread::bugStacks();
  for( int i = 0; i<sList.count(); i++ ) {
    str += "---- thread ";
    str += tList.at(i);
    str += "\n";
    str += QtbTraceBug::backTrace(sList.at(i));
  }
  QtbThread::unlockThreads();
  return str;
}

QString QtbCoreApplicationPrivate::bugInfo( QtbMessage::Severity sev, const QString& msg )
{
  QtbXMLStringStream s(QIODevice::WriteOnly);
  QtbXMLClass::writeProperty( s, "severity", QtbMessage::severityString( sev ) );
  QtbXMLClass::writeProperty( s, "message", msg );
  QtbXMLClass::writeProperty( s, "time", QDateTime::currentDateTime().toString("yyyy-MM-dd hh:mm:ss.zzz") );
  QtbXMLClass::writeProperty( s, "threads", QtbThread::threadNames().join(",") );
  QtbXMLClass::writeProperty( s, "currentThread", _crashingThread );
  QtbXMLClass::writeProperty( s, "stack", backTrace() );
  QtbXMLClass::writeProperty( s, "bugStack", backTraceBug() );
  return s.toString();
}

/*!
  Returns name of currently running thread
*/
QString QtbCoreApplicationPrivate::currentThreadName()
{
  QtbThread * t = qobject_cast<QtbThread *>(QThread::currentThread());
  if (t) {
    return t->objectName();
  } else {
    return "main thread";
  }
}

QString QtbCoreApplicationPrivate::systemInfo()
{
#if defined(Q_WS_WIN)
  switch (QSysInfo::WindowsVersion) {
  case QSysInfo::WV_DOS_based:
  case QSysInfo::WV_32s: return "Windows 3.1";
  case QSysInfo::WV_95: return "Windows 95";
  case QSysInfo::WV_98: return "Windows 98";
  case QSysInfo::WV_Me: return "Windows Me";
  case QSysInfo::WV_NT_based:
  case QSysInfo::WV_NT: return "Windows NT";
  case QSysInfo::WV_2000: return "Windows 2000";
  case QSysInfo::WV_XP: return "Windows XP";
  case QSysInfo::WV_2003: return "Windows Server 2003";
  case QSysInfo::WV_VISTA: return "Windows Vista";
  case QSysInfo::WV_CE_based:
  case QSysInfo::WV_CE: return "Windows CE\n";
  case QSysInfo::WV_CE_5: return "Windows CE 5\n";
  case QSysInfo::WV_CE_6: return "Windows CE 6\n";
  case QSysInfo::WV_CENET: return "Windows CE .NET";
  }
  return "Windows ??";
#elif defined(Q_WS_MAC)
  switch (QSysInfo::MacintoshVersion) {
  case QSysInfo::MV_9: return "Mac OS 9";
  case QSysInfo::MV_CHEETAH: return "Mac OS X Cheetah (10.0)";
  case QSysInfo::MV_PUMA: return "Mac OS X Puma (10.1)";
  case QSysInfo::MV_JAGUAR: return "Mac OS X Jaguar (10.2)";
  case QSysInfo::MV_PANTHER: return "Mac OS X Panther (10.3)";
  case QSysInfo::MV_TIGER: return "Mac OS X Tiger (10.4)";
  case QSysInfo::MV_LEOPARD: return "Mac OS X Leopard (10.5)";
  case QSysInfo::MV_Unknown: break;
  }
  return "Mac OS X ??";
#else
  QProcess uname;
  uname.start("uname");
  if (uname.waitForFinished(-1)) {
    return uname.readAll().data();
  } else {
    return "Linux";
  }
#endif
}

QString QtbCoreApplicationPrivate::appInfo()
{
  QtbXMLStringStream s(QIODevice::WriteOnly);
  QtbXMLClass::writeProperty( s, "application", QCoreApplication::applicationName() );
  QtbXMLClass::writeProperty( s, "version", version() );
  QtbXMLClass::writeProperty( s, "system", systemInfo() );
  return s.toString();
}

void QtbCoreApplicationPrivate::showHelp( int& i, int argc, char ** argv, QtbApplicationHelp * ( *help ) () )
{
  QtbApplicationHelp * h = help();
  setHelp( h );
  if ( checkOptionArg(i, argc, argv, false) ) {
    h->exec(argv[i]);
  } else {
    h->exec();
  }
  delete h;
}

void QtbCoreApplicationPrivate::setHelp( QtbApplicationHelp * h )
{
  QList<QByteArray> sections = h->sections();
  qSort(sections);
  QString str = translate("QtbCoreApplicationPrivate","Show help about options. SECTION may be empty or: all, wiki, latex");
  for (QList<QByteArray>::iterator it = sections.begin();it!=sections.end();it++) {
    str+= ", " + *it;
  }
  h->addOption(translate("QtbCoreApplicationPrivate","-help <SECTION>"), str);
  h->addOption(translate("QtbCoreApplicationPrivate","-version"),
              translate("QtbCoreApplicationPrivate","Show version information"));
  h->addOption(translate("QtbCoreApplicationPrivate","-app-version"),
              translate("QtbCoreApplicationPrivate","Show short version information"));
  h->addOption(translate("QtbCoreApplicationPrivate","-reportbug"),
              translate("QtbCoreApplicationPrivate","Start bug report dialog, information about bug is passed through stdin. This "
                                                    "option is used internally to report bugs if option -nobugreport is not "
                                                    "specified."));
  h->addOption(translate("QtbCoreApplicationPrivate","-reportint"),
              translate("QtbCoreApplicationPrivate","Start bug report dialog, information about interruption is passed through stdin. This "
                                                    "option is used internally to report interruptions if option -nobugreport is not "
                                                    "specified."));
  h->addOption(translate("QtbCoreApplicationPrivate","-nobugreport"),
              translate("QtbCoreApplicationPrivate","Do not generate bug report in case of error"));
}

/*!
  Check if option argument at \a i is available. If not and if \a mandatory is true, the application exits
  with error code 2. Else, it returns true if an argument is available.

  This function can be used even without a QtbCoreApplication running.
*/
bool QtbCoreApplicationPrivate::checkOptionArg( int& i, int argc, char ** argv, bool mandatory )
{
  i++;
  if (i<argc) {
    bool ok;
    if ( argv[i][0] == '-' ) {
      QByteArray option( argv[i]);
      option.toDouble(&ok);
      if (ok) return true;
    } else return true;
  }
  if (mandatory) {
    // Do not use qtbApp because it is not sure whether it is already created.
    QTextStream sOut(stderr);
    sOut << translate("QtbCoreApplicationPrivate","Option %1 requires an argument, see -help for details.").arg( argv[i-1] ) << endl;
    ::exit(2);
  }
  return false;
}

QString QtbCoreApplicationPrivate::getStdin()
{
  QTB_TRACE;
  QString str;
  int n=256;
  char * buf=new char[n];
  while(true) {
    if (QtbFile::readLine(buf, n, stdin)) {
      if (strcmp(buf,"END\n")==0) break;
      str+=buf;
    }
  }
  delete [] buf;
  return str;
}

/*!
  Convert all remaining arguments into file names (accepting wildchars)
*/
QStringList QtbCoreApplicationPrivate::getFileList(int argc, char ** argv)
{
  QStringList files;
  for (int index = 1; index < argc; index++)
    files.append(argv[index]);
  // Replace file names that include "*" with all their corresponding entries
  QDir d;
  QStringList result;
  QString file;
  foreach(file, files) {
    QFileInfo fi (file);
    QDir dloc(d);
    dloc.cd(fi.dir().path());
    QStringList filters;
    filters << fi.fileName();
    QStringList list=dloc.entryList(filters);
    foreach(file, list) result << dloc.absoluteFilePath(file);
  }
  return result;
}

void QtbCoreApplicationPrivate::addGlobalObject(QtbGlobalObject * ptr)
{
  qtbApp._globalPointers.append(ptr);
}

void QtbCoreApplicationPrivate::removeGlobalObject(QtbGlobalObject * ptr)
{
  int i = qtbApp._globalPointers.indexOf(ptr);
  if (i>-1) qtbApp._globalPointers.removeAt(i);
}

/*!
  Return an integer that "uniquely" identify the user. In fact this is just random number stored in QSettings.
  It is initialized only if not found in QSettings. This is mainly used to identify different users behind
  a proxy server. Each user is then identified by its IP address or http proxy address (if any) and this integer number.
*/
int QtbCoreApplicationPrivate::userId()
{
#ifdef Q_WS_MAC
  QSettings reg(QCoreApplication::organizationDomain());
#else
  QSettings reg(QCoreApplication::organizationName());
#endif
  if (reg.contains("id")) {
    return reg.value("id").toInt();
  } else {
    int id = rand();
    reg.setValue("id",id);
    return id;
  }
}

/*!
  Return a translated path. If not successuful (user cancel), the string is empty.
*/
QString QtbCoreApplicationPrivate::translatePath(const QString& file, const QString& title, const QString& fileType)
{
  if (!_pathTranslator) {
    _pathTranslator = new QtbPathTranslator;
  }
  return _pathTranslator->translate( file, title, fileType );
}

int QtbCoreApplicationPrivate::terminalRows() const
{
  if (_terminalRows==0) {
    terminalSize();
  }
  return _terminalRows;
}

int QtbCoreApplicationPrivate::terminalCols() const
{
  if (_terminalCols==0) {
    terminalSize();
  }
  return _terminalCols;
}

/*!
  Get the terminal size

  Originaly copied from ps command written by Albert Cahalan (Copyright 1998-2002),
  file global.c, package procps-3.2.7

  Comments of the original function (set_screen_size(void)):

  The rules:
    1. Defaults are implementation-specific. (ioctl,termcap,guess)
    2. COLUMNS and LINES override the defaults. (standards compliance)
    3. Command line options override everything else.
    4. Actual output may be more if the above is too narrow.

  SysV tends to spew semi-wide output in all cases. The args
  will be limited to 64 or 80 characters, without regard to
  screen size. So lines of 120 to 160 chars are normal.
  Tough luck if you want more or less than that! HP-UX has a
  new "-x" option for 1024-char args in place of comm that
  we'll implement at some point.

  BSD tends to make a good effort, then fall back to 80 cols.
  Use "ww" to get infinity. This is nicer for "ps | less"
  and "watch ps". It can run faster too.
*/
void QtbCoreApplicationPrivate::terminalSize() const
{
/*
 * Try not to overflow the output buffer:
 *    32 pages for env+cmd
 *    64 kB pages on IA-64
 *    4 chars for "\377", or 1 when mangling to '?'  (ESC_STRETCH)
 *    plus some slack for other stuff
 * That is about 8.5 MB on IA-64, or 0.6 MB on i386
 *
 * Sadly, current kernels only supply one page of env/command data.
 * The buffer is now protected with a guard page, and via other means
 * to avoid hitting the guard page.
 */

/* output buffer size */
// since we mangle to '?' this is 1 (would be 4 for octal escapes)
#define ESC_STRETCH 1
#define OUTBUF_SIZE (2 * 64*1024 * ESC_STRETCH)
#ifdef Q_WS_WIN
  _terminalCols = 80;
  _terminalRows = 1000;
#else
  struct winsize ws;
  char *columns; /* Unix98 environment variable */
  char *lines;   /* Unix98 environment variable */

  do{
    int fd;
    if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) != -1 && ws.ws_col>0 && ws.ws_row>0) break;
    if(ioctl(STDERR_FILENO, TIOCGWINSZ, &ws) != -1 && ws.ws_col>0 && ws.ws_row>0) break;
    if(ioctl(STDIN_FILENO,  TIOCGWINSZ, &ws) != -1 && ws.ws_col>0 && ws.ws_row>0) break;
    fd = open("/dev/tty", O_NOCTTY|O_NONBLOCK|O_RDONLY);
    if(fd != -1){
      int ret = ioctl(fd, TIOCGWINSZ, &ws);
      close(fd);
      if(ret != -1 && ws.ws_col>0 && ws.ws_row>0) break;
    }
    // TODO: ought to do tgetnum("co") and tgetnum("li") here
    ws.ws_col = 80;
    ws.ws_row = 24;
  }while(0);
  _terminalCols = ws.ws_col;  // hmmm, NetBSD subtracts 1
  _terminalRows = ws.ws_row;
  //fprintf(stderr,"Terminal size %i*%i\n",_terminalCols,_terminalRows);

  // TODO: delete this line
  //if(!isatty(STDOUT_FILENO)) _terminalCols = OUTBUF_SIZE;

  columns = getenv("COLUMNS");
  if(columns && *columns){
    long t;
    char *endptr;
    t = strtol(columns, &endptr, 0);
    if(!*endptr && (t>0) && (t<(long)OUTBUF_SIZE)) _terminalCols = (int)t;
  }

  lines   = getenv("LINES");
  if(lines && *lines){
    long t;
    char *endptr;
    t = strtol(lines, &endptr, 0);
    if(!*endptr && (t>0) && (t<(long)OUTBUF_SIZE)) _terminalRows = (int)t;
  }

  if((_terminalCols<9) || (_terminalRows<2))
    fprintf(stderr,"Your %dx%d screen size is bogus. Expect trouble.\n",
      _terminalCols, _terminalRows
    );
#endif
}
