/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-02-08
**
***************************************************************************/

#ifndef QTB_COREAPPLICATIONPRIVATE_H
#define QTB_COREAPPLICATIONPRIVATE_H

#include "qtbtoolsdllexport.h"
#include "qtbpackageinfo.h"
#include "qtblogstream.h"
#include "qtbmessage.h"
#include "qtbapplicationhelp.h"

class QtbPathTranslator;
class QtbApplicationHelp;

class QTBTOOLS_EXPORT QtbGlobalObject
{
public:
  QtbGlobalObject() {}
  virtual ~QtbGlobalObject() {}
};

class QTBTOOLS_EXPORT QtbCoreApplicationPrivate
{
protected:
  QtbCoreApplicationPrivate() { int argc = 0; init(argc, 0 ); }
public:
  QtbCoreApplicationPrivate( int & argc, char ** argv ) { init(argc, argv ); }
  virtual ~QtbCoreApplicationPrivate();

  static QString version();
  static QString version( QString item );
  static QString copyright();
  virtual bool hasGui() const = 0;

  static void checkBlocks( bool showLib, bool showUndeleted, char* file=0, int line=0 );
  static bool isBlockDamaged( void* v, char* file, int line );
  static void printBlockInfo(void * ptr, char* file, int line );
  static void setMemoryVerbose( bool v );

  static bool checkOptionArg( int& i, int argc, char ** argv, bool mandatory = true );
  void showHelp( int& i, int argc, char ** argv, QtbApplicationHelp * ( *help ) () );

  static QtbCoreApplicationPrivate * instance() { return _self; }

  static QString backTrace();
  static QString backTraceBug();
  static QString bugInfo( QtbMessage::Severity sev, const QString& msg );
  static QString systemInfo();
  static QString appInfo();
  static void osSignal( int sigNum );
  static void messageOutput(QtMsgType type, const char *msg);
  static void messageOutputAbort(QtMsgType type, const char *msg);
  void reportBug( QtbMessage::Severity sev, const char * msg );
  void reportBugNow( QtbMessage::Severity sev, const char * msg );
  QThread * mainThread() const { return _thread; }
  static void sleep( int ms );
  static QString currentThreadName();

  QtbMessage * messageHandler() const { return _messageHandler; }
  QString translatePath(const QString& file, const QString& title, const QString& fileType);

  void clearStreams();
  void initStream( QThread * thread = 0 );
  void setStream( QtbLogStream * stream, QThread * thread = 0 );
  void freezeStream( bool b ) const;
  void streamDestroyed( QtbLogStream * stream );
  void setStreamPrefix( const QString& prefix ) { currentStream().setPrefix( prefix ); }
  inline QtbLogStream& currentStream() const;
  QtbLogStream& stream( QThread * thread ) const;
  void redirectStream( QtbLogStream * stream );
  void restoreStream();

  QtbLogStream& operator<<( const char * val ) const {return currentStream() << val;}
  QtbLogStream& operator<<( const QString& val ) const {return currentStream() << val;}
  QtbLogStream& operator<<( double val ) const {return currentStream() << val;}
  QtbLogStream& operator<<( int val ) const {return currentStream() << val;}

  static void addGlobalObject(QtbGlobalObject * ptr);
  static void removeGlobalObject(QtbGlobalObject * ptr);

  static QString getStdin();
  static QStringList getFileList(int argc, char ** argv);
  static int userId();
  int terminalRows() const;
  int terminalCols() const;
protected:
  void setTerminalSize();
  void askForUserInterrupt();
  void setMessageHandler( QtbMessage * m );
  static QString constructorApplicationName();
  void destructorCleanUp();
  void initInternalDebugger(bool reportBugs);
  void initTranslations();
  void checkAtomicOperations();
  virtual void setHelp( QtbApplicationHelp * h );
private:
  void terminalSize() const;
  void init ( int & argc, char ** argv );

  static QtbCoreApplicationPrivate * _self;
  static QString _crashingThread;

  QMap<QThread *, QtbLogStream *> _originalStreams;
  QMap<QThread *, QtbLogStream *> _streams;
  QThread * _thread;
  QtbMessage * _messageHandler;
  QtbPathTranslator * _pathTranslator;
  QStringList _errors;
  QList<QtbGlobalObject *> _globalPointers;
  mutable int _terminalCols, _terminalRows;
};

inline QtbLogStream& QtbCoreApplicationPrivate::currentStream() const
{
  QMap<QThread *, QtbLogStream *>::ConstIterator it = _streams.find(QThread::currentThread());
  if (it!=_streams.end()) return *it.value(); else return *_streams[_thread];
}

inline QtbLogStream& endl(QtbCoreApplicationPrivate &s)
{
  QtbLogStream& ls = s.currentStream();
  ls << endl;
  return ls;
}

inline QtbLogStream& operator<<(QtbCoreApplicationPrivate &s, QTextStreamFunction f)
{
  QtbLogStream& ls=s.currentStream();
  (*f)(ls);
  return ls;
}

inline QtbLogStream& operator<<(QtbCoreApplicationPrivate &s, QTextStreamManipulator m)
{
  QtbLogStream& ls=s.currentStream();
  m.exec(ls);
  return ls;
}

// Convenient macro definitions with code localization
#ifdef QTB_MEMORY_CHECKER
#define isBlockDamaged(ptr) QtbCoreApplicationPrivate::isBlockDamaged(ptr, __FILE__, __LINE__)
#define printBlockInfo(ptr) QtbCoreApplicationPrivate::printBlockInfo(ptr,__FILE__, __LINE__)
#endif

#define qtbApp (*QtbCoreApplicationPrivate::instance())

#endif // QTB_COREAPPLICATIONPRIVATE_H
