/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-09-07
**
***************************************************************************/

#include "qtbcurve.h"

/*!
  \class QtbCurve qtbcurve.h
  \brief Template class for a curve
  It is a vector of "points" that may eventually contain extra data. The "point" type must at least have the
  following functions:
 
  \li double x() const;
  \li void setX(double);
  \li double y() const;
  \li void setY(double);
  \li void interpole( double valX, const pointType& p1, const pointType& p2 );
  \li void average(const pointType& p) const;
  \li void setValid(bool v);
  \li bool isValid() const;
*/

/*!
  \fn void QtbCurve::setFunction()
  Sort points by increasing x and remove duplicates if any, so that only one Y corresponds to each X.
*/

/*!
  \fn int QtbCurve<pointType>::indexAfter ( double valX )
  Returns the index for which x() is just greater or equal to valX.
  Returns indexes from 0 to count(). Uses dichotomy, best for functions with
  more than 10 values. Assumes that points are
  sorted (setFunction() called before).
*/

/*!
  \fn void QtbCurve::resample( int n, double min, double max, Qtb::SamplingOptions options, double valX, double valY )
  Resamples the curve with a constant adimensional distance between samples.
  \a valX and \a valY are representative values used to adimensionalize.
  Select identical values if X and Y have the same dimensions.
  If options has Function flag, \a n samples are taken starting from \a min (x) to \a max (x).
  Extrapolation is performed but resulting new samples are set as invalid.

  Accepted \a options are Function, InversedScale and LogScale.
*/

/*!
  \fn void QtbCurve::resample( int n, double min, double max, Qtb::SamplingOptions options )
  Resamples the curve with \a n samples from \a min to \a max

  Accepted \a options are InversedScale and LogScale, option Function is mandatory and you must call setFunction() before.
*/

/*!
  \fn void QtbCurve::resample( const QtbCurve<pointType>& xModel, Qtb::SamplingOptions options )
  Resample the function at x defined by xModel (.x() of xModel)

  Accepted \a options are Interpole, option Function is mandatory and you must call setFunction() before.
*/

/*!
  \fn void QtbCurve::resample( const QVector<double>& xModel, Qtb::SamplingOptions options )
  Resample the function at x defined by xModel (e.g. from another curve)

  Accepted \a options are Interpole, option Function is mandatory and you must call setFunction() before.
*/

/*!
  \fn void QtbCurve::cut( double min, double max, Qtb::SamplingOptions options )
  Cut function between \a min and \a max considering these limits expressed in a linear or
  inversed scale.

  Accepted \a options are LinearScale or InversedScale, option Function is mandatory and you must call setFunction() before.
*/

/*!
  \fn void QtbCurve::smooth( double min, double max, const QtbSmoothingParameters& param )
  Smooth function between two x values
*/

/*!
  \fn void QtbCurve::average( const QtbCurve<pointType>& curve )
  Average this function with \a curve and put the result in this curve.
*/

/*!
  \fn void QtbCurve::setValid( bool v )
  Set all point as valid or not valid according to \a v.
*/

/*!
  \fn void QtbCurve::removeInvalid()
  Remove invalid points
*/

/*!
  \fn void QtbCurve::swapXY()
  Swap x() and y(), if the curve is a function, mandatory to call setFunction() again, except if monotoneous.
*/

/*!
  \fn void QtbCurve::log10()
  Order not modified, just take _x=log10(_x)
*/

/*!
  \fn void QtbCurve::exp10()
  Order not modified, just take _x=10^_x
*/

/*!
  \fn QtbCurve::inverse( Qtb::SamplingOptions options )
  Order of _x values must be changed, inverse x values
*/

/*!
  \fn const pointType& QtbCurve::minimumY()
  Return the point at minimum Y value
*/

/*!
  \fn const pointType& QtbCurve::maximumY()
  Return the point at maximum Y value
*/

/*! \fn void QtbCurve::line(double x1, double y1, double x2, double y2)
  Construct a basic line curve between (\a x1 , \a y1 ) and ( \a x2 , \a y2 )

  The resulting curve is not necessarely a function.
*/

/*! \fn void QtbCurve::closestMax(int index)
  Return the local maxima close to index either on the left or right side
*/

/*!
  \fn int QtbCurve::max()
  Return the index of the maximum value. Return -1 if the curve is empty.
*/
