/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-09-07
**
***************************************************************************/

#ifndef QTB_CURVE_H
#define QTB_CURVE_H

#include <math.h>

#include <QVector>

#include "qtbglobal.h"
#include "qtbtools.h"
#include "qtbpoint2d.h"
#include "qtbsmoothingparameters.h"
#include "qtbtoolsdllexport.h"

namespace Qtb {
  enum SamplingOption { LinearScale=0, LogScale=1, InversedScale=2, Interpole=4, Function=8};
  Q_DECLARE_FLAGS(SamplingOptions, SamplingOption)
};
Q_DECLARE_OPERATORS_FOR_FLAGS(Qtb::SamplingOptions)

template <class pointType>
class QTBTOOLS_EXPORT QtbCurve : private QVector<pointType>
{
public:
  QtbCurve() {}
  QtbCurve( int n ) : QVector<pointType>(n) {}
  QtbCurve( const QtbCurve<pointType>& o ) : QVector<pointType>(o) {}
  QtbCurve( const QVector<pointType>& o ) : QVector<pointType>(o) {}

  void setFunction();

  // Vector functions
  bool operator==(const QtbCurve<pointType>& o) const { return QVector<pointType>::operator==(o); }
  int count() const { return QVector<pointType>::count(); }
  bool isEmpty() const { return QVector<pointType>::isEmpty(); }
  void resize( int n ) { QVector<pointType>::resize(n); }
  int indexAfter ( double valX ) const;
  pointType at( double x ) const;
  const pointType& at( int index ) const { return QVector<pointType>::at(index); }
  pointType& operator[]( int index ) { return QVector<pointType>::operator[](index); }
  const pointType& operator[]( int index ) const { return QVector<pointType>::at(index); }
  const pointType& first() const { return QVector<pointType>::first(); }
  const pointType& last() const { return QVector<pointType>::last(); }
  pointType& first() { return QVector<pointType>::first(); }
  pointType& last() { return QVector<pointType>::last(); }
  void remove( int index ) { QVector<pointType>::remove(index); }
  void insert( int index, const pointType& p ) { QVector<pointType>::insert( index, p ); }
  void insert( const pointType& p );
  void clear() { QVector<pointType>::clear(); }
  void append( const pointType& p ) { QVector<pointType>::append(p); }
  void append( const QtbCurve<pointType>& c ) { QVector<pointType>::operator+=(c); }
  typedef pointType * iterator;
  typedef const pointType * const_iterator;
  const_iterator begin() const { return QVector<pointType>::begin(); }
  const_iterator end() const { return QVector<pointType>::end(); }
  iterator begin() { return QVector<pointType>::begin(); }
  iterator end() { return QVector<pointType>::end(); }

  void xLog10();
  void xExp10();
  void xMultiply(double factor, Qtb::SamplingOptions options = Qtb::Function);
  void xInverse( Qtb::SamplingOptions options = Qtb::Function );

  void yLog10();
  void yExp10();
  void yMultiply( double factor );
  void yInverse();

  void line( double x1, double y1, double x2, double y2 );
  void resample( int n, double min, double max, Qtb::SamplingOptions options, double valX, double valY );
  void resample( int n, double min, double max, Qtb::SamplingOptions options );
  void resample( const QtbCurve<pointType>& xModel, Qtb::SamplingOptions options );
  void resample( const QVector<double>& xModel, Qtb::SamplingOptions options );
  void cut( double min, double max, Qtb::SamplingOptions options );
  void smooth( double min, double max, const QtbSmoothingParameters& param );
  void average( const QtbCurve<pointType>& o );
  QtbCurve<QtbPoint2D> leastSquare() const;
  void swapXY();
  QtbCurve<pointType> derivative() const;
  void setValid( bool v );
  void removeInvalid();
  int minimumX() const;
  int maximumX() const;
  int minimumY() const;
  int maximumY() const;
  int closestMax(int index);

  QVector<double> xVector() const;
  void toStream( QTextStream& s ) const;
private:
  inline static bool lessThan(const pointType& p1, const pointType& p2);
  double sumX() const;
  double sumY() const;
  double sumXY() const;
  double sumX2() const;
  double sumY2() const;
};

template <class pointType>
void QtbCurve<pointType>::setFunction()
{
  // modified Qtb::uniq()
  qSort( begin(), end(), lessThan );
  for ( int i = 1;i < count(); ) {
    if ( at( i - 1 ).x() == at( i ).x() )
      remove( i );
    else
      i++;
  }
}

template <class pointType>
inline bool QtbCurve<pointType>::lessThan(const pointType& p1, const pointType& p2)
{
  return p1.x()<p2.x();
}

template <class pointType>
int QtbCurve<pointType>::indexAfter ( double valX ) const
{
  QTB_TRACE;
  QTB_ASSERT( count()>1 );
  if ( valX <= at(0).x() ) return 0;
  int lasti=count()-1;
  if ( valX > at(lasti).x() ) return lasti+1;
  int i = 1;
  while ( i < count() ) i = i << 1;
  int step2=i >> 1;
  while (step2>0) {
    if ( i>lasti ) i-=step2;
    else if ( valX <= at(i).x() ) {
      if ( valX > at(i-1).x() ) break;
      i-=step2;
    } else i+=step2;
    step2=step2 >> 1;
  }
  return i;
}

template <class pointType>
pointType QtbCurve<pointType>::at( double x ) const
{
  if (count()<2) {
    if ( count() == 1 ) {
      return at(0);
    } else {
      return pointType();
    }
  } else {
    int i = indexAfter(x);
    if (i==0) i++;                // Force interpolation if outside range
    else if ( i==count() ) i--;
    pointType p;
    p.interpole( x, at( i-1 ), at( i ) );
    return p;
  }
}

template <class pointType>
void QtbCurve<pointType>::insert( const pointType& p )
{
  QTB_TRACE;
  if ( count() < 2 ) {
    if ( count() == 0 ) {
      append( p );
    } else if ( p.x() < at( 0 ).x() ) {
      prepend( p );
    } else if ( p.x() > at( 0 ).x() ) {
      append( p );
    } else {
      (*this)[0] = p;
    }
  } else {
    int i = indexAfter( p.x() );
    if ( i<count() && at(i).x() == p.x() ) {
      (*this)[i] = p;
    } else {
      if ( at( p.x() ).y() != p.y() ) {
        insert( i, p );
      }
    }
  }
}

template <class pointType>
void QtbCurve<pointType>::resample( int n, double min, double max, Qtb::SamplingOptions options, double valX, double valY )
{
  QTB_TRACE;
  // Transforms X scale according to options (log and inv)
  if ( options & Qtb::InversedScale ) {
    xInverse( options );
    valX = 1.0 / valX;
  }
  if ( options & Qtb::LogScale ) {
    xLog10();
    valX = ::log10( valX );
    if ( options & Qtb::Function ) {
      min = ::log10( min );
      max = ::log10( max );
    }
  }
  if ( options & Qtb::Function ) {
    if ( min < first().x() )
      min = first().x();
    if ( max > last().x() )
      max = last().x();
  }
  int currentN = count();

  // Calculate the total "distance" curve
  QtbCurve<QtbPoint2D> distances( currentN );
  distances[ 0 ].setX(at( 0 ).x());
  distances[ 0 ].setY(0.0);
  for ( int i = 1; i < currentN; i++ ) {
    const pointType& p1 = at( i - 1 );
    const pointType& p2 = at( i );
    distances[ i ].setX( p2.x() );
    double deltaX = ( p2.x() - p1.x() ) / valX;
    double deltaY = ( p2.y() - p1.y() ) / valY;
    distances[ i ].setY( distances[ i - 1 ].y() + sqrt( deltaX * deltaX + deltaY * deltaY ) );
  }
  distances.setFunction();

  // Calculate the distance of min and max, and constant step
  double distMin, distMax;
  if ( options & Qtb::Function ) {
    QtbPoint2D p;
    int i;
    i = distances.indexAfter( min );
    p.interpole( min, distances.at( i-1 ), distances.at( i ) );
    distMin = p.y();
    i = distances.indexAfter( max );
    p.interpole( max, distances.at( i-1 ), distances.at( i ) );
    distMax = p.y();
  } else {
    distMin = 0;
    distMax = distances[ currentN - 1 ].y();
  }
  double cstStep = ( distMax - distMin ) / ( currentN - 1 );

  // Map x as a function of distance, function monotoneous, then no need to setFunction()
  distances.swapXY();

  // Intepolate x for constant step distance
  QVector<double> x( n );
  double dist = distMin;
  QtbPoint2D p;
  for ( int i = 0; i < n; i++, dist += cstStep ) {
    int iDist = distances.indexAfter( dist  );
    p.interpole( dist, distances.at( iDist-1 ), distances.at( iDist ) );
    x[ i ] = p.y();
  }
  resample( x, options );

  // Re-Transforms axis according options (log and inv)
  if ( options & Qtb::LogScale )
    xExp10();
  if ( options & Qtb::InversedScale )
    xInverse( options);
}

template <class pointType>
void QtbCurve<pointType>::resample( int n, double min, double max, Qtb::SamplingOptions options )
{
  QTB_TRACE;
  QTB_ASSERT( options & Qtb::Function );
  // Transforms X scale according to options (log and inv)
  if ( options & Qtb::InversedScale ) {
    xInverse( options );
  }
  if ( options & Qtb::LogScale ) {
    xLog10();
    min = ::log10( min );
    max = ::log10( max );
  }
  // Calculate the constant step
  double cstStep = ( max - min ) / ( n - 1 );
  QVector<double> x( n );
  for ( int i = 0;i < n;i++ )
    x[ i ] = min + (double)i*cstStep;
  resample( x, options );

  // Re-Transforms axis according options (log and inv)
  if ( options & Qtb::LogScale )
    xExp10();
  if ( options & Qtb::InversedScale )
    xInverse( options);
}

template <class pointType>
void QtbCurve<pointType>::resample( const QtbCurve<pointType>& xModel, Qtb::SamplingOptions options )
{
  QTB_TRACE;
  QTB_ASSERT( options & Qtb::Function );
  QVector<double> x( xModel.count() );
  for ( int i = 0;i < xModel.count();i++ )
    x[ i ] = xModel.at(i).x();
  resample( x, options );
}

template <class pointType>
void QtbCurve<pointType>::resample( const QVector<double>& xModel, Qtb::SamplingOptions options )
{
  QTB_TRACE;
  QTB_ASSERT( options & Qtb::Function );
  int nCurve = count();
  int nModel = xModel.count();
  if(nCurve<2 || nModel<2) return;
  setFunction(); // Make sure values are sorted
  QtbCurve<pointType> interpolated( nModel );
  // Insert values inside xModel
  int iCurve = 1;
  double min = at( 0 ).x()*(1-1e-15);
  double max = at( nCurve-1 ).x()*(1+1e-15);
  for (int i =0; i<nModel;i++) {
    double x = xModel.at( i );
    while (iCurve<nCurve-1 && at( iCurve ).x()<x) iCurve++;
    interpolated[i].interpole( x, at(iCurve-1), at(iCurve) );
    if ( x<min || x>max || !at(iCurve-1).isValid() || !at(iCurve).isValid()) interpolated[i].setValid(false);
  }
  // Insert all values outside xModel
  if(nModel>1) {
    min = xModel.at( 0 )*(1-1e-15);
    max = xModel.at( nModel-1 )*(1+1e-15);
    for(int i = 0;i<nCurve;i++) {
      const pointType& p = at(i);
      if( p.x()<min || max<p.x() ) {
        interpolated.append(p);
      }
    }
  }
  interpolated.setFunction();
  *this = interpolated;
}

template <class pointType>
void QtbCurve<pointType>::cut( double min, double max, Qtb::SamplingOptions options )
{
  QTB_TRACE;
  QTB_ASSERT( options & Qtb::Function );
  // Assert that count()>1 is made by indexAfter()
  // Transforms X scale according to options (log and inv)
  if ( options & Qtb::InversedScale ) {
    xInverse( options );
  }
  if ( min < first().x() )
    min = first().x();
  if ( max > last().x() )
    max = last().x();
  // Find the first point
  int ix = indexAfter( min );
  if ( ix>0 ) {
    if( fabs( min - at( ix-1 ).x() ) < 1e-15) {
      ix--;
    } else if ( (options & Qtb::Interpole) && fabs( min - at( ix ).x() ) > 1e-15 ) {
      ix--;
      (*this)[ ix ].interpole( min, at( ix ), at( ix+1 ) );
    }
  }
  // Remove everything before
  for (int i = 0; i<ix; i++) remove(0);
  // Find the last point
  ix = indexAfter( max );
  if ( ix < count() ) {
    if (fabs( max - at( ix ).x() ) < 1e-15) {
      ix++;
    } else if ( (options & Qtb::Interpole) && fabs( max - at( ix-1 ).x() ) > 1e-15 ) {
      (*this)[ ix ].interpole( max, at( ix-1 ), at( ix ) );
      ix++;
    }
  }
  // Remove everything after
  for (int i = count()-1; i>=ix; i--) remove( i );
  // Re-Transforms axis according options (log and inv)
  if ( options & Qtb::InversedScale )
    xInverse( options);
}

template <class pointType>
void QtbCurve<pointType>::smooth( double min, double max, const QtbSmoothingParameters& param )
{
  QTB_TRACE;
  
}

template <class pointType>
void QtbCurve<pointType>::average( const QtbCurve<pointType>& o  )
{
  QTB_TRACE;
  if(o.count()<2) return;
  if(count()==0) {
    *this = o;
    return;
  }
  // Set the same sampling for both curves
  QVector<double> x = xVector();
  x+=o.xVector();
  Qtb::uniq(x);
  resample( x, Qtb::Function );
  QtbCurve<pointType> oResamp = o;
  oResamp.resample( x, Qtb::Function );
  // Take the average of the two curves
  for (int i = x.count()-1; i>=0; i-- ) {
    operator[](i).average( oResamp.at(i) );
  }
  setFunction();
}

template <class pointType>
void QtbCurve<pointType>::setValid( bool v )
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->setValid(v);
  }
}

template <class pointType>
void QtbCurve<pointType>::removeInvalid()
{
  QTB_TRACE;
  for ( int i = count() - 1; i>=0; i-- ) {
    if ( !at(i).isValid() ) {
      remove(i);
    }
  }
}

template <class pointType>
void QtbCurve<pointType>::swapXY()
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    double tmp = it->x();
    it->setX( it->y() );
    it->setY( tmp );
  }
}

template <class pointType>
QtbCurve<pointType> QtbCurve<pointType>::derivative() const
{
  QTB_TRACE;
  int n = count()-1;
  QTB_ASSERT(n>1);
  QtbCurve<pointType> d(n-1);
  for ( int i = 1; i<n; i++ ) {
    double v = ( at(i).y()-at(i-1).y() ) / ( at(i).x()-at(i-1).x() );
    v+=( at(i+1).y()-at(i).y() ) / ( at(i+1).x()-at(i).x() );
    v*=2.0;
    pointType& p = d[i-1];
    p.setX( at(i).x() );
    p.setY( v );
  }
  return d;
}

template <class pointType>
void QtbCurve<pointType>::xLog10()
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    if (it->x()>0.0) it->setX( ::log10( it->x() ) );
  }
}

template <class pointType>
void QtbCurve<pointType>::xExp10()
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->setX( pow(10, it->x() ) );
  }
}

template <class pointType>
void QtbCurve<pointType>::yLog10()
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    if (it->y()>0.0) it->setY( ::log10( it->y() ) );
  }
}

template <class pointType>
void QtbCurve<pointType>::yExp10()
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->setY( pow(10, it->y() ) );
  }
}

template <class pointType>
void QtbCurve<pointType>::xMultiply(double factor, Qtb::SamplingOptions options )
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->setX( it->x()*factor );
  }
  if (factor<0.0 && options & Qtb::Function ) setFunction();
}

template <class pointType>
void QtbCurve<pointType>::yMultiply(double factor)
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->setY( it->y()*factor );
  }
}

template <class pointType>
void QtbCurve<pointType>::xInverse( Qtb::SamplingOptions options )
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->setX( 1.0 / it->x() );
  }
  if (options & Qtb::Function ) setFunction();
}

template <class pointType>
void QtbCurve<pointType>::yInverse()
{
  QTB_TRACE;
  iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->setY( 1.0 / it->y() );
  }
}

template <class pointType>
void QtbCurve<pointType>::line(double x1, double y1, double x2, double y2)
{
  QTB_TRACE;
  resize( 2 );
  (*this)[0].setX( x1 );
  (*this)[0].setY( y1 );
  (*this)[1].setX( x2 );
  (*this)[1].setY( y2 );
}

template <class pointType>
QVector<double> QtbCurve<pointType>::xVector() const
{
  int n = count();
  QVector<double> x(n);
  for ( int i = 0; i<n; i++ ) {
    x[i]=at(i).x();
  }
  return x;
}

template <class pointType>
int QtbCurve<pointType>::minimumX() const
{
  int n = count();
  const pointType * p = &first();
  int iMin = 0;
  for ( int i = 1; i<n; i++ ) {
    if ( at(i).x() < p->x() ) {
      p=&at(i);
      iMin = i;
    }
  }
  return iMin;
}

template <class pointType>
int QtbCurve<pointType>::maximumX() const
{
  int n = count();
  const pointType * p = &first();
  int iMax = 0;
  for ( int i = 1; i<n; i++ ) {
    if ( at(i).x() > p->x() ) {
      p=&at(i);
      iMax = i;
    }
  }
  return iMax;
}

template <class pointType>
int QtbCurve<pointType>::minimumY() const
{
  int n = count();
  const pointType * p = &first();
  int iMin = 0;
  for ( int i = 1; i<n; i++ ) {
    if ( at(i).y() < p->y() ) {
      p=&at(i);
      iMin = i;
    }
  }
  return iMin;
}

template <class pointType>
int QtbCurve<pointType>::maximumY() const
{
  int n = count();
  const pointType * p = &first();
  int iMax = 0;
  for ( int i = 1; i<n; i++ ) {
    if ( at(i).y() > p->y() ) {
      p=&at(i);
      iMax = i;
    }
  }
  return iMax;
}

template <class pointType>
int QtbCurve<pointType>::closestMax(int index)
{
  QTB_TRACE;
  int n=count();
  int di;

  if (index==0) {
    di=1;
    index=1;
  }
  else if (index>=n) {
    index=n;
    di=-1;
  } else if (at(index).y()>at(index-1).y()) di=1;
  else if (at(index).y()<at(index-1).y()) di=-1;
  else return index;

  if (di>0) {
    while (index<n) {
      if(at(index).y()<at(index-1).y()) {
        index--;
        break;
      }
      index+=di;
    }
    if (index==n) index--;
  } else {
    while (index>0) {
      if(at(index).y()>at(index-1).y()) break;
      index+=di;
    }
  }
  return index;
}

template <class pointType>
void QtbCurve<pointType>::toStream( QTextStream& s ) const
{
  QTB_TRACE;
  for ( const_iterator it=begin(); it!=end(); ++it ) {
    s << it->toString(16) << "\n";
  }
  s << flush;
}

/*!
  Sum of all x (used in least square method)
*/
template <class pointType>
double QtbCurve<pointType>::sumX() const
{
  QTB_TRACE;
  double sum=0;
  const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x();
  return sum;
}

/*!
  Sum of all y (used in least square method)
*/
template <class pointType>
double QtbCurve<pointType>::sumY() const
{
  QTB_TRACE;
  double sum=0;
  const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->y();
  return sum;
}

/*!
  Sum of all x*y (used in least square method)
*/
template <class pointType>
double QtbCurve<pointType>::sumXY() const
{
  QTB_TRACE;
  double sum=0;
  const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x()*it->y();
  return sum;
}

/*!
  Sum of all x*x (used in least square method)
*/
template <class pointType>
double QtbCurve<pointType>::sumX2() const
{
  QTB_TRACE;
  double sum=0;
  const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x()*it->x();
  return sum;
}

/*!
  Sum of all y*y (used in least square method)
*/
template <class pointType>
double QtbCurve<pointType>::sumY2() const
{
  QTB_TRACE;
  double sum=0;
  const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->y()*it->y();
  return sum;
}

template <class pointType>
QtbCurve<QtbPoint2D> QtbCurve<pointType>::leastSquare() const
{
  QTB_TRACE;
  int n=count();
  if (!n) return QtbCurve<QtbPoint2D>();
  QtbCurve<QtbPoint2D> line;
  if (n>1) {
    double sx=sumX();
    double sy=sumY();
    double sxy=sumXY();
    double sx2=sumX2();
    double denom=n*sx2-sx*sx;
    if (denom!=0) {
      double A=(n*sxy-sy*sx)/denom;
      double C=(sy*sx2-sxy*sx)/denom;
      line.append( QtbPoint2D( first().x(), A*first().x()+C) );
      line.append( QtbPoint2D( last().x(), A*last().x()+C) );
    } else {
      line.append( QtbPoint2D( first().x(), first().y() ) );
      line.append( QtbPoint2D( last().x(), last().y()) );
    }
  } else {
    line.append( QtbPoint2D( first().x(), first().y() ) );
    line.append( QtbPoint2D( last().x(), last().y()) );
  }
  return line;
}

#endif // QTB_CURVE_H
