/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-04-10
**
***************************************************************************/

#ifndef QTB_EXPRESSIONACTION_H
#define QTB_EXPRESSIONACTION_H

#include <QVariant>
#include <QVector>

#include "qtbxmlclass.h"
#include "qtbexpressionstorage.h"
#include "qtbexpressioncontext.h"
#include "qtbglobal.h"
#include "qtbtoolsdllexport.h"

class QtbExpressionContext;

class QTBTOOLS_EXPORT QtbExpressionAction : public QtbXMLClass
{
public:
  enum Type {NoOperator, Assign, Sum, SumAssign, Subtract, SubtractAssign, Mult, MultAssign, FloatDiv,
             FloatDivAssign, IntMod, IntDiv, Coma, OpenBracket, CloseBracket, EndAction, Function, Equal,
             NotEqual, GreaterThan, GreaterOrEqual, LessThan, LessOrEqual, OpenSquareBracket, CloseSquareBracket,
             VariableArray };
  QtbExpressionAction() {_parent=0;}
  virtual ~QtbExpressionAction() {}

  virtual const QString& xml_tagName() const { return xmlExpressionActionTag;}
  static const QString xmlExpressionActionTag;

  virtual bool isReadOnly() const {return true;}
  virtual QVariant value() const=0;
  virtual void setValue(const QVariant& ) {}

  virtual int priority() const {return 4;}
  virtual QtbExpressionAction * parent() const {return _parent;}
  void setParent(QtbExpressionAction * p) {_parent=p;}
  QtbExpressionAction * superParent();
  virtual QtbExpressionAction * close( QtbExpressionAction::Type t );

  bool addAction(QtbExpressionAction * a);
  virtual bool addArgument(QtbExpressionAction * a) = 0;
  virtual void replaceArgument(QtbExpressionAction * oldA, QtbExpressionAction * newA) = 0;
protected:
  static QVariant sum(const QVariant& op1, const QVariant& op2);
  static QVariant subtract(const QVariant& op1, const QVariant& op2);
  static QVariant mult(const QVariant& op1, const QVariant& op2);
  static QVariant floatDiv(const QVariant& op1, const QVariant& op2);
  static int compare(const QVariant& op1, const QVariant& op2, bool& ok);

  QtbExpressionAction * _parent;
protected:
  //virtual QtbXMLMember xml_member( XML_MEMBER_ARGS );
};

class QTBTOOLS_EXPORT QtbExpressionVariable : public QtbExpressionAction
{
public:
  QtbExpressionVariable(QString name, QtbExpressionContext * c);
  ~QtbExpressionVariable() { delete _index; }

  virtual const QString& xml_tagName() const { return xmlExpressionVariableTag;}
  static const QString xmlExpressionVariableTag;

  virtual QtbExpressionAction * parent() const {if (_open) return 0; else return _parent;}
  virtual bool isReadOnly() const;
  virtual QVariant value() const;
  virtual void setValue(const QVariant& val);
  virtual bool addArgument(QtbExpressionAction * a);
  virtual void replaceArgument(QtbExpressionAction *, QtbExpressionAction *);
  virtual QtbExpressionAction * close( Type t);
protected:
  virtual void xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const;
  virtual void xml_writeChildren( XML_WRITECHILDREN_ARGS ) const;
  //virtual bool xml_setProperty( XML_SETPROPERTY_ARGS );
private:
  QtbExpressionAction * _index;
  QtbExpressionContext * _context;
  bool _open;
  int _varIndex;
};

class QTBTOOLS_EXPORT QtbExpressionValue : public QtbExpressionAction
{
public:
  QtbExpressionValue(QVariant val) {_val=val;}

  virtual const QString& xml_tagName() const { return xmlExpressionValueTag;}
  static const QString xmlExpressionValueTag;

  virtual QVariant value() const {return _val;}
  virtual bool addArgument(QtbExpressionAction *) {return false;}
  virtual void replaceArgument(QtbExpressionAction *, QtbExpressionAction *) { QTB_ASSERT(false); }
private:
  QVariant _val;
protected:
  virtual void xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const;
  //virtual bool xml_setProperty( XML_SETPROPERTY_ARGS );
};

class QTBTOOLS_EXPORT QtbExpressionBracket : public QtbExpressionAction
{
public:
  QtbExpressionBracket() {_open=true; _arg=0;}
  ~QtbExpressionBracket() {delete _arg;}

  virtual const QString& xml_tagName() const { return xmlExpressionBracketTag;}
  static const QString xmlExpressionBracketTag;

  virtual QtbExpressionAction * parent() const {if (_open) return 0; else return _parent;}
  virtual bool isReadOnly() const {return _arg->isReadOnly();}
  virtual QVariant value() const {return _arg ? _arg->value() : QVariant();}
  virtual QtbExpressionAction * close( Type t );
  virtual bool addArgument(QtbExpressionAction *);
  virtual void replaceArgument(QtbExpressionAction * oldA, QtbExpressionAction * newA);
private:
  QtbExpressionAction * _arg;
  bool _open;
protected:
  virtual void xml_writeChildren( XML_WRITECHILDREN_ARGS ) const;
  virtual void xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const;
  //virtual bool xml_setProperty( XML_SETPROPERTY_ARGS );
};

class QTBTOOLS_EXPORT QtbExpressionOperator : public QtbExpressionAction
{
public:
  QtbExpressionOperator(Type type) {_type=type;_operand1=0;_operand2=0;}
  virtual ~QtbExpressionOperator() {delete _operand1; delete _operand2;}

  virtual const QString& xml_tagName() const { return xmlExpressionOperatorTag;}
  static const QString xmlExpressionOperatorTag;

  virtual QVariant value() const;
  virtual int priority() const {
    switch (_type) {
    case Mult:
    case FloatDiv:
    case IntMod:
    case IntDiv:
      return 3;
    case Sum:
    case Subtract:
      return 2;
    default:
      return 1;
    }
  }
  virtual bool addArgument(QtbExpressionAction * a);
  virtual void replaceArgument(QtbExpressionAction * oldA, QtbExpressionAction * newA);
private:
  Type _type;
  QtbExpressionAction * _operand1;
  QtbExpressionAction * _operand2;
protected:
  virtual void xml_writeChildren( XML_WRITECHILDREN_ARGS ) const;
  virtual void xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const;
  //virtual bool xml_setProperty( XML_SETPROPERTY_ARGS );
};

class QTBTOOLS_EXPORT QtbExpressionFunction : public QtbExpressionAction
{
public:
  QtbExpressionFunction(QString name, QtbExpressionContext * context ) {_name=name; _open=true; _currentArg=0; _context = context; }
  virtual ~QtbExpressionFunction() {qDeleteAll(_args);}

  virtual const QString& xml_tagName() const { return xmlExpressionFunctionTag;}
  static const QString xmlExpressionFunctionTag;

  virtual QtbExpressionAction * parent() const {if (_open) return 0; else return _parent;}
  virtual QVariant value() const { return _context->functionValue( _name, _args ); }
  virtual bool addArgument(QtbExpressionAction * a);
  virtual void replaceArgument(QtbExpressionAction * oldA, QtbExpressionAction * newA);
  virtual QtbExpressionAction * close( Type t );
private:
  bool _open;
  QString _name;
  QtbExpressionAction * _currentArg;
  QVector<QtbExpressionAction *> _args;
  QtbExpressionContext * _context;
protected:
  virtual void xml_writeChildren( XML_WRITECHILDREN_ARGS ) const;
  virtual void xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const;
  //virtual bool xml_setProperty( XML_SETPROPERTY_ARGS );
};

#endif // QTB_EXPRESSIONACTION_H
