/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-04-10
**
***************************************************************************/

#include "qtbexpressionparser.h"
#include "qtbcoreapplication.h"
#include "qtbmemorychecker.h"

/*!
  Parse a text code and return a list of QtbExpressionAction strutures which are able to effectively execute the expressions.
*/
QtbExpressionActions QtbExpressionParser::parse(QString code, QtbExpressionContext * context)
{
  QTB_TRACE;

  QtbExpressionActions rootActions (context);
  _row=0;
  _ptr=code.data();
  _beginRowPtr=_ptr;
  while (_ptr->unicode()!=0x0000) {
    QtbExpressionBracket * a=new QtbExpressionBracket;
    if (parse(a, context)) {
      rootActions << a;
      //a->xml_saveFile("/home/mwathele/equ.xml");
    } else {
      delete a;
      break;
    }
  }
  return rootActions;
}

bool QtbExpressionParser::parse(QtbExpressionAction * parent, QtbExpressionContext * context)
{
  QTB_TRACE;
  ArgumentType argType;
  QtbExpressionAction::Type opType;
  QtbExpressionAction * a=0;
  bool doubleOperator = false;
  QString arg;

  while (true) {

    // doubleOperator is true when 2 operators can be consecutive: e.g. )+ or ]*
    if (!doubleOperator) {
      skipBlanks();
      const QChar * argBegin=_ptr;
      if (!skipArgument(argType)) return false;
      arg=QString(argBegin, _ptr-argBegin);
    } else {
      doubleOperator=false;
      argType=NoArgument;
    }

    skipBlanks();
    if (!skipOperator(opType)) return false;

    switch (argType) {
    case NoArgument:
      a=0;
      break;
    case Numeric:
      a=new QtbExpressionValue(arg.toDouble());
      break;
    case String: // A string may be interrupted by comments, hence arg may contain more than just the string
      // Get the cleaned and recomposed string obtained during the last call to getArgument()
      a=new QtbExpressionValue(_lastString);
      break;
    case AlphaNumeric:
      if (opType==QtbExpressionAction::OpenSquareBracket) {
        a=0;
        opType=QtbExpressionAction::VariableArray;
      } else if (opType==QtbExpressionAction::OpenBracket && context->isValidFunction(arg)) {
        a=0;
        opType=QtbExpressionAction::Function;
      } else { // Define new variable if it does not exist yet
        a=new QtbExpressionVariable(arg, context);
      }
      break;
    }

    if (a) {
      if (!parent->addAction(a)) {
        qtbApp << tr("Not a valid argument, syntax error %1").arg(bufferLocation()) << endl;
        return false;
      }
    }

    switch (opType) {
    case QtbExpressionAction::Function:
      a = new QtbExpressionFunction(arg, context);
      break;
    case QtbExpressionAction::OpenBracket:
      if (argType==NoArgument) a=new QtbExpressionBracket; else a=0;
      break;
    case QtbExpressionAction::CloseBracket:
      parent=parent->close( QtbExpressionAction::CloseBracket );
      if (!parent) {
        qtbApp << tr("Unmatched bracket %1").arg(bufferLocation()) << endl;
        return false;
      }
      a=0;
      doubleOperator=true;
      break;
    case QtbExpressionAction::VariableArray:
      a = new QtbExpressionVariable(arg, context);
      break;
    case QtbExpressionAction::OpenSquareBracket:
      qtbApp << tr("Square brackets allowed only for arrays: %1").arg(bufferLocation()) << endl;
      return false;
    case QtbExpressionAction::CloseSquareBracket:
      parent=parent->close( QtbExpressionAction::CloseSquareBracket );
      if (!parent) {
        qtbApp << tr("Unmatched square bracket %1").arg(bufferLocation()) << endl;
        return false;
      }
      a=0;
      doubleOperator=true;
      break;
    case QtbExpressionAction::Coma:
      parent=parent->close( QtbExpressionAction::Coma );
      if (!parent) {
        qtbApp << tr("No coma expected here %1").arg(bufferLocation()) << endl;
        return false;
      }
      a=0;
      break;
    case QtbExpressionAction::EndAction:
      return true;
    case QtbExpressionAction::NoOperator:
      return false;
    default:
      a=new QtbExpressionOperator(opType);
      break;
    }

    if (a) {
      if (!parent->addAction(a)) {
        qtbApp << tr("Not a valid operator, syntax error %1").arg(bufferLocation()) << endl;
        return false;
      }
      parent=a;
    }
  }
}

bool QtbExpressionParser::skipArgument( ArgumentType& type)
{
  QTB_TRACE;
  enum ScientificPos {Mantisse, ExponentSign, Exponent};
  ScientificPos scientificPos=Mantisse;
  bool decimal=false;
  bool hasFrontSign=false;
  type=NoArgument;
  while ( true ) {
    switch ( _ptr->unicode() ) {
    case 0x000A:   // '\n'
      _beginRowPtr=_ptr+1;
      _row++;
    case 0x0020:   // ' '
    case 0x0009:   // '\t'
    case 0x000D:   // '\r'
    case 0x0000:   // '\0'
    case 0x002A:   // '*'
    case 0x002F:   // '/'
    case 0x003D:   // '='
    case 0x0021:   // '!'
    case 0x003C:   // '<'
    case 0x003E:   // '>'
    case 0x003B:   // ';'
    case 0x002C:   // ','
    case 0x0028:   // '('
    case 0x005B:   // '['
    case 0x005D:   // ']'
    case 0x0029:   // ')'
      return true;
    case 0x004D:   // 'MID'
      if (_ptr[1].unicode()==0x0049 && _ptr[2].unicode()==0x0044 && _ptr[3].isSpace()) {
        return true;
      }
      break;
    case 0x0044:  // 'DIV'
      if (_ptr[1].unicode()==0x0049 && _ptr[2].unicode()==0x0056 && _ptr[3].isSpace()) {
        return true;
      }
      break;
    case 0x0022:   // '"'
      if (type==NoArgument) {
        _lastString="";
        QtbStringSection section;
        while (_ptr->unicode()==0x0022) {
          _ptr++;
          section.set(_ptr,0);
          if (!skipString()) return false;
          section.setEnd(_ptr);
          section.appendTo(_lastString);
          _ptr++;
          skipBlanks();
        }
        type=String;
        return true;
      } else return badCharacterError();
    case 0x005F:   // '_'
      if (type==Numeric) {
        if (!hasFrontSign) type=AlphaNumeric; else return badCharacterError();
      }
      break;
    case 0x002B:   // '+'
    case 0x002D:   // '-'
      if (type==Numeric && scientificPos==ExponentSign) {
        scientificPos=Exponent;
      } else if (type==NoArgument) {
        hasFrontSign=true; // generate error if something else than Numeric
        type=Numeric;
      } else return true;
      break;
    case 0x0065:   // 'e'
    case 0x0045:   // 'E'
      if (type==Numeric && scientificPos==Mantisse) {
        scientificPos=ExponentSign;
      } else if (!hasFrontSign) type=AlphaNumeric;
      else return badCharacterError();
      break;
    case 0x002E:   // '.'
      if (type==Numeric && !decimal) decimal=true; else return badCharacterError();
      break;
    default:
      if ((_ptr->unicode()>=0x0041 && _ptr->unicode()<=0x005A) ||
          (_ptr->unicode()>=0x0061 && _ptr->unicode()<=0x007A)) {
        switch (type) {
        case Numeric:
          if (!hasFrontSign) type=AlphaNumeric; else return badCharacterError();
          break;
        case NoArgument:
          type=AlphaNumeric;
          break;
        default:
          break;
        }
      } else if (_ptr->unicode()>=0x0030 && _ptr->unicode()<=0x0039) {
        if (type==NoArgument) type=Numeric;
      } else return badCharacterError();
      break;
    }
    _ptr++;
  }
}

bool QtbExpressionParser::skipOperator( QtbExpressionAction::Type& type)
{
  QTB_TRACE;
  switch ( _ptr->unicode() ) {
  case 0x0000:   // '\0'
    type=QtbExpressionAction::EndAction;
    return true;
  case 0x003B:   // ';'
    type=QtbExpressionAction::EndAction;
    break;
  case 0x002A:   // '*'
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::MultAssign;
    } else type=QtbExpressionAction::Mult;
    break;
  case 0x002F:   // '/'
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::FloatDivAssign;
    } else type=QtbExpressionAction::FloatDiv;
    break;
  case 0x003D:   // '='
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::Equal;
    } else type=QtbExpressionAction::Assign;
    break;
  case 0x0021:   // '!'
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::NotEqual;
    } else {
      type=QtbExpressionAction::NoOperator;
      return badCharacterError();
    }
    break;
  case 0x003C:   // '<'
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::LessOrEqual;
    } else type=QtbExpressionAction::LessThan;
    break;
  case 0x003E:   // '>'
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::GreaterOrEqual;
    } else type=QtbExpressionAction::GreaterThan;
    break;
  case 0x002C:   // ','
    type=QtbExpressionAction::Coma;
    break;
  case 0x0028:   // '('
    type=QtbExpressionAction::OpenBracket;
    break;
  case 0x0029:   // ')'
    type=QtbExpressionAction::CloseBracket;
    break;
  case 0x005B:   // '['
    type=QtbExpressionAction::OpenSquareBracket;
    break;
  case 0x005D:   // ']'
    type=QtbExpressionAction::CloseSquareBracket;
    break;
  case 0x002B:   // '+'
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::SumAssign;
    } else type=QtbExpressionAction::Sum;
    break;
  case 0x002D:   // '-'
    if (_ptr[1].unicode()==0x003D) {
      _ptr++;
      type=QtbExpressionAction::SubtractAssign;
    } else type=QtbExpressionAction::Subtract;
    break;
  case 0x004D:   // "MOD"
    if (_ptr[1].unicode()==0x004F && _ptr[2].unicode()==0x0044 && _ptr[3].isSpace()) {
      _ptr+=2;
      type=QtbExpressionAction::IntMod;
    } else {
      type=QtbExpressionAction::NoOperator;
      return badCharacterError();
    }
    break;
  case 0x0044:   // "DIV"
    if (_ptr[1].unicode()==0x0049 && _ptr[2].unicode()==0x0056 && _ptr[3].isSpace()) {
      _ptr+=2;
      type=QtbExpressionAction::IntDiv;
    } else {
      type=QtbExpressionAction::NoOperator;
      return badCharacterError();
    }
    break;
  default:
    type=QtbExpressionAction::NoOperator;
    return badCharacterError();
  }
  _ptr++;
  return true;
}

bool QtbExpressionParser::badCharacterError()
{
  QTB_TRACE;
  qtbApp << tr("Unrecognized character %1").arg(bufferLocation()) << endl;
  return false;
}

QString QtbExpressionParser::bufferLocation()
{
  QTB_TRACE;
  return tr("at line %1 and column %2").arg(_row+1).arg(_ptr-_beginRowPtr+1);
}

bool QtbExpressionParser::skipString()
{
  QTB_TRACE;
  while ( true ) {
    switch ( _ptr->unicode() ) {
    case 0x0000:
      qtbApp << tr( "Unmatched \" before reaching the end of buffer" ) << endl;
      return false;
    case 0x005C:   // '\\'
      _ptr++;
      break;
    case 0x000A:   // '\n'
      _beginRowPtr=_ptr+1;
      _row++;
      break;
    case 0x0022:   // '"'
      return true;
    }
    _ptr++;
  }
}

void QtbExpressionParser::skipBlanks( )
{
  QTB_TRACE;
  while ( true ) {
    switch ( _ptr->unicode() ) {
    case 0x0000:   // '\0'
      return;
    case 0x000A:   // '\n'
      _beginRowPtr=_ptr+1;
      _row++;
    case 0x0020:   // ' '
    case 0x0009:   // '\t'
    case 0x000D:   // '\r'
      break;
    case 0x002F:   // '/'
      switch ((_ptr+1)->unicode()) {
      case 0x002F: // line comment
        _ptr+=2;
        skipLineComment();
        if ( _ptr->unicode() == 0x0000 ) return;
        break;
      case 0x002A: // multi line comment
        _ptr+=2;
        skipMultiLineComment();
        if ( _ptr->unicode() == 0x0000 ) return;
        break;
      default:
        return; // else operator '/'
      }
      break;
    default:
      return ;
    }
    _ptr++;
  }
}

/*!
  Skip a C++ like comment over one lines
*/
void QtbExpressionParser::skipLineComment( )
{
  QTB_TRACE;
  while (true) {
    switch (_ptr->unicode()) {
    case 0x000A:   // '\n'
      _beginRowPtr=_ptr+1;
      _row++;
    case 0x0000:
      return;
    default:
      break;
    }
    _ptr++;
  }
}

/*!
  Skip a C++ like comment over several lines
*/
void QtbExpressionParser::skipMultiLineComment( )
{
  QTB_TRACE;
  while (true) {
    switch (_ptr->unicode()) {
    case 0x000A:   // '\n'
      _beginRowPtr=_ptr+1;
      _row++;
      break;
    case 0x0000:
      return;
    case 0x002A:
      if ((_ptr+1)->unicode()==0x002F) {
        _ptr++;
        return;
      }
    default:
      break;
    }
    _ptr++;
  }
}
