/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-08-23
**
***************************************************************************/

#include <math.h>

#include "qtbgridsearch.h"
#include "qtbglobal.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbGridSearch qtbgridsearch.h
  \brief Search a grid for maxima (global or local)
*/

/*!
*/
QtbGridSearch::QtbGridSearch()
{
  QTB_TRACE;
  _precisionX=0.001;
  _precisionY=0.001;
  _nx = 0;
  _ny = 0;
  _x = 0;
  _y = 0;
}

/*!
  Description of destructor still missing
*/
QtbGridSearch::~QtbGridSearch()
{
  QTB_TRACE;
  delete [] _x;
  delete [] _y;
}

void QtbGridSearch::setGrid(double minX, double maxX, double dX,
                            double minY, double maxY, double dY)
{
  QTB_TRACE;
  QTB_ASSERT(_function);

  int nx=(int)floor((maxX-minX)/dX+0.5)+1;
  int ny=(int)floor((maxY-minY)/dY+0.5)+1;
  if (nx != _nx) {
    delete [] _x;
    _x=new double [nx];
    _nx = nx;
  }
  if (ny != _ny) {
    delete [] _y;
    _y=new double [ny];
    _ny = ny;
  }
  for (int i=0;i<_nx;i++) _x[i]=minX+dX*(double)i;
  for (int i=0;i<_ny;i++) _y[i]=minY+dY*(double)i;

  _function->initGrid(_nx * _ny);
  int k = 0;
  for (int iy=0;iy<_ny;iy++) {
    for (int ix=0;ix<_nx;ix++, k++) {
      _function->initGrid(_x[ix], _y[iy], k);
    }
  }
}

/*!
  Search all grid for its maximum and refine it
*/
double QtbGridSearch::globalMax()
{
  QTB_TRACE;
  int maxix=0, maxiy=0;
  double val,maxVal=-1e99;
  int k = 0;
  for (int iy=0;iy<_ny;iy++) {
    for (int ix=0;ix<_nx;ix++, k++) {
      val=_function->value(_x[ix], _y[iy], k);
      if (val>maxVal) {
        maxVal=val;
        maxix=ix;
        maxiy=iy;
      }
    }
  }
  // Refine search by recursion and re-regridding
  if (_ny==1) {
    _pos.setY(_y[0]);
    return refineMax(lowLimit(maxix, _x),highLimit(maxix, _x, _nx));
  }
  else return refineMax(lowLimit(maxix, _x), highLimit(maxix, _x, _nx),
                        lowLimit(maxiy, _y), highLimit(maxiy, _y, _ny));
}

QtbFunction2MaximaIterator QtbGridSearch::localMax(int nMax, double absThres, double relThres)
{
  QTB_TRACE;
  double * map=new double [_nx*_ny];
  int k=0;
  int iy;
  for (iy=0;iy<_ny;iy++) {
    for (int ix=0;ix<_nx;ix++,k++) {
      map[k]=_function->value(_x[ix], _y[iy], k);
    }
  }
  _localMaxima.clear();
  int nxm=_nx-1;
  int nym=_ny-1;
  k=_nx+1;
  for (iy=1;iy<nym;iy++) {
    for (int ix=1;ix<nxm;ix++,k++) {
      if (map[k-1]>0 && map[k]>map[k-1] &&
          map[k+1]>0 && map[k]>map[k+1] &&
          map[k-_nx-1]>0 && map[k]>map[k-_nx-1] &&
          map[k-_nx]>0 && map[k]>map[k-_nx] &&
          map[k-_nx+1]>0 && map[k]>map[k-_nx+1] &&
          map[k+_nx-1]>0 && map[k]>map[k+_nx-1] &&
          map[k+_nx]>0 && map[k]>map[k+_nx] &&
          map[k+_nx+1]>0 && map[k]>map[k+_nx+1]) {
        double val = refineMax(_x[ix-1],_x[ix+1], _y[iy-1],_y[iy+1]);
        _localMaxima.append(QtbFunction2SearchMaximum(val, _pos));
      }
    }
    k+=2;
  }
  delete [] map;
  return QtbFunction2Search::localMax(nMax, absThres, relThres);
}

inline void QtbGridSearch::checkForMax(double x, double y, double& maxVal, double& maxX, double& maxY)
{
  QTB_TRACE;
  double val=_function->value(x, y);
  if (val>maxVal) {
    maxVal=val;
    maxX=x;
    maxY=y;
  }
}

inline void QtbGridSearch::checkForMax(double x, double& maxVal, double& maxX)
{
  QTB_TRACE;
  double val=_function->value(x, _pos.y());
  if (val>maxVal) {
    maxVal=val;
    maxX=x;
  }
}

/*!
  Refines the maximum of the grid in an area defined by x and y limits.
  return value of the refined maximum
*/
double QtbGridSearch::refineMax(double minX, double maxX, double minY, double maxY)
{
  QTB_TRACE;
  double x=(minX+maxX)*0.5;
  double dx=(maxX-minX)*0.5;
  double y=(minY+maxY)*0.5;
  double dy=(maxY-minY)*0.5;

  // get value for (x, y) supposed to be the maxVal
  double maxVal=_function->value(x, y);
  double absPrecisionX, absPrecisionY;
  if (_relative) {
    absPrecisionX=fabs(_precisionX*x);
    absPrecisionY=fabs(_precisionY*y);
  } else {
    absPrecisionX=_precisionX;
    absPrecisionY=_precisionY;
  }

  while (dx>absPrecisionX || dy>absPrecisionY) {

    dx*=0.5;
    dy*=0.5;

    double x1=x-dx;
    double x2=x+dx;

    double y1=y-dy;
    double y2=y+dy;

    checkForMax(x1, y1, maxVal, x, y);
    checkForMax(x2, y1, maxVal, x, y);
    checkForMax(x1, y2, maxVal, x, y);
    checkForMax(x2, y2, maxVal, x, y);

    // In case the first cell center was null, correct the precision for the current estimate
    if (absPrecisionX==0.0) absPrecisionX=fabs(_precisionX*x);
    if (absPrecisionY==0.0) absPrecisionY=fabs(_precisionY*y);
  }
  _pos.setX(x);
  _pos.setY(y);
  return maxVal;
}

/*!
  Refines the maximum of the grid in an area defined by x limits.
  return value of the refined maximum
*/
double QtbGridSearch::refineMax(double minX, double maxX)
{
  QTB_TRACE;
  double x=(minX+maxX)*0.5;
  double dx=(maxX-minX)*0.5;

  // get value for (x) supposed to be the maxVal
  double maxVal=_function->value(x, _pos.y());
  double absPrecisionX;
  if (_relative) {
    absPrecisionX=fabs(_precisionX*x);
  } else {
    absPrecisionX=_precisionX;
  }

  while (dx>absPrecisionX) {
    dx*=0.5;

    double x1=x-dx;
    double x2=x+dx;

    checkForMax(x1, maxVal, x);
    checkForMax(x2, maxVal, x);
  }
  _pos.setX(x);
  return maxVal;
}
