/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2008-07-15
**
***************************************************************************/

#ifndef QTB_IRREGULARGRID2DDATA_H
#define QTB_IRREGULARGRID2DDATA_H

#include <QSharedData>

#include "qtbpoint2d.h"
#include "qtbcurve.h"
#include "qtbsegment2d.h"

namespace Qtb {
  enum AxisType { XAxis, YAxis, ZAxis };
  Q_DECLARE_FLAGS(AxisTypes, AxisType)
};
Q_DECLARE_OPERATORS_FOR_FLAGS(Qtb::AxisTypes)

class QtbGridSearch;

class QTBTOOLS_EXPORT QtbIrregularGrid2DData : public QSharedData
{
  QTB_TRANSLATIONS( "QtbIrregularGrid2DData" );
public:
  QtbIrregularGrid2DData();
  QtbIrregularGrid2DData( int nx, int ny );
  QtbIrregularGrid2DData( const QtbIrregularGrid2DData& o );
  ~QtbIrregularGrid2DData();

  void operator=( const QtbIrregularGrid2DData& o );
  void operator=( const QtbGridSearch& o );
  void operator+=( const QtbIrregularGrid2DData& o );

  int nx() const { return _nx; }
  int ny() const { return _ny; }
  void clear();
  void init( int nx, int ny );
  void init( double value = 0.0 );
  inline void init( int nx, int ny, double value);
  bool isLike( QtbIrregularGrid2DData& o ) const;
  // Function to set _x and _y values
  void setLinear(Qtb::AxisType axis,double min, double max);
  void setLog(Qtb::AxisType axis,double min, double max);
  // Resample the grid at x defined by xModel (e.g. from another curve)
  void resample( Qtb::AxisType axis, const QVector<double>& xModel, Qtb::SamplingOptions options );
  void resample( Qtb::AxisType axis, int n, double min, double max, Qtb::SamplingOptions options );
  // Cut function between two x values
  void cut ( Qtb::AxisType axis, double min, double max, Qtb::SamplingOptions options );
  void inverse( Qtb::AxisType axis);
  void log10( Qtb::AxisType axis );
  void log10();
  void pow10( Qtb::AxisType axis );
  void pow10();
  void multiplyCoordinates( Qtb::AxisType axis, double fac );
  void multiplyValues(Qtb::AxisType along, int index, double fac);
  void multiplyValues(double fac);
  const double * axisData( Qtb::AxisType axis ) const { int n; return getAxis(axis,n); }
  // Function to acces the grid
  inline double value(int ix, int iy) const;
  inline void setValue(int ix, int iy, double val);

  double x(int ix) const { return _x[ix]; }
  double y(int iy) const { return _y[iy]; }
  void setX(int ix, double val) { _x[ix]=val; }
  void setY(int iy, double val) { _y[iy]=val; }
  QVector<double> x() const;
  QVector<double> y() const;
  inline double * valuePointer(int ix, int iy);
  inline const double * valuePointer(int ix, int iy) const;
  const double * xPointer() const { return _x; }
  double * xPointer() { return _x; }
  const double * yPointer() const { return _y; }
  double * yPointer() { return _y; }
  inline QtbPoint2D coordinates(int ix, int iy) const;
  int indexOfX( double x ) const { return indexOf( _x, _nx, _n2x, x ); }
  int indexOfY( double y ) const { return indexOf( _y, _ny, _n2y, y ); }

  double minimumValue() const;
  double maximumValue() const;
  double minimumValue(int& ix, int& iy) const;
  double maximumValue(int& ix, int& iy) const;
  double minimumAxis( Qtb::AxisType axis ) const;
  double maximumAxis( Qtb::AxisType axis ) const;
  double sum( Qtb::AxisType along, int index ) const;
  double mean( Qtb::AxisType along, int index ) const;
  double median( Qtb::AxisType along, int index ) const;
  double mode( Qtb::AxisType along, int index ) const;
  double variance( Qtb::AxisType along, int index ) const;
  double variance( Qtb::AxisType along, int index, double average ) const;
  double normalize( Qtb::AxisType along, int index );
  void smooth( Qtb::AxisType along, int index );
  double smoothInvalidCells( Qtb::AxisType along, int index,
                             double invalidValue, double sigma2,
                             double RVal, bool modify );
  void printValues( Qtb::AxisType along, int index );
  double * iterator( Qtb::AxisType along, int index, int &di, int& n ) const;
  /* Get values of a row or a column of the grid as a QtbCurve object
      Return a row if along Qtb::X_Axis, a column if along Qtb::YAxis **/
  QtbCurve<QtbPoint2D> crossSection( Qtb::AxisType axis, int ixy ) const;
  QtbCurve<QtbPoint2D> crossSection( QtbSegment2D seg, double deltaX = 0 ) const;
  QtbCurve<QtbPoint2D> crossSection( const QtbCurve<QtbPoint2D>& path, double deltaX = 0 ) const;
  QtbCurve<QtbPoint2D> integralCrossSection( Qtb::AxisType axis,int index ) const;
  template <class CurveClass, class PointClass> CurveClass followMaximumX() const;
  template <class CurveClass, class PointClass> CurveClass followMaximumX( int ixMin, int ixMax, int iyMax ) const;
  template <class CurveClass, class PointClass> void followMaximumX( CurveClass& curve, double min, double max ) const;
  template <class CurveClass, class PointClass> CurveClass followMaximumY() const;
  template <class CurveClass, class PointClass> CurveClass followMaximumY( int iyMin, int iyMax, int ixMax ) const;
  template <class CurveClass, class PointClass> void followMaximumY( CurveClass& curve, double min, double max ) const;
  // Geometries of cells
  double width(int ix) const { return widthOrHeight(_x, _nx,ix); }
  double height(int iy) const { return widthOrHeight(_y, _ny,iy); }
  double left(int ix) const { return leftOrBottom(_x, _nx,ix); }
  double right(int ix) const { return rightOrTop(_x, _nx,ix); }
  double bottom(int iy) const { return leftOrBottom(_y, _ny,iy); }
  double top(int iy) const { return rightOrTop(_y, _ny,iy); }
private:
  // Init number of cells rounded to power of 2
  void initLookup();
  int indexOf( double * axis, int n, int n2, double xory ) const;
  // Convert an axis type into its pointer and number of values
  double * getAxis( Qtb::AxisType axis, int& n ) const;
  // General geometries of cells
  double leftOrBottom(double * axis, int n, int ival) const;
  double rightOrTop(double * axis, int n, int ival) const;
  double widthOrHeight(double * axis, int n, int ival) const;

  // Values at each grid node
  double *_values;
  // Vector of x coordinades of nodes
  double * _x;
  // Vector of y coordinades of nodes
  double * _y;
  // Number of cells in Y direction
  int _ny;
  // Number of cells in X direction
  int _nx;
  // Number of cells in Y direction rounded to next power of 2, used for fast lookup
  int _n2y;
  // Number of cells in X direction rounded to next power of 2, used for fast lookup
  int _n2x;
};

inline double QtbIrregularGrid2DData::value(int ix, int iy) const
{
  return(_values[iy*_nx+ix]);
}

inline void QtbIrregularGrid2DData::setValue(int ix, int iy, double val)
{
  _values[iy*_nx+ix]=val;
}

inline const double * QtbIrregularGrid2DData::valuePointer(int ix, int iy) const
{
  return _values+(iy*_nx+ix);
}

inline double * QtbIrregularGrid2DData::valuePointer(int ix, int iy)
{
  return _values+(iy*_nx+ix);
}

inline QtbPoint2D QtbIrregularGrid2DData::coordinates(int ix, int iy) const
{
  return QtbPoint2D(_x[ix],_y[iy]);
}

inline void QtbIrregularGrid2DData::init( int nx, int ny, double value )
{
  init( nx, ny );
  init( value );
}

template <class CurveClass, class PointClass>
CurveClass QtbIrregularGrid2DData::followMaximumX() const
{
  QTB_TRACE;
  CurveClass curve;
  if (_nx==0 || _ny==0) return curve;
  // Search the maximum value of grid
  int ix, iy;
  maximumValue( ix, iy );
  curve.append(followMaximumX<CurveClass,PointClass>(ix, _nx-1, iy));
  curve.append(followMaximumX<CurveClass,PointClass>(ix, 0, iy));
  curve.setFunction();
  return curve;
}

template <class CurveClass, class PointClass>
CurveClass QtbIrregularGrid2DData::followMaximumX( int ixMin, int ixMax, int iyMax) const
{
  QTB_TRACE;
  CurveClass curve;
  curve.append( PointClass( x( ixMin ), y( iyMax ) ) );
  int dx = ixMin<ixMax ? 1 : -1;
  ixMax+=dx;
  for (int i = ixMin;i != ixMax;i+=dx ) {
    QtbCurve<QtbPoint2D> cs=crossSection( Qtb::YAxis, i );
    iyMax = cs.closestMax( iyMax );
    curve.append( PointClass( x( i ), y( iyMax ) ) );
    curve.last().setValid(true);
  }
  return curve;
}

template <class CurveClass, class PointClass>
void QtbIrregularGrid2DData::followMaximumX( CurveClass& curve, double min, double max ) const
{
  QTB_TRACE;
  QTB_ASSERT( nx()>1 ); // Needed by QtbCurve::indexAfter()
  QTB_ASSERT( ny()>1 ); // Needed by QtbCurve::indexAfter()
  curve.resample( x(), Qtb::Function );
  for (int i = 0;i<_nx;i++ ) {
    if (min<=x(i) && x(i)<=max) {
      int curveIndex = curve.indexAfter(x(i));
      PointClass& p = curve[curveIndex];
      QtbCurve<QtbPoint2D> cs=crossSection( Qtb::YAxis, i );
      int iMax = cs.closestMax( cs.indexAfter(p.y()) );
      p.setY(y(iMax));
      p.setValid(true);
    }
  }
}

template <class CurveClass, class PointClass> CurveClass QtbIrregularGrid2DData::followMaximumY() const
{
  CurveClass curve;
  if (_nx==0 || _ny==0) return curve;
  // Search the maximum value of grid
  int ix, iy;
  maximumValue( ix, iy );
  curve.append(followMaximumY<CurveClass,PointClass>(iy, _ny-1, ix));
  curve.append(followMaximumY<CurveClass,PointClass>(iy, 0, ix));
  curve.setFunction();
  return curve;
}

template <class CurveClass, class PointClass> CurveClass QtbIrregularGrid2DData::followMaximumY( int iyMin, int iyMax, int ixMax) const
{
  CurveClass curve;
  curve.append( PointClass( y( iyMin ), x( ixMax ) ) );
  int dy = iyMin<iyMax ? 1 : -1;
  iyMax+=dy;
  for (int i = iyMin;i != iyMax;i+=dy ) {
    QtbCurve<QtbPoint2D> cs=crossSection( Qtb::XAxis, i );
    ixMax = cs.closestMax( ixMax );
    curve.append( PointClass( y( i ), x( ixMax ) ) );
    curve.last().setValid(true);
  }
  return curve;
}

template <class CurveClass, class PointClass> void QtbIrregularGrid2DData::followMaximumY( CurveClass& curve, double min, double max ) const
{
  QTB_ASSERT( nx()>1 ); // Needed by QtbCurve::indexAfter()
  QTB_ASSERT( ny()>1 ); // Needed by QtbCurve::indexAfter()
  curve.resample( y(), Qtb::Function );
  for (int i = 0;i<_ny;i++ ) {
    if (min<=y(i) && y(i)<=max) {
      int curveIndex = curve.indexAfter(y(i));
      PointClass& p = curve[curveIndex];
      QtbCurve<QtbPoint2D> cs=crossSection( Qtb::XAxis, i );
      int iMax = cs.closestMax( cs.indexAfter(p.y()) );
      p.setY(x(iMax));
      p.setValid(true);
    }
  }
}

#endif // QTB_IRREGULARGRID2DDATA_H
