/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-05-17
**
***************************************************************************/

#ifndef QTB_Matrix_H
#define QTB_Matrix_H

#include <QVector>

#include "qtbmatrixdata.h"
#include "qtbpoint.h"
#include "qtbcomplex.h"

template <typename T> class QtbMatrixIterator;
template <typename T> class QtbMutableMatrixIterator;
template <typename T> class QtbMatrixRowIterator;
template <typename T> class QtbMutableMatrixRowIterator;
template <typename T> class QtbMatrixColumnIterator;
template <typename T> class QtbMutableMatrixColumnIterator;

template <typename T> class QTBTOOLS_EXPORT QtbMatrix
{
public:
  QtbMatrix() { _d = new QtbMatrixData<T>; }
  QtbMatrix( int ndim ) { _d = new QtbMatrixData<T>; _d->resize(ndim); }
  QtbMatrix( int nrow, int ncol ) { _d = new QtbMatrixData<T>; _d->resize(nrow, ncol); }

  QtbMatrix<T> operator=( const QtbMatrix<T>& m ) { *_d = *m._d; return *this; }
  bool operator==( const QtbMatrix<T>& m ) const { return _d->operator==( *m._d ); }
  QtbMatrix<T> operator+( const QtbMatrix<T>& m ) const { return _d->operator+( *m._d ); }
  void operator+=( const QtbMatrix<T>& m );
  QtbMatrix<T> operator*( const QtbMatrix<T>& m ) const { return _d->operator*( *m._d ); }
  void operator*=( const QtbMatrix<T>& m );
  void zero() { _d->zero(); }
  void identity() { _d->identity(); }
  T& at( int row, int col ) { return _d->at(row,col); }
  const T& at( int row, int col ) const { return _d->at(row,col); }
  QtbMatrix<T> transposed() const { return _d->transposed(); }
  void transpose() { *this = transposed(); }

  int columnCount() const { return _d->columnCount(); }
  int rowCount() const { return _d->rowCount(); }
  void resize( int ndim ) { _d->resize(ndim); }
  void resize( int nrow, int ncol ) { _d->resize(nrow, ncol); }
protected:
  friend class QtbMatrixData<T>;
  friend class QtbMatrixIterator<T>;
  friend class QtbMutableMatrixIterator<T>;
  friend class QtbMatrixRowIterator<T>;
  friend class QtbMutableMatrixRowIterator<T>;
  friend class QtbMatrixColumnIterator<T>;
  friend class QtbMutableMatrixColumnIterator<T>;

  QSharedDataPointer< QtbMatrixData<T> > _d;
};


template <typename T> QtbMatrix<T> QtbMatrixData<T>::operator+( const QtbMatrixData<T>& m ) const
{
  QTB_ASSERT(_nrow == m._nrow && _ncol == m._ncol);
  QtbMatrix<T> r( _nrow, _ncol );
  T * rValues = r._d->values();
  for (int i=_nrow * _ncol-1; i>=0;i--) rValues[i] = _values[i] + m._values[i];
  return r;
}

template <typename T> QtbMatrix<T> QtbMatrixData<T>::operator*( const QtbMatrixData<T>& m ) const
{
  QTB_ASSERT( _ncol == m._nrow );
  QtbMatrix<T> r( _nrow, m._ncol );
  r.zero();
  if (QtbMatrixMultiply::isValid(_nrow, _ncol, m._ncol)) {
    const QtbMatrixMultiply& multiplier = *QtbMatrixMultiply::begin(_nrow, _ncol, m._ncol);
    T * rValues = r._d->_values;
    const T * thisValues = _values;
    const T * mValues = m._values;
    int n = multiplier.indexCount();
    for(int i=0;i<n;i++) {
      const QtbMatrixMultiply::IndexMap& index = multiplier.indexMap(i);
      rValues[index.resultIndex()]+=thisValues[index.factor1Index()]*mValues[index.factor2Index()];
    }
    QtbMatrixMultiply::end(&multiplier);
  } else {
    QTB_ASSERT(false);
    // TODO big matrix multiplication: call to LAPACK?
  }
  return r;
}

template <typename T> QtbMatrix<T> QtbMatrixData<T>::transposed() const
{
  QtbMatrix<T> t(_ncol, _nrow);
  T * pt = t._d->values();
  const T * prow;
  int j;
  for( int i = 0; i < _nrow; i++ ) {
    prow = _values + i * _ncol;
    for ( j=0; j < _ncol; j++ ) {
      (*pt++) = prow[j];
    }
  }
  return t;
}

template <typename T> inline void QtbMatrix<T>::operator*=( const QtbMatrix<T>& m )
{
  *this = _d->operator*( *m._d );
}

template <typename T> inline void QtbMatrix<T>::operator+=( const QtbMatrix<T>& m )
{
  *this = _d->operator+( *m._d );
}

#endif
