/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2008-07-15
**
***************************************************************************/

#ifndef QTB_MATRIXITERATOR_H
#define QTB_MATRIXITERATOR_H

#include "qtbtoolsdllexport.h"
#include "qtbmatrix.h"

template <typename T> class QTBTOOLS_EXPORT QtbMatrixIterator
{
public:
  inline QtbMatrixIterator( const QtbMatrix<T>& m );

  inline const T * next() { _cur += _offset; return _cur; }
  inline const T * previous() { _cur -= _offset; return _cur; }
  inline const T& operator*() { QTB_ASSERT(_cur >= _start && _cur <= _end); return *_cur; }
  bool hasNext() const { return _cur < _end; }
  bool hasPrevious() const { return _cur > _start; }
  void toBack() { return _cur = _end; }
  void toFront() { return _cur = _start; }
protected:
  QtbMatrixIterator() {}

  const T * _cur, * _start, * _end;
  int _offset;
};

template <typename T> class QTBTOOLS_EXPORT QtbMutableMatrixIterator
{
public:
  inline QtbMutableMatrixIterator( QtbMatrix<T>& m );

  inline T * next() { _cur += _offset; return _cur; }
  inline T * previous() { _cur -= _offset; return _cur; }
  inline T& operator*() { QTB_ASSERT(_cur >= _start && _cur <= _end); return *_cur; }
  bool hasNext() const { return _cur < _end; }
  bool hasPrevious() const { return _cur > _start; }
  void toBack() { return _cur = _end; }
  void toFront() { return _cur = _start; }
protected:
  QtbMutableMatrixIterator() {}

  T * _cur, * _start, * _end;
  int _offset;
};

template <typename T> class QTBTOOLS_EXPORT QtbMatrixRowIterator : public QtbMatrixIterator<T>
{
public:
  inline QtbMatrixRowIterator( const QtbMatrix<T>& m, int iRow );
};

template <typename T> class QTBTOOLS_EXPORT QtbMatrixColumnIterator : public QtbMatrixIterator<T>
{
public:
  inline QtbMatrixColumnIterator( const QtbMatrix<T>& m, int iCol );
};

template <typename T>
inline QtbMatrixIterator<T>::QtbMatrixIterator( const QtbMatrix<T>& m )
{
  QTB_TRACE;
  _offset = 1;
  _start = m._d->values();
  _end = _start + m.rowCount()*m.columnCount() -1;
  _cur = _start - _offset;
}

template <typename T>
inline QtbMutableMatrixIterator<T>::QtbMutableMatrixIterator( QtbMatrix<T>& m )
{
  QTB_TRACE;
  _offset = 1;
  _start = m._d->values();
  _end = _start + m.rowCount()*m.columnCount() -1;
  _cur = _start - _offset;
}

template <typename T>
inline QtbMatrixRowIterator<T>::QtbMatrixRowIterator( const QtbMatrix<T>& m, int iRow )
   : QtbMatrixIterator<T>()
{
  QTB_TRACE;
  QTB_ASSERT( iRow >= 0 && iRow < m.rowCount() );
  QtbMatrixIterator<T>::_offset = m.rowCount();
  QtbMatrixIterator<T>::_start = m._d->values() + iRow;
  QtbMatrixIterator<T>::_end = QtbMatrixIterator<T>::_start +  m.rowCount() * (m.columnCount() - 1);
  QtbMatrixIterator<T>::_cur = QtbMatrixIterator<T>::_start - m.rowCount();
}

template <typename T>
inline QtbMatrixColumnIterator<T>::QtbMatrixColumnIterator( const QtbMatrix<T>& m, int iCol )
   : QtbMatrixIterator<T>()
{
  QTB_TRACE;
  QTB_ASSERT( iCol >= 0 && iCol < m.columnCount() );
  QtbMatrixIterator<T>::_offset = 1;
  QtbMatrixIterator<T>::_start = m._d->values() + m.rowCount() * iCol;
  QtbMatrixIterator<T>::_end = QtbMatrixIterator<T>::_start + m.rowCount() - 1;
  QtbMatrixIterator<T>::_cur = QtbMatrixIterator<T>::_start - 1;
}

template <typename T> class QTBTOOLS_EXPORT QtbMutableMatrixRowIterator : public QtbMatrixIterator<T>
{
public:
  inline QtbMutableMatrixRowIterator( const QtbMatrix<T>& m, int iRow );
};

template <typename T>
inline QtbMutableMatrixRowIterator<T>::QtbMutableMatrixRowIterator( const QtbMatrix<T>& m, int iRow )
   : QtbMutableMatrixIterator<T>()
{
  QTB_TRACE;
  QTB_ASSERT( iRow >= 0 && iRow < m.rowCount() );
  QtbMutableMatrixIterator<T>::_offset = m.rowCount();
  QtbMutableMatrixIterator<T>::_start = m._data.data() + iRow;
  QtbMutableMatrixIterator<T>::_end = QtbMutableMatrixIterator<T>::_start +  m.rowCount() * (m.columnCount() - 1);
  QtbMutableMatrixIterator<T>::_cur = QtbMutableMatrixIterator<T>::_start - m.rowCount();
}

template <typename T> class QTBTOOLS_EXPORT QtbMutableMatrixColumnIterator : public QtbMatrixIterator<T>
{
public:
  inline QtbMutableMatrixColumnIterator( const QtbMatrix<T>& m, int iCol );
};

template <typename T>
inline QtbMutableMatrixColumnIterator<T>::QtbMutableMatrixColumnIterator( const QtbMatrix<T>& m, int iCol )
   : QtbMutableMatrixIterator<T>()
{
  QTB_TRACE;
  QTB_ASSERT( iCol >= 0 && iCol < m.columnCount() );
  QtbMutableMatrixIterator<T>::_offset = 1;
  QtbMutableMatrixIterator<T>::_start = m._data.data() + m.rowCount() * iCol;
  QtbMutableMatrixIterator<T>::_end = QtbMutableMatrixIterator<T>::_start + m.rowCount() - 1;
  QtbMutableMatrixIterator<T>::_cur = QtbMutableMatrixIterator<T>::_start - 1;
}

#endif // QTB_MATRIXITERATOR_H
