/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2005-10-04
**
***************************************************************************/

#include <QHash>
#include <QProcess>
#include <QTime>

#include "qtbmessage.h"
#include "qtbfile.h"
#include "qtbcoreapplication.h"

/*!
  \class QtbMessage qtbmessage.h
  \brief Show messages to user
  If not graphical user interface is running this class send messages to and get information from the user through
  the console. In graphical user interface, a QtbGuiMessage class is also available (in library qtbguitools), which
  transparently (for the developer) send the message to dialog boxes. Those two classes have eaxctly the same API.
  So always use this class for send messages and automatically the messages will be displayed in the appropriate way.

  The console is considered is its largest sense here. In fact this is a log stream which can be redirected anywhere:
  true unix terminal, inside a general log view, to a specific log (attached to a thread), or simply to trash.
*/

/*!
  Init base context for answers to messages
*/
QtbMessage::QtbMessage()
{
  QTB_TRACE;
  beginContext( false );
}

/*!
  Define a new context for automatic answers, stack this context

  \sa endContext(), QtbMessageContext
*/
void QtbMessage::beginContext( bool quiet )
{
  QTB_TRACE;
  QMutexLocker ml(&_contextMutex);
  _contexts.push( Context( quiet ) );
}

/*!
  Delete current context for automatic answers, and restore old one

  \sa beginContext(), QtbMessageContext
*/
void QtbMessage::endContext()
{
  QTB_TRACE;
  QMutexLocker ml(&_contextMutex);
  _contexts.pop();
}

void QtbMessage::setAutoAnswer( uint msgId, Answer a )
{
  QTB_TRACE;
  QMutexLocker ml(&_contextMutex);
  _contexts.top().answers.insert(msgId, a);
}

/*!
  Print auto answer to log console and return the automatic answer for message with id \a msgId
*/
QtbMessage::Answer QtbMessage::autoAnswer( uint msgId,
                                           Severity sev,
                                           QString caption,
                                           const QString & text,
                                           QString answer0,
                                           QString answer1,
                                           QString answer2 )
{
  qtbApp << severityString(sev) << QTime::currentTime().toString("hh:mm:ss") << " -- " << caption << "----" << endl;
  qtbApp << text << "\n ==> ";
  _contextMutex.lock();
  Answer a=_contexts.top().answers[msgId];
  _contextMutex.unlock();
  switch (a) {
  case Answer0:
    qtbApp << answer0 << endl;
    break;
  case Answer1:
    qtbApp << answer1 << endl;
    break;
  case Answer2:
    qtbApp << answer2 << endl;
    break;
  }
  return a;
}

/*!
  Generate a unique message id from position in source file.

  Use C macros __FILE__ and __LINE__ as argument of this function.
*/
uint QtbMessage::messageId(const char * fileName, int lineNumber)
{
  QTB_TRACE;
  return qHash(QByteArray(fileName))-lineNumber;
}

/*!
  Translate severity \a sev into a localized string.
*/
QString QtbMessage::severityString(Severity sev)
{
  QTB_TRACE;
  switch (sev) {
  case Information:
    return translate("QtbMessage","-----INFO----- ");
  case Warning:
    return translate("QtbMessage","----WARNING--- ");
  case Critical:
    return translate("QtbMessage","----ERROR----- ");
  case Fatal:
    return translate("QtbMessage","--FATAL ERROR- ");
  case Question:
    return translate("QtbMessage","---QUESTION--- ");
  default:
    return translate("QtbMessage","-------------- ");
  }
}

/*!
  General message function with interactions with console
*/
QtbMessage::Answer QtbMessage::message( uint msgId,
                                        Severity sev,
                                        QString caption,
                                        const QString & text,
                                        QString answer0,
                                        QString answer1,
                                        QString answer2,
                                        bool againOption )
{
  if ( againOption && hasAutoAnswer(msgId) ) {
    return autoAnswer(msgId, sev, caption, text, answer0, answer1, answer2 );
  }

  qtbApp << severityString(sev) << caption << "----" << endl;
  qtbApp << text << endl;
  if ( answer1.isEmpty() && answer2.isEmpty() ) return Answer0;
  qtbApp << "  1. " << answer0 << tr( " <-- default" ) << endl;
  qtbApp << "  2. " << answer1 << endl;
  if ( !answer2.isEmpty() ) qtbApp << "  3. " << answer2 << endl;
  qtbApp << tr( "? " ) << flush;

  QString rep = QtbFile::readLine(stdin, true);
  rep = rep.toLower();
  Answer a;
  if ( rep == "1" ) a=Answer0;
  else if ( rep == "2" ) a=Answer1;
  else if ( rep == "3" ) a=Answer2;
  else if ( answer0.toLower().startsWith(rep) ) a=Answer0;
  else if ( answer1.toLower().startsWith(rep) ) a=Answer1;
  else if ( answer2.toLower().startsWith(rep) ) a=Answer2;
  else a=Answer0;

  if (againOption) {
    qtbApp << tr( "Show this message again? [y]/n " ) << flush;
    QString rep = QtbFile::readLine(stdin, true);
    if ( rep.toLower().startsWith("n") ) setAutoAnswer(msgId, a);
  }
  return a;
}

/*!
  Show an information message.
  \a msgId has to be the macro MSG_ID.
*/
QtbMessage::Answer QtbMessage::information( uint msgId, const QString & caption,
                                            const QString & text,
                                            QString answer0,
                                            QString answer1,
                                            QString answer2,
                                            bool againOption)
{
  return qtbApp.messageHandler()->message( msgId, Information, caption, text,
                                           answer0, answer1, answer2, againOption );
}

/*!
  Show a warning message.
  \a msgId has to be the macro MSG_ID.
*/
QtbMessage::Answer QtbMessage::warning( uint msgId, const QString & caption,
                                        const QString & text,
                                        QString answer0,
                                        QString answer1,
                                        QString answer2,
                                        bool againOption)
{
  return qtbApp.messageHandler()->message( msgId, Warning, caption, text,
                                           answer0, answer1, answer2, againOption );
}

/*!
  Show a critical message.
  \a msgId has to be the macro MSG_ID.
*/
QtbMessage::Answer QtbMessage::critical( uint msgId, const QString & caption,
                                         const QString & text,
                                         QString answer0,
                                         QString answer1,
                                         QString answer2,
                                         bool againOption)
{
  return qtbApp.messageHandler()->message( msgId, Critical, caption, text,
                                           answer0, answer1, answer2, againOption );
}

/*!
  Ask a question.
  \a msgId has to be the macro MSG_ID.
*/
QtbMessage::Answer QtbMessage::question( uint msgId, const QString & caption,
                                         const QString & text,
                                         QString answer0,
                                         QString answer1,
                                         QString answer2,
                                         bool againOption )
{
  return qtbApp.messageHandler()->message( msgId, Question, caption, text,
                                           answer0, answer1, answer2, againOption );
}

/*!
  get an existing file name
*/
QString QtbMessage::getOpenFileName( const QString & caption,
                                     const QString & filter,
                                     const QString & directory )
{
  QTB_TRACE;
  return qtbApp.messageHandler()->getOpenFileNameInternal( caption, filter, directory );
}

/*!
  get a new file name
*/
QString QtbMessage::getSaveFileName( const QString & caption,
                                     const QString & filter,
                                     const QString & selection)
{
  QTB_TRACE;
  return qtbApp.messageHandler()->getSaveFileNameInternal( caption, filter, selection );
}

/*!
  get existing file names
*/
QStringList QtbMessage::getOpenFileNames( const QString & caption,
                                          const QString & filter,
                                          const QString & directory )
{
  QTB_TRACE;
  return qtbApp.messageHandler()->getOpenFileNamesInternal( caption, filter, directory );
}

/*!
  get an existing directory
*/
QString QtbMessage::getExistingDirectory( const QString & caption,
                                          const QString & dir)
{
  QTB_TRACE;
  return qtbApp.messageHandler()->getExistingDirectoryInternal( caption, dir );
}

/*!
  get an existing file name
*/
QString QtbMessage::getOpenFileNameInternal( const QString & caption,
                                     const QString & filter,
                                     const QString & directory )
{
  QTB_TRACE;
  qtbApp << "---- " << caption << " ----" << endl;
  qtbApp << tr("  Filter: ") << filter << endl;
  qtbApp << tr("  Current directory: ") << directory << endl;
  qtbApp << tr("File to open: ") << flush;
  return QtbFile::readLine(stdin, true);
}

/*!
  get a new file name
*/
QString QtbMessage::getSaveFileNameInternal( const QString & caption,
                                     const QString & filter,
                                     const QString & selection)
{
  QTB_TRACE;
  qtbApp << "---- " << caption << " ----" << endl;
  qtbApp << tr("  Filter: ") << filter << endl;
  qtbApp << tr("  Selection: ") << selection << endl;
  qtbApp << tr("File to save: ") << flush;
  return QtbFile::readLine(stdin, true);
}

/*!
  get existing file names
*/
QStringList QtbMessage::getOpenFileNamesInternal( const QString & caption,
                                          const QString & filter,
                                          const QString & directory )
{
  QTB_TRACE;
  qtbApp << "---- " << caption << " ----" << endl;
  qtbApp << tr("  Filter: ") << filter << endl;
  qtbApp << tr("  Current directory: ") << directory << endl;
  QString fn;
  QStringList fileNames;
  while(true) {
    qtbApp << tr("File %1 to open: ").arg(fileNames.count()+1) << flush;
    QString fn = QtbFile::readLine(stdin, true);
    if (fn.isEmpty()) break;
    fileNames.append(fn);
  }
  return fileNames;
}

/*!
  get an existing directory
*/
QString QtbMessage::getExistingDirectoryInternal( const QString & caption,
                                                  const QString & dir)
{
  QTB_TRACE;
  qtbApp << "---- " << caption << " ----" << endl;
  qtbApp << tr("  Dir: ") << dir << endl;
  qtbApp << tr("Directory to choose: ") << flush;
  return QtbFile::readLine(stdin, true);
}

/*!
  Call this function to print a user friendly message in case of error while parsing a formated text
  from a QTextStream. If graphical user interface is available, it will be printed in a dialog box, if not to the terminal
*/
void QtbMessage::wrongTextFormat(QTextStream& sErr, const QString& caption)
{
  QTB_TRACE;
  QFile * fErr=static_cast<QFile *>(sErr.device());
  QString str;
  if (fErr) {
    str = translate("QtbMessage","Error parsing text file %1\n").arg(fErr->fileName());
  } else {
    str = translate("QtbMessage","Error parsing text\n");
  }
  qint64 errorLocation=sErr.pos ();
  qint64 errorShowStart;
  if (errorLocation<200) errorShowStart=0; else errorShowStart=errorLocation-200;
  if (sErr.seek(errorShowStart)) {
    str+=translate("QtbMessage","Offset from beginning of file: %1 bytes.\n").arg(errorShowStart);
    QString strLine=sErr.readLine();
    qint64 lastLinePos=errorShowStart;
    qint64 currentPos=sErr.pos();
    while (currentPos<errorLocation) {
      str+=translate("QtbMessage","    " )+strLine+"\n";
      strLine=sErr.readLine();
      lastLinePos=currentPos;
      currentPos=sErr.pos();
    }
    int errorCharPos=errorLocation-lastLinePos;
    str+=translate("QtbMessage","==> %1\n").arg( strLine.left(errorCharPos) );
    QString blankStr;
    blankStr.fill(' ',errorCharPos+4);
    if (strLine.length()>=errorCharPos)
      str+=blankStr+strLine.right(strLine.length()-errorCharPos)+"\n";
    str+=blankStr+"^\n";
    while (!sErr.atEnd() && currentPos<errorLocation+200) {
      str+=translate("QtbMessage","    " )+sErr.readLine()+"\n";
      currentPos=sErr.pos();
    }
    sErr.seek(errorLocation);
  } else {
    str+=translate("QtbMessage","Impossible to print the context of the parsing error\n");
  }
  critical( MSG_ID, caption, str, cancel() );
}

/*!
  Call this function to print a user friendly message in case of error while reading a text file.
  If graphical user interface is available, it will be printed in a dialog box, if not to the terminal
*/
void QtbMessage::wrongTextFormat( FILE * fErr, const QString& caption, const QString& fileName )
{
  QTB_TRACE;
  QTextStream sOut( stdout );
  QString str;
  if (fileName.isEmpty() ) {
    str=translate("QtbMessage","Error parsing text file\n");
  } else {
    str=translate("QtbMessage","Error parsing text file %1\n").arg(fileName);
  }
  uint errorLocation=ftell(fErr);
  uint errorShowStart;
  if (errorLocation<200) errorShowStart=0; else errorShowStart=errorLocation-200;
  // TODO check if possible to know if it can seek correctly
  fseek(fErr,errorShowStart,SEEK_SET);
  int strLineLen=256;
  char * strLine=new char[strLineLen];
  QtbFile::readLine( strLine, strLineLen, fErr );
  uint lastLinePos=errorShowStart;
  uint currentPos=ftell(fErr);
  while (currentPos<errorLocation) {
    str+=translate("QtbMessage","    " )+strLine+"\n";
    QtbFile::readLine( strLine, strLineLen, fErr );
    lastLinePos=currentPos;
    currentPos=ftell(fErr);
  }
  uint errorCharPos=errorLocation-lastLinePos;
  char c=strLine[errorCharPos];
  strLine[errorCharPos]='\0';
  str+=translate("QtbMessage","==> %1\n").arg( strLine );
  strLine[errorCharPos]=c;
  QString blankStr;
  blankStr.fill(' ',errorCharPos+4);
  if (strlen(strLine)>=errorCharPos)
    str+=blankStr+(strLine+errorCharPos)+"\n";
  str+=blankStr+"^\n";
  while (!feof(fErr) && currentPos<errorLocation+200) {
    QtbFile::readLine( strLine, strLineLen, fErr );
    str+=translate("QtbMessage","    " )+strLine+"\n";
    currentPos=ftell(fErr);
  }
  fseek(fErr,errorLocation,SEEK_SET);
  critical( MSG_ID, caption, str, cancel() );
}

