/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2008-03-28
**
***************************************************************************/

#include <math.h>

#include <qglobal.h>

#include "qtbnumber.h"

/*!
  \fn double toDouble( QString str, Type t, bool * ok );
  Convert any valid string into a double value using \a type
*/

/*!
  \fn QString toString( double val, Type t, int precision )
  Convert any value into a valid string using \a type and \a precision
*/

/*!
  Convert char format to type

  \li 'f' for Fixed
  \li 'e' for Scientific
  \li 't' for Time
*/
QtbNumber::Type QtbNumber::type( char t )
{
  switch(t) {
  case 'e':
    return Scientific;
  case 't':
    return Time;
  default:
    return Fixed;
  }
}


/*!
  \fn double QtbNumer::toDouble( double a )
  Does nothing else than returning \a. Offers a transparent conversion from any real type to double.

  \sa toDouble( mpf_class a ), toDouble ( float a )
*/

/*!
  \fn double QtbNumber::toDouble( mpf_class a )
  Converts arbitrary precision \a to a double. Offers a transparent conversion from any real type to double.

  \sa toDouble( double a ), toDouble ( float a )
*/

/*!
  Converts a float to a double keeping track of rounding errors.
  Converting a float to a double in a basic way may lead to strange results.

  \pre
  float f = 0.002;
  double d = f;
  // d is 0.0020000000949949026108
  d = 0.002;
  // d is 0.0020000000000000000416
  \pre

  Tests made with this function:

  0.002 returns 0.0020000000000000000416 instead of 0.0020000000949949026108
  0.0020405 returns 0.0020405000000000002129 instead of 0.0020405000541359186172
  0.002040567 returns 0.0020405670000000000752 instead of 0.0020405671093612909317
  204056700000 returns 204056705154.90594482 instead of 204056707072

  Offers a transparent conversion from any real type to double.

  \sa toDouble( double a ), toDouble ( float a )
*/
double QtbNumber::toDouble( float f )
{
  if ( f == 0.0 ) return 0.0;
  // Extract the exponent value of the float (bits 2-9)
  // Generate a warning difficult to avoid... any other solution?
#if Q_BYTE_ORDER == Q_BIG_ENDIAN
  int exp=( (*(reinterpret_cast<unsigned short*>(&f)) & 0x7F80 ) >> 7) - 127;
#else
  int exp=( (*(reinterpret_cast<unsigned short*>(&f)+1) & 0x7F80 ) >> 7) - 127;
#endif
  // log10(2) = 0.301029995663981198
  // Precision for float mantissa is 1/2^23, look for power of 10 to multiply f so that
  // 1/2^23 stand after decimal point, to remove it by rounding.
  float fac = pow(10.0,floor( 0.301029995663981198*(22-exp) ) );
  return ( (double)round( f * fac ) / fac );
}

/*!
  Return all divisors of \a n
*/
QList<int> QtbNumber::divisors( int n )
{
  // max number to analyse: 157609 = 397*397
  static const int primes[] = {   2,   3,   5,   7,  11,  13,  17,  19,  23,  29,  31,  37,  41,  43,  47,
                                 53,  59,  61,  67,  71,  73,  79,  83,  89,  97, 101, 103, 107, 109, 113,
                                127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197,
                                199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281,
                                283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 353, 359, 367, 373, 379,
                                383, 389, 397 };
  QList<int> ds;
  int q = 398, d = 2, i = 0;
  while( d<q ) {
    q = n / d;
    if ( (q*d) == n ) {
      n = q;
      ds << d;
    } else {
      i++;
    }
    d = primes[i];
  }
  ds << n;
  return ds;
}
