/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-06-20
**
***************************************************************************/

#include "qtbparallelloop.h"
#include "qtbcoreapplication.h"
#include "qtbglobal.h"
#include "qtbtrace.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbLoopTask qtbparallelloop.h
  \brief Task of a parallel loop
*/

void QtbLoopTask::run()
{
  QTB_TRACE;
  while (true) {
    if (terminated()) break;
    _indexMutex->lock();
    int i = *_index;
    (*_index)++;
    _indexMutex->unlock();
    if (i<_endIndex)
      run(i);
    else
      break;
  }
}

/*!
  \class QtbParallelLoop qtbparallelloop.h
  \brief Class that helps to parallelize loops
*/

/*!
  \fn QtbParallelLoop::QtbParallelLoop( Parameters * param )
  Initialize a loop. The parameters \a param will be provided as a pointer to each task.
*/

void QtbParallelLoop::terminate()
{
  QTB_TRACE;
  for (QMap<QtbLoopTask *,TaskInfo>::iterator it = _tasks.begin(); it!=_tasks.end(); it++) {
    if (!it.key()->terminated()) it.key()->terminate();
  }
}

void QtbParallelLoop::taskFinished()
{
  QTB_TRACE;
  QtbLoopTask * t = static_cast<QtbLoopTask *>(sender());
  QMap<QtbLoopTask *,TaskInfo>::iterator it = _tasks.find(t);
  if ( it!=_tasks.end() ) {
    emit statusChanged( it.value().index, tr("Finished") );
    emit progressChanged( it.value().index, it.value().maximumProgress );
    _tasks.erase(it);
    delete t;
    if (_tasks.isEmpty()) emit finished();
  }
}

void QtbParallelLoop::waitFinished()
{
  while (!_tasks.isEmpty()) {
    QtbCoreApplication::sleep(500);
    QCoreApplication::processEvents();
  }
}

inline QtbLoopTask * QtbParallelLoop::newTask( int iEnd )
{
  QtbLoopTask * t = newTask();
  connect( t, SIGNAL(finished()), this, SLOT(taskFinished()));
  connect( t, SIGNAL(statusChanged(QString)), this, SLOT(statusChanged(QString)));
  connect( t, SIGNAL(progressInit(int)), this, SLOT(progressInit(int)));
  connect( t, SIGNAL(progressChanged(int)), this, SLOT(progressChanged(int)));
  t->setIndex( &_index, &_indexMutex );
  t->setEndIndex( iEnd );
  return t;
}

/*!
  Start loop from \a iStart to \a iEnd-1
*/
void QtbParallelLoop::start(int iStart, int iEnd, bool forceParallel)
{
  QTB_TRACE;
  int nThreads = QThread::idealThreadCount();
  _index = iStart;
  QtbLoopTask * t;
  if (nThreads==1 && !forceParallel) {
    t = newTask( iEnd );
    t->startNoThread();
  } else {
    TaskInfo infos;
    for (int i = 0; i<nThreads; i++) {
      t = newTask( iEnd );
      t->setObjectName("parallelLoop");
      infos.index = i;
      infos.maximumProgress = 0;
      _tasks.insert(t, infos);
      t->setIndex( &_index, &_indexMutex );
      t->start();
    }
  }
}

void QtbParallelLoop::statusChanged(QString msg)
{
  QTB_TRACE;
  QtbLoopTask * t = qobject_cast<QtbLoopTask *>(sender());
  QMap<QtbLoopTask *,TaskInfo>::iterator it = _tasks.find(t);
  if ( it!=_tasks.end() ) {
    emit statusChanged( it.value().index, msg );
  }
}

void QtbParallelLoop::progressChanged( int value )
{
  QTB_TRACE;
  QtbLoopTask * t = qobject_cast<QtbLoopTask *>(sender());
  QMap<QtbLoopTask *,TaskInfo>::iterator it = _tasks.find(t);
  if ( it!=_tasks.end() ) {
    emit progressChanged( it.value().index, value );
  }
}

void QtbParallelLoop::progressInit( int maximumValue )
{
  QTB_TRACE;
  QtbLoopTask * t = qobject_cast<QtbLoopTask *>(sender());
  QMap<QtbLoopTask *,TaskInfo>::iterator it = _tasks.find(t);
  if ( it!=_tasks.end() ) {
    // Store maximum value for taskFinished()
    it.value().maximumProgress = maximumValue;
    emit progressInit( it.value().index, maximumValue );
  }
}
