/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-06-20
**
***************************************************************************/

#ifndef QTB_PARALLELLOOP_H
#define QTB_PARALLELLOOP_H

#include <time.h>

#include <QObject>
#include <QMap>

#include "qtbtoolsdllexport.h"
#include "qtbcompatqt43.h"
#include "qtbthread.h"

class QtbLoopTask;

class QTBTOOLS_EXPORT QtbParallelLoop : public QObject
{
  Q_OBJECT
public:
  QtbParallelLoop( QObject * parent = 0 ) : QObject(parent) {}

  void start(int iStart, int iEnd, bool forceParallel = false);
  void terminate();
  void waitFinished();

protected:
  virtual QtbLoopTask * newTask() = 0;
protected slots:
  virtual void taskFinished();
private slots:
  void statusChanged(QString msg);
  void progressChanged(int value);
  void progressInit(int maximumValue);
signals:
  void statusChanged( int processIndex, QString msg );
  void progressChanged( int processIndex, int progress );
  void progressInit( int processIndex, int maximumValue);
  void finished();
private:
  inline static int idealThreadCount();
  inline QtbLoopTask * newTask( int iEnd );
  struct TaskInfo {
    int index;
    int maximumProgress;
  };
  QMap<QtbLoopTask *, TaskInfo> _tasks;
  QMutex _indexMutex;
  int _index;
};

class QTBTOOLS_EXPORT QtbLoopTask : public QtbThread
{
  Q_OBJECT
public:
  QtbLoopTask() : QtbThread() { _terminated = false; }

  void terminate() { _terminated.fetchAndStoreOrdered(true); }
  bool terminated() { return _terminated.testAndSetOrdered(true,true); }

  void startNoThread() { run(); }

  void setIndex(int * index, QMutex * mutex) { _index = index; _indexMutex = mutex; }
  void setEndIndex(int index) { _endIndex = index; }
protected:
  virtual void run();
  virtual void run( int index ) = 0;
  void setProgressValue( int value );
  void setProgressMaximum( int value ) { emit progressInit( value ); }
  void setStatus( QString msg ) { emit statusChanged( msg ); }
signals:
  void statusChanged( QString msg );
  void progressInit( int maximumValue );
  void progressChanged( int value );
private:
  QAtomicInt _terminated;
  QMutex * _indexMutex;
  int * _index;
  int _endIndex;
  QAtomicInt _lastProgressTime;
};

inline void QtbLoopTask::setProgressValue( int value )
{
  QTB_TRACE;
  int now = time(0);
  if ( _lastProgressTime.fetchAndStoreOrdered(now) != now ) {
    progressChanged( value );
  }
}

#endif // QTB_PARALLELLOOP_H
