/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-06-08
**
***************************************************************************/

#include <QDir>

#include "qtbpathtranslator.h"
#include "qtbmessage.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbPathTranslator
  \brief

  Utility to ease access to file which path is stored as absolute. It can translate obsolete
  paths into new ones, for instance, if the user change the organisation of his file tree.
  The new path is only asked one time. For all other requests, the obsolete path is searched
  in a map and if found, the file path is modified.
*/

/*!
  Return true if the path has been successfully changed

  File type must contain '%1' to be replace by file name
*/
QString QtbPathTranslator::translate( const QString& originalFile, const QString& title, const QString& fileType )
{
  QTB_TRACE;
  QString path = platformIndependantPath( originalFile );
  QFileInfo file(path);
  path=file.path();
  QString relPath;
  iterator it;
  //qDebug("Entering QtbPathTranslator::translate(%s,%s,%s)",path.toAscii().data(),title.toAscii().data(),
  //                                                         fileType.toAscii().data());
  while ( true ) {
    it = find ( path );
    if ( it!=end() ) {
      path = it.value();
      //qDebug("Path replaced by %s",path.toAscii().data());
      path += relPath;
      //qDebug("definitive path %s",path.toAscii().data());
      QDir d ( path );
      if ( d.exists() ) {
        QFileInfo fi( d.absoluteFilePath( file.fileName() ) );
        if ( fi.exists() ) return fi.absoluteFilePath();
      }
      return askPath( file, title, fileType, d.absolutePath() );
    }
    if ( isRoot( path ) ) break;
    QString name = dirName( path );
    path.chop( name.count() + 1 );
    relPath=QDir::separator() + name + relPath;
    //qDebug("Choped path: %s",path.toAscii().data());
  }
  return askPath( file, title, fileType );
}

/*!
  Add the different root between \a obsoletePath and \a newPath to this translator
*/
void QtbPathTranslator::addPath( QString obsoletePath, QString newPath )
{
  QTB_TRACE;
  //qDebug("Entering QtbPathTranslator::addPath( %s, %s )",obsoletePath.toAscii().data(),newPath.toAscii().data());
  while ( !isRoot( obsoletePath ) && !isRoot( newPath ) ) {
    QString obsName = dirName( obsoletePath );
    QString newName = dirName( newPath );
    if ( obsName != newName ) {
      //qDebug("Adding translation for %s to %s", obsoletePath.toAscii().data(), newPath.toAscii().data());
      insert ( obsoletePath, newPath );
      return;
    }
    obsoletePath.chop( obsName.count() + 1 );
    newPath.chop( newName.count() + 1 );
    //qDebug("Choped obsoletePath: %s",obsoletePath.toAscii().data());
    //qDebug("Choped newPath: %s",newPath.toAscii().data());
  }
  if (obsoletePath != newPath ) {
    //qDebug("Adding translation for %s to %s", obsoletePath.toAscii().data(), newPath.toAscii().data());
    insert ( obsoletePath, newPath );
  }
}

QString QtbPathTranslator::platformIndependantPath( QString path )
{
  QTB_TRACE;
#ifdef Q_WS_WIN
  if ( path[0] == '/' ) {
    path = "C:\\" + path.mid( 1 );
  }
  return path;
#else
  QRegExp winPath("[a-zA-Z]:[\\\\/]");
  if ( winPath.indexIn(path)==0 ) {
    path = "/" + path.mid( 3 );
    path = path.replace("\\", "/");
  }
  return path;
#endif
}

/*!
  Ask for \a file from directory \a d
*/
QString QtbPathTranslator::askPath ( const QFileInfo& file, const QString& title, const QString& fileType, const QString& directory  )
{
  QTB_TRACE;
  switch ( QtbMessage::warning( MSG_ID, title, tr( "File %1 does not exist.\n"
                                "The path might have changed. Would you like to manually select its new location?" )
                                .arg( file.filePath() ), QtbMessage::yes(), QtbMessage::no(), QString::null, true ) ) {
  case QtbMessage::Answer1:
    return QString::null;
  case QtbMessage::Answer0: {
      QString newFileName = QtbMessage::getOpenFileName( title, fileType. arg( file.fileName() ),
                                                         directory );
      if ( !newFileName.isEmpty() ) {
        QFileInfo finew( newFileName );
        if ( finew.fileName() == file.fileName() ) {
          addPath( file.absolutePath(), finew.absolutePath() );
          return newFileName;
        } else {
          QtbMessage::warning( MSG_ID, title, tr( "Selected file is not the correct file. Aborting." ), QtbMessage::ok() );
          return QString::null;
        }
      }
    }
  default:
    return QString::null;
  }
}

/*!
  Return the directory name (without the path).
*/
QString QtbPathTranslator::dirName( QString path )
{
  QTB_TRACE;
  int i = path.lastIndexOf( QDir::separator() );
  if ( i != -1 )
    return path.right( path.count() - i - 1 );
  else
    return path;
}

/*!
  Return true if root or empty path.
*/
bool QtbPathTranslator::isRoot( QString path )
{
  QTB_TRACE;
  if ( path.isEmpty() ) return true;
#ifdef Q_WS_WIN
  if ( path.count() == 2 && path[ 1 ] == ':' )
    return true;
#endif
  return false;
}
