/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2002-04-15
**
***************************************************************************/

#include <math.h>

#include <QVector>

#include "qtbpoint.h"
#include "qtbpoint2d.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbPoint qtbpoint.h
  
  Basic properties of a 3D point (coordinates in double precision)

  Conversion from ECEF XYZ to Lat., Long. and Height
  http://www.colorado.edu/geography/gcraft/notes/datum/gif/xyzllh.gif
  http://www.colorado.edu/geography/gcraft/notes/datum/gif/llhxyz.gif
*/

/*!
  \fn QtbPoint::QtbPoint()
  Default constructor
*/

/*!
  \fn QtbPoint::QtbPoint( double x, double y, double z )
  Constructor setting \a x and \a y, and optionally \a z
*/

/*!
  \fn QtbPoint::QtbPoint(const QtbPoint2D& p)
  Copy constructor from QtbPoint2D
*/

/*!
  \fn void QtbPoint::interpole( double valX, const QtbPoint& p1, const QtbPoint& p2 )
  Interpole between two points
*/

/*!
  \fn void QtbPoint::setValid(bool v)
  Does nothing. Used by QtbCurve.
*/

/*!
  \fn bool QtbPoint::isValid() const
  Return always true. Used by QtbCurve.
*/

QString QtbPoint::toString(int precision) const
{
  QTB_TRACE;
  QString tmp;
  tmp+=QString::number(_x,'g',precision);
  tmp+=" ";
  tmp+=QString::number(_y,'g',precision);
  tmp+=" ";
  tmp+=QString::number(_z,'g',precision);
  return tmp;
}

bool QtbPoint::fromString(QtbStringSection str)
{
  QTB_TRACE;
  const QChar * ptr=0;
  QtbStringSection f;
  f=str.firstField(ptr);
  if (f.isValid()) _x=f.toDouble(); else return false;
  f=str.nextField(ptr);
  if (f.isValid()) _y=f.toDouble(); else return false;
  f=str.nextField(ptr);
  if (f.isValid()) _z=f.toDouble(); else _z=0.0;
  return true;
}

bool QtbPoint::isHit(const QtbPoint &p,double tolX, double tolY) const
{
  QTB_TRACE;
  if ((_x-tolX<=p.x() && p.x()<_x+tolX)
      && (_y-tolY<=p.y() && p.y()<_y+tolY))
    return true;
  else
    return false;
}

double QtbPoint::distanceTo(const QtbPoint &p) const
{
  QTB_TRACE;
  double tmp, dist;

  dist=0.;
  tmp=_x-p.x();
  dist+=tmp*tmp;
  tmp=_y-p.y();
  dist+=tmp*tmp;
  tmp=_z-p.z();
  dist+=tmp*tmp;
  return sqrt(dist);
}

/*!
  Returns the azimuth from this point to \a p in radians in the mathematical sense (from 0 to 2 PI, 0 is East).
*/
double QtbPoint::azimuth(const QtbPoint &p) const
{
  QTB_TRACE;
  double a=atan2(p.y()-_y, p.x()-_x);
  if (a<0) a+=2*M_PI;
  return a;
}

double QtbPoint::distanceToSegment(const QtbPoint &p1, const QtbPoint &p2) const
{
  QTB_TRACE;
  double d1=distanceTo(p1);
  double d2=distanceTo(p2);
  double d12=p1.distanceTo(p2);
  QtbPoint dir1((_x-p1.x())/d1, (_y-p1.y())/d1, (_z-p1.z())/d1);
  QtbPoint dir2((_x-p2.x())/d2, (_y-p2.y())/d2, (_z-p2.z())/d2);
  QtbPoint dir12((p2.x()-p1.x())/d12, (p2.y()-p1.y())/d12, (p2.z()-p1.z())/d12);
  double costheta1=dir1.x()*dir12.x()+dir1.y()*dir12.y()+dir1.z()*dir12.z();
  double costheta2=-dir2.x()*dir12.x()-dir2.y()*dir12.y()-dir2.z()*dir12.z();
  if ((costheta1 < 0.) || (costheta2 < 0.))
    return -1.;
  double d=d1*sqrt(1.-costheta1*costheta1);
  return d;
}

/*void QtbPoint::xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const
{
  writeProperty(s, "x",x);
  writeProperty(s, "y",y);
  writeProperty(s, "z",z);
}
 
QtbXMLMember QtbPoint::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  if (tag=="x") return QtbXMLMember(0);
  else if (tag=="y") return QtbXMLMember(1);
  else if (tag=="z") return QtbXMLMember(2);
  return QtbXMLMember(QtbXMLMember::Unknown);
}
 
bool QtbPoint::xml_setProperty(XML_SETPROPERTY_ARGS)
                               QtbStringSection& content)
{
  QTB_TRACE;
  switch (memberID) {
  case 0: x=content.toDouble(); return true;
  case 1: y=content.toDouble(); return true;
  case 2: z=content.toDouble(); return true;
  default: return false;
  }
}*/

#define E_RAD 6378.163
#define E_FLAT  298.26
#define DRAD  1.7453292e-2
#define DRLT  9.9330647e-1

/*!
  Convert point from geographical coordinates (long and lat, WGS84) to cartesian coordinates in meters,
  using reference point \a ref (in geographical coordinates). Longitude is along X axis, latitude is along Y axis.
  Z axis is not touched.

  \a rotation is an angle to rotate north reference (clockwize).

  Kindly provided by Matthias Ohrnberger.
*/
void QtbPoint::geographicalToCartesian( const QtbPoint& ref, double rotation )
{
  double olon;
  double olat;
  double lat_fac;  // conversion factor for latitude in km
  double lon_fac;  // conversion factor for longitude in km
  double snr;      // sin of rotation angle
  double csr;      // cos of rotation angle
  double dlt1;
  double dlt2;
  double del;
  double radius;
  double tmp;
  double tmp_x, tmp_y;

  /* convert everything to minutes */
  olon = 60.0 * ref.x();
  olat = 60.0 * ref.y();

  /* latitude */
  dlt1 = atan(DRLT * tan((double)olat * DRAD/60.0));
  dlt2 = atan(DRLT * tan(((double)olat +1.0) * DRAD/60.0));
  del  = dlt2 - dlt1;
  radius = E_RAD * (1.0 - (sin(dlt1)*sin(dlt1) / E_FLAT));
  lat_fac = radius * del;

  /* longitude */
  del = acos(1.0 - (1.0 - cos(DRAD/60.0)) * cos(dlt1) * cos(dlt1));
  dlt2 = radius * del;
  lon_fac = dlt2 / cos(dlt1);

  /* rotation */
  snr = sin((double)rotation * DRAD);
  csr = cos((double)rotation * DRAD);

  _y *= 60.0; 
  _x *= 60.0;

  tmp_x = _x - olon;
  tmp_y = _y - olat;
  tmp   = atan(DRLT * tan(DRAD * (_y+olat)/120.0));
  tmp_x = (double)tmp_x * lon_fac * cos(tmp);
  tmp_y = (double)tmp_y * lat_fac;

  _x = ( csr*tmp_x - snr*tmp_y )*1000.0;
  _y = ( csr*tmp_y + snr*tmp_x )*1000.0;
}



/*!
  Convert point from cartesian coordinates in meters to geographical coordinates (lat and long, WGS84),
  using reference point \a ref (in geographical coordinates). Longitude is along X axis, latitude is along Y axis.
  Z axis is not touched.

  \a rotation is an angle to rotate north reference (clockwize).

  Kindly provided by Matthias Ohrnberger.
*/
void QtbPoint::cartesianToGeographical( const QtbPoint& ref, double rotation )
{
  double olon;
  double olat;
  double lat_fac;  // conversion factor for latitude in km
  double lon_fac;  // conversion factor for longitude in km
  double snr;      // sin of rotation angle
  double csr;      // cos of rotation angle
  double dlt1;
  double dlt2;
  double del;
  double radius;
  double tmp;
  double tmp_x, tmp_y;

  // Convert to km
  _x*=0.001;
  _y*=0.001;

  /* convert everything to minutes */
  olon = 60.0 * ref.x();
  olat = 60.0 * ref.y();

  /* latitude */
  dlt1 = atan(DRLT * tan((double)olat * DRAD/60.0));
  dlt2 = atan(DRLT * tan(((double)olat +1.0) * DRAD/60.0));
  del  = dlt2 - dlt1;
  radius = E_RAD * (1.0 - (sin(dlt1)*sin(dlt1) / E_FLAT));
  lat_fac = radius * del;

  /* longitude */
  del = acos(1.0 - (1.0 - cos(DRAD/60.0)) * cos(dlt1) * cos(dlt1));
  dlt2 = radius * del;
  lon_fac = dlt2 / cos(dlt1);

  /* rotation */
  snr = sin((double)rotation * DRAD);
  csr = cos((double)rotation * DRAD);
  tmp_x = snr* _y + csr* _x;
  tmp_y = csr* _y - snr* _x;

  tmp_y = tmp_y/lat_fac;
  tmp_y += olat;

  tmp = atan(DRLT * tan(DRAD * (tmp_y+olat)/120.0));
  tmp_x = tmp_x / (lon_fac * cos(tmp));
  tmp_x += olon;

  _x = tmp_x/60.0;
  _y = tmp_y/60.0;
}

/*!
  Rotate point by \a angle degree clockwize
*/
void QtbPoint::rotate(double angle)
{
  angle = -angle * M_PI / 180.0;
  double cosa = cos( angle );
  double sina = sin( angle );
  double x = cosa * _x - sina * _y;
  double y = sina * _x + cosa * _y;
  _x=x;
  _y=y;
}

/*!
  Sum of all x (used in least square method)
*/
double QtbPointList::sumX() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x();
  return sum;
}

/*!
  Sum of all y (used in least square method)
*/
double QtbPointList::sumY() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->y();
  return sum;
}

/*!
  Sum of all x*y (used in least square method)
*/
double QtbPointList::sumXY() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x()*it->y();
  return sum;
}

/*!
  Sum of all x*x (used in least square method)
*/
double QtbPointList::sumX2() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x()*it->x();
  return sum;
}

/*!
  Sum of all y*y (used in least square method)
*/
double QtbPointList::sumY2() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->y()*it->y();
  return sum;
}

/*!
  Performs a least square regression on all points in the horizontal plane.
 
  Return the azimuth of the regression line.
*/
double QtbPointList::azimuth() const
{
  QTB_TRACE;
  int n=count();
  if (n>1) {
    double sx=sumX();
    double sy=sumY();
    double sxy=sumXY();
    double sx2=sumX2();
    double denom=n*sx2-sx*sx;
    if (denom!=0) return atan((n*sxy-sy*sx)/denom); else return 0.5*M_PI;
  } else return 0;
}

/*!
  Performs a least square regression on all points in the horizontal plane.
  
  Returns the distances along the profile
*/
QVector<double> QtbPointList::project2D() const
{
  QTB_TRACE;
  int n=size();
  if (!n) return QVector<double>();
  QVector<double> projections(n);
  if (n>1) {
    double sx=sumX();
    double sy=sumY();
    double sxy=sumXY();
    double sx2=sumX2();
    double denom=n*sx2-sx*sx;
    QList<QtbPoint>::const_iterator it;
    if (denom!=0) {
      double A=(n*sxy-sy*sx)/denom;
      double C=(sy*sx2-sxy*sx)/denom;
      double min=1e99;
      int i=0;
      for(it=begin();it!=end();++it, i++) {
        double denom=A*A+1;
        double cp=-it->x()-A*it->y();
        double xp=(-cp-A*C)/denom;
        double yp=(-A*cp+C)/denom;
        yp-=C;
        double d=sqrt(xp*xp+yp*yp);
        projections[i]=(xp>=0)? d : -d;
        if(projections[i]<min) min=projections[i];
      }
      // set min to abcsisse=0
      for (i=0; i<n;i++) projections[i]-=min;
    } else {
      int i=0;
      for(it=begin();it!=end();++it, i++)
        projections[i]=it->y();
    }
  } else projections[0]=0;
  return projections;
}

QTextStream& operator<< (QTextStream& s, const QtbPointList& plist)
{
  QTB_TRACE;
  s << "%x      y      z"<< endl;
  QList<QtbPoint>::const_iterator it;
  for(it=plist.begin();it!=plist.end();++it) s << *it << endl;
  return s;
}

QTextStream& operator>> (QTextStream& s, QtbPointList& plist)
{
  QTB_TRACE;
  QString str;
  plist.clear();
  while (!s.atEnd()) {
    str=s.readLine();
    while (str[0]!='%') str=s.readLine();
    QTextStream currentLine(&str, QIODevice::ReadOnly);
    QtbPoint * p=new QtbPoint;
    currentLine >> *p;
    plist.append(*p);
  }
  return s;
}

QTextStream& operator<< (QTextStream& s, const QtbPoint& p)
{
  QTB_TRACE;
  s.setRealNumberNotation(QTextStream::SmartNotation);
  s.setRealNumberPrecision(20);
  s << p.x() << " " << p.y() << " " << p.z();
  return s;
}

QTextStream& operator>> (QTextStream& s, QtbPoint& p)
{
  QTB_TRACE;
  double x, y, z;
  s >> x >> y >> z;
  p.setX( x );
  p.setY( y );
  p.setZ( z );
  return s;
}

QDataStream& operator<< (QDataStream& s, const QtbPoint& p)
{
  QTB_TRACE;
  s << p.x() << p.y() << p.z();
  return s;
}

QDataStream& operator>> (QDataStream& s, QtbPoint& p)
{
  QTB_TRACE;
  double x, y, z;
  s >> x >> y >> z;
  p.setX( x );
  p.setY( y );
  p.setZ( z );
  return s;
}
