/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2002-04-15
**
***************************************************************************/

#ifndef QTB_Point_H
#define QTB_Point_H

#include <QString>
#include "qtbstringsection.h"
#include "qtbtoolsdllexport.h"

class QtbPoint2D;

class QTBTOOLS_EXPORT QtbPoint
{
public:
  inline QtbPoint();
  inline QtbPoint(double x, double y, double z=0.);
  inline QtbPoint(const QtbPoint2D& p);

  double x() const {return _x;}
  double y() const {return _y;}
  double z() const {return _z;}
  void setX( double v ) { _x = v;}
  void setY( double v ) { _y = v;}
  void setZ( double v ) { _z = v;}
  void setValid(bool) {}
  bool isValid() const { return true; }
  void average(const QtbPoint& p ) { _y= 0.5*(_y+p._y); _z= 0.5*(_z+p._z); }

  void geographicalToCartesian( const QtbPoint& ref, double rotation = 0.0 );
  void cartesianToGeographical( const QtbPoint& ref, double rotation = 0.0 );
  void translate(const QtbPoint& t) {operator+=(t);}
  void rotate(double angle);
  // Copy operators
  inline QtbPoint& operator=(const QtbPoint& p);
  inline QtbPoint& operator=(const QtbPoint2D& p);
  inline void set(double xi, double yi, double zi=0.);

  // Comparison operators
  inline bool operator<  (const QtbPoint& p) const;
  inline bool operator== (const QtbPoint& p) const;
  bool operator<= (const QtbPoint& obj) const
    { return (*this == obj || *this < obj);};
  bool operator>  (const QtbPoint& obj) const
    {return (! (*this <= obj));} ;
  bool operator>= (const QtbPoint& obj) const
    {return (! (*this < obj));};
  bool operator!= (const QtbPoint& obj) const
    {return (! (*this == obj));};
  inline short compare (const QtbPoint& p) const;

  // Arithmetic operation
  inline QtbPoint operator+ (const QtbPoint& p) const;
  inline QtbPoint operator- (const QtbPoint& p) const;
  inline void operator+= (const QtbPoint& p);
  inline void operator-= (const QtbPoint& p);
  inline QtbPoint operator* (double mul) const;
  inline QtbPoint operator* (const QtbPoint& p) const;
  inline QtbPoint operator/ (double div) const;
  inline QtbPoint operator/ (const QtbPoint& p) const;
  inline void operator*= (const QtbPoint& p);
  inline void operator*= (double mul);
  inline void operator/= (const QtbPoint& p);
  inline void operator/= (double div);
  inline double scalarProduct (const QtbPoint& p) const;
  inline void vectorialProduct (const QtbPoint& p1,const QtbPoint& p2);
  inline void interpole( double valX, const QtbPoint& p1, const QtbPoint& p2 );

  // I/O functions
  bool fromString(QtbStringSection str);
  QString toString(int precision=6) const;

  // Miscellaneous
  // Checks whether p is in the neighbourhood of point (+/- tolX,+/- tolY)
  bool isHit(const QtbPoint &p,double tolX, double tolY) const;
  // Calculate Distance to p
  double distanceTo(const QtbPoint &p) const;
  // Calculate the distance to a segment defined by p1 and p2
  double distanceToSegment(const QtbPoint &p1, const QtbPoint &p2) const;
  // Calculate azimuth to p from 0 to 2*pi (OX=0) anti-clockwize
  double azimuth(const QtbPoint &p) const;
private:
  // Coordinates of the point
  double _x, _y, _z;

};

#include "qtbpoint2d.h"

// Constructors
inline QtbPoint::QtbPoint()
{
  QTB_TRACE;
  _x=0.;
  _y=0.;
  _z=0.;
}

inline QtbPoint::QtbPoint(double xi, double yi, double zi)
{
  QTB_TRACE;
  _x=xi;
  _y=yi;
  _z=zi;
}

inline QtbPoint::QtbPoint(const QtbPoint2D& p)
{
  QTB_TRACE;
  _x=p.x();
  _y=p.y();
  _z=0;
}

// Copy operators
inline QtbPoint& QtbPoint::operator= (const QtbPoint& p)
{
  QTB_TRACE;
  _x=p._x;
  _y=p._y;
  _z=p._z;
  return *this;
}

inline QtbPoint& QtbPoint::operator= (const QtbPoint2D& p)
{
  QTB_TRACE;
  _x=p.x();
  _y=p.y();
  _z=0;
  return *this;
}

inline void QtbPoint::set(double xi, double yi, double zi)
{
  QTB_TRACE;
  _x=xi;
  _y=yi;
  _z=zi;
}

// Comparison operators
inline bool QtbPoint::operator<  (const QtbPoint& p) const
{
  QTB_TRACE;
  if (_x < p._x) return true;
  else if (_x==p._x) {
    if (_y < p._y) return true;
    else if (_y==p._y && _z < p._z) return true;
  }
  return false;
}

inline bool QtbPoint::operator== (const QtbPoint& p) const
{
  QTB_TRACE;
  if ((_x == p._x) &&
      (_y == p._y) &&
      (_z == p._z))
    return true;
  else
    return false;
}

inline short QtbPoint::compare (const QtbPoint& p) const
{
  QTB_TRACE;
  if (_x < p._x) return(-1);
  else if (_x > p._x) return(1);
  else if (_y < p._y) return(-1);
  else if (_y > p._y) return(1);
  else if (_z < p._z) return(-1);
  else if (_z > p._z) return(1);
  else return(0);
}

// Arithmetic operations
inline QtbPoint QtbPoint::operator+ (const QtbPoint& p) const
{
  QTB_TRACE;
  return QtbPoint (_x+p._x, _y+p._y, _z+p._z);
}

inline QtbPoint QtbPoint::operator- (const QtbPoint& p) const
{
  return QtbPoint (_x-p._x, _y-p._y, _z-p._z);
}

inline void QtbPoint::operator+= (const QtbPoint& p)
{
  QTB_TRACE;
  _x+=p._x; _y+=p._y; _z+=p._z;
}

inline void QtbPoint::operator-= (const QtbPoint& p)
{
  _x-=p._x; _y-=p._y; _z-=p._z;
}

inline void QtbPoint::operator*= (const QtbPoint& p)
{
  QTB_TRACE;
  _x*=p._x;
  _y*=p._y;
  _z*=p._z;
}

inline void QtbPoint::operator*= (double mul)
{
  QTB_TRACE;
  _x*=mul;
  _y*=mul;
  _z*=mul;
}

inline void QtbPoint::operator/= (const QtbPoint& p)
{
  QTB_TRACE;
  _x/=p._x;
  _y/=p._y;
  _z/=p._z;
}

inline void QtbPoint::operator/= (double div)
{
  QTB_TRACE;
  _x/=div;
  _y/=div;
  _z/=div;
}

inline QtbPoint QtbPoint::operator* (double mul) const
{
  QTB_TRACE;
  return QtbPoint (_x*mul, _y*mul, _z*mul);
}

inline QtbPoint QtbPoint::operator* (const QtbPoint& p) const
{
  QTB_TRACE;
  return QtbPoint (_x*p._x, _y*p._y, _z*p._z);
}

inline QtbPoint QtbPoint::operator/ (double div) const
{
  QTB_TRACE;
  return QtbPoint (_x/div, _y/div, _z/div);
}

inline QtbPoint QtbPoint::operator/ (const QtbPoint& p) const
{
  QTB_TRACE;
  return QtbPoint (_x/p._x, _y/p._y, _z/p._z);
}

inline double QtbPoint::scalarProduct (const QtbPoint& p) const
{
  QTB_TRACE;
  return _x*p._x+_y*p._y+_z*p._z;
}

inline void QtbPoint::vectorialProduct (const QtbPoint& p1,const QtbPoint& p2)
{
  QTB_TRACE;
  _x=p1._y*p2._z-p2._y*p1._z;
  _y=-p1._x*p2._z+p2._x*p1._z;
  _z=p1._x*p2._y-p2._x*p1._y;
}

inline void QtbPoint::interpole( double valX, const QtbPoint& p1, const QtbPoint& p2 )
{
  QTB_TRACE;
  double factor = ( valX - p1.x() ) / ( p2.x()-p1.x() );
  setX( valX );
  setY( p1.y()+ ( p2.y()-p1.y() ) * factor );
  setZ( p1.z()+ ( p2.z()-p1.z() ) * factor );
}

class QTBTOOLS_EXPORT QtbPointList : public QList<QtbPoint>
{
public:
  QVector<double> project2D() const;
  double azimuth() const;
protected:
  double sumX() const;
  double sumY() const;
  double sumXY() const;
  double sumX2() const;
  double sumY2() const;
};

QTBTOOLS_EXPORT QTextStream& operator<< (QTextStream& s, const QtbPointList& plist);
QTBTOOLS_EXPORT QTextStream& operator>> (QTextStream& s, QtbPointList& plist);
QTBTOOLS_EXPORT QTextStream& operator<< (QTextStream& s, const QtbPoint& p);
QTBTOOLS_EXPORT QTextStream& operator>> (QTextStream& s, QtbPoint& p);
QTBTOOLS_EXPORT QDataStream& operator<< (QDataStream& s, const QtbPoint& p);
QTBTOOLS_EXPORT QDataStream& operator>> (QDataStream& s, QtbPoint& p);

#endif // QtbPoint
