/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2002-04-15
**
***************************************************************************/

#include <math.h>

#include "qtbpoint2d.h"
#include "qtbpoint.h"
#include "qtbstringsection.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbPoint2D qtbpoint2d.h
  
  Basic properties of a 2D point (coordinates in double precision)
*/

/*!
  \fn QtbPoint2D::QtbPoint2D()
  Default constructor
*/

/*!
  \fn QtbPoint2D::QtbPoint2D( double x, double y )
  Constructor setting x and y
*/

/*!
  \fn QtbPoint2D::QtbPoint2D(const QtbPoint& p)
  Copy constructor from QtbPoint
*/

/*!
  \fn void QtbPoint2D::interpole( double valX, const QtbPoint2D& p1, const QtbPoint2D& p2 )
  Interpole between two points
*/

/*!
  \fn void QtbPoint2D::setInvalid()
  Does nothing,. Used by QtbCurve.
*/

/*!
  \fn bool QtbPoint2D::isValid() const
  Always return true. Used by QtbCurve.
*/

QString QtbPoint2D::toString(int precision) const
{
  QTB_TRACE;
  QString tmp;
  tmp+=QString::number(_x,'g',precision);
  tmp+=" ";
  tmp+=QString::number(_y,'g',precision);
  return tmp;
}

bool QtbPoint2D::fromString(QtbStringSection str)
{
  QTB_TRACE;
  const QChar * ptr=0;
  QtbStringSection f;
  f=str.firstField(ptr);
  if (f.isValid()) _x=f.toDouble(); else return false;
  f=str.nextField(ptr);
  if (f.isValid()) _y=f.toDouble(); else return false;
  return true;
}

/*void QtbPoint2D::xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const
{
  writeProperty(s, "x",x);
  writeProperty(s, "y",y);
}
 
QtbXMLMember QtbPoint2D::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  if (tag=="x") return QtbXMLMember(0);
  else if (tag=="y") return QtbXMLMember(1);  
  return QtbXMLMember(QtbXMLMember::Unknown);
}
 
bool QtbPoint2D::xml_setProperty(XML_SETPROPERTY_ARGS)
                                 QtbStringSection& content)
{
  QTB_TRACE;
  switch (memberID) {
  case 0: x=content.toDouble(); return true;
  case 1: y=content.toDouble(); return true;
  default: return false;
  }
}*/

bool QtbPoint2D::isHit(const QtbPoint2D &p,double tolX, double tolY) const
{
  QTB_TRACE;
  if ((_x-tolX<=p.x() && p.x()<_x+tolX)
      && (_y-tolY<=p.y() && p.y()<_y+tolY))
    return true;
  else
    return false;
}

double QtbPoint2D::distanceTo(const QtbPoint2D &p) const
{
  QTB_TRACE;
  double tmp, dist;

  dist=0.;
  tmp=_x-p.x();
  dist+=tmp*tmp;
  tmp=_y-p.y();
  dist+=tmp*tmp;
  return sqrt(dist);
}

double QtbPoint2D::azimuth(const QtbPoint2D &p) const
{
  QTB_TRACE;
  double a=atan2(p.y()-_y, p.x()-_x);
  if (a<0) a+=2*M_PI;
  return a;
}

double QtbPoint2D::distanceToSegment(const QtbPoint2D &p1, const QtbPoint2D &p2) const
{
  QTB_TRACE;
  // Find the intersection between segment and its perpendicular passing at this
  double dys=p2.y()-p1.y();
  if (dys==0) {
    if (p1.x()<p2.x()) {
      if (_x>p2.x()) return distanceTo(p2);
      if (_x<p1.x()) return distanceTo(p1);
    } else {
      if (_x>p1.x()) return distanceTo(p1);
      if (_x<p2.x()) return distanceTo(p2);
    }
    return fabs(_y-p1.y());
  }
  double dxs=p2.x()-p1.x();
  if (dxs==0) {
    if (p1.y()<p2.y()) {
      if (_y>p2.y()) return distanceTo(p2);
      if (_y<p1.y()) return distanceTo(p1);
    } else {
      if (_y>p1.y()) return distanceTo(p1);
      if (_y<p2.y()) return distanceTo(p2);
    }
    return fabs(_x-p1.x());
  }
  double coeff=dys/dxs;
  double coeffp=dxs/dys;
  double interx=(coeff*p1.x()-coeffp*_x+_y-p1.y())/(coeff-coeffp);
  if (p1.x()<p2.x()) {
    if (interx>p2.x()) return distanceTo(p2);
    if (interx<p1.x()) return distanceTo(p1);
  } else {
    if (interx>p1.x()) return distanceTo(p1);
    if (interx<p2.x()) return distanceTo(p2);
  }
  double intery=coeffp*(interx-_x)+_y;
  if (p1.y()<p2.y()) {
    if (intery>p2.y()) return distanceTo(p2);
    if (intery<p1.y()) return distanceTo(p1);
  } else {
    if (intery>p1.y()) return distanceTo(p1);
    if (intery<p2.y()) return distanceTo(p2);
  }
  return distanceTo(QtbPoint2D(interx,intery));
}

void QtbPoint2D::walkAzimuth(double dir,double d)
{
  QTB_TRACE;
  _x+=d*cos(dir);
  _y+=d*sin(dir);
}

QTextStream& operator<< (QTextStream& s,const QtbPoint2D& p)
{
  QTB_TRACE;
  s.setRealNumberNotation(QTextStream::SmartNotation);
  s.setRealNumberPrecision(20);
  s << p.x() << " " << p.y();
  return s;
}

QTextStream& operator>> (QTextStream& s,QtbPoint2D& p)
{
  QTB_TRACE;
  double x, y;
  s >> x >> y;
  p.setX( x );
  p.setY( y );
  return s;
}

QDataStream& operator<< (QDataStream& s,const QtbPoint2D& p)
{
  QTB_TRACE;
  s << p.x() << p.y();
  return s;
}

QDataStream& operator>> (QDataStream& s,QtbPoint2D& p)
{
  QTB_TRACE;
  double x, y;
  s >> x >> y;
  p.setX( x );
  p.setY( y );
  return s;
}

QTextStream& operator<< (QTextStream& s, const QList<QtbPoint2D>& plist)
{
  QTB_TRACE;
  s << "%x      y"<< endl;
  QList<QtbPoint2D>::const_iterator it;
  for(it=plist.begin();it!=plist.end();++it) s << *it << endl;
  return s;
}

QTextStream& operator>> (QTextStream& s, QList<QtbPoint2D>& plist)
{
  QTB_TRACE;
  QString str;
  plist.clear();
  while (!s.atEnd()) {
    str=s.readLine();
    while (str[0]!='%') str=s.readLine();
    QTextStream currentLine(&str, QIODevice::ReadOnly);
    QtbPoint * p=new QtbPoint;
    currentLine >> *p;
    plist.append(*p);
  }
  return s;
}

/*!
  Sum of all x (used in least square method)
*/
double QtbPoint2DList::sumX() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint2D>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x();
  return sum;
}

/*!
  Sum of all y (used in least square method)
*/
double QtbPoint2DList::sumY() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint2D>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->y();
  return sum;
}

/*!
  Sum of all x*y (used in least square method)
*/
double QtbPoint2DList::sumXY() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint2D>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x()*it->y();
  return sum;
}

/*!
  Sum of all x*x (used in least square method)
*/
double QtbPoint2DList::sumX2() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint2D>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->x()*it->x();
  return sum;
}

/*!
  Sum of all y*y (used in least square method)
*/
double QtbPoint2DList::sumY2() const
{
  QTB_TRACE;
  double sum=0;
  QList<QtbPoint2D>::const_iterator it;
  for(it=begin();it!=end();++it) sum+=it->y()*it->y();
  return sum;
}

/*!
  Performs a least square regression on all points: y = a*x+b
*/
void QtbPoint2DList::regression( double& a, double& b) const
{
  QTB_TRACE;
  double n=size();
  if (n<2.0) {
    a = 1.0;
    b = 0.0;
    return;
  } else {
    double sx=sumX();
    double sy=sumY();
    double sxy=sumXY();
    double sx2=sumX2();
    double denom=n*sx2-sx*sx;
    if (denom!=0) {
      a=(n*sxy-sy*sx)/denom;
      b=(sy-a*sx)/n;
    } else { // All X are = ==> a=inifinity
      a = 1.0;
      b = 0.0;
    }
  }
}
