/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2002-04-15
**
***************************************************************************/

#ifndef QTB_POINT2D_H
#define QTB_POINT2D_H

#include <QDataStream>
#include <QTextStream>
#include <QList>
#include <QString>

#include "qtbtoolsdllexport.h"

class QtbPoint;
class QtbStringSection;

class QTBTOOLS_EXPORT QtbPoint2D
{
public:
  inline QtbPoint2D();
  inline QtbPoint2D(double x, double y);
  inline QtbPoint2D(const QtbPoint& p);

  double x() const { return _x; }
  double y() const { return _y; }
  void setX( double v ) { _x = v;}
  void setY( double v ) { _y = v;}
  void setValid(bool) {}
  bool isValid() const { return true; }
  void average(const QtbPoint2D& p ) { _y= 0.5*(_y+p._y); }

  // Copy operators
  inline QtbPoint2D& operator=(const QtbPoint& p);
  inline QtbPoint2D& operator=(const QtbPoint2D& p);
  inline void set(double xi, double yi);

  // Comparison operators
  inline bool operator< (const QtbPoint2D& p) const;
  inline bool operator== (const QtbPoint2D& p) const;
  bool operator<= (const QtbPoint2D& obj) const
    { return (*this == obj || *this < obj);};
  bool operator>  (const QtbPoint2D& obj) const
    {return (! (*this <= obj));} ;
  bool operator>= (const QtbPoint2D& obj) const
    {return (! (*this < obj));};
  bool operator!= (const QtbPoint2D& obj) const
    {return (! (*this == obj));};
  inline short compare (const QtbPoint2D& p) const;

  // I/O functions
  bool fromString(QtbStringSection str);
  QString toString(int precision=6) const;

  // Arithmetic operation
  inline QtbPoint2D operator+ (const QtbPoint2D& p) const;
  inline QtbPoint2D operator- (const QtbPoint2D& p) const;
  inline void operator+= (const QtbPoint2D& p);
  inline void operator-= (const QtbPoint2D& p);
  inline QtbPoint2D operator* (double mul) const;
  inline QtbPoint2D operator* (const QtbPoint2D& p) const;
  inline QtbPoint2D operator/ (double div) const;
  inline QtbPoint2D operator/ (const QtbPoint2D& p) const;
  inline void operator*= (const QtbPoint2D& p);
  inline void operator*= (double mul);
  inline void operator/= (const QtbPoint2D& p);
  inline void operator/= (double div);
  inline double scalarProduct (const QtbPoint2D& p) const;
  inline void interpole( double valX, const QtbPoint2D& p1, const QtbPoint2D& p2 );

  // Miscellaneous
  // Checks whether p is in the neighbourhood of point (+/- tolX,+/- tolY)
  bool isHit(const QtbPoint2D &p,double tolX, double tolY) const;
  // Calculate Distance to p
  double distanceTo(const QtbPoint2D &p) const;
  // Calculate the distance to a segment defined by p1 and p2
  double distanceToSegment(const QtbPoint2D &p1, const QtbPoint2D &p2) const;
  // Calculate azimuth to p from 0 to 2*pi (OX=0) anti-clockwize
  double azimuth(const QtbPoint2D &p) const;
  // Displace the point by distance d in direction dir
  void walkAzimuth(double dir,double d);
private:
  // Coordinates of the point
  double _x, _y;
};

#include "qtbpoint.h"

// Constructors
inline QtbPoint2D::QtbPoint2D()
{
  QTB_TRACE;
  _x=0.;
  _y=0.;
}

inline QtbPoint2D::QtbPoint2D(double xi, double yi) {
  QTB_TRACE;_x=xi; _y=yi;}

inline QtbPoint2D::QtbPoint2D(const QtbPoint& p)
{
  QTB_TRACE;
  _x=p.x();
  _y=p.y();
}
// Copy operators
inline QtbPoint2D& QtbPoint2D::operator= (const QtbPoint2D& p)
{
  QTB_TRACE;
  _x=p._x;
  _y=p._y;
  return *this;
}

inline QtbPoint2D& QtbPoint2D::operator= (const QtbPoint& p)
{
  QTB_TRACE;
  _x=p.x();
  _y=p.y();
  return *this;
}

inline void QtbPoint2D::set(double xi, double yi)
{
  QTB_TRACE;
  _x=xi;
  _y=yi;
}
// Comparison operators
inline bool QtbPoint2D::operator<  (const QtbPoint2D& p) const
{
  QTB_TRACE;
  if (_x < p._x) return true;
  else if (_x==p._x && _y<p._y) return true;
  return false;
}

inline bool QtbPoint2D::operator== (const QtbPoint2D& p) const
{
  QTB_TRACE;
  if ((_x==p._x) &&
      (_y==p._y)) return true;
  return false;
}

inline short QtbPoint2D::compare (const QtbPoint2D& p) const
{
  QTB_TRACE;
  if (_x<p._x) return(-1);
  else if (_x>p._x) return(1);
  else if (_y<p._y) return(-1);
  else if (_y>p._y) return(1);
  else return(0);
}
// Arithmetic operations
inline QtbPoint2D QtbPoint2D::operator+ (const QtbPoint2D& p) const
{
  QTB_TRACE;
  return QtbPoint2D(_x+p._x, _y+p._y);
}

inline QtbPoint2D QtbPoint2D::operator- (const QtbPoint2D& p) const
{
  return QtbPoint2D (_x-p._x, _y-p._y);
}

inline void QtbPoint2D::operator+= (const QtbPoint2D& p)
{
  QTB_TRACE;
  _x+=p._x;
  _y+=p._y;
}

inline void QtbPoint2D::operator-= (const QtbPoint2D& p)
{
  _x-=p._x;
  _y-=p._y;
}

inline void QtbPoint2D::operator*= (const QtbPoint2D& p)
{
  QTB_TRACE;
  _x*=p._x;
  _y*=p._y;
}

inline void QtbPoint2D::operator*= (double mul)
{
  QTB_TRACE;
  _x*=mul;
  _y*=mul;
}

inline void QtbPoint2D::operator/= (const QtbPoint2D& p)
{
  QTB_TRACE;
  _x/=p._x;
  _y/=p._y;
}

inline void QtbPoint2D::operator/= (double div)
{
  QTB_TRACE;
  _x/=div;
  _y/=div;
}

inline QtbPoint2D QtbPoint2D::operator* (double mul) const
{
  QTB_TRACE;
  return QtbPoint2D (_x*mul, _y*mul);
}

inline QtbPoint2D QtbPoint2D::operator* (const QtbPoint2D& p) const
{
  QTB_TRACE;
  return QtbPoint2D (_x*p._x, _y*p._y);
}

inline QtbPoint2D QtbPoint2D::operator/ (double div) const
{
  QTB_TRACE;
  return QtbPoint2D (_x/div, _y/div);
}

inline QtbPoint2D QtbPoint2D::operator/ (const QtbPoint2D& p) const
{
  QTB_TRACE;
  return QtbPoint2D (_x/p._x, _y/p._y);
}

inline double QtbPoint2D::scalarProduct (const QtbPoint2D& p) const
{
  QTB_TRACE;
  return _x*p._x+_y*p._y;
}

inline void QtbPoint2D::interpole( double valX, const QtbPoint2D& p1, const QtbPoint2D& p2 )
{
  QTB_TRACE;
  double factor = ( valX - p1.x() ) / ( p2.x()-p1.x() );
  setX( valX );
  setY( p1.y()+ ( p2.y()-p1.y() ) * factor );
}

class QTBTOOLS_EXPORT QtbPoint2DList : public QList<QtbPoint2D>
{
public:
  void regression( double& a, double& b ) const;
protected:
  double sumX() const;
  double sumY() const;
  double sumXY() const;
  double sumX2() const;
  double sumY2() const;
};

QTBTOOLS_EXPORT QTextStream& operator<< (QTextStream& s, const QList<QtbPoint2D>& plist);
QTBTOOLS_EXPORT QTextStream& operator>> (QTextStream& s, QList<QtbPoint2D>& plist);
QTBTOOLS_EXPORT QTextStream& operator<< (QTextStream& s, const QtbPoint2D& p);
QTBTOOLS_EXPORT QTextStream& operator>> (QTextStream& s, QtbPoint2D& p);
QTBTOOLS_EXPORT QDataStream& operator<< (QDataStream& s, const QtbPoint2D& p);
QTBTOOLS_EXPORT QDataStream& operator>> (QDataStream& s, QtbPoint2D& p);

#endif // QtbPOINT2D_H
