/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2002-04-15
**
***************************************************************************/

#include <math.h>

#include "qtbrect.h"
#include "qtbmemorychecker.h"

const QString QtbRect::xmlRectTag="Rect";

/*!
  \class QtbRect qtbrect.h
  \brief Storage for a rectangle with coordinates in double precision
  A rectangle is defined by top-left corner (x1,y1) and bottom-right corner (x2,y2).
  All function that changes the coordinates of these two points check for validity.

*/

QtbRect::QtbRect()
{
  QTB_TRACE;
  setLimits(0,0,0,0);
}

QtbRect::QtbRect(double nx1, double ny1, double nx2, double ny2)
{
  QTB_TRACE;
  setLimits(nx1,ny1,nx2,ny2);
}

QtbRect::QtbRect(const QtbPoint& p1, const QtbPoint& p2)
{
  QTB_TRACE;
  setLimits(p1.x(), p1.y(), p2.x(), p2.y());
}

/*!
  Adjust the top-left and the bottom-rigth corners to include the points (nx1,xy1)
  and (nx2,ny2).
*/
void QtbRect::setLimits (double nx1, double ny1, double nx2, double ny2)
{
  QTB_TRACE;
  double temp;
  _x1=nx1;
  _y1=ny1;
  _x2=nx2;
  _y2=ny2;
  if (_x2<_x1) {
    temp=_x2;
    _x2=_x1;
    _x1=temp;
  }
  if (_y2<_y1) {
    temp=_y2;
    _y2=_y1;
    _y1=temp;
  }
  /*if (_x2-_x1<1e-10) { // avoid minuscule rectangles
    _x2+=1e-10;
    _x1-=1e-10;
  }
  if (_y2-_y1<1e-10) {
    _y2+=1e-10;
    _y1-=1e-10;
  }*/
}

/*!
  Test whether the rectangle r is entirely inside the rectangle
*/
bool QtbRect::includes (const QtbRect & r) const
{
  QTB_TRACE;
  return (_x1 <= r._x1 && _y1 <= r._y1 && _x2 >= r._x2 && _y2 >= r._y2);
}

/*!
  Test whether the point is inside the rectangle
*/
bool QtbRect::includes (const QtbPoint & p) const
{
  QTB_TRACE;
  return (_x1 <= p.x() && p.x() <= _x2 && _y1 <= p.y() && p.y() <= _y2);
}

/*!
  Test whether the point is inside the rectangle
*/
bool QtbRect::includes (const QtbPoint2D & p) const
{
  QTB_TRACE;
  return (_x1 <= p.x() && p.x() <= _x2 && _y1 <= p.y() && p.y() <= _y2);
}

/*!
  Test whether the point is inside the speified area of the rectangle
*/
bool QtbRect::includes (const QtbPoint2D & p, Area inArea) const
{
  QTB_TRACE;
  bool inRect=(_x1 <= p.x() && p.x() <= _x2 && _y1 <= p.y() && p.y() <= _y2);
  if (!inRect) return false;
  double xLoc=(p.x()-_x1) / (_x2-_x1);
  double yLoc=(p.y()-_y1) / (_y2-_y1);
  switch (inArea) {
  case AllRect:
    return true;
  case AboveAscDiag:
    return yLoc>xLoc;
  case BelowAscDiag:
    return yLoc<xLoc;
  case AboveDescDiag:
    return yLoc>(1.0-xLoc);
  case BelowDescDiag:
    return yLoc<(1.0-xLoc);
  }
  return false;
}

void QtbRect::setX1(double nx1)
{
  QTB_TRACE;
  double temp;
  _x1=nx1;
  if (_x2<_x1) {
    temp=_x2;
    _x2=_x1;
    _x1=temp;
  }
}

void QtbRect::setY1(double ny1)
{
  QTB_TRACE;
  double temp;
  _y1=ny1;
  if (_y2<_y1) {
    temp=_y2;
    _y2=_y1;
    _y1=temp;
  }
}

void QtbRect::setX2(double nx2)
{
  QTB_TRACE;
  double temp;
  _x2=nx2;
  if (_x2<_x1) {
    temp=_x2;
    _x2=_x1;
    _x1=temp;
  }
}

void QtbRect::setY2(double ny2)
{
  QTB_TRACE;
  double temp;
  _y2=ny2;
  if (_y2<_y1) {
    temp=_y2;
    _y2=_y1;
    _y1=temp;
  }
}

/*!
  Change rectangle size to include the new point, if already included, does nothing.

  It also does nothing if the point is outside rectangle (-1e99, -1e99, 1e99, 1e99).
*/
void QtbRect::add(double x, double y)
{
  QTB_TRACE;
  if ( (x>-1e99 && x<_x1) || _x1<=-1e99) _x1=x;
  if ( (x<1e99 && x>_x2) || _x2>=1e99) _x2=x;
  if ( (y>-1e99 && y<_y1) || _y1<=-1e99) _y1=y;
  if ( (y<1e99 && y>_y2) || _y2>=1e99) _y2=y;
}

/*!
  Change rectangle size to include the new point, if already included, does nothing.

  It also does nothing if the point is outside rectangle (-1e99, -1e99, 1e99, 1e99).
*/
void QtbRect::add(const QtbPoint& p)
{
  QTB_TRACE;
  add( p.x(), p.y());
}

/*!
  Change rectangle size to include the new point, if already included, does nothing.

  It also does nothing if the point is outside rectangle (-1e99, -1e99, 1e99, 1e99).
*/
void QtbRect::add(const QtbPoint2D& p)
{
  QTB_TRACE;
  add( p.x(), p.y());
}

/*!
  Change rectangle size to include the new rectangle, if already included, does nothing.

  It also does nothing if the rectangle is outside rectangle (-1e99, -1e99, 1e99, 1e99).
*/
void QtbRect::add(const QtbRect& r)
{
  QTB_TRACE;
  add(r._x1, r._y1);
  add(r._x2, r._y2);
}

/*!
  Find intersection rect between two rects
*/
QtbRect QtbRect::intersect(const QtbRect& r) const
{
  QTB_TRACE;
  QtbRect rint;
  if (_x1<r._x1) rint._x1=r._x1; else rint._x1=_x1;
  if (_y1<r._y1) rint._y1=r._y1; else rint._y1=_y1;
  if (_x2<r._x2) rint._x2=_x2; else rint._x2=r._x2;
  if (_y2<r._y2) rint._y2=_y2; else rint._y2=r._y2;
  if (rint._x1>rint._x2) {
    rint._x2 = rint._x1;
  }
  if (rint._y1>rint._y2) {
    rint._y2 = rint._y1;
  }
  return rint;
}

/*!
  Test whether the segment defined by x1, y1 and x2, y2 is inside or crosses the rectangle
*/
bool QtbRect::includes ( double x1, double y1, double x2, double y2 ) const
{
  QTB_TRACE;
  if (x1>_x1 && x1<_x2 && y1>_y1 && y1<_y2) return true;
  if (x2>_x1 && x2<_x2 && y2>_y1 && y2<_y2) return true;
  // p1 is outside, p2 is outside
  // Try to find one intersection with sides of the rectangle
  if (x2!=x1) {
    double coef=(y2-y1)/(x2-x1);
    double yi=y1+coef*(_x2-x1);
    if (_y1<=yi && yi<=_y2 &&
        ((x1<x2 && x1<=_x2 && _x2<=x2) ||
         (x1>x2 && x2<=_x2 && _x2<=x1))) return true;
    yi=y1+coef*(_x1-x1);
    if (_y1<=yi && yi<=_y2 &&
        ((x1<x2 && x1<=_x1 && _x1<=x2) ||
         (x1>x2 && x2<=_x1 && _x1<=x1))) return true;
  } else {
    if (_x1<=x1 && x1<=_x2 && 
        ((y1<=y2 && y1<_y1 && y2>_y2) ||
         (y1>y2 && y2<_y1 && y1>_y2))) return true;
    else return false;
  }
  if (y2!=y1) {
    double coef=(x2-x1)/(y2-y1);
    double xi=x1+coef*(_y2-y1);
    if (_x1<=xi && xi<=_x2 &&
        ((y1<y2 && y1<=_y2 && _y2<=y2) ||
         (y1>y2 && y2<=_y2 && _y2<=y1))) return true;
    xi=x1+coef*(_y1-y1);
    if (_x1<=xi && xi<=_x2 &&
        ((y1<y2 && y1<=_y1 && _y1<=y2) ||
         (y1>y2 && y2<=_y1 && _y1<=y1))) return true;
  } else {
    if (_y1<=y1 && y1<=_y2 && 
        ((x1<=x2 && x1<_x1 && x2>_x2) ||
         (x1>x2 && x2<_x1 && x1>_x2))) return true;
    else return false;
  }
  return false;
}

/*!
  \fn bool QtbRect::includes (const QtbPoint2D& p1, const QtbPoint2D& p2) const
  Test whether the segment defined by p1 and p2 is inside or crosses the rectangle
*/

/*!
  Enlarge rectange by \a dx and \a dy in both direction. Add or multiply according to the scale type.
  Inversed scales are considered like linear scales
*/
void QtbRect::enlarge( double dx, double dy, Qtb::SamplingOption xScale, Qtb::SamplingOption yScale )
{
  QTB_TRACE;
  switch(xScale) {
  case Qtb::LogScale:
    _x1 *= 1.0-dx;
    _x2 *= 1.0+dx;
    break;
  case Qtb::InversedScale:
    _x1 = 1.0 / ( 1.0/_x1 + 1.0/dx );
    _x2 = 1.0 / ( 1.0/_x2 - 1.0/dx );
    break;
  default:
    _x1-=dx;
    _x2+=dx;
    break;
  }
  switch(yScale) {
  case Qtb::LogScale:
    _y1*=1.0-dy;
    _y2*=1.0+dy;
    break;
  case Qtb::InversedScale:
    _y1 = 1.0 / ( 1.0/_y1 + 1.0/dy );
    _y2 = 1.0 / ( 1.0/_y2 - 1.0/dy );
    break;
  default:
    _y1-=dy;
    _y2+=dy;
    break;
  }
}

/*!
  Enlarge rectangle by width and height times \a ratio times in both directions.
*/
void QtbRect::enlarge( double ratio, Qtb::SamplingOption xScale, Qtb::SamplingOption yScale )
{
  QTB_TRACE;
  double s1, s2;
  switch(xScale) {
  case Qtb::LogScale:
    s1 = pow(_x2/_x1, ratio );
    _x1 /= s1;
    _x2 *= s1;
    break;
  case Qtb::InversedScale:
    {
      double t1 = 1/_x1;
      double t2 = 1/_x2;
      s1 = ratio*(t1-t2);
      _x1 = 1/(t1+s1);
      _x2 = 1/(t2-s1);
    }
    /*s1 = ratio*(_x2-_x1);
    s2 = _x1*_x2;
    _x1 = s2/(_x2+s1);
    _x2 = s2/(_x1-s1);*/
    break;
  default:
    s1 = ratio*(_x2-_x1);
    _x1 -= s1;
    _x2 += s1;
    break;
  }
  switch(yScale) {
  case Qtb::LogScale:
    s1 = pow(_y2/_y1, ratio );
    _y1 /= s1;
    _y2 *= s1;
    break;
  case Qtb::InversedScale:
    s1 = ratio*(_y2-_y1);
    s2 = _y1*_y2;
    _y1 = s2/(_x2+s1);
    _y2 = s2/(_x1-s1);
    break;
  default:
    s1 = ratio*(_y2-_y1);
    _y1 -= s1;
    _y2 += s1;
    break;
  }
}

/*!
  Find intersection points between a rect and a segment
*/
void QtbRect::intersect(QtbPoint& p1, QtbPoint& p2) const
{
  QTB_TRACE;
  QtbRect seg(p1.x(), p1.y(), p2.x(), p2.y());
  double c,dx,dy,x,y;
  dx=p2.x()-p1.x();
  dy=p2.y()-p1.y();
  if (fabs(dx)>1e-15) {
    c=dy/dx;
    y=p1.y()+c*(_x1-p1.x());
    if ((y>=_y1 && y<=_y2) && (y>=seg._y1 && y<=seg._y2)) {
      if (p1.x()<=_x1) {p1.setX(_x1);p1.setY(y);}
      else {p2.setX(_x1);p2.setY(y);}
    }
    y=p1.y()+c*(_x2-p1.x());
    if ((y>=_y1 && y<=_y2) && (y>=seg._y1 && y<=seg._y2)) {
      if (p1.x()>=_x2) {p1.setX(_x2);p1.setY(y);}
      else {p2.setX(_x2);p2.setY(y);}
    }
  }
  if (fabs(dy)>1e-15) {
    c=dx/dy;
    x=p1.x()+c*(_y1-p1.y());
    if ((x>=_x1 && x<=_x2) && (x>=seg._x1 && x<=seg._x2)) {
      if (p1.y()<=_y1) {p1.setX(x);p1.setY(_y1);}
      else {p2.setX(x);p2.setY(_y1);}
    }
    x=p1.x()+c*(_y2-p1.y());
    if ((x>=_x1 && x<=_x2) && (x>=seg._x1 && x<=seg._x2)) {
      if (p1.y()>=_y2) {p1.setX(x);p1.setY(_y2);}
      else {p2.setX(x);p2.setY(_y2);}
    }
  }
}

void QtbRect::printDebug() const
{
  printf("(%lg,%lg) to (%lg,%lg)\n",_x1, _y1, _x2, _y2);
}

void QtbRect::xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const
{
  QTB_TRACE;
  Q_UNUSED(context);
  writeProperty(s, "x",_x1);
  writeProperty(s, "y",_y1);
  writeProperty(s, "w",_x2-_x1);
  writeProperty(s, "h",_y2-_y1);
}

QtbXMLMember QtbRect::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  if (tag=="x1") return QtbXMLMember(0);
  else if (tag=="y1") return QtbXMLMember(1);
  else if (tag=="x2") return QtbXMLMember(2);
  else if (tag=="y2") return QtbXMLMember(3);
  else return QtbXMLMember(QtbXMLMember::Unknown);
}  

bool QtbRect::xml_setProperty(XML_SETPROPERTY_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  double val=content.toDouble();
  switch (memberID) {
  case 0:
    _x1=val;
    return true;
  case 1:
    _y1=val;
    return true;
  case 2:
    _x2=val;
    return true;
  case 3:
    _y2=val;
    return true;
  default:
    return false;
  }
}
