/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-21
**
***************************************************************************/

#include "qtbstatvalue.h"
#include "qtbvalue.h"
#include "qtbcoreapplication.h"
#include "qtbmemorychecker.h"

const QString QtbStatValue::xmlStatValueTag = "StatValue";
const QString QtbStatValueVector::xmlStatValueVectorTag = "StatValueVector";

QDataStream& operator>> (QDataStream& s, QtbStatValue& p)
{
  QTB_TRACE;
  double v;
  s >> v;
  p.setMean(v);
  s >> v;
  p.setStddev(v);
  s >> v;
  p.setWeight(v);
  return s;
}

QDataStream& operator<< (QDataStream& s, const QtbStatValue& p)
{
  QTB_TRACE;
  s << p.mean() << p.stddev() << p.weight();
  return s;
}

void QtbStatValue::xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const
{
  QTB_TRACE;
  Q_UNUSED(context);
  writeProperty(s,"mean",_mean);
  writeProperty(s,"stddev",_stddev);
  writeProperty(s,"weight",_weight);
  writeProperty(s,"valid",_valid);
}

QtbXMLMember QtbStatValue::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  if (tag=="mean") return QtbXMLMember(0);
  else if (tag=="stddev") return QtbXMLMember(1);
  else if (tag=="weight") return QtbXMLMember(2);
  else if (tag=="valid") return QtbXMLMember(3);
  else return false;
}

bool QtbStatValue::xml_setProperty(XML_SETPROPERTY_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  switch (memberID) {
  case 0: _mean=content.toDouble(); return true;
  case 1: _stddev=content.toDouble(); return true;
  case 2: _weight=content.toDouble(); return true;
  case 3: _valid=content.toBool(); return true;
  default: return false;
  }
}

QString QtbStatValue::misfitTypeString( MisfitType type )
{
  QTB_TRACE;
  switch (type) {
  case L1:
    return "L1";
  case L1_Normalized:
    return "L1_Normalized";
  case L2:
    return "L2";
  case L2_Normalized:
    return "L2_Normalized";
  case Akaike:
    return "Akaike";
  case AkaikeFewSamples:
    return "AkaikeFewSamples";
  }
  return QString::null;
}

QtbStatValue::MisfitType QtbStatValue::misfitType( QString type )
{
  QTB_TRACE;
  if (type.count()>1) {
    switch( type[1].unicode() ) {
    case '1':
      if (type=="L1") return L1; else return L1_Normalized;
    case '2':
      if (type=="L2") return L2; else return L2_Normalized;
    case 'k':
      if (type=="Akaike") return Akaike; else return AkaikeFewSamples;
      break;
    default:
      break;
    }
  }
  return L2_Normalized;
}

/*!
  Return a misfit value comparing \a val with the mean and stddev.

  \a nValues is incremented by one if the value and this point are valid.
  \a nData is incremented by one if this point is valid.

  If \a min is greater than 0, the difference between \a val and mean (eventually normalized) is clipped to \a min.
*/
double QtbStatValue::misfit( int& nValues, int& nData, const QtbValue& val, MisfitType type, double min ) const
{
  QTB_TRACE;
  if ( isValid() ) {
    nData++;
    if ( val.isValid() ) {
      nValues++;
      double diff;
      switch (type) {
      case L1:
        diff = fabs( _mean - val.value() );
        if (diff<min) diff = min;
        return diff;
      case L1_Normalized:
        if ( _stddev>0.0 ) {
          diff = fabs( ( _mean - val.value() ) / _stddev );
        } else {
          diff = fabs( ( _mean - val.value() ) / _mean );
        }
        if (diff<min) diff = min;
        return diff;
      case L2:
        diff = fabs( _mean - val.value() );
        if (diff<min) diff = min;
        return diff*diff;
      case L2_Normalized:
        if ( _stddev>0.0 ) {
          diff = fabs( ( _mean - val.value() ) / _stddev );
        } else {
          diff = fabs( ( _mean - val.value() ) / _mean );
        }
        if (diff<min) diff = min;
        return diff*diff;
      case Akaike:
      case AkaikeFewSamples:
        diff = fabs( _mean - val.value() );
        if (diff<min) diff = min;
        return diff*diff;
      }
      return 0.0;
    } else {
      return 0.0;
    }
  } else {
    return 0.0;
  }
}

QVector<double> * QtbStatValueVector::meanVector() const
{
  QVector<double> * list=new QVector<double>;
  for (int i=0;i<size();i++) list->append(at(i).mean());
  return list;
}

QVector<double> * QtbStatValueVector::stddevVector() const
{
  QVector<double> * list=new QVector<double>;
  for (int i=0;i<size();i++) list->append(at(i).stddev());
  return list;
}

QVector<double> * QtbStatValueVector::weightVector() const
{
  QVector<double> * list=new QVector<double>;
  for (int i=0;i<size();i++) list->push_back(at(i).weight());
  return list;
}

QtbXMLMember QtbStatValueVector::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes)
  Q_UNUSED(context);
  if (tag=="StatValue" || tag=="DataPoint") { // backwards compatibility
    append(QtbStatValue());
    return QtbXMLMember(&last());
  } else return QtbXMLMember(QtbXMLMember::Unknown);
}

void QtbStatValueVector::xml_writeChildren(XML_WRITECHILDREN_ARGS) const
{
  QTB_TRACE;
  static const QString key("index");
  QtbXMLSaveAttributes att;
  att.add(key);
  QString& value = att.value(0);
  for (int i=0;i<size();i++) {
    value = QString::number(i);
    at(i).xml_save(s, context, att);
  }
}
