/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-21
**
***************************************************************************/

#ifndef QTB_STATVALUE_H
#define QTB_STATVALUE_H

#include <math.h>

#include <QDataStream>
#include <QVector>

#include "qtbxmlclass.h"
#include "qtbtoolsdllexport.h"

class QtbValue;

class QTBTOOLS_EXPORT QtbStatValue : public QtbXMLClass
{
public:
  inline QtbStatValue();
  inline QtbStatValue ( double mean, double stddev = 0.0, double weight = 1.0, bool valid = true );

  virtual const QString& xml_tagName() const { return xmlStatValueTag;}
  static const QString xmlStatValueTag;

  inline void operator=( const QtbStatValue& o );
  inline bool operator==( const QtbStatValue& o ) const;

  void setMean( double v ) { _mean = v;}
  void setStddev( double v ) { _stddev = v;}
  void setWeight( double v ) { _weight = v;}

  double mean() const { return _mean;}
  double stddev() const { return _stddev;}
  double weight() const { return _weight;}

  void setValid( bool f ) { _valid = f; }
  bool isValid() const { return _valid; }

  enum MisfitType { L1, L1_Normalized, L2, L2_Normalized, Akaike, AkaikeFewSamples };
  static QString misfitTypeString( MisfitType type );
  static MisfitType misfitType( QString type );

  double misfit( int& nValues, int& nData, const QtbValue& value, MisfitType type, double min ) const;

  inline void average( const QtbStatValue& p );
protected:
  double _mean;
  double _stddev;
  double _weight;
  bool _valid;
protected:
  virtual void xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const;
  virtual QtbXMLMember xml_member( XML_MEMBER_ARGS );
  virtual bool xml_setProperty( XML_SETPROPERTY_ARGS );
};

inline QtbStatValue::QtbStatValue()
{
  QTB_TRACE;
  _mean = 0.0;
  _stddev = 0.0;
  _weight = 1.0;
  _valid = true;
}

inline QtbStatValue::QtbStatValue( double mean, double stddev, double weight, bool valid )
{
  QTB_TRACE;
  _mean = mean;
  _stddev = stddev;
  _weight = weight;
  _valid = valid;
}

inline void QtbStatValue::operator=( const QtbStatValue& o )
{
  QTB_TRACE;
  _mean = o._mean;
  _stddev = o._stddev;
  _weight = o._weight;
  _valid = o._valid;
}

inline bool QtbStatValue::operator==( const QtbStatValue& o ) const
{
  return _mean==o._mean && _stddev == o._stddev && _weight == o._weight && _valid == o._valid;
}

inline void QtbStatValue::average(const QtbStatValue& o )
{
  QTB_TRACE;
  if(isValid()) {
    if (o.isValid()) {
      double mu2 = _mean*_mean;
      double wt = _weight + o._weight;
      _mean = ( _weight*_mean + o._weight*o._mean ) / wt;
      double sigma2 = _weight * mu2;
      if (_weight>1.0)
        sigma2 += ( _weight -1.0 )* _stddev*_stddev;
      else
        sigma2 += _weight* _stddev*_stddev;
      sigma2 += o._weight * o._mean * o._mean;
      if (o._weight>1.0)
        sigma2 += (o._weight-1.0) * o._stddev*o._stddev;
      else
        sigma2 += o._weight * o._stddev*o._stddev;
      sigma2 -= _mean*_mean * wt;
      sigma2 /= wt-1;
      _stddev = sqrt(sigma2);
      _weight = wt;
    }
  } else if(o.isValid()) {
    *this = o;
  }
}

QTBTOOLS_EXPORT QDataStream& operator<< ( QDataStream& s, const QtbStatValue& p );
QTBTOOLS_EXPORT QDataStream& operator>> ( QDataStream& s, QtbStatValue& p );

class QTBTOOLS_EXPORT QtbStatValueVector : public QVector<QtbStatValue>,
      public QtbXMLClass
{
public:
  QVector<double> * meanVector() const;
  QVector<double> * stddevVector() const;
  QVector<double> * weightVector() const;
protected:
  virtual const QString& xml_tagName() const { return xmlStatValueVectorTag;}
  static const QString xmlStatValueVectorTag;
  virtual void xml_writeChildren( XML_WRITECHILDREN_ARGS ) const;
  virtual QtbXMLMember xml_member( XML_MEMBER_ARGS );
};

#endif
