/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-21
**
***************************************************************************/

#include "qtbstringsection.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbStringSection qtbstringsection.h
  \brief Section of a QString

  Usefull for parser
*/

/*!
  Default separators: lf, cr, tab and space
*/
const QString QtbStringSection::defaultSeparators = "\n\r\t ";
/*!
  Default static null string
*/
const QtbStringSection QtbStringSection::null;

/*!
  Global Hash function to store QtbStringSection into QHash

  This code was copied from qHash(QString key)
*/
uint qHash( QtbStringSection key )
{
  QTB_TRACE;
  const QChar * p = key.data();
  int n = key.size();
  uint h = 0;
  uint g;

  while ( n-- ) {
    h = ( h << 4 ) + ( *p++ ).unicode();
    if ( ( g = ( h & 0xf0000000 ) ) != 0 )
      h ^= g >> 23;
    h &= ~g;
  }
  return h;
}

int QtbStringSection::size( const QChar * str )
{
  QTB_TRACE;
  int i;
  for ( i = 0; str[i]!= 0; i++ ) {}
  return i;
}

int QtbStringSection::compare( const QChar * str1, const QChar * str2, int n )
{
  QTB_TRACE;
  if ( ! str1 ) { // Put invalid strings after valid ones
    if ( ! str2 ) {
      return 0;
    } else {
      return 1;
    }
  } else if ( ! str2 ) {
    return -1;
  }
  for ( int i = 0;i < n;i++ ) {
    if ( str1[ i ] < str2[ i ] ) return -1;
    else if ( str1[ i ] > str2[ i ] ) return 1;
  }
  return 0;
}

int QtbStringSection::compare( const QtbStringSection &o ) const
{
  QTB_TRACE;
  int r;
  if ( _size < o._size ) {
    r = compare( _str, o._str, _size );
    if ( r == 0 ) return -1; else return r;
  } else if ( _size > o._size ) {
    r = compare( _str, o._str, o._size );
    if ( r == 0 ) return 1; else return r;
  } else return compare( _str, o._str, _size );
}

const QChar * QtbStringSection::find( QChar c ) const
{
  const QChar * ptr = _str;
  const QChar * ptrE = _str + _size;
  while ( ptr < ptrE ) {
    if ( *ptr == c ) break;
    ptr++;
  }
  return ptr;
}

int QtbStringSection::find( const QtbStringSection target, bool reverse ) const
{
  QTB_TRACE;
  if ( !isValid() ) return -1;
  const QChar * ptrS = _str;
  const QChar * ptrE = _str + _size;
  int nTarget = target.size(), i, n;
  const QChar * targetPtr = target.data();
  QChar initChar;
  if ( reverse ) {
    targetPtr += nTarget - 1;
    initChar = *targetPtr;
    ptrE--;
    while ( ptrE >= ptrS ) {
      if ( initChar == *ptrE ) {
        n = ptrS - ptrE;
        if ( n < -nTarget ) n = -nTarget;
        for ( i = -1; targetPtr[i] == ptrE[i] && i > n; i-- ) {}
        if ( targetPtr[i] == 0 ) return ptrE -_str - nTarget;
      }
      ptrE--;
    }
  } else {
    initChar = *targetPtr;
    while ( ptrS < ptrE ) {
      if ( initChar == *ptrS ) {
        n = ptrE - ptrS;
        if ( n > nTarget ) n = nTarget;
        for ( i = 1; targetPtr[i] == ptrS[i] && i < n; i++ ) {}
        if ( targetPtr[i] == 0 ) return ptrS -_str;
      }
      ptrS++;
    }
  }
  return -1;
}

/*!
  Remove white space at the end and at the beginning
*/
QtbStringSection& QtbStringSection::trimmed()
{
  QTB_TRACE;
  if ( !isValid() ) return *this;
  const QChar * ptrE = _str + _size;
  while ( _str < ptrE ) {
    if ( !_str->isSpace() ) break;
    _str++;
  }
  while ( _str < ptrE ) {
    ptrE--;
    if ( !ptrE->isSpace() ) {
      ptrE++;
      break;
    }
  }
  if ( _str < ptrE ) _size = ptrE - _str; else _size = 0;
  return *this;
}

/*!
  Extract the first non-blank field of this string starting at \a ptr.
  If \a ptr = 0, it is set to the first character of the string.
  It returns an invalid QtbStringSection if no more field can be found.
  A QtbStringSection is returned starting at first character of the field and width
  the correct length. No data is re-allocated nor copied.
*/
QtbStringSection QtbStringSection::firstField( const QChar *& ptr, const QString& sep, bool skipEmpty ) const
{
  QTB_TRACE;
  if ( !isValid() ) return QtbStringSection();
  const QChar * ptrE = _str + _size;
  if ( !ptr ) ptr = _str;
  if (skipEmpty) {
    while ( ptr < ptrE ) {
      if ( !sep.contains( *ptr ) ) break;
      ptr++;
    }
  }
  if ( ptr < ptrE ) {
    QtbStringSection field( ptr, 0 );
    while ( ptr < ptrE ) {
      if ( sep.contains( *ptr ) ) break;
      ptr++;
    }
    field.setEnd( ptr );
    return field;
  } else return QtbStringSection();
}

/*!
  Same as above, except if a '\\n' or '\\r' is encountered before the first
  field, \a ptr is moved to the end of the line and it returns an invalid string.

  This function is used when blank lines are not tolerated or after a first
  call to firstField().

  Here is an example of the interaction with firstField() to parse a buffer
  with data ordered in columns.

  \code
      const QString sep="= \t\n\r";
      int i;
      QtbStringSection f;
      const QChar * ptr=0;
      for (i=0;i<n;i++) {
        f=content.firstField(ptr, sep);
        if (f.isValid()) x1=f.toDouble(); else break;
        f=content.nextField(ptr, sep);
        if (f.isValid()) x2=f.toDouble(); else break;
        f=content.nextField(ptr, sep);
        if (f.isValid()) x3=f.toDouble(); else break;
      }
      if (i<n) fprintf(stderr,"Error");
  \endcode
*/
QtbStringSection QtbStringSection::nextField( const QChar *& ptr, const QString& sep, bool skipEmpty ) const
{
  QTB_TRACE;
  if ( !isValid() ) return QtbStringSection();
  const QChar * ptrE = _str + _size;
  if ( !ptr ) ptr = _str;
  if (skipEmpty) {
    while ( ptr < ptrE ) {
      if ( *ptr == '\r' || *ptr == '\n' ) return QtbStringSection();
      else if ( !sep.contains( *ptr ) ) break;
      ptr++;
    }
  } else {
    ptr++; // Go to character just after separator
  }
  if ( ptr < ptrE ) {
    QtbStringSection field( ptr, 0 );
    while ( ptr < ptrE ) {
      if ( sep.contains( *ptr ) ) break;
      ptr++;
    }
    field.setEnd( ptr );
    return field;
  } else return QtbStringSection();
}
