/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-12-26
**
***************************************************************************/

#include <QStringList>
#include <QMutexLocker>
#include <QSet>

#include "qtbthread.h"
#include "qtbtraceinfo.h"

/*!
  \class QtbThread qtbthread.h
  \brief Thread with an internal stack to record trace points.
*/

QMutex QtbThread::_threadsMutex;
QList<QtbThread *> QtbThread::_threads;
QSet<QtbThread *> QtbThread::_runningThreads;

QtbThread::QtbThread( QObject * parent )
   : QThread(parent)
{
  QMutexLocker ml(&_threadsMutex);
  _threads.append(this);
  _sleeping = false;
}

QtbThread::~QtbThread()
{
  QMutexLocker ml(&_threadsMutex);
  int index = _threads.indexOf(this);
  if (index>=0) {
    _threads.removeAt(index);
  }
}

QStringList QtbThread::threadNames()
{
  QMutexLocker ml(&_threadsMutex);
  QStringList list;
  list.append("main");
  for(QList<QtbThread *>::iterator it = _threads.begin(); it!=_threads.end(); it++ ) {
    list.append( (*it)->objectName() );
  }
  return list;
}

#ifdef QTB_TRACE_ENABLED
QList< QStack<const QtbTraceStamp *> * > QtbThread::stacks()
{
  QList< QStack<const QtbTraceStamp *> * > list;
  list.append( QtbTrace::mainStack() );
  for(QList<QtbThread *>::iterator it = _threads.begin(); it!=_threads.end(); it++ ) {
    list.append( (*it)->stack() );
  }
  return list;
}
#endif

QList< QStack<QtbTraceBug *> * > QtbThread::bugStacks()
{
  QList< QStack<QtbTraceBug *> * > list;
  list.append( QtbTraceBug::mainStack() );
  for(QList<QtbThread *>::iterator it = _threads.begin(); it!=_threads.end(); it++ ) {
    list.append( (*it)->bugStack() );
  }
  return list;
}

/*!
  An implementation that tries to take already running threads into
  account. The return number is always less or equal to the number of
  physical cores. Never returns a number less than 1.
*/
int QtbThread::idealThreadCount()
{
  int nThreads = QThread::idealThreadCount();
  if (nThreads<2) return 1; // Exit as soon as possible for single processor machines
  QSet<QtbThread *>::iterator it;
  QMutexLocker ml(&_threadsMutex);
  for(it = _runningThreads.begin(); it!=_runningThreads.end(); it++) {
    if (!(*it)->sleeping()) nThreads--;
  }
  return nThreads>0 ? nThreads : 1;
}

void QtbThread::start()
{
  _threadsMutex.lock();
  _runningThreads.insert( this );
  _threadsMutex.unlock();
  QThread::start();
  _threadsMutex.lock();
  _runningThreads.remove( this );
  _threadsMutex.unlock();
}

void QtbThread::sleep( unsigned long secs )
{
  QtbThread * t = currentThread();
  _threadsMutex.unlock();
  if (_runningThreads.contains( t ) ) {
    _threadsMutex.unlock();
    t->setSleeping( true );
    QThread::sleep( secs );
    t->setSleeping( false );
  } else {
    QThread::sleep( secs );
  }
}

void QtbThread::msleep( unsigned long msecs )
{
  QtbThread * t = currentThread();
  _threadsMutex.lock();
  if ( msecs>10 && _runningThreads.contains( t ) ) {
    _threadsMutex.unlock();
    t->setSleeping( true );
    QThread::msleep( msecs );
    t->setSleeping( false );
  } else {
    _threadsMutex.unlock();
    QThread::msleep( msecs );
  }
}

void QtbThread::usleep( unsigned long usecs )
{
  QtbThread * t = currentThread();
  _threadsMutex.unlock();
  if ( usecs>10000 && _runningThreads.contains( t ) ) {
    _threadsMutex.unlock();
    t->setSleeping( true );
    QThread::usleep( usecs );
    t->setSleeping( false );
  } else {
    _threadsMutex.unlock();
    QThread::usleep( usecs );
  }
}
