/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2002-09-18
**
***************************************************************************/

#include <math.h>

#include <QRegExp>

#include "qtbtime.h"
#include "qtbcoreapplication.h"
#include "qtbtrace.h"
#include "qtbmemorychecker.h"

/*!
  Convert \a sec seconds into a string with format ".w.d.h.n.s"
*/
QString QtbTime::secondsToTime( double sec, int numberPrecision )
{
  QTB_TRACE;
  QString str;
  if (sec<0) {
    sec=-sec;
    str+="-";
  }
  if (sec>=604800.0) {
    int week=(int)floor(sec/604800.0);
    sec=sec-(double)week*604800.0;
    str+=QString::number(week);
    str+="w";
  }
  if (sec>=86400.0) {
    int day=(int)floor(sec/86400.0);
    sec=sec-(double)day*86400.0;
    str+=QString::number(day);
    str+="d";
  }
  if (sec>=3600.0) {
    int hour=(int)floor(sec/3600.0);
    sec=sec-(double)hour*3600.0;
    str+=QString::number(hour);
    str+="h";
  }
  if (sec>=60.0) {
    int min=(int)floor(sec/60.0);
    sec=sec-(double)min*60.0;
    str+=QString::number(min);
    str+="m";
  }
  if (sec>0.0) {
    str+=QString::number(sec,'f',numberPrecision);
    str+="s";
  } else if (sec==0.0 && str.isEmpty()) {
    str = "0s";
  }
  return str;
}

double QtbTime::timeToSecondsError(QString str, int pos)
{
  QTB_TRACE;
  qtbApp << translate("QtbTime","Error parsing time '%1' at position %2").arg(str).arg(pos+1) << endl;
  return 0.0;
}

inline bool QtbTime::timeToSecondsSign( const QChar * p, int& i, int n )
{
  QTB_TRACE;
  for( i = 0; i<n; i++ ) {
    switch(p[i].unicode()) {
    case ' ':
      break;
    case '-':
      i++;
      return true;
    default:
      return false;
    }
  }
  return false;
}

/*!
  Convert a string with format ".w.d.h.n.s" or "h:m:s" into a number of seconds
*/
double QtbTime::timeToSeconds(QString str)
{
  QTB_TRACE;
  const QChar * p = str.data();
  int n = str.count();
  QString number;
  bool ok;
  int vali;
  double vald;
  double sec = 0.0;
  enum ColonState { Unknown, NoColon, Hours, Minutes, Seconds };
  ColonState colonState = Unknown;
  // Skip first blanks and check sign
  int i;
  bool neg = timeToSecondsSign( p, i, n );
  for( ; i<n; i++ ) {
    switch(p[i].unicode()) {
    case ' ':
      break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
    case '.':
      number+=p[i];
      break;
    case ':':
      if ( number.isEmpty() ) return timeToSecondsError( str, i );
      switch( colonState ) {
      case Unknown:
      case Hours:
        vali = number.toInt(&ok);
        if (!ok) return timeToSecondsError(str, i);
        sec+=vali*3600.0;
        colonState = Minutes;
        break;
      case Minutes:
        vali = number.toInt(&ok);
        if (!ok) return timeToSecondsError(str, i);
        sec+=vali*60.0;
        colonState = Seconds;
        break;
      case Seconds:
        vald = number.toDouble(&ok);
        if (!ok) return timeToSecondsError(str, i);
        sec+=vald;
        colonState = Seconds;
        break;
      case NoColon:
        return timeToSecondsError( str, i );
      }
      number.clear();
      break;
    case 'w':
      switch (colonState) {
      case NoColon:
        break;
      case Unknown:
        colonState = NoColon;
        break;
      default:
        return timeToSecondsError(str, i);
      }
      vali = number.toInt(&ok);
      if (!ok) return timeToSecondsError(str, i);
      sec+=vali*604800.0;
      number.clear();
      break;
    case 'd':
      switch (colonState) {
      case NoColon:
        break;
      case Unknown:
        colonState = NoColon;
        break;
      default:
        return timeToSecondsError(str, i);
      }
      vali = number.toInt(&ok);
      if (!ok) return timeToSecondsError(str, i);
      sec+=vali*86400.0;
      number.clear();
      break;
    case 'h':
      switch (colonState) {
      case NoColon:
        break;
      case Unknown:
        colonState = NoColon;
        break;
      default:
        return timeToSecondsError(str, i);
      }
      vali = number.toInt(&ok);
      if (!ok) return timeToSecondsError(str, i);
      sec+=vali*3600.0;
      number.clear();
      break;
    case 'm':
      switch (colonState) {
      case NoColon:
        break;
      case Unknown:
        colonState = NoColon;
        break;
      default:
        return timeToSecondsError(str, i);
      }
      vali = number.toInt(&ok);
      if (!ok) return timeToSecondsError(str, i);
      sec+=vali*60.0;
      number.clear();
      break;
    case 's':
      switch (colonState) {
      case NoColon:
        break;
      case Unknown:
        colonState = NoColon;
        break;
      default:
        return timeToSecondsError(str, i);
      }
      vald = number.toDouble(&ok);
      if (!ok) return timeToSecondsError(str, i);
      sec+=vald;
      number.clear();
      break;
    default:
      return timeToSecondsError(str, i);
    }
  }
  if (!number.isEmpty()) {
    vald = number.toDouble(&ok);
    if (!ok) return timeToSecondsError(str, i);
    sec+=vald;
  }
  return neg ? -sec : sec;
}

TIME QtbTime::addSecs(QDateTime dt,double secs)
{
  QTB_TRACE;
  // Positive number rounded to inferior integer, neg to superior
  int nSecs=(int)floor(secs);
  // msecs is allways positive and less than 1
  double msecs=secs-nSecs;
  dt=dt.addSecs(nSecs);
  TIME t;
  t.yr=dt.date().year();
  t.mo=dt.date().month();
  t.day=dt.date().day();
  t.hr=dt.time().hour();
  t.mn=dt.time().minute();
  t.sec=(float)dt.time().second()+msecs;
  return t;
}

/*!
  Return a string with format compatible with cap (from Matthias Ohrnberger)
*/
QString QtbTime::addSecs(QDateTime dt,double secs,QString format)
{
  QTB_TRACE;
  // Positive number rounded to inferior integer, neg to superior
  int nSecs=(int)floor(secs);
  // msecs is allways positive and less than 1000
  int msecs=(int) floor((secs-nSecs)*1000+0.5);
  if (msecs>999) {msecs=0; nSecs++;}
  dt=dt.addSecs(nSecs);
  QString str=dt.toString(format);
  if (msecs<10) str+=QString(".00%1").arg(msecs);
  else if (msecs<100) str+=QString(".0%1").arg(msecs);
  else str+=QString(".%1").arg(msecs);
  return str;
}
