/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-11-24
**
***************************************************************************/

#include "qtbtreeitem.h"
#include "qtbglobal.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbTreeItem qtbtreeitem.h
  \brief Brief description of class still missing

  Full description of class still missing
*/

const QString QtbTreeItem::xmlTreeItemTag = "TreeItem";

/*!
  Description of constructor still missing
*/
QtbTreeItem::QtbTreeItem( QtbTreeItem * parent )
{
  QTB_TRACE;
  _modified=false;
  _folder=false;
  _removed=false;
  // As we are in constructor, parent cannot be this and children list is empty => no circular
  if (parent && parent->isFolder()) {
    parent->addChild( this );
    _parent = parent;
  } else {
    _parent = 0;
  }
}

/*!
  Destructor remove all children
*/
QtbTreeItem::~QtbTreeItem()
{
  QTB_TRACE;
  qDeleteAll(_children);
}

/*!
  Mark this item to be removed or not
*/
void QtbTreeItem::setRemoved( bool e )
{
  QTB_TRACE;
  if (_removed != e) {
    _removed=e;
    _modified=true;
    if (_parent) _parent->_modified = true;
  }
}

/*!
  Remove child, does nothing if child is not marked as removed
*/
void QtbTreeItem::removeChild( int index )
{
  QTB_TRACE;
  QtbTreeItem * item = _children[index];
  if (item->_removed) {
    _children.removeAt(index);
    delete item;
  }
}

/*!
  Remove all children and delete them.
*/
void QtbTreeItem::removeChildren()
{
  QTB_TRACE;
  qDeleteAll(_children);
  _children.clear();
}

/*!
  Remove all children but do not delete them.
*/
QList<QtbTreeItem *> QtbTreeItem::takeChildren()
{
  QTB_TRACE;
  QList<QtbTreeItem *> l = _children;
  _children.clear();
  return l;
}

bool QtbTreeItem::isParentRemoved() const
{
  QTB_TRACE;
  if (_parent) {
    if (_removed)
      return true;
    else
      return _parent->isParentRemoved();
  } else
    return _removed;
}

/*!
  Change parent of this item
*/
void QtbTreeItem::setParent( QtbTreeItem * parent )
{
  QTB_TRACE;
  if (!parent || !parent->isFolder()) return;
  // simply ignore it to avoid circular references
  if (parent==this) return;
  // Check if parent is not a children of this item
  QtbTreeItem * p = parent;
  while (p) {
    if (p==this) return;
    p=p->parent();
  }
  if (_parent) {
    int i=_parent->_children.indexOf(this);
    QTB_ASSERT(i>-1);
    _parent->_children.removeAt(i);
  }
  parent->addChild(this);
  _parent=parent;
  _modified = true;
}

/*!
  Count the number of children that are folder (no ids attached directly to them).

  \sa folderAt(), indexOfFolder(()
*/
int QtbTreeItem::folderCount() const
{
  QTB_TRACE;
  int n=0;
  for (const_iterator it=begin(); it!=end(); ++it) {
    if ((*it)->isFolder()) n++;
  }
  return n;
}

/*!
  Return the child folder at position \a index.

  \sa folderCount(), indexOfFolder(()
*/
QtbTreeItem * QtbTreeItem::folderAt( int index )
{
  QTB_TRACE;
  int i=0;
  for (const_iterator it=begin(); it!=end(); ++it) {
    if ((*it)->isFolder()) {
      if (i == index ) {
        return *it;
      }
      i++;
    }
  }
  return 0;
}

/*!
  Return the index of child folder \a g.

  \sa folderCount(), folderAt()
*/
int QtbTreeItem::indexOfFolder( QtbTreeItem * g ) const
{
  QTB_TRACE;
  const QtbTreeItem * parent = g->parent();
  if (!parent) return -1;
  int i=0;
  for (const_iterator it=parent->begin(); it!=parent->end(); ++it) {
    if ((*it)->isFolder()) {
      if (*it == g ) {
        return i;
      }
      i++;
    }
  }
  return -1;
}

int QtbTreeItem::childRow()
{
  QTB_TRACE;
  if ( parent() )
    return parent()->indexOfChild( this );
  else
    return 0;
}

int QtbTreeItem::folderRow()
{
  QTB_TRACE;
  if ( parent() )
    return parent()->indexOfFolder( this );
  else
    return 0;
}

void QtbTreeItem::setModified( bool m )
{
  QTB_TRACE;
  for ( const_iterator it = begin();it != end();++it ) {
    (*it)->setModified(m);
  }
  _modified=m;
}

bool QtbTreeItem::isModified() const
{
  QTB_TRACE;
  if (_modified) return true;
  for ( const_iterator it = begin();it != end();++it ) {
    if ( (*it)->isModified() ) return true;
  }
  return false;;
}

void QtbTreeItem::xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const
{
  QTB_TRACE;
  Q_UNUSED(context);
  writeProperty(s, "folder", isFolder());
}

void QtbTreeItem::xml_writeChildren(XML_WRITECHILDREN_ARGS) const
{
  QTB_TRACE;
  Q_UNUSED(context);
  for (const_iterator it=begin(); it!=end(); ++it) {
    (*it)->xml_save(s, context);
  }
}

QtbXMLMember QtbTreeItem::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  if (tag == "folder") return QtbXMLMember(0);
  if (tag=="TreeItem") {
    return QtbXMLMember( createNewItem( this ) );
  } else return QtbXMLMember(QtbXMLMember::Unknown);
}

bool QtbTreeItem::xml_setProperty(XML_SETPROPERTY_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  switch (memberID) {
  case 0:
    _folder=content.toBool();
    return true;
  default:
    return false;
  }
}

void QtbTreeItem::xml_polish(XML_POLISH_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(context);
  if (!_parent) setModified( false );
}

