/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-20
**
***************************************************************************/

#include <QDate>
#include <QFile>
#include <QDataStream>
#include <QCoreApplication>
#include <QTextCodec>
#include <QMetaProperty>

#include "qtbxmlclass.h"
#include "qtbxmlparser.h"
#include "qtbcoreapplication.h"
#include "qtbtoolsversion.h"
#include "qtbversion.h"
#include "qtbmemorychecker.h"

const QString QtbXMLClass::xmlXMLClassTag = "XMLClass";

const QString QtbXMLClass::openBeginTagStr = "<";
const QString QtbXMLClass::closeBeginTagStr = "</";
const QString QtbXMLClass::endTagStr = ">";
const QString QtbXMLClass::closeEndTagStr = "/>";
const QString QtbXMLClass::trueStr = "true";
const QString QtbXMLClass::falseStr = "false";
int QtbXMLClass::fileIndex = 1000;
const QString QtbXMLClass::binDataTag = "binDataFile";
const QString QtbXMLClass::binData200510Tag = "binDataOffset";

const QtbXMLSaveAttributes QtbXMLClass::nullSaveAttributes;

/*!
  \class QtbXMLMember qtbxmlclass.h
  \brief QtbXMLMember is used by QtbXMLClass to receive the type of tag to parse

  QtbXMLMember is used only in re-implemetations of
  QtbXMLClass::xml_member( QtbStringSection& tag, QtbXMLRestoreAttributes& map )
*/

/*!
  \fn QtbXMLMember::QtbXMLMember(QtbXMLMember::Unknown)

  Return an QtbXMLMember constructed with this constructor if the member is not valid
*/

/*!
  \fn QtbXMLMember::QtbXMLMember(int memberID)

  Return an QtbXMLMember constructed with this constructor if the member is a property
*/

/*!
  \fn QtbXMLMember::QtbXMLMember(QtbXMLClass * child, bool temp )

  Return an QtbXMLMember constructed with this constructor if the member is another QtbXMLClass object (child).

  If \a temp is true the \a child is automatically destroyed once it is restored. \a temp is false by default.
*/

/*!
  \class QtbXMLClass qtbxmlclass.h
  \brief QtbXMLClass is a parser for a simplified XML syntax
 
*/

/*!
  \fn void QtbXMLClass::xml_writeBinaryData( QDataStream& s ) const

  This function must be re-implemented in all classes dealing with binary data
  that cannot be saved in an ASCII xml file (e.g. due to the amount of data ).

  The way binary data is stored drastically changed in November 2006 with the
  introduction of tar.gz structures for xml files. Each class willing to store
  binary data can automatically generate a new file (with an automatic file name) in the
  .tar.gz structure by sending bytes to s.

  See also xml_setBinaryData().

  For the arguments of this function use Macro XML_WRITEBINARYDATA_ARGS.
*/

/*!
  \fn bool QtbXMLClass::xml_setBinaryData( QDataStream& s )

  This function must be re-implemented in all classes dealing with binary data
  that cannot be saved in an ASCII xml file (e.g. due to the amount of data ).

  See also xml_writeBinaryData().

  The difference between xml_setBinaryData() and xml_setBinaryData200410()
  is detected by the type of tag at the beginning of each block if it can be read with
  QString ==> 200510, else try with a normal C string, if it match the current tag
  then execute xml_setBinaryData200411().

  See also xml_setBinaryData200411() to maintain compatibility
  with previous versions of xml storages.

  For the arguments of this function use Macro XML_SETBINARYDATA_ARGS.
*/

/*!
  \fn bool QtbXMLClass::xml_setBinaryData200411( QDataStream& s )

  This function must be re-implemented in all classes dealing with binary data
  that cannot be saved in an ASCII xml file (e.g. due to the amount of data ).

  See also xml_setBinaryData().

  For the arguments of this function use Macro XML_SETBINARYDATA_ARGS.

  This function must not used in new codes. This is only kept for compatibility
  reasons, to be able to read all xml.bin files generated by releases before
  November 2006.
*/

/*!
  \fn QtbXMLMember QtbXMLClass::xml_member( QtbStringSection& tag, QtbXMLRestoreAttributes& map )

  Re-implement this function to offer XML restore (children and properties) support to your class.

  From \a tag and \a map (with contains the attibute value) return a unique identifier under
  the format of a QtbXMLMember. QtbXMLMember is initialized with 3 types of contructors:

  \li An integer: id number of a property
  \li A QtbXMLClass * : a child of this object identified by \a tag
  \li Default constructor: error, unknow child or property

  Map of attributes can be inspected in this way (can be achived also in xml_setProperty() ):

  \code
    static const QString tmp("childrenName");
    QtbXMLRestoreAttributeIterator it=map.find(tmp);
    if (it!=map.end()) {
      // found attribute "childrenName"
    }
  \endcode

  If the map of attributes is not used:

  \code
    Q_UNUSED(attributes);
    if (tag=="x1") return QtbXMLMember(0);
    else if (tag=="y1") return QtbXMLMember(1);
    else if (tag=="x2") return QtbXMLMember(2);
    else if (tag=="y2") return QtbXMLMember(3);
    else return QtbXMLMember(QtbXMLMember::Unknown);
  \endcode

  Arithmetic operations + and - apply to QtbXMLMember to avoid confusion of property id numbers
  between inherited objects. Offset 3 corresponds to the number of properties defined in this object.

  \code
    if (tag=="anInteger") return QtbXMLMember(0);
    else if (tag=="aString") return QtbXMLMember(1);
    else if (tag=="aDouble") return QtbXMLMember(2);
    return QtbAbstractLine::xml_member(tag, attributes, context)+3;
  \endcode

  For the arguments of this function use Macro XML_MEMBER_ARGS.
*/

/*!
  \fn QtbXMLMember QtbXMLClass::xml_setProperty( int memberID, QtbXMLRestoreAttributes& map, QtbStringSection& content )

  Re-implement this function to offer XML restore properties support to your class.

  From \a memberID set the corresponding property with value \a content. The \a map of attributes
  is given as a supplementary information (not useful in all cases).

  For a general case:

  \code
  Q_UNUSED(attributes);
  double val=content.toDouble();
  switch (memberID) {
  case 0:
    _x1=val;
    return true;
  case 1:
    _y1=val;
    return true;
  case 2:
    _x2=val;
    return true;
  case 3:
    _y2=val;
    return true;
  default:
    return false;
  }
  \endcode

  For classes inheriting other classes (see also xml_member())
  
  \code
  switch (memberID) {
  case 0:
    _anInteger=content.toString();
    return true;
  case 1:
    _aString=content.toInt();
    return true;
  case 2:
    _aDouble=content.toDouble();
    return true;
  default:
    return QtbAbstractLine::xml_setProperty(memberID-3, map, content);
  \endcode

  For the arguments of this function use Macro XML_SETPROPERTY_ARGS.
*/

QTB_IMPL_XML_WRITE_PROPERTY( const char *, saveSpecChar( content ) )
QTB_IMPL_XML_WRITE_PROPERTY( QString, saveSpecChar( content ) )
QTB_IMPL_XML_WRITE_PROPERTY( QChar, saveSpecChar( content ) )
QTB_IMPL_XML_WRITE_PROPERTY( int, QString::number( content ) )
QTB_IMPL_XML_WRITE_PROPERTY( qint64, QString::number( content ) )
QTB_IMPL_XML_WRITE_PROPERTY( double, QString::number( content, 'g', 20 ) )
QTB_IMPL_XML_WRITE_PROPERTY( bool, content ? trueStr : falseStr )
QTB_IMPL_XML_WRITE_PROPERTY_ATT( const char *, saveSpecChar( content ) )
QTB_IMPL_XML_WRITE_PROPERTY_ATT( QString, saveSpecChar( content ) )
QTB_IMPL_XML_WRITE_PROPERTY_ATT( QChar, saveSpecChar( content ) )
QTB_IMPL_XML_WRITE_PROPERTY_ATT( int, QString::number( content ) )
QTB_IMPL_XML_WRITE_PROPERTY_ATT( qint64, QString::number( content ) )
QTB_IMPL_XML_WRITE_PROPERTY_ATT( double, QString::number( content, 'g', 20 ) )
QTB_IMPL_XML_WRITE_PROPERTY_ATT( bool, content ? trueStr : falseStr )

void QtbXMLClass::writeProperty(QtbXMLStream& s, QString name, const QtbXMLSaveAttributes& attributes) \
{
  QTB_TRACE;
  QString tmp = s.indent();
  tmp+=openBeginTagStr;
  tmp+=name;
  tmp+=attributes.toEncodedString();
  tmp+=closeEndTagStr;
  s << tmp << endl;
}

/*!
  Convenient function to "manually" write some XML content
*/
void QtbXMLClass::writeChildren(QtbXMLStream& s, QString name, const QtbXMLSaveAttributes& attributes, QString children)
{
  QTB_TRACE;
  QString childIndent = "\n"+s.indent()+"  ";
  QString tmp = s.indent();
  tmp+=openBeginTagStr;
  tmp+=name;
  tmp+=attributes.toEncodedString();
  tmp+=endTagStr;
  tmp+="\n";
  tmp+=children;
  tmp+=s.indent() + closeBeginTagStr;
  tmp+=name;
  tmp+=endTagStr;
  s << tmp << endl;
}


/*!
  \fn QtbXMLClass::xml_save(QtbXMLStream &s, QString attributes=QString::null) const
  All output must execute this function that add header and footer to the XML
*/

/*!
  Called by each child to export its data into the XML stream.

  \li s            current active XML stream
  \li attributes   any string with format keyword1="value1" keyword2="value2" ..., may be null.

  Do not call this function directly with an empty QTextStream: the produced XML will not contain the header.
*/
void QtbXMLClass::xml_save(QtbXMLStream& s, QtbXMLContext * context, const QtbXMLSaveAttributes& attributes) const
{
  QTB_TRACE;
  QString tmp = s.indent();
  tmp += openBeginTagStr;
  tmp += xml_tagName();
  tmp+=attributes.toEncodedString();
  tmp += endTagStr;
  s << tmp << endl;

  s.incrementIndent();
  xml_writeProperties(s, context);
  xml_writeChildren(s, context);
  s.decrementIndent();

  tmp = s.indent();
  tmp += closeBeginTagStr;
  tmp += xml_tagName();
  tmp += endTagStr;
  s << tmp << endl;
}

QtbXMLClass::Error QtbXMLClass::restore( const QChar *& ptr, int& line, QtbXMLStream * s, QtbXMLContext * context)
{
  QtbXMLParser parser(s, context);
  parser.setLine(line);
  parser.setPointer(ptr);
  return parser.parse(this);
}

void QtbXMLClass::skipBlanks( const QChar *& ptr, int& line )
{
  QTB_TRACE;
  while ( *ptr != 0 ) {
    switch ( ptr->unicode() ) {
    case 0x000A:    // '\n'
      line++;
    case 0x0020:    // ' '
    case 0x0009:    // '\t'
    case 0x000D:    // '\r'
      break;
    default:
      return ;
    }
    ptr++;
  }
}

/*!
  This function is called when special characters are suspected. The
  following special characters are replaced by XML sequences (see
  http://en.wikipedia.org/wiki/List_of_XML_and_HTML_character_entity_references for
  a list XML predefined entities):

  Replace:
  \li '&amp;' by '\&amp;'
  \li '&lt;' by '\&lt;'
  \li '&gt;' by '\&gt;'
  \li '&quot;' by '\&quot;'
  \li '&apos;' by '\&apos;'

  Replacements are inlined due the small number of special characters to replace
  and to avoid time losses.
*/
QString QtbXMLClass::saveSpecChar( QString content )
{
  QTB_TRACE;
  QString dest;
  const QChar * ptr = content.data();
  QtbStringSection cur ( ptr, 0 );
  while ( true ) {
    switch ( ptr->unicode() ) {
    case 0x0000:
      if ( cur.data() == content.data() ) { // no special character found
        return content;
      } else {
        cur.setEnd( ptr );
        cur.appendTo( dest );
        //qtbApp << "Content before replacements ---" << content << "---" << endl;
        //qtbApp << "Content after replacements ---" << dest << "---" << endl;
        return dest;
      }
    case 0x0026:    // '&'
      cur.setEnd( ptr );
      cur.appendTo( dest );
      dest += "&amp;";
      cur.set( ptr + 1, 0 );
      break;
    case 0x003C:    // '<'
      cur.setEnd( ptr );
      cur.appendTo( dest );
      dest += "&lt;";
      cur.set( ptr + 1, 0 );
      break;
    case 0x003E:    // '>'
      cur.setEnd( ptr );
      cur.appendTo( dest );
      dest += "&gt;";
      cur.set( ptr + 1, 0 );
      break;
    case 0x0027:    // '''
      cur.setEnd( ptr );
      cur.appendTo( dest );
      dest += "&apos;";
      cur.set( ptr + 1, 0 );
      break;
    case 0x0022:    // '"'
      cur.setEnd( ptr );
      cur.appendTo( dest );
      dest += "&quot;";
      cur.set( ptr + 1, 0 );
      break;
    default:
      break;
    }
    ptr++;
  }
}

void QtbXMLClass::writeBinaryData(QtbXMLStream & s, QtbXMLContext * context) const
{
  QTB_TRACE;
  if ( s.isMultiFile() ) {
    QString fileName = QString("bin_data_%1").arg(fileIndex++);
    writeProperty( s, binDataTag, fileName );
    QByteArray binArray;
    QDataStream binStream( &binArray, QIODevice::WriteOnly );
    binStream << xml_tagName();
    xml_writeBinaryData(binStream, context);
    s.addFile( fileName, binArray );
  }
}

bool QtbXMLClass::errorMessage(Error err, int line)
{
  QTB_TRACE;
  QString msg;
  switch (err) {
    case NoError:
      return true;
    case ErrorFileNotOpen:
      msg=translate("QtbXMLClass", "Cannot open file. Check file permissions");
      break;
    case ErrorWritingFile:
      msg=translate("QtbXMLClass", "Cannot write to file. Check disk space");
      break;
    case ErrorNoDocType:
      msg=translate("QtbXMLClass", "No doc type found at the beginning of the XML");
      break;
    case ErrorNoVersion:
      msg=translate("QtbXMLClass", "No version found at the beginning of the XML");
      break;
    case ErrorUnmatchedAmpSemiColon:
      msg=translate("QtbXMLClass", "Unmatched ampersand and semi colon for XML special characters");
      break;
    case ErrorUnknowSpecialCharacter:
      msg=translate("QtbXMLClass", "Unknwown XML special character");
      break;
    case ErrorEmptyTag:
      msg=translate("QtbXMLClass", "Empty XML tag");
      break;
    case ErrorUnmatchedTag:
      msg=translate("QtbXMLClass", "Unmatched XML tag");
      break;
    case ErrorUnmatchedTopLevelTag:
      msg=translate("QtbXMLClass", "Unmatched top level XML tag");
      break;
    case ErrorEmptyContextStack:
      msg=translate("QtbXMLClass", "Empty context stack");
      break;
    case ErrorParsingContent:
      msg=translate("QtbXMLClass", "Error parsing tag content");
      break;
    case ErrorParsingAttributes:
      msg=translate("QtbXMLClass", "Error parsing tag attributes");
      break;
    case ErrorSettingAttributes:
      msg=translate("QtbXMLClass", "Error setting tag attributes");
      break;
    case ErrorSettingBinaryData:
      msg=translate("QtbXMLClass", "Error setting binary data");
      break;
    case ErrorWrongBinaryOffset:
      msg=translate("QtbXMLClass", "Wrong offset in binary file");
      break;
    case ErrorWrongBinaryTag:
      msg=translate("QtbXMLClass", "Wrong binary tag");
      break;
    case ErrorBinaryFileNotFound:
      msg=translate("QtbXMLClass", "Binary file not found");
      break;
    case ErrorEndTruncatedTag:
      msg=translate("QtbXMLClass", "Truncated tag at the end of XML");
      break;
    case ErrorEndTruncatedString:
      msg=translate("QtbXMLClass", "Truncated string at the end of XML");
      break;
    case ErrorEndTruncatedContext:
      msg=translate("QtbXMLClass", "Truncated context at the end of XML");
      break;
    case ErrorEndStillInside:
      msg=translate("QtbXMLClass", "Object not completely restored at the end of XML");
      break;
    case ErrorEndTagNotFound:
      msg=translate("QtbXMLClass", "Object not found in XML");
      break;
  }
  if(line>0) msg=translate("QtbXMLClass", "%1 at line %2.").arg(msg).arg(line);
  QtbMessage::critical( MSG_ID, translate("QtbXMLClass", "Parsing xml"), msg, true);
  return false;
}

/*!
  This function is designed to save as XML properties all QObject properties.
*/
void QtbXMLClass::qobject_writeProperties( const QObject * o, const QtbXMLClass * xmlo, XML_WRITEPROPERTIES_ARGS )
{
  QTB_TRACE;
  Q_UNUSED(context);
  const QMetaObject * meta=o->metaObject();
  // Ignore property with index 0 (objectName from QObject, replaced by objectName of the first base class from Qtb library)
  int n=meta ->propertyCount();
  for ( int i = meta ->indexOfProperty("objectName");i < n; i++ ) {
    QMetaProperty prop = meta ->property( i );
    if (prop.isStored()) {
      QVariant val = prop.read( o );
      if ( val.type() == QVariant::String )
        xmlo->writeProperty( s, prop.name(), val.toString() );
      else xmlo->writeProperty( s, prop.name(), val.toString() );
    }
  }
}

/*!
  This function is designed to restore from XML properties to QObject properties.
*/
QtbXMLMember QtbXMLClass::qobject_member( QObject * o, XML_MEMBER_ARGS )
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  const QMetaObject * meta = o->metaObject();
  int id = meta->indexOfProperty( tag.toString().toAscii().data() );
  if (id==-1) {
    return QtbXMLMember(QtbXMLMember::Unknown);
  } else return QtbXMLMember( id );
}

/*!
  This function is designed to restore from XML properties to QObject properties.
*/
bool QtbXMLClass::qobject_setProperty( QObject * o, XML_SETPROPERTY_ARGS )
{
  QTB_TRACE;
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  const QMetaObject * meta = o->metaObject();
  QMetaProperty p = meta->property( memberID );
  if ( !p.isValid() ) {
    return false;
  }
  if ( !p.isWritable() ) {
    qtbApp << translate( "QtbXMLClass", "read-only property : %1" ).arg(p.name()) << endl;
    return false;
  }
  if ( p.type() == QVariant::Bool ) {
    QVariant v( content == "true" );
    return p.write( o, v );
  } else {
    QVariant v( content.toString() );
    return p.write( o, v );
  }
}
