/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-20
**
***************************************************************************/

#ifndef QTB_XMLClass_H
#define QTB_XMLClass_H

#include <QString>
#include <QHash>
#include <QStack>
#include <QTextStream>
#include <QCoreApplication>

#include "qtbtoolsdllexport.h"
#include "qtbxmlstream.h"
#include "qtbxmlcontext.h"
#include "qtbxmlmember.h"
#include "qtbxmlsaveattributes.h"
#include "qtbtranslations.h"
#include "qtbstringsection.h"

#define QTB_IMPL_XML_WRITE_PROPERTY(type, convert) \
void QtbXMLClass::writeProperty(QtbXMLStream& s, QString name, type content) \
{ \
  QTB_TRACE; \
  QString tmp = s.indent(); \
  tmp+=openBeginTagStr; \
  tmp+=name; \
  tmp+=endTagStr; \
  tmp+=convert; \
  tmp+=closeBeginTagStr; \
  tmp+=name; \
  tmp+=endTagStr; \
  s << tmp << endl; \
}

#define QTB_IMPL_XML_WRITE_PROPERTY_ATT(type, convert) \
void QtbXMLClass::writeProperty(QtbXMLStream& s, QString name, const QtbXMLSaveAttributes& attributes, type content) \
{ \
  QTB_TRACE; \
  QString tmp = s.indent(); \
  tmp+=openBeginTagStr; \
  tmp+=name; \
  tmp+=attributes.toEncodedString(); \
  tmp+=endTagStr; \
  tmp+=convert; \
  tmp+=closeBeginTagStr; \
  tmp+=name; \
  tmp+=endTagStr; \
  s << tmp << endl; \
}

#define QTB_DECL_XML_WRITE_PROPERTY(type) \
static void writeProperty(QtbXMLStream& s, QString name, type content);
#define QTB_DECL_XML_WRITE_PROPERTY_ATT(type) \
static void writeProperty(QtbXMLStream& s, QString name, const QtbXMLSaveAttributes& attributes, type content);

#define XML_WRITEPROPERTIES_ARGS QtbXMLStream& s, QtbXMLContext * context
#define XML_WRITECHILDREN_ARGS QtbXMLStream& s, QtbXMLContext * context
#define XML_MEMBER_ARGS QtbStringSection& tag, const QtbXMLRestoreAttributes& attributes, QtbXMLContext * context
#define XML_SETPROPERTY_ARGS int memberID, const QtbXMLRestoreAttributes& attributes, QtbStringSection& content, QtbXMLContext * context
#define XML_POLISH_ARGS QtbXMLContext * context
#define XML_POLISHCHILD_ARGS QtbXMLClass * child, QtbXMLContext * context
#define XML_POLISHED_ARGS
#define XML_WRITEBINARYDATA_ARGS QDataStream& s, QtbXMLContext * context
#define XML_SETBINARYDATA_ARGS QDataStream& s, QtbXMLContext * context
#define XML_ATTRIBUTES_ARGS QtbXMLSaveAttributes& attributes, QtbXMLContext * context
#define XML_SETATTRIBUTES_ARGS const QtbXMLRestoreAttributes& attributes, QtbXMLContext * context

typedef QHash<QtbStringSection, QtbStringSection> QtbXMLRestoreAttributes;
typedef QHash<QtbStringSection, QtbStringSection>::const_iterator QtbXMLRestoreAttributeIterator;

class QTBTOOLS_EXPORT QtbXMLClass
{
  QTB_TRANSLATIONS( "QtbXMLClass" )
public:
  QtbXMLClass() {}
  virtual ~QtbXMLClass() {}

  enum Error {
    NoError,
    ErrorFileNotOpen,
    ErrorWritingFile,
    ErrorNoDocType,
    ErrorNoVersion,
    ErrorUnmatchedAmpSemiColon,
    ErrorUnknowSpecialCharacter,
    ErrorEmptyTag,
    ErrorUnmatchedTag,
    ErrorUnmatchedTopLevelTag,
    ErrorEmptyContextStack,
    ErrorParsingContent,
    ErrorParsingAttributes,
    ErrorSettingAttributes,
    ErrorSettingBinaryData,
    ErrorWrongBinaryOffset,
    ErrorWrongBinaryTag,
    ErrorBinaryFileNotFound,
    ErrorEndTruncatedTag,
    ErrorEndTruncatedString,
    ErrorEndTruncatedContext,
    ErrorEndStillInside,
    ErrorEndTagNotFound
  };

  enum FileType {
    TarFile,
    XmlFile
  };

  void xml_save(QtbXMLStream& s, QtbXMLContext * context, const QtbXMLSaveAttributes& = nullSaveAttributes) const;
  QtbXMLClass::Error restore( const QChar *& ptr, int& line, QtbXMLStream * s, QtbXMLContext * context);

  inline virtual void xml_attributes( XML_ATTRIBUTES_ARGS ) const;
  inline virtual bool xml_setAttributes( XML_SETATTRIBUTES_ARGS );
  inline virtual QtbXMLMember xml_member( XML_MEMBER_ARGS );
  inline virtual void xml_polishChild( XML_POLISHCHILD_ARGS );

  virtual const QString& xml_tagName() const { return xmlXMLClassTag;}
  static const QString xmlXMLClassTag;

  static void skipBlanks(const QChar *& ptr, int& line);
  static QString saveSpecChar(QString content);
  static bool errorMessage(Error err, int line=0);

  // Functions that can be use to write information inside re-implemented functions (see below)
  QTB_DECL_XML_WRITE_PROPERTY( const char * )
  QTB_DECL_XML_WRITE_PROPERTY( QString )
  QTB_DECL_XML_WRITE_PROPERTY( QChar )
  QTB_DECL_XML_WRITE_PROPERTY( double )
  QTB_DECL_XML_WRITE_PROPERTY( int )
  QTB_DECL_XML_WRITE_PROPERTY( qint64 )
  QTB_DECL_XML_WRITE_PROPERTY( bool )
  QTB_DECL_XML_WRITE_PROPERTY_ATT( const char * )
  QTB_DECL_XML_WRITE_PROPERTY_ATT( QString )
  QTB_DECL_XML_WRITE_PROPERTY_ATT( QChar )
  QTB_DECL_XML_WRITE_PROPERTY_ATT( double )
  QTB_DECL_XML_WRITE_PROPERTY_ATT( int )
  QTB_DECL_XML_WRITE_PROPERTY_ATT( qint64 )
  QTB_DECL_XML_WRITE_PROPERTY_ATT( bool )
  static void writeProperty( QtbXMLStream& s, QString name, const QtbXMLSaveAttributes& attributes );
  static void writeChildren( QtbXMLStream& s, QString name, QString children ) {
    writeChildren( s, name, nullSaveAttributes, children );
  }
  static void writeChildren(QtbXMLStream& s, QString name, const QtbXMLSaveAttributes& attributes, QString children);
  void writeBinaryData(QtbXMLStream& s, QtbXMLContext * context) const;
protected:
  friend class QtbXMLParser;
  // Functions that can be re-implemented
  inline virtual void xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const;
  inline virtual void xml_writeChildren( XML_WRITECHILDREN_ARGS ) const;
  inline virtual bool xml_setProperty( XML_SETPROPERTY_ARGS );
  inline virtual void xml_polish( XML_POLISH_ARGS );
  inline virtual bool xml_polished(XML_POLISHED_ARGS);
  inline virtual void xml_writeBinaryData( XML_WRITEBINARYDATA_ARGS ) const;
  inline virtual bool xml_setBinaryData( XML_SETBINARYDATA_ARGS );
  inline virtual bool xml_setBinaryData200411( XML_SETBINARYDATA_ARGS ) {return xml_setBinaryData(s, context);}
protected:
  static void qobject_writeProperties( const QObject * o, const QtbXMLClass * xmlo, XML_WRITEPROPERTIES_ARGS );
  static QtbXMLMember qobject_member( QObject * o, XML_MEMBER_ARGS );
  static bool qobject_setProperty( QObject * o, XML_SETPROPERTY_ARGS );
private:
  QtbStringSection getPropertyName( const QtbStringSection& tag, int& line );

  static const QString openBeginTagStr;
  static const QString closeBeginTagStr;
  static const QString endTagStr;
  static const QString closeEndTagStr;
  static const QString trueStr;
  static const QString falseStr;
  static const QString binDataTag;
  static const QString binData200510Tag;
  static int fileIndex;
  static const QtbXMLSaveAttributes nullSaveAttributes;
};

inline void QtbXMLMember::release()
{
  if (_child && _memberID==TemporaryChild) {
    delete _child;
    _child=0;
  }
}

inline void QtbXMLClass::xml_writeProperties( XML_WRITEPROPERTIES_ARGS ) const
{
  Q_UNUSED(s);
  Q_UNUSED(context);
}

inline void QtbXMLClass::xml_writeChildren( XML_WRITECHILDREN_ARGS ) const
{
  Q_UNUSED(s);
  Q_UNUSED(context);
}

inline QtbXMLMember QtbXMLClass::xml_member( XML_MEMBER_ARGS )
{
  Q_UNUSED(tag);
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  return QtbXMLMember(QtbXMLMember::Unknown);
}

inline  bool QtbXMLClass::xml_setProperty( XML_SETPROPERTY_ARGS )
{
  Q_UNUSED(memberID);
  Q_UNUSED(attributes);
  Q_UNUSED(content);
  Q_UNUSED(context);
  return false;
}

inline void QtbXMLClass::xml_polish( XML_POLISH_ARGS )
{
  Q_UNUSED(context);
}

inline void QtbXMLClass::xml_polishChild( XML_POLISHCHILD_ARGS )
{
  Q_UNUSED( child );
  Q_UNUSED(context);
}

inline bool QtbXMLClass::xml_polished( XML_POLISHED_ARGS )
{
  return false;
}

inline void QtbXMLClass::xml_writeBinaryData( XML_WRITEBINARYDATA_ARGS ) const
{
  Q_UNUSED(s);
  Q_UNUSED(context);
}

inline bool QtbXMLClass::xml_setBinaryData( XML_SETBINARYDATA_ARGS )
{
  Q_UNUSED(s);
  Q_UNUSED(context);
  return false;
}

inline void QtbXMLClass::xml_attributes( XML_ATTRIBUTES_ARGS ) const
{
  Q_UNUSED(attributes);
  Q_UNUSED(context);
}

inline bool QtbXMLClass::xml_setAttributes( XML_SETATTRIBUTES_ARGS )
{
  Q_UNUSED(attributes);
  Q_UNUSED(context);
  return true;
}

inline QString QtbXMLSaveAttributes::toEncodedString() const
{
  QString tmp;
  for(const_iterator it = begin();it!=end();it++) {
    tmp += " ";
    tmp += it->key();
    tmp += "=\"";
    tmp += QtbXMLClass::saveSpecChar( it->value() );
    tmp += "\"";
  }
  return tmp;
}

#endif
