/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbtools.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-11-27
**
***************************************************************************/

#include "qtbxmlclassfactory.h"
#include "qtbcoreapplication.h"

/*!
  \fn virtual QtbXMLClass * QtbXMLClassCreator::create() const
  Abstract function called by QtbXMLClassFactory for creating a new instance.
*/

/*!
  \class QtbXMLClassFactory qtbxmlclassfactory.h
  \brief QtbXMLClassFactory provides creation of instances of classes from their XML tag
*/

QtbXMLClassFactory QtbXMLClassFactory::_self;

QtbXMLClassFactory::~QtbXMLClassFactory()
{
  qDeleteAll(_workshops);
}

/*!
  Creates a new class of tag \a tag.
  Null is returned in case of unregistered class.
*/
QtbXMLClass * QtbXMLClassFactory::create( QString workshopName, QString tag ) const
{
  //printf("Request for '%s' in workshop '%s'\n", tag.toAscii().data(), workshopName.toAscii().data() );
  QtbXMLClassCreator * c = creator( workshopName, tag );
  if (c)
    return c->create();
  else
    return 0;
}

/*!
  Returns the creator corresponding to \a tag or null if it is not registered.
*/
QtbXMLClassCreator * QtbXMLClassFactory::creator( QString workshopName, QString tag ) const
{
  QtbXMLClassWorkshop * w = workshop( workshopName );
  if (w) return w->creator( tag ); else return 0;
}

/*!
  Returns the list of all registered tags
*/
QStringList QtbXMLClassFactory::registeredTags( QString workshopName ) const
{
  QtbXMLClassWorkshop * w = workshop( workshopName );
  if (w) return workshop( workshopName )->registeredTags(); else return QStringList();
}

void QtbXMLClassFactory::registerClass( QString workshopName, QString tag, QtbXMLClassCreator * creator )
{
  //printf("Register '%s' in workshop '%s'\n", tag.toAscii().data(), workshopName.toAscii().data() );
  return workshop( workshopName )->registerClass( tag, creator );
}

void QtbXMLClassFactory::registerSynonym( QString workshopName, QString syn, QString tag )
{
  workshop( workshopName )->registerSynonym( syn, tag );
}

QtbXMLClassWorkshop * QtbXMLClassFactory::workshop( QString workshopName )
{
  QMap<QString, QtbXMLClassWorkshop *>::iterator it;
  it = _workshops.find( workshopName );
  if ( it != _workshops.end() ) {
    return it.value();
  } else {
    QtbXMLClassWorkshop * w = new QtbXMLClassWorkshop;
    _workshops.insert( workshopName, w );
    return w;
  }
}

QtbXMLClassWorkshop * QtbXMLClassFactory::workshop( QString workshopName ) const
{
  QMap<QString, QtbXMLClassWorkshop *>::const_iterator it;
  it = _workshops.find( workshopName );
  if ( it != _workshops.end() ) {
    return it.value();
  } else {
    return 0;
  }
}

void QtbXMLClassWorkshop::registerClass( QString tag, QtbXMLClassCreator * creator )
{
  _registeredClasses.insert( tag, creator );
}

void QtbXMLClassWorkshop::registerSynonym( QString syn, QString tag )
{
  _registeredSynonyms.insert( syn, tag );
}

/*!
  Returns the list of all registered tags
*/
QStringList QtbXMLClassWorkshop::registeredTags() const
{
  QStringList l;
  QMap<QString, QtbXMLClassCreator *>::const_iterator it;
  for(it = _registeredClasses.begin(); it!=_registeredClasses.end(); it++) {
    l.append( it.key() );
  }
  return l;
}

/*!
  Returns the creator corresponding to \a tag or null if it is not registered.
*/
QtbXMLClassCreator * QtbXMLClassWorkshop::creator( QString tag ) const
{
  QMap<QString, QtbXMLClassCreator *>::const_iterator itTag;
  itTag = _registeredClasses.find( tag );
  if ( itTag != _registeredClasses.end() ) {
    return itTag.value();
  } else {
    // Look if it is a synonym of another tag
    QMap<QString, QString>::const_iterator itSyn;
    itSyn = _registeredSynonyms.find( tag );
    if ( itSyn != _registeredSynonyms.end() ) {
      itTag = _registeredClasses.find( itSyn.value() );
      if ( itTag != _registeredClasses.end() ) {
        return itTag.value();
      }
    }
  }
  return 0;
}
