/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2007-05-22
**
***************************************************************************/

#include "qtbautocorrcurves.h"
#include "qtbautocorrdispersion.h"
#include "qtbcoreapplication.h"

/*!
  \class QtbAutocorrCurves qtbautocorrcurves.h
  \brief Brief description of class still missing

  Full description of class still missing
*/

const QString QtbAutocorrCurves::xmlAutocorrCurvesTag= "AutocorrCurves";

/*!
  Add one ring (for convenience)
*/
void QtbAutocorrCurves::addRing( const QtbAutocorrRing& r )
{
  QTB_TRACE;
  _rings.append( r );
}

/*!
  Remove ring with index \a index
*/
void QtbAutocorrCurves::removeRing( int index )
{
  QTB_TRACE;
  _rings.remove( index );
}

QList<QtbModalCurve> QtbAutocorrCurves::ringCurves( int index ) const
{
  QList<QtbModalCurve> l;
  for (QList<QtbModalCurve>::ConstIterator it = _curves.begin(); it != _curves.end(); it++ ) {
    const QtbModalCurve& c = *it;
    if ( c.modes().first().ringIndex() == index ) {
      l.append(c);
    }
  }
  return l;
}

/*!
  Adds a new curve. Make sure ring index fits with the number of rings.
  At least one ring must be available.
*/
void QtbAutocorrCurves::addCurve( const QtbModalCurve& c )
{
  QTB_TRACE;
  QTB_ASSERT(_rings.count()>0);
  _curves.append(c);
  QList<QtbMode>& modes = _curves.last().modes();
  int iRing;
  if (modes.isEmpty()) {
    _curves.last().addMode( QtbMode( QtbMode::Vertical, 0, 0 ) );
    iRing = 0;
  } else {
    iRing = modes.first().ringIndex();
    if ( iRing < 0 || iRing >= _rings.count() ) {
      qtbApp << tr("Bad ring index for curve %1: %2 while max is %3").arg(c.name()).arg(iRing).arg(_rings.count()-1) << endl;
    }
    for (QList<QtbMode>::Iterator it = modes.begin(); it!=modes.end(); it++ ) {
      it->setRingIndex( iRing );
    }
  }
}

/*!
  Adds all curves and rings from \a c into this object. Common rings are smartly merged
*/
void QtbAutocorrCurves::add( const QtbAutocorrCurves& curves )
{
  QTB_TRACE;
  for (QList<QtbModalCurve>::ConstIterator it = curves._curves.begin(); it != curves._curves.end(); it++ ) {
    const QtbModalCurve& c = *it;
    int iCurvesRing = c.modes().first().ringIndex();
    if ( iCurvesRing >= curves._rings.count() ) {
      qtbApp << tr("Bad ring index for curve %1: %2 while max is %3").arg(c.name()).arg(iCurvesRing).arg(curves._rings.count()-1) << endl;
    } else {
      const QtbAutocorrRing& r = curves._rings.at(iCurvesRing);
      int iRing = indexOf(r);
      if (iRing == -1 ) {
        iRing = ringCount();
        addRing( r );
      }
      _curves.append(c);
      QList<QtbMode>& modes = _curves.last().modes();
      for (QList<QtbMode>::Iterator it = modes.begin(); it!=modes.end(); it++ ) {
        it->setRingIndex(iRing);
      }
    }
  }
}

void QtbAutocorrCurves::sort()
{
  for (QList<QtbModalCurve>::Iterator it = _curves.begin(); it != _curves.end(); it++ ) {
    it->sort();
  }
}

void QtbAutocorrCurves::xml_writeChildren(XML_WRITECHILDREN_ARGS) const
{
  QTB_TRACE;
  int nRings = _rings.count();
  static const QString key("type");
  static const QString value("autocorr");
  QtbXMLSaveAttributes att;
  att.add( key, value );
  for ( int iRing=0;iRing<nRings;iRing++ ) {
    _rings.at(iRing).xml_save(s, context);
    QList<QtbModalCurve> curves = ringCurves(iRing);
    for (QList<QtbModalCurve>::Iterator it = curves.begin(); it!=curves.end(); it++ ) {
      it->xml_save(s, context, att);
    }
  }
}

QtbXMLMember QtbAutocorrCurves::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(context);
  if ( tag == "ModalCurve" ) {
    static const QString typeAtt("type");
    QtbXMLRestoreAttributeIterator it=attributes.find(typeAtt);
    if (it!=attributes.end()) {
      if (it.value() == "autocorr") {
        _curves.append(QtbModalCurve());
        return QtbXMLMember( &_curves.last() );
      } else {
        qtbApp << tr("Wrong type of modal curve %1").arg(it.value().toString()) << endl;
        return QtbXMLMember(QtbXMLMember::Unknown);
      }
    } else {
      qtbApp << tr("No type defined for modal curve") << endl;
      return QtbXMLMember(QtbXMLMember::Unknown);
    }
  } else if (tag=="AutocorrRing") {
    addRing();
    return QtbXMLMember( &_rings.last() );
  } else return QtbXMLMember(QtbXMLMember::Unknown);
}

/*!
  Calculate the dispersion curve corresponding to curve \a index.
  The solutions are searched between \a kmin and \a kmax. The returned curve
  may contain various values for the same frequency.
*/
QtbModalCurve QtbAutocorrCurves::dispersionCurve( int index, double kmin, double kmax ) const
{
  const QtbModalCurve& autocorr = _curves.at( index );
  int nx = autocorr.count();
  QtbModalCurve dispersion;
  QtbAutocorrDispersion eng;
  eng.setRing( _rings.at( autocorr.modes().first().ringIndex() ) );
  for (int ix = 0; ix<nx; ix++ ) {
    const QtbFactoryPoint& p = autocorr.at(ix);
    if ( p.isValid() ) {
      QVector<double> solutions = eng.dispersion( p, kmin, kmax );
      for ( QVector<double>::iterator it = solutions.begin(); it!= solutions.end(); it++ ) {
        dispersion.append( QtbFactoryPoint( p.x(), *it ) );
      }
    }
  }
  dispersion.addMode( QtbMode( QtbMode::Phase, QtbMode::Rayleigh, 0 ) );
  return dispersion;
}
