/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2009-02-04
**
***************************************************************************/

#ifndef QTB_AUTOCORRENGINE_H
#define QTB_AUTOCORRENGINE_H

#include "qtbwavedllexport.h"
#include "qtbautocorrring.h"
#include "qtbvalue.h"

class QTBWAVE_EXPORT QtbAutocorrEngine
{
public:
  inline void setVerticalRing( const QtbAutocorrRing& r );
  inline void setHorizontalRing( const QtbAutocorrRing& r );

  bool isThinRing() { return fabs( _dr ) < 0.02 * _r0; }

  inline QtbValue verticalThinRing( double omega, const QtbValue& dispR );
  inline QtbValue verticalThickRing( double omega, const QtbValue& dispR );

  inline double verticalThinRing( double k );
  inline double verticalThickRing( double k );

  inline bool initHorizontalThinRing( double omega, const QtbValue& dispR , const QtbValue& dispL );
  inline bool initHorizontalThickRing( double omega, const QtbValue& dispR , const QtbValue& dispL );

  inline QtbValue radial( double alphaRayleigh, double alphaLove );
  inline QtbValue transverse( double alphaRayleigh, double alphaLove );
private:
  double _r1, _r2, _r0, _dr, _rFactor;
  double _jRR, _jLR, _jRT, _jLT;
};

inline void QtbAutocorrEngine::setVerticalRing( const QtbAutocorrRing& ring )
{
  QTB_TRACE;
  _r1 = ring.minRadius();
  _r2 = ring.maxRadius();
  _r0 = 0.5 * ( _r1 + _r2 );
  _dr = _r2 - _r1;
  _rFactor = _r0 * _dr;  // (r2^2-r1^2)/2
}

inline void QtbAutocorrEngine::setHorizontalRing( const QtbAutocorrRing& ring )
{
  QTB_TRACE;
  _r1 = ring.minRadius();
  _r2 = ring.maxRadius();
  _r0 = 0.5 * ( _r1 + _r2 );
  _dr = _r2 - _r1;
  _rFactor = 2.0 / ( _r0 * _dr );  // 4/(r2^2-r1^2)
}

inline QtbValue QtbAutocorrEngine::verticalThinRing( double omega, const QtbValue& dispR )
{
  if ( dispR.isValid() ) {
    return QtbValue( verticalThinRing( omega *  dispR.value() ) );
  } else {
    return QtbValue();
  }
}

inline double QtbAutocorrEngine::verticalThinRing( double k )
{
  return j0( k * _r0 );
}

inline QtbValue QtbAutocorrEngine::verticalThickRing( double omega, const QtbValue& dispR )
{
  QTB_TRACE;
  if ( dispR.isValid() ) {
    return QtbValue( verticalThickRing( omega * dispR.value() ) );
  } else {
    return QtbValue();
  }
}

inline double QtbAutocorrEngine::verticalThickRing( double k )
{
  return ( _r2 * j1( k * _r2 ) - _r1 * j1( k * _r1 ) ) / ( k * _rFactor );
}

inline bool QtbAutocorrEngine::initHorizontalThinRing( double omega,
                                                       const QtbValue& dispR,
                                                       const QtbValue& dispL )
{
  if ( dispR .isValid() && dispL.isValid() ) {
    double omegaR = omega * _r0;
    double argR = omegaR * dispR.value();
    double j0R = j0( argR );
    double j2R = jn( 2, argR );
    double argL = omegaR * dispL.value();
    double j0L = j0( argL );
    double j2L = jn( 2, argL );
    _jRR = j0R - j2R;
    _jLR = j0L + j2L;
    _jRT = j0R + j2R;
    _jLT = j0L - j2L;
    return true;
  } else {
    return false;
  }
}

inline bool QtbAutocorrEngine::initHorizontalThickRing( double omega,
                                                        const QtbValue& dispR,
                                                        const QtbValue& dispL )
{
  if ( dispR .isValid() && dispL.isValid() ) {
    double omegaR1 = omega * _r1;
    double omegaR2 = omega * _r2;

    double invKR = 1.0 / ( omega * dispR.value() );
    double argR1 = omegaR1 * dispR.value();
    double argR2 = omegaR2 * dispR.value();
    double j0R1 = j0( argR1 );
    double j1R1 = j1( argR1 );
    double j0R2 = j0( argR2 );
    double j1R2 = j1( argR2 );

    double invKL = 1.0 / ( omega * dispL.value() );
    double argL1 = omegaR1 * dispL.value();
    double argL2 = omegaR2 * dispL.value();
    double j0L1 = j0( argL1 );
    double j1L1 = j1( argL1 );
    double j0L2 = j0( argL2 );
    double j1L2 = j1( argL2 );

    _jRR = _rFactor * invKR *( invKR * ( j0R2 - j0R1 ) + _r2 * j1R2 - _r1 * j1R1 );
    _jLR = _rFactor * invKL * invKL * ( j0L1 - j0L2 );
    _jRT = _rFactor * invKR * invKR * ( j0R1 - j0R2 );
    _jLT = _rFactor * invKL *( invKL * ( j0L2 - j0L1 ) + _r2 * j1L2 - _r1 * j1L1 );
    return true;
  } else {
    return false;
  }
}

inline QtbValue QtbAutocorrEngine::radial( double alphaRayleigh, double alphaLove )
{
  return QtbValue( alphaRayleigh * _jRR + alphaLove * _jLR );
}

inline QtbValue QtbAutocorrEngine::transverse( double alphaRayleigh, double alphaLove )
{
  return QtbValue( alphaRayleigh * _jRT + alphaLove * _jLT );
}

#endif // QTB_AUTOCORRENGINE_H
