/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-21
**
***************************************************************************/

#include <math.h>

#include "qtbdispersion.h"
#include "qtbellipticity.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbDispersion qtbdispersion.h
  \brief Calculate Love and Rayleigh dispersion curves
  
  The main function used to calculate the dispersion curve is calculate().
*/

QtbDispersion::QtbDispersion( int nModes, const QVector<double>* x )
  : QtbModalStorage( nModes, x)
{
  QTB_TRACE;
  _precision = 1e-7;
}

/*!
  \fn void QtbDispersion::setPrecision( double p )
  Set precision in slowness for dispersion computation. If you need more than 1e-16, you must use the big number version of calculate().
*/

/*!
  \fn bool QtbDispersion::calculate( QtbRayleigh& model, QtbEllipticity * ell = 0 )
  Calculate the dispersion for Rayleigh modes.
*/

/*!
  \fn bool QtbDispersion::calculate( QtbRayleighBigNum& model, QtbEllipticity * ell = 0 )
  Calculate the dispersion for Rayleigh modes with high precision.
*/

/*!
  \fn bool QtbDispersion::calculate( QtbLove& model )
  Calculate the dispersion for Love modes.
*/

/*!
  \fn bool QtbDispersion::calculate(QtbModelType& model, QtbEllipticity * ell)

  Calculates the dispersion curve for the input model modelS. The type of curve
  calculated, either Love or Rayleigh, depends upon the class type for models.
  (see QtbLoveSlowness or QtbRayleighSlowness).

  ell is optional, if non-null and if Rayleigh is computed, the ellipticity
  of Rayleigh modes are computed

  Internal use only.\n
*/

/*!
  Used by the ellipticity computation, it allows the refinement of the ellipticity
  peak. The method for dispersion computation is the same as calculate().
  It is just a resume to a particular frequency sample. \a lastIndex is the x index of the fixed and already computed
  curve. It avoids searching in the x list every time this function is called.
  \a lastIndex must be less than xCount()-1.
*/
bool QtbDispersion::refine( int modeIndex, int lastIndex, double omega, QtbRootSolver<QtbRayleigh>& solver, QtbEllipticity& ell )
{
  QTB_TRACE;
  QTB_ASSERT(lastIndex<xCount()-1);
  QtbModalRefine& dispRef = refineAdd( omega );
  QtbModalRefine& ellRef = ell.refineAdd( omega );
  QtbValue * values = mode( 0 )+lastIndex;

  QtbRayleigh * waveModel = solver.values();
  const QtbLayeredModel * model = waveModel->model();
  double minSlow = model->minSlowS();
  double maxSlow, curSlow, lastSupBound = model->maxSlowR(), newCurSlow;
  QTB_SAFE_UNINITIALIZED(curSlow,0);
  bool forbitVelocityInversion = model->checkVelocityInversion() == -1;
  // Initialize equation's polarity
  waveModel->setOmega( 0.05 ); // the sign of polarity is the same at all frequencies
  solver.setPolarity( model->maxSlowR() );

  for ( int iMode = 0; iMode <= modeIndex; iMode++ ) {
    solver.setPrecision( _precision );
    if ( iMode == 0 ) {
      maxSlow = model->maxSlowR();
    } else {
      maxSlow = curSlow;
      solver.inversePolarity();
    }
    while ( true ) {
      waveModel->setOmega(omega);
      //printf("\nFrequency = %.15lg Hz\n\n",0.5 * omega / M_PI);
      bool errorDetected = false;
      curSlow = values[iMode*xCount()+1].value();
      if ( ! solver.searchDown( curSlow, minSlow, maxSlow ) ) {
        if ( iMode == 0 ) {
          // for fundamental mode this condition can never occur unless a mode jumping occured
          qtbApp << tr( "** Warning ** : mode jumping for mode %1 (end), reducing "
                        "step ratio to %2" ).arg( iMode ).arg( QtbNumber::toDouble( solver.searchStep() * 0.1 ) ) << endl;
          errorDetected = true;
        } else {
          dispRef.setValid( iMode, false );
          ellRef.setValid( iMode, false );
          curSlow = minSlow;
        }
      } else {
        solver.neville();
        newCurSlow = solver.lower();
        //printf("%lg %lg-->%lg %lg\n",minSlow, curSlow, newCurSlow, maxSlow);
        waveModel->y( 0.5 * ( solver.upper() + newCurSlow ) );
        ellRef.setValue( iMode, fabs( waveModel->ellipticity() ) );
        if ( newCurSlow - curSlow > _precision ) {
          if ( forbitVelocityInversion ) {
            errorDetected = true;
            qtbApp << tr( "** Warning ** : retrograde dispersion not allowed (mode %1), "
                          "reducing step ratio to %2" ).arg( iMode ).arg( QtbNumber::toDouble( solver.searchStep() * 0.1 ) ) << endl;
          } else {
            // Inversions are allowed, check at last point that no mode exist at a higher slowness
            waveModel->setOmega( x( lastIndex ) );
            solver.setRelativeStep( solver.searchStep() * 0.1 );
            if ( solver.searchUp( lastSupBound, maxSlow ) ) {
              // recalculating curve because of mode jumping
              qtbApp << tr( "** Warning ** : mode jumping for mode %1 (middle), "
                            "reducing step ratio to %2" ).arg( iMode ).arg( QtbNumber::toDouble( solver.searchStep() * 0.1 ) ) << endl;
              errorDetected = true;
            }
            solver.setRelativeStep( solver.searchStep() * 10.0 );
          }
        }
        if (!errorDetected) {
          lastSupBound = solver.upper();
          curSlow = newCurSlow;
          dispRef.setValue( iMode, curSlow );
          // When calculating the dispersion curve, stepRatio is positive to search downwards
          // (From maxSlow towards minSlow)
          // Here we want to search carefully upwards: from curSlow to maxSlow
          // If at least one root is found, this proves that mode jumping occured
          if ( curSlow > minSlow ) {
            // The curSlow was calculated using _solve and _x2 is the value returned by
            // Solve (slightly greater than the true root).
            // Here we need the _x1 value, e.g. slightly less than the true root, in
            // order to be sure that the root eventually found will not be the same
            // as curSlow
            curSlow = solver.upper();
          }
          // Contrary to normal dispersion curve computation, here we know the next point on the curve
          // Hence, we simulate its computation and we must find the same number, if not mode jumping occured.
          waveModel->setOmega( x( lastIndex ) );
          //printf("Check refine jumpig: frequency = %.15lg Hz\n",0.5 * x( lastIndex ) / M_PI);
          double correctSlow = values[iMode*xCount()].value();
          bool ok;
          if (iMode==0) {
            ok = solver.searchDown( curSlow, minSlow, maxSlow );
          } else {
            double lastMaxSlow = values[(iMode-1)*xCount()].value();
            if (curSlow>lastMaxSlow) curSlow = lastMaxSlow;
            ok = solver.searchDown( curSlow, minSlow, lastMaxSlow );
          }
          if ( ok ) {
            solver.neville();
            newCurSlow = solver.lower();
          }
          //printf("%lg %lg-->%lg %lg = %lg %lg %lg\n",minSlow, curSlow, newCurSlow, lastMaxSlow, correctSlow, newCurSlow-correctSlow, _precision);
          if ( !ok || fabs(newCurSlow-correctSlow)>_precision ) {
            ok = true;
            qtbApp << tr( "** Warning ** : mode jumping for mode %1 (refine), reducing "
                          "step ratio to %2" ).arg( iMode ).arg( QtbNumber::toDouble( solver.searchStep() * 0.1 ) ) << endl;
            errorDetected = true;
          }
        }
      }
      if ( errorDetected ) {
        // recalculating curve because of mode jumping
        if ( solver.searchStep() < 1e-4 ) {
          qtbApp << tr( "** Warning ** : dispersion curve impossible to obtain "
                        "for mode %1" ).arg( iMode ) << endl;
          return false;
        }
        solver.setPrecision( solver.precision() * 0.1 );
        solver.setRelativeStep( solver.searchStep() * 0.1 );
      } else break;
    }
  }
  return true;
}

/*!
  Convert phase slownesses into group slownesses
*/
void QtbDispersion::setGroupSlowness()
{
  QTB_TRACE;
  int nx=xCount()-1;
  QtbValue tmp[ xCount() ];
  for (int im=0;im<modeCount();im++) {
    QtbValue * point=mode(im);
    for(int i = 1;i<nx;i++) {
      if (point[i-1].isValid() && point[i].isValid() && point[i+1].isValid()) {
        double val = point[i].value();
        double  derslow=0.5*((val-point[i-1].value()) / (x(i)-x(i-1))+
                            (point[i+1].value()-val) / (x(i+1)-x(i)));
        derslow*=x(i);
        tmp[ i ] = val+derslow;
      } else {
        tmp[ i ].setValid(false);
      }
    }
    point[0].setValid(false);
    for(int i=1;i<nx;i++) {
      point[i]=tmp[i];
    }
    point[nx].setValid(false);
  }
}

