/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-21
**
***************************************************************************/

#ifndef QTB_MODALDISPERSION_H
#define QTB_MODALDISPERSION_H

#include <math.h>

#include "qtbellipticity.h"
#include "qtbmodalstorage.h"
#include "qtbcompatqt43.h"
#include "qtblove.h"
#include "qtbrayleigh.h"
#include "qtbwavedllexport.h"
#include "qtbtranslations.h"
#include "qtbcoreapplication.h"

class QtbDispersion : public QtbModalStorage
{
  QTB_TRANSLATIONS("QtbDispersion");
public:
  QTBWAVE_EXPORT QtbDispersion( int nModes, const QVector<double>* x );

  QTBWAVE_EXPORT void setGroupSlowness();

  bool calculate( QtbRayleigh * waveModel, QtbEllipticity * ell = 0, QAtomicInt * terminated = 0 )
  {
    return calculate<double, QtbRayleigh>( waveModel, ell, terminated );
  }
#ifdef QTB_BIGNUM
  bool calculate( QtbRayleighBigNum * waveModel, QtbEllipticity * ell = 0, QAtomicInt * terminated = 0 )
  {
    return calculate<mp_real, QtbRayleighBigNum>( waveModel, ell, terminated );
  }
#endif
  bool calculate( QtbLove * waveModel, QAtomicInt * terminated = 0 )
  {
    return calculate<double, QtbLove>( waveModel, 0, terminated );
  }
  void setPrecision( double p ) { _precision = p; }
  QTBWAVE_EXPORT bool refine( int modeIndex, int lastIndex, double omega, QtbRootSolver<QtbRayleigh>& modelS, QtbEllipticity& ell );
private:
  template <class QtbRealType, class QtbWaveModel>
  bool calculate( QtbWaveModel * waveModel, QtbEllipticity * ell, QAtomicInt * terminate );
  double _precision;
};

template <class QtbRealType, class QtbWaveModel>
bool QtbDispersion::calculate( QtbWaveModel * waveModel, QtbEllipticity * ell, QAtomicInt * terminated )
{
  QTB_TRACE;
  QtbRealType * curMode = new QtbRealType [ xCount() ];
  QtbRealType * lastMode = new QtbRealType [ xCount() ];
  if ( ell ) ell->refineClear();
  refineClear();
//#define calculate_DEBUG

  const QtbLayeredModel * model = waveModel->model();
  QtbRootSolverTemplate<QtbRealType, QtbWaveModel> solver( waveModel );
  QtbRealType minSlow = model->minSlowS();
  QtbRealType maxSlow, curSlow, lastSupBound = model->maxSlowR(), newCurSlow;
  bool forbitVelocityInversion = model->checkVelocityInversion() == -1;
  // Initialize equation's polarity
  waveModel->setOmega( 0.05 ); // the sign of polarity is the same at all frequencies
  solver.setPolarity( model->maxSlowR() );
  int lastOmega = xCount() - 1;

  for ( int iMode = 0; iMode < modeCount(); iMode++ ) {
#ifdef calculate_DEBUG
    printf( "***************** Mode %i ******************\n", iMode );
#endif
    solver.setPrecision( _precision );
    if ( iMode == 0 ) {
      maxSlow = model->maxSlowR();
      // 2009-02-15 Reduce factor from 0.5 to 0.25 to avoid too many tries
      //            With 0.5 it appears that mode jumping occurs quite a lot
      solver.setRelativeStep( 0.25 * ( model->maxSlowR() - model->maxSlowS() ) / maxSlow );
    } else {
      QtbRealType * tmpMode = lastMode;
      lastMode = curMode;
      curMode = tmpMode;
      maxSlow = lastMode[ lastOmega ];
      solver.setRelativeStep( 0.5 * ( model->maxSlowR() - model->maxSlowS() ) / maxSlow * ( 2 * M_PI / x( lastOmega ) ) );
      solver.inversePolarity();
    }
    QtbValue * ellValues;
    if ( ell && iMode<ell->modeCount() ) {
      ellValues = ell->mode( iMode );
    } else {
      ellValues = 0;
    }
    QtbValue * slownessValues = mode( iMode );

    while ( true ) {
      bool errorYetDetected = false;
      curSlow = maxSlow;
      for ( int i = lastOmega;i >= 0;i-- ) {
#ifdef calculate_DEBUG
        printf( "***************** Frequency %i omega %lf f %lf\n", i, x(i), x(i)/(2*M_PI) );
#endif
        if ( terminated && terminated->testAndSetOrdered(true,true)) {
          // Catch user interrupt
          return false;
        }
        waveModel->setOmega( x( i ) );
        if ( iMode > 0 ) {
          maxSlow = lastMode[ i ];
          if ( curSlow > maxSlow ) curSlow = maxSlow;
        }
        //printf("%20.10lf %20.10lf %20.10lf\n",dble(curSlow),dble(_minSlow),dble(maxSlow));
        if ( solver.searchDown( curSlow, minSlow, maxSlow ) ) {
          solver.neville();
          newCurSlow = solver.lower();
        } else {
          if ( iMode == 0 ) {
            // for fundamental mode this condition can never occur unless a mode jumping occured
            errorYetDetected = true;
            qtbApp << tr( "** Warning ** : mode jumping for mode %1 (end), reducing "
                          "step ratio to %2" ).arg( iMode ).arg( QtbNumber::toDouble( solver.searchStep() * 0.1 ) ) << endl;
            break;
          } else {
            // The end of the curve, the mode does not exist at this period (no root)
            for ( int j = i;j >= 0;j-- ) {
              slownessValues[ j ].setValid( false );
              curMode[ j ] = minSlow;
            }
            if (ellValues) {
              for ( int j = i;j >= 0;j-- ) ellValues[ j ].setValid( false );
            }
            curSlow = minSlow;
            break;
          }
        }
        if ( ellValues ) {
          /* Take a mid slowness between the refined bounds and recompute the
           propagator matrices to get the corresponding sub-determinents necessary
           for ellipticity computation. */
          waveModel->y( 0.5 * ( solver.upper() + newCurSlow ) );
          ellValues[ i ].setValue( fabs( waveModel->ellipticity() ) );
          // Get ellipticity with its sign
          //ellValues[ i ].setValue( model.ellipticity() );
        }
        double newLastSupBound = solver.upper(); // searchUp may alter it
        //printf("%20.10lf %20.10lf %20.10lf\n",x( i )/(2*M_PI), modelS.infBoundRoot(), modelS.supBoundRoot());
        if ( i < lastOmega && newCurSlow - curSlow > _precision ) {
          if ( forbitVelocityInversion ) {
            errorYetDetected = true;
            qtbApp << tr( "** Warning ** : retrograde dispersion not allowed (mode %1), "
                          "reducing step ratio to %2" ).arg( iMode ).arg( QtbNumber::toDouble( solver.searchStep() * 0.1 ) ) << endl;
            break;
          } else {
            // Inversions are allowed, check at last point that no mode exist at a higher slowness
            waveModel->setOmega( x( i + 1 ) );
            solver.setRelativeStep( solver.searchStep() * 0.1 );
            if ( solver.searchUp( lastSupBound, iMode>0 ? lastMode[ i + 1 ] : maxSlow ) ) {
              // recalculating curve because of mode jumping
              qtbApp << tr( "** Warning ** : mode jumping for mode %1 (middle at %3 Hz), "
                            "reducing step ratio to %2" )
                        .arg( iMode )
                        .arg( QtbNumber::toDouble( solver.searchStep() ) )
                        .arg( x( i )/ (2*M_PI) ) << endl;
              errorYetDetected = true;
              solver.setRelativeStep( solver.searchStep() * 10.0 );
              break;
            }
            solver.setRelativeStep( solver.searchStep() * 10.0 );
          }
        }
        lastSupBound = newLastSupBound;
        curSlow = newCurSlow;
        curMode[ i ] = curSlow;
        slownessValues[ i ].setValue( QtbNumber::toDouble( curSlow ) );
        slownessValues[ i ].setValid( true );
#ifdef calculate_DEBUG
        printf( "%30.20lf %30.20lf\n", 0.5 * x(i) / M_PI, slownessValues[ i ].value() );
#endif
      }
      // When calculating the dispersion curve, stepRatio is positive to search downwards
      // (From maxSlow towards minSlow)
      // Here we want to search carefully upwards: from curSlow to maxSlow
      // If at least one root is found, this proves that mode jumping occured
      if ( curSlow > minSlow ) {
        // The curSlow was calculated using _solve and _x2 is the value returned by
        // Solve (slightly greater than the true root).
        // Here we need the _x1 value, e.g. slightly less than the true root, in
        // order to be sure that the root eventually found will not be the same
        // as curSlow
        curSlow = solver.upper();
      }
      if ( errorYetDetected || solver.searchUp( curSlow, maxSlow ) ) {
        // recalculating curve because of mode jumping
        if ( solver.searchStep() < 1e-4 ) {
          qtbApp << tr( "** Warning ** : dispersion curve impossible to obtain "
                        "for mode %1" ).arg( iMode ) << endl;
          for ( ; iMode < modeCount(); iMode++ ) {
            QtbValue * slownessValues = mode( iMode );
            for ( int j = lastOmega;j >= 0;j-- ) {
              slownessValues[ j ].setValid( false );
            }
          }
          return false;
        }
        solver.setPrecision( solver.precision() * 0.1 );
        solver.setRelativeStep( solver.searchStep() * 0.1 );
        if ( !errorYetDetected )
          qtbApp << tr( "** Warning ** : mode jumping for mode %1 (end), reducing "
                        "step ratio to %2" ).arg( iMode ).arg( QtbNumber::toDouble( solver.searchStep() ) ) << endl;
        if ( iMode > 0 ) maxSlow = lastMode[ lastOmega ];
      } else break;
    }
  }
  delete [] curMode;
  delete [] lastMode;
  return true;
}

#endif // QTB_MODALDISPERSION_H
