/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-09-26
**
***************************************************************************/

#include "qtbdispersionfactory.h"
#include "qtbellipticityfactory.h"
#include "qtbautocorrfactory.h"
#include "qtbdispersion.h"
#include "qtbmode.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbDispersionFactory qtbdispersionfactory.h
  \brief Brief description of class still missing

  Full description of class still missing
*/

QtbDispersionFactory::QtbDispersionFactory()
{
  QTB_TRACE;
  _phaseRayleigh = 0;
  _phaseLove = 0;
  _groupRayleigh = 0;
  _groupLove = 0;
}

QtbDispersionFactory::~QtbDispersionFactory()
{
  QTB_TRACE;
  delete _phaseRayleigh;
  delete _phaseLove;
  delete _groupRayleigh;
  delete _groupLove;
}

/*!
  Validate modes, make sure that the number of modes for group slowness is greater than for
  phase slowness.
*/
void QtbDispersionFactory::validate( QtbEllipticityFactory * ell, QtbAutocorrFactory * autocorr )
{
  QTB_TRACE;
  if ( ell && ell->values() ) {
    upgrade<QtbDispersion>( _phaseRayleigh, ell->values()->modeCount()-1 );
  }
  if (autocorr) {
    if (autocorr->verticalModeCount()>0) {
      upgrade<QtbDispersion>( _phaseRayleigh, autocorr->verticalModeCount()-1 );
    }
    if (autocorr->horizontalModeCount()>0) {
      upgrade<QtbDispersion>( _phaseRayleigh, autocorr->horizontalModeCount()-1 );
      upgrade<QtbDispersion>( _phaseLove, autocorr->horizontalModeCount()-1 );
    }
  }
  if (_groupRayleigh) {
    upgrade<QtbDispersion>( _phaseRayleigh, _groupRayleigh->modeCount()-1 );
  }
  if (_groupLove) {
    upgrade<QtbDispersion>( _phaseLove, _groupLove->modeCount()-1 );
  }
}

void QtbDispersionFactory::setMode ( const QtbMode& m )
{
  QTB_TRACE;
  switch ( m.slowness() ) {
  case QtbMode::Phase:
    switch ( m.polarisation() ) {
    case QtbMode::Vertical:
    case QtbMode::Radial:
    case QtbMode::Rayleigh:
      upgrade<QtbDispersion>( _phaseRayleigh, m.index() ); break;
    case QtbMode::Love:
    case QtbMode::Transverse:
      upgrade<QtbDispersion>( _phaseLove, m.index() ); break;
    }
    break;
  case QtbMode::Group:
    switch ( m.polarisation() ) {
    case QtbMode::Vertical:
    case QtbMode::Radial:
    case QtbMode::Rayleigh:
      upgrade<QtbDispersion>( _groupRayleigh, m.index() ); break;
    case QtbMode::Love:
    case QtbMode::Transverse:
      upgrade<QtbDispersion>( _groupLove, m.index() ); break;
    default:
      QTB_ASSERT(false); break;
    }
    break;
  }
}

const QtbValue * QtbDispersionFactory::mode( const QtbMode& m) const
{
  switch ( m.slowness() ) {
  case QtbMode::Phase:
    switch ( m.polarisation() ) {
    case QtbMode::Vertical:
    case QtbMode::Radial:
    case QtbMode::Rayleigh:
      return _phaseRayleigh ? _phaseRayleigh->mode( m.index() ) : 0;
    case QtbMode::Love:
    case QtbMode::Transverse:
      return _phaseLove ? _phaseLove->mode( m.index() ) : 0;
    }
    break;
  case QtbMode::Group:
    switch ( m.polarisation() ) {
    case QtbMode::Vertical:
    case QtbMode::Radial:
    case QtbMode::Rayleigh:
      return _groupRayleigh ? _groupRayleigh->mode( m.index() ) : 0;
    case QtbMode::Love:
    case QtbMode::Transverse:
      return _groupLove ? _groupLove->mode( m.index() ) : 0;
    }
    break;
  }
  return 0; // just to avoid warning
}

/*!
  Calculate theoretical curves
*/
bool QtbDispersionFactory::calculate( const QtbLayeredModel * m, QtbEllipticityFactory * ellFactory )
{
  QTB_TRACE;
  if (_phaseRayleigh) {
    QtbRayleigh rayleigh( m );
    QtbEllipticity * ell;
    if (ellFactory) ell = ellFactory->values(); else ell=0;
    if ( !_phaseRayleigh->calculate( &rayleigh, ell ) ) return false;
    if (_groupRayleigh) {
      *_groupRayleigh = *_phaseRayleigh;
      _groupRayleigh->setGroupSlowness();
    }
  }
  if (_phaseLove) {
    QtbLove love( m );
    if ( !_phaseLove->calculate( &love ) ) return false;
    if (_groupLove) {
      *_groupLove = *_phaseLove;
      _groupLove->setGroupSlowness();
    }
  }
  return true;
}
