/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-10-22
**
***************************************************************************/

#include "qtbellipticity.h"
#include "qtbdispersion.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbEllipticity qtbellipticity.h
  \brief Calculate Rayleigh ellipticity curves

  The main function used to calculate the ellipticity curve is calculate().

  Experience has shown that a 1e-10 precision on dc is necessary to refine the ellipticity down to 1e-3
  The precision used to compute the original curve must be the same as for the refines.
*/

inline double QtbEllipticity::refineMax( int modeIndex, int iOmega, QtbDispersion& disp,
                                         QtbRootSolver<QtbRayleigh>& solver, bool& badSampling )
{
  QTB_TRACE;
//#define QTB_REFINEMAX_DEBUG
  QtbValue * point=mode(modeIndex);
  double x1 = x(iOmega-2), ell1 = point[iOmega-2].value();
  double x2 = x(iOmega-1), ell2 = point[iOmega-1].value();
  double x3              , ell3;
  double x4 = x(iOmega)  , ell4 = point[iOmega].value();
  QTB_SAFE_UNINITIALIZED(ell3,0);
  QTB_SAFE_UNINITIALIZED(x3,0);

  // Initialize step ratio, contrary to normal dispersion computation, step ratio is not reset before each mode
  QtbRayleigh * waveModel = solver.values();
  const QtbLayeredModel * model = waveModel->model();
  solver.setRelativeStep( 0.5 * ( model->maxSlowR() - model->maxSlowS() ) / model->maxSlowR() * ( 2 * M_PI / x2 ) );

  double x20 = x2;
  double minEll=ell1;
  if (ell4<minEll) minEll=ell4;
  minEll-=minEll*1e-10;
  badSampling=false;
#ifdef QTB_REFINEMAX_DEBUG
  printf("#---- %15lg %15lg %15lg %15lg\n", 0.5/M_PI*x1, 0.5/M_PI*x2, 0.5/M_PI*x3, 0.5/M_PI*x4 );
  printf("#     %15lg %15lg %15lg %15lg\n", ell1, ell2, ell3, ell4 );
#endif
  while (x4-x1>1e-3) {
    if (x2-x1>x4-x2) {
      // insert a point between x1 and x2
      x3=x2;
      ell3=ell2;
      x2=0.5*(x3+x1);
      if (!disp.refine( modeIndex, x2<=x20 ? iOmega-2 : iOmega-1, x2, solver, *this)) break;
      ell2=lastRefine().value(modeIndex).value();
      if (ell2<minEll) badSampling=true;
    } else {
      x3=0.5*(x4+x2);
      if (!disp.refine( modeIndex, x3<=x20 ? iOmega-2 : iOmega-1, x3, solver, *this)) break;
      ell3=lastRefine().value(modeIndex).value();
      if (ell3<minEll) badSampling=true;
    }
#ifdef QTB_REFINEMAX_DEBUG
    printf("#---- %15lg %15lg %15lg %15lg\n", 0.5/M_PI*x1, 0.5/M_PI*x2, 0.5/M_PI*x3, 0.5/M_PI*x4 );
    printf("#     %15lg %15lg %15lg %15lg\n", ell1, ell2, ell3, ell4 );
#endif
    if (ell3<ell2) {
      x4=x3;
#ifdef QTB_REFINEMAX_DEBUG
      ell4=ell3;
#endif
    }
    else {
      x1=x2;
      x2=x3;
#ifdef QTB_REFINEMAX_DEBUG
      ell1=ell2;
#endif
      ell2=ell3;
    }
  }
  disp.refineSort();
  refineSort();
  if (ell3>ell2) return x3; else return x2;
}

/*!
  Return a list of the frequency of all peaks within the sampled range
*/
QList<double> QtbEllipticity::peaks( int modeIndex, QtbDispersion& disp, QtbRayleigh * waveModel)
{
  QList<double> peaks;
  double fac = 0.5 / M_PI;

  QtbRootSolver<QtbRayleigh> solver(waveModel);
  QtbValue * point=mode(modeIndex);
  int n=xCount();

  int i,j;
  for (j=1; j<n && !point[j].isValid(); j++) {}
  if (j==n) return peaks;
  for ( i=j+1; i<n && !point[i].isValid(); i++ ) {}
  if (i==n) return peaks;
  double lastSlope=point[i].value()-point[j].value();
  j=i;
  for ( i++; i<n; i++ ) {
    if (point[i].isValid()) {
      double slope=point[i].value()-point[j].value();
      j=i;
      if ( slope<=0 && lastSlope>0 && lastSlope-slope>1e-10) {
        // Found a maximum in the sampled curve that is at distance > 0.01 the ellipticity of the half space
        bool badSampling;
        double xr=refineMax( modeIndex, i, disp, solver, badSampling );
        peaks << fac*xr;
        if (badSampling)
          qtbApp << tr("** Warning ** : probably a missing peak, bad initial sampling") << endl;
      }
      lastSlope = slope;
    }
  }
  return peaks;
}

/*!
  Find the frequency of the closest peak to \a val and return the misfit
*/
double QtbEllipticity::peakMisfit( int modeIndex, const QtbStatValue& val, QtbDispersion& disp, QtbRayleigh * waveModel)
{
  QTB_TRACE;
  double f0, devf0, misfit=1e99;
  f0=val.mean()*2.0*M_PI;
  if (val.stddev()>0) devf0=1.0/ (val.stddev()*2.0*M_PI); else devf0=1.0/f0;

  QtbRootSolver<QtbRayleigh> solver(waveModel);
  QtbValue * point=mode(modeIndex);
  int n=xCount();

  int i,j;
  for (j=1; j<n && !point[j].isValid(); j++) {}
  if (j==n) return 0.0;
  for ( i=j+1; i<n && !point[i].isValid(); i++ ) {}
  if (i==n) return 0.0;
  double lastSlope=point[i].value()-point[j].value();
  j=i;
  for ( i++; i<n; i++ ) {
    if (point[i].isValid()) {
      double slope=point[i].value()-point[j].value();
      j=i;
      if (slope<=0 && lastSlope>0) {
        double diff;
        // Found a maximum in the sampled curve that is at distance > 0.01 the ellipticity of the half space
        if (((diff=(f0-x(i))*devf0)>=0 && diff<misfit) ||
            ((diff=(x(i-2)-f0)*devf0)>=0 && diff<misfit) ||
            (x(i-2)<f0 && f0<x(i))) {
          // There's a chance to get a better misfit
          bool badSampling;
          double xr=refineMax( modeIndex, i, disp, solver, badSampling );
          //qtbApp << tr("Refined peak frequency = %1").arg(0.5*xr/M_PI) << endl;
          // The peak is sufficiently refined, adjust the misfit if necessary
          diff=fabs(xr-f0)*devf0;
          if (diff<misfit) misfit=diff;
          if (badSampling)
            qtbApp << tr("** Warning ** : probably a missing peak, bad initial sampling") << endl;
        }
      }
      lastSlope = slope;
    }
  }
  if (misfit==1e99) {
    // monotonous curve, evaluate fundamental frequency by V0/4H
    double roughFreq = 2.0 * M_PI * waveModel->model()->roughFundamentalFrequency();
    //qtbApp << tr("Rough fundamental frequency = %1").arg(0.5 * roughFreq / M_PI ) << endl;
    if (roughFreq>=x(0) && roughFreq<=x(n-1)) {
      qtbApp << tr("** Warning ** : estimated rough fundamental frequency inside range") << endl;
      double midRange=0.5*(x(0)+x(n-1));
      if (roughFreq>midRange)
        misfit=fabs(x(n-1)-f0)*devf0;
      else
        misfit=fabs(x(0)-f0)*devf0;
    } else
      misfit=fabs(roughFreq-f0)*devf0;
  }
  refineSort();
  disp.refineSort();
  return floor(misfit*1000+0.5)*0.001; // frequency is estimated down to 1e-3, avoid classification within precision limits
}
