/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-04-28
**
***************************************************************************/

#include <QString>
#include <math.h>

#include "qtblayeredmodel.h"
#include "qtbpoint2d.h"
#include "qtbfile.h"
#include "qtbcoreapplication.h"
#include "qtbmemorychecker.h"

#define twopi 2*M_PI

/*!
  \class QtbLayeredModel qtblayeredmodel.h
  \brief Storage class for an horizontaly layered ground model.

  QtbLayeredModel stores data vectors for thicknesses, P-wave and S-wave
  slownesses, and densities of each layer. A series of checks and information
  functions are implemented.

  The same data vectors can be shared automatically using the copy constructor between
  several instances of QtbLayeredModel.
*/

/*!
  Default constructor

  Only called by QtbAnelasticLayeredModel.
*/
QtbLayeredModel::QtbLayeredModel()
{
  QTB_TRACE;
  _layerCount=0;
  initMembers();
}

/*!
  Copy constructor
*/
QtbLayeredModel::QtbLayeredModel( const QtbLayeredModel& o )
{
  QTB_TRACE;
  _layerCount=0;
  initMembers();
  operator=(o);
}

/*!
  Constructor: inits the number of layers including half space.
*/
QtbLayeredModel::QtbLayeredModel( int layersNumber )
{
  QTB_TRACE;
  // Sharing initialization
  _layerCount=layersNumber;
  initMembers();
  allocateData();
}

/*!
  Destructor: remove data vectors.
*/
QtbLayeredModel::~QtbLayeredModel()
{
  QTB_TRACE;
  deleteData();
}

/*!
  Initialize internal members

  Internal use only.
*/
void QtbLayeredModel::initMembers()
{
  _h=0;
  _slowP=0;
  _slowS=0;
  _rho=0;
  _mu=0;
  _qp=0;
  _qs=0;
}

/*!
  Allocates all data vectors (_layerCount must be initialized before)

  Internal use only.
*/
void QtbLayeredModel::allocateData()
{
  QTB_TRACE;
  deleteData();
  // allocate the data vectors
  if (_layerCount>1) {
    _h=new double [_layerCount-1];
    _slowP=new double [_layerCount];
    _slowS=new double [_layerCount];
    _rho=new double [_layerCount];
    _mu=new double [_layerCount];
    _qp=new double [_layerCount];
    _qs=new double [_layerCount];
  }
}

/*!
  Delete all data vectors

  Internal use only.
*/
void QtbLayeredModel::deleteData()
{
  QTB_TRACE;
  delete [ ] _h;
  delete [ ] _slowP;
  delete [ ] _slowS;
  delete [ ] _rho;
  delete [ ] _mu;
  delete [ ] _qp;
  delete [ ] _qs;
  _h=0;
  _slowP=0;
  _slowS=0;
  _rho=0;
  _mu=0;
  _qp=0;
  _qs=0;
}

/*!
  Copy oprator
*/
void QtbLayeredModel::operator=( const QtbLayeredModel& o )
{
  deleteData();
  _layerCount=o._layerCount;
  allocateData();
  int i = _layerCount-1;
  _slowP[i] = o._slowP[i];
  _slowS[i] = o._slowS[i];
  _rho[i] = o._rho[i];
  _mu[i] = o._mu[i];
  _qp[i] = o._qp[i];
  _qs[i] = o._qs[i];
  for (i--;i>=0;i--) {
    _h[i] = o._h[i];
    _slowP[i] =o. _slowP[i];
    _slowS[i] = o._slowS[i];
    _rho[i] = o._rho[i];
    _mu[i] = o._mu[i];
    _qp[i] = o._qp[i];
    _qs[i] = o._qs[i];
  }
  // Calculated variables:
  _slowSMin=o._slowSMin;
  _slowSMax=o._slowSMax;
  _maxRayleighSlowness=o._maxRayleighSlowness;
}

/*!
  Calculates the basic parameters for future calculations and test model
  validity.
*/
bool QtbLayeredModel::initCalculation()
{
  QTB_TRACE;
  bool ret=true;
  // Initialize calculation variables:
  for (int i=0;i<_layerCount;i++) {
    if (_slowS[i]!=0) _mu[i]=_rho[i]/ (_slowS[i]*_slowS[i]);
  }
  // Gets index of the layer with beta min or _slowS max
  int iSlowSMax;
  slowMinMax( iSlowSMax );
  if(!checkPoissonRatio()) ret=false;
  _maxRayleighSlowness = 1.05 * halfSpaceRayleighSlowness( iSlowSMax );
  return ret;
}

/*!
  Checks that the model respects the physic conditions over Poisson's ratio.
  Most geological materials have a positive Poisson's ratio less than 0.5

  Poisson'ratio=\f$(V_s^2-0.5*V_p^2)/(V_s^2-V_p^2)\f$

  This means that \f$V_s\f$ should be between \f$0\f$ and \f$\sqrt{2} V_p\f$.
  A null value for \f$V_s\f$ is not accepted in this case (no liquid layer).

  If the model satisfy these condition, this function returns -1 else the layer index where the condition is not satisfied.

  Internal use only.\n
*/
int QtbLayeredModel::checkPoissonRatio()
{
  QTB_TRACE;
#define sqrt2 1.41421356237
  for (int i=0;i<_layerCount;i++) {
    if (_slowS[i]<=0 || _slowP[i]*sqrt2>=_slowS[i]) {
      return i;
    }
  }
  return -1;
}

/*!
  Finds the layer index with the minimum shear-wave Velocity (or maximum shear-wave slowness).

  Internal use only.\n
*/
void QtbLayeredModel::slowMinMax( int& iSlowSMax )
{
  QTB_TRACE;
  _slowSMin=1e99;
  _slowSMax=-1e99;
  for(int i=0;i<_layerCount;i++) {
    if (_slowS[i]>_slowSMax) {
      _slowSMax=_slowS[i];
      iSlowSMax=i;
    }
    // Not meaningful, only the slowness of the half space is critical
    // Still don't know found cases with LVZ where half space slowness is not the minimum ofdispersion curve
    if (_slowS[i]<_slowSMin) _slowSMin=_slowS[i];
  }
  // The minimum slowness usefull for dispersion computation is the slowness of the half space
  //_slowSMin=_slowS[_layerCount-1];
}

/*!
  Calculates the Rayleigh-wave velovity as if the model was homogeneous and
  had the velocity of the layer \a iLayer.

  sq_cob means (c/b)^2 and sq_boa means (b/a)^2.

  Equation : (2-sq_cob)-4*sqrt(1-sq_boa*sq_cob)*sqrt(1-sq_cob)=0

  The sqrt can be elevated to square, interiors are positive:
  [(2-x)]=16* (1-sq_boa*x) * (1-x), where x=sq_cob

  Calculating all terms:
  x * [x-8*x+(24-16*sq_boa) *x+(16*sq_boa-16)]=0

  x=0 is not a valid solution
  Thus we are looking for the roots of:
  x-8*x+a3*x+a4
  where,
     a4=16*sq_boa-16
     a3=8-a4

  3rd degree poly which solutions are calculated by newton raphson
  using the derivative : 3*x-16*x+a3

  This poly has a simple root in the neighborhood of the starting
  value (for all boa), thus there is no problem of convergence
*/
double QtbLayeredModel::halfSpaceRayleighSlowness( int iLayer ) const
{
  QTB_TRACE;
  double y,x=0.9025;
  /* (c/beta)=0.95 solution for a=infinity, this is a maximum bound
     value for the Rayleigh velocity
  */
  double boa=_slowP[iLayer]/_slowS[iLayer];
  double a4=16*boa*boa-16;
  double a3=8-a4;
  do {
    double sqx=x*x;
    y=x*sqx-8*sqx+a3*x+a4;
    double dy=3*sqx-16*x+a3;
    x=x-y/dy;
  } while (fabs(y)>1e-10);
  // This condition over y achieves a precision over x of about 1e-15
  return _slowS[iLayer]/sqrt(x);
}

/*!
  Checks velocity inversions on Vp and Vs.
  The return value is -1 if no LVZ is found, else the first layer index where the condition is not satisfied.
*/
int QtbLayeredModel::checkVelocityInversion() const
{
  QTB_TRACE;
  for (int i=1;i<_layerCount;i++) {
    if (_slowS[i]>_slowS[i-1] || _slowP[i]>_slowP[i-1]) {
      return i;
    }
  }
  return -1;
}

/*!
 \fn  int QtbLayeredModel::n() const
 Returns the number of layers.
*/

/*!
 \fn  double * QtbLayeredModel::h( int iLayer )
 Returns the thickness of layer \a iLayer.
*/

/*!
 \fn  double * QtbLayeredModel::slowP( int iLayer )
 Returns the P-wave slownes of layer \a iLayer.
*/

/*!
 \fn  double * QtbLayeredModel::slowS( int iLayer )
 Returns the S-wave slowness of layer \a iLayer.
*/

/*!
 \fn  double * QtbLayeredModel::rho( int iLayer )
 Returns the density of layer \a iLayer.
*/

/*!
 \fn  double * QtbLayeredModel::mu( int iLayer )
 Returns the  of layer \a iLayer.
*/

/*!
  Returns the depth of the top of layer with index \a ilayer.
*/
double QtbLayeredModel::depth( int iLayer ) const
{
  QTB_TRACE;
  if (iLayer>=_layerCount) iLayer=_layerCount-1;
  double d=0;
  for (int i=0;i<iLayer;i++) d+=_h[i];
  return d;
}

/*!
  When a sub-stack of layers has been generated with a power law exponent,
  this function tries to recalculate the exponent from the values of S-wave
  slowness and thicknesses.
*/
double QtbLayeredModel::expGrad( int iLayer ) const
{
  QTB_TRACE;
  if (iLayer>=_layerCount-1) iLayer=_layerCount-2;
  double v1=1.0/_slowS[iLayer];
  double v2=1.0/_slowS[iLayer+1];
  double z0=depth(iLayer)+1;
  double z1=z0+0.5*_h[iLayer];
  double z2=z0+_h[iLayer]+0.5*_h[iLayer+1];
  double v21=v2/v1;
  double v211=v21-1;
  /* The problem is to solve f(alpha)=z2^alpha-v2/v1*z1^alpha+(v2/v1-1)*z0^alpha=v2/v1-1
     By Newton-Raphson the process is not stable enough, if transformed into log => too slow
     As f(0)<c and f(1)>c, and that the df/dalpha!=0 there's only one root, the solution
     By bisection it'is the best.
  */
  double a1=0;
  double a2=1;
  double a3=0.5;
  double f3=pow(z2,a3)-v21*pow(z1,a3)+v211*pow(z0,a3);
  do {
    if (f3<v211) a1=a3; else a2=a3;
    a3=0.5*(a1+a2);
    f3=pow(z2,a3)-v21*pow(z1,a3)+v211*pow(z0,a3);
  } while (f3!=v211 && a2-a1>1e-5);
  return a3;
}

/*!
  \fn QtbLayeredModel::roughFundamentalFrequency()
  Returns an estimation of the fundamental resonnance frequency

  \f[\frac {V_{soft}}{4 H_t}\f]

  \f$V_{soft}\f$ is the average velocity

  \f$H_t\f$ is the total thickness of layers except the half space.
*/
double QtbLayeredModel::roughFundamentalFrequency() const
{
  // TODO: replace by a more accurate estimation of resonance frequency
  QtbProfile p = slowsProfile();
  p.average();
  return p.values()[_layerCount-1] / (4*depth(_layerCount-1));
}

/*!
  Returns Poisson's ratio for layer \a iLayer
*/
double QtbLayeredModel::poisson( int iLayer ) const
{
  QTB_TRACE;
  double slowP2 = _slowP[iLayer];
  slowP2 *= slowP2;
  double slowS2 = _slowS[iLayer];
  slowS2 *= slowS2;
  return (slowP2-0.5*slowS2)/(slowP2-slowS2);
}

/*!
  Returns true if model is elastic (quality factors all set to 0)
*/
bool QtbLayeredModel::elastic() const
{
  for (int i=_layerCount-1;i>=0;i--) {
    if( _qp[i]>0.0 || _qs[i]>0.0 ) return false;
  }
  return true;
}

bool QtbLayeredModel::setValue( double& var, const QtbStringSection& field, int iLayer, const QString& varName, bool optional )
{
  QTB_TRACE;
  if (field.isValid()) {
    var = field.toDouble();
    if (var>0.0) {
      return true;
    } else {
      qtbApp << tr("Bad value for %1 at layer %2: %3").arg(varName).arg(iLayer).arg(var) << endl;
    }
  } else if (!optional) {
    qtbApp << tr("Missing value for %1 at layer %2").arg(varName).arg(iLayer) << endl;
  }
  return false;
}

QString QtbLayeredModel::formatHelp()
{
  QTB_TRACE;
  return "  Line 1    <number of layers including half-space for first model>\n"
         "  Line 2    <thickness (m)> <Vp (m/s)> <Vs (m/s)> <Density (kg/m3)>[ <Qp> <Qs>]\n"
         "  ....\n"
         "  Line n    0 <Vp (m/s)> <Vs (m/s)> <Density (kg/m3)>[ <Qp> <Qs>]\n"
         "  Line n+1  <number of layers including half-space for second model>\n"
         "  ....";
}

/*!
  Loads model from a text stream

    Format:
      \verbatim
      layerCount
      h Vp Vs rho Qp Qs (layer 1)
      ...
      0 Vp Vs rho Qp Qs (layer n)
      \endverbatim

  Qp and Qs are currently ignored.

  Returns false if an error occured during reading or if the read model is not consistent.
  It returns true even if the number of layers is null or cannot be read. This is the responsability
  of the caller to test if the number of layers is greater than 1 before doing anything else.
*/
bool QtbLayeredModel::fromStream( QTextStream& s, QString * comments )
{
  QTB_TRACE;
  // read the number of layers
  QtbStringSection field, fields;
  const QChar * ptr;
  QString buf;
  buf = QtbFile::readLineNoComments( s, comments );
  fields.set(buf);
  fields.trimmed();
  ptr = 0;
  field = fields.nextField( ptr );
  if (field.isValid()) {
    _layerCount = field.toInt();
  } else {
    _layerCount = 0;
    return true;  // This is just a blank line, before doing anything, you must test if the number of layer is >0
  }
  if (_layerCount>0 && !s.atEnd()) {
    allocateData();
    // read the data
    for (int i=0;i<_layerCount;i++) {
      if (s.atEnd()) {
        qtbApp << tr("Layered model: reaching the end of file, %1 layers found, expecting %2")
                  .arg(i).arg(_layerCount) << endl;
        _layerCount = 0;
        return false;
      }
      buf = QtbFile::readLineNoComments( s, comments );
      fields.set(buf);
      fields.trimmed();
      ptr = 0;
      if (i<_layerCount-1) {
        field = fields.nextField( ptr );
        if (!setValue( _h[i], field, i, tr("thickness"), false ) ) return false;
      } else {
        fields.nextField( ptr );
      }
      field = fields.nextField( ptr );
      if (!setValue( _slowP[i], field, i, tr("Vp"), false ) ) return false;
      field = fields.nextField( ptr );
      if (!setValue( _slowS[i], field, i, tr("Vs"), false ) ) return false;
      field = fields.nextField( ptr );
      if (!setValue( _rho[i], field, i, tr("density"), false ) ) return false;
      // Not mandatory quality factors
      field = fields.nextField( ptr );
      if (!setValue( _qp[i], field, i, tr("Qp"), true ) ) _qp[i] = 0.0;;
      field = fields.nextField( ptr );
      if (!setValue( _qs[i], field, i, tr("Qs"), true ) ) _qs[i] = 0.0;;
      // Switch to slowness
      _slowP[i]=1/_slowP[i];
      _slowS[i]=1/_slowS[i];
    }
    return initCalculation();
  } else {
    _layerCount = 0;
    return false;
  }
}

/*!
  Saves model to a text stream

    Format:
      \verbatim
      nLayers
      h Vp Vs rho Qp Qs (layer 1)
      ...
      0 Vp Vs rho Qp Qs (layer n)
      \endverbatim
*/
void QtbLayeredModel::toStream( QTextStream& s ) const
{
  QTB_TRACE;
  static const QString str4("%1 %2 %3 %4\n");
  static const QString str6("%1 %2 %3 %4 %5 %6\n");
  s << _layerCount << "\n";
  if (elastic()) {
    for (int i=0;i<_layerCount-1;i++)
      s << str4.arg( _h[i], 0, 'g', 20 )
               .arg( _slowP[i]==0.0 ? 0.0 : 1/_slowP[i], 0, 'g', 20)
               .arg( _slowS[i]==0.0 ? 0.0 : 1/_slowS[i], 0, 'g', 20)
               .arg( _rho[i], 0, 'g', 20);
    s << str4.arg( 0.0, 0, 'g', 20 )
             .arg( _slowP[_layerCount-1]==0.0 ? 0.0 : 1/_slowP[_layerCount-1], 0, 'g', 20)
             .arg( _slowS[_layerCount-1]==0.0 ? 0.0 : 1/_slowS[_layerCount-1], 0, 'g', 20)
             .arg( _rho[_layerCount-1], 0, 'g', 20) << flush;
   } else {
    for (int i=0;i<_layerCount-1;i++)
      s << str6.arg( _h[i], 0, 'g', 20 )
               .arg( _slowP[i]==0.0 ? 0.0 : 1/_slowP[i], 0, 'g', 20)
               .arg( _slowS[i]==0.0 ? 0.0 : 1/_slowS[i], 0, 'g', 20)
               .arg( _rho[i], 0, 'g', 20)
               .arg( _qp[i], 0, 'g', 20)
               .arg( _qs[i], 0, 'g', 20);
    s << str6.arg( 0.0, 0, 'g', 20 )
             .arg( _slowP[_layerCount-1]==0.0 ? 0.0 : 1/_slowP[_layerCount-1], 0, 'g', 20)
             .arg( _slowS[_layerCount-1]==0.0 ? 0.0 : 1/_slowS[_layerCount-1], 0, 'g', 20)
             .arg( _rho[_layerCount-1], 0, 'g', 20)
             .arg( _qp[_layerCount-1], 0, 'g', 20)
             .arg( _qs[_layerCount-1], 0, 'g', 20) << flush;
   }
}

/*!
  Returns model as a string
*/
QString QtbLayeredModel::toString() const
{
  QTB_TRACE;
  QString str, valErrStr("%1%2");
  str=QString("%1 layers\n").arg(_layerCount);
  str+="          H(m)       Vp(m/s)       Vs(m/s)    Rho(kg/m3)\n";
  for (int i=0;i<_layerCount-1;i++) {
    str+=QString::number(_h[i],'f',1).leftJustified(14,' ',true);
    str+=QString::number(1/_slowP[i],'f',1).leftJustified(14,' ',true);
    str+=QString::number(1/_slowS[i],'f',1).leftJustified(14,' ',true);
    str+=QString::number(_rho[i],'f',1).leftJustified(14,' ',true);
    str+="\n";
  }
  str+=QString("-").leftJustified(14,' ',true);
  str+=QString::number(1/_slowP[_layerCount-1],'f',1).leftJustified(14,' ',true);
  str+=QString::number(1/_slowS[_layerCount-1],'f',1).leftJustified(14,' ',true);
  str+=QString::number(_rho[_layerCount-1],'f',1).leftJustified(14,' ',true);
  str+="\n";
  return str;
}

/*!
  Saves model to Herrmann's format
*/
void QtbLayeredModel::exportHerrmann( QTextStream& s ) const
{
  QTB_TRACE;
  static const QString str("%1 %2 %3 %4 1\n");
  s << "2 1 0 2.25 250 1\n";
  for (int i=0;i<_layerCount-1;i++)
    s << str.arg( _h[i], 0, 'f', 15 )
            .arg( 1/_slowP[i], 0, 'f', 15)
            .arg( 1/_slowS[i], 0, 'f', 15)
            .arg( _rho[i], 0, 'f', 15);
  s << str.arg( 0.0, 0, 'f', 20 )
          .arg( 1/_slowP[_layerCount-1], 0, 'f', 20)
          .arg( 1/_slowS[_layerCount-1], 0, 'f', 20)
          .arg( _rho[_layerCount-1], 0, 'f', 20) << flush;
}

/*!
  Saves model to a binary stream
*/
void QtbLayeredModel::toStream( QDataStream& s ) const
{
  QTB_TRACE;
  s << _layerCount;
  for (int i=0;i<_layerCount-1;i++)
  {
    s << _h[i];
    s << _slowP[i];
    s << _slowS[i];
    s << _rho[i];
  }
  s << _slowP[_layerCount-1];
  s << _slowS[_layerCount-1];
  s << _rho[_layerCount-1];
}

/*!
  Loads model from a binary stream
*/
void QtbLayeredModel::fromStream( QDataStream& s )
{
  QTB_TRACE;
  int n;
  s >> n;
  if (n>_layerCount) n=_layerCount;
  QTB_ASSERT(n>0);
  for (int i=0;i<n-1;i++)
  {
    s >> _h[i];
    s >> _slowP[i];
    s >> _slowS[i];
    s >> _rho[i];
  }
  s >> _slowP[n-1];
  s >> _slowS[n-1];
  s >> _rho[n-1];
}

/*!
  Returns the P-wave slowness profile as a QtbProfile
*/
QtbProfile QtbLayeredModel::slowpProfile() const
{
  QTB_TRACE;
  QtbProfile p;
  p.resize( _layerCount );
  double d = 0;
  int i;
  for (i=0;i<_layerCount-1;i++) {
    d+= _h[i];
    p.setDepth( i, d );
    p.setValue( i, _slowP[i] );
  }
  p.setDepth( i, 1e99 );
  p.setValue( i, _slowP[i] );
  return p;
}

/*!
  Returns the S-wave slowness profile as a QtbProfile
*/
QtbProfile QtbLayeredModel::slowsProfile() const
{
  QTB_TRACE;
  QtbProfile p;
  p.resize( _layerCount );
  double d = 0;
  int i;
  for (i=0;i<_layerCount-1;i++) {
    d+= _h[i];
    p.setDepth( i, d );
    p.setValue( i, _slowS[i] );
  }
  p.setDepth( i, 1e99 );
  p.setValue( i, _slowS[i] );
  return p;
}

/*!
  Returns the density profile as a QtbProfile
*/
QtbProfile QtbLayeredModel::rhoProfile() const
{
  QTB_TRACE;
  QtbProfile p;
  p.resize( _layerCount );
  double d = 0;
  int i;
  for (i=0;i<_layerCount-1;i++) {
    d+= _h[i];
    p.setDepth( i, d );
    p.setValue( i, _rho[i] );
  }
  p.setDepth( i, 1e99 );
  p.setValue( i, _rho[i] );
  return p;
}

/*!
  Returns the Poisson's ratio profile as a QtbProfile
*/
QtbProfile QtbLayeredModel::poissonProfile() const
{
  QTB_TRACE;
  QtbProfile p;
  p.resize( _layerCount );
  double d = 0;
  int i;
  for (i=0;i<_layerCount-1;i++) {
    d+= _h[i];
    p.setDepth( i, d );
    p.setValue( i, poisson(i) );
  }
  p.setDepth( i, 1e99 );
  p.setValue( i, poisson(i) );
  return p;
}

/*!
  Returns the impedance (density*S-wave velocity) profile as a QtbProfile
*/
QtbProfile QtbLayeredModel::impedanceProfile() const
{
  QTB_TRACE;
  QtbProfile p;
  p.resize( _layerCount );
  double d = 0;
  int i;
  for (i=0;i<_layerCount-1;i++) {
    d+= _h[i];
    p.setDepth( i, d );
    p.setValue( i, _rho[i] / _slowS[i] );
  }
  p.setDepth( i, 1e99 );
  p.setValue( i, _rho[i] / _slowS[i] );
  return p;
}

double QtbLayeredModel::vpAt(double depth) const
{
  QTB_TRACE;
  double d=0;
  for (int i=0;i<_layerCount-1;i++) {
    d+=_h[i];
    if (depth<d) return 1/_slowP[i];
  }
  return 1/_slowP[_layerCount-1];
}

double QtbLayeredModel::vsAt(double depth) const
{
  QTB_TRACE;
  double d=0;
  for (int i=0;i<_layerCount-1;i++) {
    d+=_h[i];
    if (depth<d) return 1/_slowS[i];
  }
  return 1/_slowS[_layerCount-1];
}

double QtbLayeredModel::slowPAt(double depth) const
{
  QTB_TRACE;
  double d=0;
  for (int i=0;i<_layerCount-1;i++) {
    d+=_h[i];
    if (depth<d) return _slowP[i];
  }
  return _slowP[_layerCount-1];
}

double QtbLayeredModel::slowSAt(double depth) const
{
  QTB_TRACE;
  double d=0;
  for (int i=0;i<_layerCount-1;i++) {
    d+=_h[i];
    if (depth<d) return _slowS[i];
  }
  return _slowS[_layerCount-1];
}

double QtbLayeredModel::rhoAt(double depth) const
{
  QTB_TRACE;
  double d=0;
  for (int i=0;i<_layerCount-1;i++) {
    d+=_h[i];
    if (depth<d) return _rho[i];
  }
  return _rho[_layerCount-1];
}

double QtbLayeredModel::poissonAt(double depth) const
{
  QTB_TRACE;
  double d=0;
  for (int i=0;i<_layerCount-1;i++) {
    d+=_h[i];
    if (depth<d) return poisson(i);
  }
  return poisson(_layerCount-1);
}

double QtbLayeredModel::impedanceAt(double depth) const
{
  QTB_TRACE;
  double d=0;
  for (int i=0;i<_layerCount-1;i++) {
    d+=_h[i];
    if (depth<d) return _rho[i] / _slowS[i];
  }
  return _rho[_layerCount-1] / _slowS[_layerCount-1];
}

void QtbLayeredModel::changeHKeepOtherDepth( int iLayer, double newH )
{
  QTB_TRACE;
  if (iLayer>=_layerCount-1) return;
  if (iLayer==_layerCount-2 || newH>=_h[iLayer]+_h[iLayer+1]) {
    _h[iLayer]=newH;
    return;
  }
  double dH=newH-_h[iLayer];
  _h[iLayer]=newH;
  _h[iLayer+1]-=dH;
}

void QtbLayeredModel::changeVpKeepPoisson( int iLayer, double newV )
{
  QTB_TRACE;
  if (iLayer>_layerCount-1) return;
  // Get actual Poisson's ratio (equivalent to Vs over Vp)
  // ratio=Vp/Vs=(1/Vs)/(1/Vp)=slowS/slowP
  double ratio=_slowS[iLayer]/_slowP[iLayer];
  _slowP[iLayer]=1.0/newV;
  _slowS[iLayer]=ratio*_slowP[iLayer];
}

void QtbLayeredModel::changeVsKeepPoisson( int iLayer, double newV )
{
  QTB_TRACE;
  if (iLayer>_layerCount-1) return;
  // Get actual Poisson's ratio (equivalent to Vs over Vp)
  // ratio=Vs/Vp=(1/Vp)/(1/Vs)=slowP/slowS
  double ratio=_slowP[iLayer]/_slowS[iLayer];
  _slowS[iLayer]=1.0/newV;
  _slowP[iLayer]=ratio*_slowS[iLayer];
}

void QtbLayeredModel::setHFrom( const QtbLayeredModel& o )
{
  QTB_TRACE;
  if (_layerCount!=o._layerCount) return;
  for (int i=_layerCount-2;i>=0;i--) _h[i]=o._h[i];
}
