/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-09-22
**
***************************************************************************/

#include "qtbmodalcurve.h"
#include "qtbvalue.h"
#include "qtbmodalfactory.h"
#include "qtbmemorychecker.h"

const QString QtbModalCurve::xmlModalCurveTag = "ModalCurve";

/*!
  \class QtbModalCurve qtbmodalcurve.h
  \brief Brief description of class still missing

  Full description of class still missing
*/

/*!
  \fn QtbModalCurve::QtbModalCurve()
  Construct an empty curve
*/

/*!
  \fn QtbModalCurve::QtbModalCurve( const QtbModalCurve& p )
  Copy constructor
*/

/*!
  \fn QtbModalCurve::QtbModalCurve( int n )
  Construct a curve with \a n points.
*/

/*!
  \fn QtbModalCurve::QtbModalCurve( const QtbCurve<QtbStatPoint>& o )
  Construct a curve from \a o
*/

QtbModalCurve::~QtbModalCurve()
{
  QTB_TRACE;
}

void QtbModalCurve::operator=( const QtbCurve<QtbStatPoint>& o)
{
  QTB_TRACE;
  int n = o.count();
  resize( n );
  for( int i = 0; i<n; i++ ) {
    (*this)[i]=o[i];
  }
  _modes.clear();
  _log = "";
  _name = "";
}

void QtbModalCurve::clear()
{
  QTB_TRACE;
  QtbCurve<QtbFactoryPoint>::clear();
  _modes.clear();
  _log = "";
  _name = "";
}

void QtbModalCurve::addMode( const QtbMode& m )
{
  QTB_TRACE;
  _modes.append( m );
}

void QtbModalCurve::removeMode( int index )
{
  QTB_TRACE;
  if ( index < _modes.count() ) _modes.removeAt( index );
}

bool QtbModalCurve::hasMode( const QtbMode& m ) const
{
  QTB_TRACE;
  int n = _modes.count();
  for ( int i = 0; i<n; i++ ) {
    if (_modes.at(i)==m) {
      return true;
    }
  }
  return false;
}

/*!
  \fn void QtbModalCurve::sort()
  Sort all samples by increasing X. Necessary for entering into a QtbModalFactory
*/

/*!
  Set the index in vector \a x of each sample of the curve. Each sample of the curve must have its
  corresponding value in vector \a x.
*/
void QtbModalCurve::linkX( const QVector<double>& x )
{
  QTB_TRACE;
  int n = count();
  int nx = x.count();
  int ix = 0;
  for (int i = 0; i< n; i++) {
    while (ix<nx && x.at(ix) < at(i).x()) ix++;
    QTB_ASSERT( at(i).x() == x.at(ix) );
    (*this)[i].setIndex(ix);
  }
}

int QtbModalCurve::indexOf( int factoryIndex ) const
{
  QTB_TRACE;
  int n = count();
  for (int i = 0; i< n; i++) {
    const QtbFactoryPoint& p = at(i);
    if ( p.index() == factoryIndex ) return i;
  }
  return -1;
}

/*!
  Calculate the squared misfit of \a values with respect to this curve.

  \a nValues is incremented by the number of valid values where the comparison was possible.
  \a nData is incremented by the number of valid data where the comparison is possible. In
  all cases, \a nData >= \a nValues and for the best matching case, \a nValues == \a nData.

  \a minMisfit2 is the square of the minimum misfit value that can be reached for each sample individualy.
*/
double QtbModalCurve::misfit( int& nValues, int& nData, const QtbModalFactory& f, QtbStatValue::MisfitType type, double minMisfit ) const
{
  QTB_TRACE;
  double res=0;
  int nx = count();
  int nm = _modes.count();
  if ( nm == 1 ) {
    const QtbValue * values = f.mode( _modes.first() );
    for (int i=0;i<nx;i++) {
      const QtbFactoryPoint& p = at(i);
      double pointMisfit = p.misfit( nValues, nData, values[ p.index() ], type, minMisfit );
      if (pointMisfit>0.0) {
        res+=pointMisfit;
      }
    }
  } else if ( nm > 1 ) {
    for ( int i=0; i<nx; i++ ) {
      const QtbFactoryPoint& p = at(i);
      if ( !p.isValid() ) continue;
      nData++;
      double modalMisfit,minModalMisfit=1e99;
      int localNData = 0, localNValues = 0;
      for (int im=0;im<nm;im++) {
        const QtbValue * values = f.mode( _modes.at(im) );
        localNValues = 0;
        modalMisfit=p.misfit( localNData, localNValues, values[ p.index() ], type, minMisfit);
        if ( localNValues == 1 ) {
          if ( modalMisfit<minModalMisfit ) {
            minModalMisfit=modalMisfit;
          }
        }
      }
      if ( minModalMisfit!=1e99 ) {
        res+=minModalMisfit;
        nValues++;
      }
    }
  }
  return res;
}

void QtbModalCurve::xml_writeProperties(XML_WRITEPROPERTIES_ARGS) const
{
  QTB_TRACE;
  Q_UNUSED(context);
  writeProperty(s,"name",_name);
  writeProperty(s,"log",_log);
}

void QtbModalCurve::xml_writeChildren(XML_WRITECHILDREN_ARGS) const
{
  QTB_TRACE;
  for (QList<QtbMode>::ConstIterator it = _modes.begin(); it!=_modes.end(); it++ ) {
    it->xml_save(s, context);
  }
  const_iterator it;
  for ( it=begin(); it!=end(); ++it ) {
    it->xml_save(s, context);
  }
}

QtbXMLMember QtbModalCurve::xml_member(XML_MEMBER_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes)
  Q_UNUSED(context);
  if (tag=="StatPoint") {
    append( QtbFactoryPoint() );
    return QtbXMLMember( &last() );
  } else if (tag=="Mode") {
    _modes.append( QtbMode() );
    return QtbXMLMember( &_modes.last() );
  } else if (tag=="name") return QtbXMLMember(0);
  else if (tag=="log") return QtbXMLMember(1);
  else return QtbXMLMember(QtbXMLMember::Unknown);
}

bool QtbModalCurve::xml_setProperty(XML_SETPROPERTY_ARGS)
{
  QTB_TRACE;
  Q_UNUSED(attributes)
  Q_UNUSED(context);
  switch (memberID) {
  case 0: _name=content.toString(); return true;
  case 1: _log=content.toString(); return true;
  default: return false;
  }
}
