/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-09-26
**
***************************************************************************/

#include "qtbmodalfactory.h"
#include "qtbmodalstorage.h"
#include "qtbmodalcurve.h"
#include "qtbcoreapplication.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbModalFactory qtbmodalfactory.h
  \brief Gathers QtbModalStorage to compute misfit for Rayleigh/Love and Phase/Group slowness

  Initialized is achieved with three distinct steps:

  \li 1.a. Repeated calls to setX() to build a list gathering all X of the curves
  \li 1.b. Repeated calls to linkX() to link points of the curves to X list indexes
  \li 2.   Repeated calls to setModes() to set the number of modes according to curves needs.
  \li 3.   Switch to angular frequency

  Never mix these steps.
*/

/*!
  \fn QtbModalFactory::validate() = 0
  After setting omegas, you can adjust the number of modes between various storages.
  e.g. if group slowness is present force storage phase (necessary step for computing
  group slowness).
*/

/*!
  \fn QtbModalFactory::calculate() = 0
  Fill in the modal storages by calculating theoretical
*/

/*!
  Merge x vectors of the two factories
*/
void QtbModalFactory::setX ( QtbModalFactory& f )
{
  QTB_TRACE;
  _x.reserve(_x.count()+f._x.count());
  for (QVector<double>::Iterator it = f._x.begin(); it != f._x.end(); it++ ) {
    _x.append( *it );
  }
  Qtb::uniq( _x );
  f._x = _x;
}

/*!
  Repeat calls to setX()
*/
void QtbModalFactory::setX ( const QList<QtbModalCurve>& curves )
{
  QTB_TRACE;
  for(QList<QtbModalCurve>::const_iterator it = curves.begin();it!=curves.end();++it) {
    setX(*it);
  }
}

/*!
  Repeat calls to linkX()
*/
void QtbModalFactory::linkX ( QList<QtbModalCurve>& curves ) const
{
  QTB_TRACE;
  for(QList<QtbModalCurve>::iterator it = curves.begin();it!=curves.end();++it) {
    linkX(*it);
  }
}

/*!
  Repeat calls to setModes()
*/
void QtbModalFactory::setModes ( const QList<QtbModalCurve>& curves )
{
  QTB_TRACE;
  for(QList<QtbModalCurve>::const_iterator it = curves.begin();it!=curves.end();++it) {
    setModes(*it);
  }
}

/*!
  Must be called before setModes()
*/
void QtbModalFactory::setX ( const QtbModalCurve& c )
{
  QTB_TRACE;
  _x.reserve(_x.count()+c.count());
  for (QtbModalCurve::const_iterator it = c.begin(); it != c.end(); it++ ) {
    _x.append( it->x() );
  }
  Qtb::uniq( _x );
}

/*!
  Must be called before setModes()
*/
void QtbModalFactory::linkX ( QtbModalCurve& c ) const
{
  QTB_TRACE;
  if ( _x.isEmpty() ) {
    qtbApp << tr("### ERROR ### : omega list not set or empty when linking omegas.") << endl;
    return;
  }
  c.linkX( _x );
}

void QtbModalFactory::setModes ( const QtbModalCurve& c )
{
  QTB_TRACE;
  if ( _x.isEmpty() ) {
    qtbApp << tr("### ERROR ### : X list not set or empty when setting modes.") << endl;
    return;
  }
  const QList<QtbMode>& l = c.modes();
  for ( QList<QtbMode>::ConstIterator it = l.begin(); it!=l.end(); it++ ) {
    setMode( *it );
  }
}

/*!
  Switch between omega and frequency.
*/
void QtbModalFactory::setAngularFrequency()
{
  QTB_TRACE;
  double factor=2*M_PI;
  for (QVector<double>::iterator it = _x.begin(); it != _x.end(); it++ ) {
    (*it) *= factor;;
  }
}
