/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-09-25
**
***************************************************************************/

#include "qtbmodalstorage.h"

/*!
  \class QtbModalStorage qtbmodalstorage.h
  \brief Storage for modal computation
 
  Storage for modal computation such as dispersion curve, ellipticity or spac
*/

/*!
  Init a empty storage. Use set() before using this object.
*/
QtbModalStorage::QtbModalStorage()
{
  QTB_TRACE;
  _nModes = 0;
  _nX = 0;
  _values = 0;
  _x = 0;
}

/*!
  \fn QtbModalStorage::QtbModalStorage( int nModes, const QVector<double> * x )
  Init storage with \a nModes and with the list of x \a o (not owned by this object)
*/

/*!
  Init storage with \a nModes and with the list of x \a o (not owned by this object)
*/
void QtbModalStorage::set( int nModes, const QVector<double> * x )
{
  QTB_TRACE;
  _nModes = nModes;
  _nX = x->count();
  _values = new QtbValue [ _nModes * _nX ];
  _x = x;
}
/*!
  \fn QtbModalStorage::~QtbModalStorage()
  Delete stored values but not the list of x provided in constructor's arguments
*/

/*!
  Copy operator, only list of x remains shared.
*/
QtbModalStorage& QtbModalStorage::operator=( const QtbModalStorage& o )
{
  QTB_TRACE;
  _nModes = o._nModes;
  _nX = o._nX;
  _x = o._x;
  int nValues = _nModes * _nX;
  _values = new QtbValue [ nValues ];
  memcpy ( _values, o._values, nValues*sizeof(QtbValue) );
  return *this;
}

/*!
  Transform all values into its log10.
*/
void QtbModalStorage::log10()
{
  int nValues = _nModes * _nX;
  for (int i = 0; i<nValues;i++) {
    _values[i].log10();
  }
}

/*!
  Transform all values into its exp10
*/
void QtbModalStorage::exp10()
{
  int nValues = _nModes * _nX;
  for (int i = 0; i<nValues;i++) {
    _values[i].exp10();
  }
}

/*!
  \fn QtbModalStorage::iterator::iterator( const QtbModalStorage * s, int iMode )
  Initialize an iterator for fixed and refined values
*/

/*!
  \fn iterator& QtbModalStorage::iterator::operator++()
  increment iterator for fixed and refined vaues
*/

/*!
  \fn bool QtbModalStorage::iterator::operator!=( const QtbModalStorage::iterator& o ) const
  Compare two iterators
*/

/*!
  \fn QtbModalStorage::iterator::x() const
  Return current x
*/

/*!
  \fn QtbValue& QtbModalStorage::iterator::value()
  Return current value
*/

/*!
  Output to stream \a s, mixing refines and original samples, removing invalid values
*/
void QtbModalStorage::toStream( QTextStream& s, int modeCount ) const
{
  QTB_TRACE;
  static const QString fmt("%1 %2\n");
  if( modeCount == 0 || modeCount>_nModes ) modeCount=_nModes;
  double xFactor = 0.5/M_PI;
  for ( int im = 0; im < modeCount;im++ ) {
    s << "# Mode " << im << "\n";
    for ( Iterator it( this, im ); !it.atEnd(); ++it ) {
      const QtbValue & val = it.value();
      if ( val.isValid() )
        s << fmt.arg(it.x()* xFactor, 0, 'g', 15).arg( val.value(), 0, 'g', 15 );
    }
  }
}

/*!
  Return a curve for mode \a iMode, mixing refines and original samples, removing invalid values
*/
QtbCurve<QtbPoint2D> QtbModalStorage::curve( int iMode ) const
{
  double xFactor = 0.5/M_PI;
  QtbCurve<QtbPoint2D> points;
  for ( Iterator it( this, iMode ); !it.atEnd(); ++it ) {
    const QtbValue & val = it.value();
    if ( val.isValid() )
      points.append( QtbPoint2D( it.x() * xFactor, val.value() ) );
  }
  return points;
}


/*!
  Save all to report (including frequency), mixing refines and original samples, removing invalid values

  Begin with a table of the number of samples per mode

  4                 nm Modes
  For each mode:
  4                 number of samples for this mode mode, nx[im]
  nx[im]*2*8        values x,y
*/
void QtbModalStorage::writeReport( QDataStream& s ) const
{
  QTB_TRACE;
  double xFactor = 0.5/M_PI;
  s << _nModes;
  for ( int im = 0; im < _nModes;im++ ) {
    // room for sample per mode table
    int xCountReal = 0;
    qint64 nSamplesPos = s.device() ->pos();
    s << xCountReal;
    for ( Iterator it( this, im ); !it.atEnd(); ++it ) {
      const QtbValue & val = it.value();
      if ( val.isValid() ) {
        s << it.x() * xFactor << val.value();
        xCountReal++;
      }
    }
    // Write the real number of x at the beginning
    qint64 curPos = s.device() ->pos();
    s.device() ->seek( nSamplesPos );
    s << xCountReal;
    s.device() ->seek( curPos );
  }
}
