/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2006-09-25
**
***************************************************************************/

#ifndef QTB_MODALSTORAGE_H
#define QTB_MODALSTORAGE_H

#include "qtbcurve.h"
#include "qtbvalue.h"
#include "qtbmodalrefine.h"
#include "qtbwavedllexport.h"

class QTBWAVE_EXPORT QtbModalStorage
{
public:
  QtbModalStorage();
  QtbModalStorage( int nModes, const QVector<double> * x ) { set( nModes, x ); }
  ~QtbModalStorage() { delete [] _values; }

  void set( int nModes, const QVector<double> * x );
  QtbModalStorage& operator=( const QtbModalStorage& o );

  void log10();
  void exp10();

  int xCount() const { return _nX; }
  int modeCount() const { return _nModes; }
  double x( int ix ) const { return ( *_x ) [ ix ]; }
  QtbValue * mode( int index ) { return _values + index * _nX; }
  const QtbValue * mode( int index ) const { return _values + index * _nX; }

  int refineCount() const { return _refines.count();}
  QtbModalRefine& refineAdd( double x ) {
    _refines.append( QtbModalRefine( _nModes, x ) );
    _refines.last().setValid(false);
    return _refines.last();
  }
  void refineClear() {_refines.clear();}
  void refineSort() {qSort( _refines );}
  QtbModalRefine& lastRefine() { return _refines.last();}
  QtbRefineIterator refineBegin() const { return _refines.begin(); }
  QtbRefineIterator refineEnd() const { return _refines.end(); }

  void toStream( QTextStream& s, int modeCount = 0 ) const;
  QtbCurve<QtbPoint2D> curve( int iMode ) const;

  void writeReport( QDataStream& s ) const;

  class Iterator {
  public:
    inline Iterator( const QtbModalStorage * s, int iMode );
    inline bool atEnd() const;
    inline Iterator &operator++();
    inline double x() const;
    inline const QtbValue& value();
  private:
    const QtbModalStorage * _storage;
    double _nextFixed, _nextRefine;
    int _fixedIterator;
    QtbRefineIterator _refineIterator;
    int _iMode;
    const QtbValue * _values;
  };

private:
  int _nX, _nModes;
  const QVector<double> * _x;
  QtbValue * _values;
  QtbRefineList _refines;
};

inline QtbModalStorage::Iterator::Iterator( const QtbModalStorage * s, int iMode )
{
  QTB_TRACE;
  _storage = s;
  _nextFixed = _storage->x( 0 );
  _fixedIterator = 0;
  _refineIterator = _storage->refineBegin();
  if ( _refineIterator != _storage->refineEnd() ) {
    _nextRefine = _refineIterator->x();
  } else {
    _nextRefine = 1e99;
  }
  _iMode = iMode;
  _values = _storage->mode( _iMode );
}

inline QtbModalStorage::Iterator& QtbModalStorage::Iterator::operator++()
{
  if ( _nextFixed < _nextRefine ) {
    _fixedIterator++;
    if ( _fixedIterator < _storage->xCount() )
      _nextFixed = _storage->x( _fixedIterator );
    else {
      _nextFixed = 1e99;
    }
  } else {
    ++_refineIterator;
    if ( _refineIterator != _storage->refineEnd() )
      _nextRefine = _refineIterator->x();
    else {
      _nextRefine = 1e99;
    }
  }
  return *this;
}

inline bool QtbModalStorage::Iterator::atEnd() const
{
  QTB_TRACE;
  return _fixedIterator == _storage->xCount() && _refineIterator == _storage->refineEnd();
}

inline double QtbModalStorage::Iterator::x() const
{
  QTB_TRACE;
  if ( _nextFixed < _nextRefine ) {
    return _nextFixed;
  } else {
    return _nextRefine;
  }
}

inline const QtbValue& QtbModalStorage::Iterator::value()
{
  if ( _nextFixed < _nextRefine ) {
    return _values[_fixedIterator] ;
  } else {
    return _refineIterator->value( _iMode ) ;
  }
}

#endif // QTB_MODALSTORAGE_H
