/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2005-12-21
**
***************************************************************************/

#include <QDataStream>

#include "qtbprofile.h"
#include "qtbpoint2d.h"
#include "qtbcoreapplication.h"

/*!
  \class QtbProfile qtbprofile.h
  \brief Storage for a 1D profile.

  A profile is made of layers with a uniform value over the depth axis.
  Internally 2 vectors ar maintained, one for the depths and the other for
  the values of the parameter. The first depth is not null. It is the bottom
  depth of the first layer.
*/


/*!
  Set this profile as the resample of p with baseD. All depths of p must belong to baseD
*/
void QtbProfile::resample( const QtbProfile& p, const QVector<double>& baseD )
{
  QTB_TRACE;
  _depths.clear();
  _values.clear();
  QVector<double>::const_iterator itBase=baseD.begin();
  int i=0, n=p._depths.count();
  double nextDepth = p._depths.at(i);
  double value = p._values.at(i);
  double depth;
  while (true) {
    depth=*(itBase++);
    _depths.append(depth);
    _values.append(value);
    if ( depth == nextDepth ) {
      ++i;
      if (i>=n) break;
      nextDepth = p._depths.at(i);
      value = p._values.at(i);
    }
  }
  /*qtbApp << "resample( const QtbProfile& p, const QList<double>& baseD )" << endl;
  int nd=_depths.count();
  for(int i=0;i<nd;i++) {
    qtbApp << "  " << _depths.at(i) << " ; " << _values.at(i) << endl;
  }*/
}

/*!
  Add samples from \a baseD to this profile if they do not exist already. Depths must be ordered
*/
void QtbProfile::resample( const QVector<double>& baseD )
{
  QTB_TRACE;
  /*qtbApp << "resample( const QList<double>& baseD ): baseD" << endl;
  for(int i=0;i<baseD.count();i++) {
    qtbApp << "  " << baseD.at(i) << endl;
  }*/
  int i = 0, n = count();
  for ( QVector<double>::const_iterator it=baseD.begin(); it!=baseD.end(); it++ ) {
    double depth = *it;
    while( i<n && _depths.at(i)<depth ) i++;
    if (i<n) {
      if ( _depths.at(i)>depth ) {
        _depths.insert(i, depth );
        _values.insert(i, _values[i] );
        i++; n++;
      }
    } else break;
  }
  /*qtbApp << "resample( const QList<double>& baseD ): result" << endl;
  for(int i=0;i<_depths.count();i++) {
    qtbApp << "  " << _depths.at(i) << " ; " << _values.at(i) << endl;
  }*/
}

/*!
  Remove all samples that are not in \a baseD. Depths must be ordered
*/
void QtbProfile::setSamples( const QVector<double>& baseD )
{
  QTB_TRACE;
  /*qtbApp << "setSamples( const QList<double>& baseD ): baseD" << endl;
  for(int i=0;i<baseD.count();i++) {
    qtbApp << "  " << baseD.at(i) << endl;
  }*/
  int i = 0, n = count();
  for ( QVector<double>::const_iterator it=baseD.begin(); it!=baseD.end(); it++ ) {
    double depth = *it;
    while( i<n && _depths.at(i)<depth ) {
      _depths.remove( i );
      _values.remove( i );
      n--;
    }
    i++;
    if (i>=n) break;
  }
  while( i<n ) {
      _depths.remove( i );
      _values.remove( i );
      n--;
  }
  /*qtbApp << "setSamples( const QList<double>& baseD ): result" << endl;
  for(int i=0;i<_depths.count();i++) {
    qtbApp << "  " << _depths.at(i) << " ; " << _values.at(i) << endl;
  }*/
}

/*
  Returns the value at \a depth assuming continuous change of the value (interpolation).
*/
double QtbProfile::interpoleAt(double depth)
{
  QTB_TRACE;
  int i=0;
  int n=_depths.count();
  if (n<2) return 0.0;
  while (i<n && depth<_depths[i]) i++;
  if (i>=n-1) i=n-2;
  return _values[i]+(_values[i+1]-_values[i])/(_depths[i+1]-_depths[i])*(depth-_depths[i]);
}

/*
  Returns the value at \a depth assuming uniform layers.
*/
double QtbProfile::uniformAt( double depth )
{
  QTB_TRACE;
  int i=0;
  int n=_depths.count();
  while ( i<n && depth>_depths[i] ) i++;
  if (i==n) i--;
  return _values[i];
}

/*!
  Calculates the average values.
  The average is defined by a mean weighted by the
  thicknesses. If the parameter is a velocity, it is the total
  thickness down to a depth divided by the total time needed
  to cross the layer stack to that depth.
*/
void QtbProfile::average()
{
  QTB_TRACE;
  int n = _depths.count();
  if ( n == 0 ) return;
  double sum = _values[0]*_depths[0];
  for( int i = 1; i<n; i++ ) {
    sum += _values[i]*(_depths[i]-_depths[i-1]);
    _values[i] = sum / _depths[i];
  }
}

/*!
  Inverse values
*/
void QtbProfile::inverse()
{
  QTB_TRACE;
  int n = _depths.count();
  for( int i = 0; i<n; i++ ) {
    _values[i] = 1.0 / _values[i];
  }
}

void QtbProfile::report2plot(QDataStream& s, QtbPoint2D * points)
{
  QTB_TRACE;
  int n;
  s >> n;
  double lastDepth=0.0;
  double depth,value;
  for (int i=0;i<n;i++) {
    s >> depth;
    s >> value;
    points[0].setY( lastDepth );
    points[0].setX( value );
    points[1].setY( depth );
    points[1].setX( value );
    points+=2;
    lastDepth=depth;
  }
}

QtbCurve<QtbPoint2D> QtbProfile::curve() const
{
  QTB_TRACE;
  int n = count();
  double lastDepth=0.0;
  QtbCurve<QtbPoint2D> c;
  for (int i=0;i<n;i++) {
    c.append(QtbPoint2D( _values[i], lastDepth ));
    lastDepth=_depths[i];
    c.append(QtbPoint2D( _values[i], lastDepth ));
  }
  return c;
}

void QtbProfile::toStream(QTextStream& s, bool plotMode) const
{
  QTB_TRACE;
  static const QString fmt("%1 %2\n");
  int n = count();
  if(plotMode) {
    double lastDepth=0.0;
    for (int i=0;i<n;i++) {
      s << fmt.arg(_values[i], 20, 'g', 15).arg(lastDepth, 20, 'g', 15);
      lastDepth=_depths[i];
      s << fmt.arg(_values[i], 20, 'g', 15).arg(lastDepth, 20, 'g', 15);
    }
  } else {
    for (int i=0;i<n;i++) {
      s << fmt.arg(_values[i], 20, 'g', 15).arg(_depths[i], 20, 'g', 15);
    }
  }
  s << flush;
}

void QtbProfile::print() const
{
  QTB_TRACE;
  int n = count();
  for (int i=0;i<n;i++) {
    printf( "%20.15lg %20.15lg\n", _values[i], _depths[i] );
  }
}
