/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-04-28
**
***************************************************************************/

#ifndef QTB_RAYLEIGH_H
#define QTB_RAYLEIGH_H

#include "qtbwavedllexport.h"
#include "qtblayeredmodel.h"
#include "qtbnumber.h"

template<class QtbRealType>
    class QtbRayleighTemplate
{
public:
  QtbRayleighTemplate( const QtbLayeredModel * model ) { _model = model; }

  const QtbLayeredModel * model() const { return _model; }

  void setOmega(double o) { _omega=o; }
  double ellipticity() { return QtbNumber::toDouble(iF1214/F1213); }
  QtbRealType y( QtbRealType slowness );
private:
  const QtbLayeredModel * _model;
  /* Dunkin determinants for the surperficial layer (once y(double) has run,
     else undetermined) to compute ellipticity
  */
  QtbRealType F1213;
  QtbRealType iF1214;
  double _omega;
};

// Usual precision for efficient computations
typedef QtbRayleighTemplate<double> QtbRayleigh;
#ifdef QTB_BIGNUM
// Arbitrary precision for special cases or experimental purposes
typedef QtbRayleighTemplate<mp_real> QtbRayleighBigNum;
#endif

template<class QtbRealType>
    QtbRealType QtbRayleighTemplate<QtbRealType>::y(QtbRealType slowness)
{
  QtbRealType tmpAbs;
  //printf("f=%.20lg s=%.20lg v=%.20lg\n",_omega/(2*M_PI),slowness,1.0/slowness);
  QtbRealType w2=(QtbRealType)_omega*(QtbRealType)_omega;
  QtbRealType k=(QtbRealType)_omega*slowness;
  QtbRealType k2=k*k;
  QtbRealType inv_k2=(QtbRealType)1.0/k2;
  // Half space
  register int i=_model->layerCount()-1;
  QtbRealType ka=(QtbRealType)_omega*(QtbRealType)_model->slowP(i);
  QtbRealType hn2=k2-ka*ka;
  tmpAbs=hn2;
  if (tmpAbs<0) tmpAbs=-tmpAbs;
  QtbRealType hn=sqrt(tmpAbs);
  QtbRealType kb=(QtbRealType)_omega*(QtbRealType)_model->slowS(i);
  QtbRealType kn2=k2-kb*kb;
  tmpAbs=kn2;
  if (tmpAbs<0) tmpAbs=-tmpAbs;
  QtbRealType kn=sqrt(tmpAbs);
  QtbRealType ln=k2+kn2;
  QtbRealType mu=(QtbRealType)_model->mu(i);
  QtbRealType c1=hn*kn;
  // Factor used to balance the values in the plane (k,s)
  //QtbRealType invFac=w2*k2*slowness;
  QtbRealType invFac=w2;
  QtbRealType fac=1.0/invFac;
  QtbRealType F1212 = mu*fac*(ln*ln-4.0*k2*c1);
  F1213 = hn*(ln-2.0*k2);
  iF1214 = k*(ln-2.0*c1);
  // iF1223 = iF1214
  QtbRealType F1224 = kn*(2.0*k2-ln);
  QtbRealType F1234 = (k2-c1)/mu;
  for (i--;i>=0;i--) {
    ka=(QtbRealType)_omega*(QtbRealType)_model->slowP(i);
    hn2=k2-ka*ka;
    tmpAbs=hn2;
    if (tmpAbs<0) tmpAbs=-tmpAbs;
    hn=sqrt(tmpAbs);
    hn2*=inv_k2;
    kb=(QtbRealType)_omega*(QtbRealType)_model->slowS(i);
    kn2=k2-kb*kb;
    tmpAbs=kn2;
    if (tmpAbs<0) tmpAbs=-tmpAbs;
    kn=sqrt(tmpAbs);
    kn2*=inv_k2;
    // Compute the hyperbolic sin and cos
    QtbRealType SH, CH, he, exphn;
    QtbRealType dn=(QtbRealType)_model->h(i);
    exphn=hn*dn;
    if (hn2>0.0) {
      he=exphn;
      QtbRealType exp2;
      //if (exphn<21.2) { // See Love for explanations
      if (exphn<115) {
        exphn=exp(-exphn);
        exp2=exphn*exphn;
      } else {
        exphn=0.0;
        exp2=0.0;
      }
      SH=0.5*k/hn*(1.0-exp2);
      CH=0.5*(1.0+exp2);
    } else if (hn2<0.0) {
      he=0.0;
      // sinh is imaginary but hn as well, thus SH is real
      SH=k/hn*sin(exphn);
      CH=cos(exphn);
      exphn=1.0;
    } else {
      he=0.0;
      SH=0.0;
      CH=1.0;
      exphn=1.0;
    }
    QtbRealType SK, CK, ke, expkn;
    expkn=kn*dn;
    if (kn2>0.0) {
      ke=expkn;
      QtbRealType exp2;
      if (expkn<21.2) { // See Love for explanations (double precision)
      //if (expkn<115) { // See Love for explanations (big number precise until 1e-50)
        expkn=exp(-expkn);
        exp2=expkn*expkn;
      } else {
        expkn=0.0;
        exp2=0.0;
      }
      SK=0.5*k/kn*(1.0-exp2);
      CK=0.5*(1.0+exp2);
    } else if (kn2<0.0) {
      ke=0.0;
      // sinh is imaginary but kn as well, thus SK is real
      SK=k/kn*sin(expkn);
      CK=cos(expkn);
      expkn=1;
    } else {
      ke=0.0;
      SK=0.0;
      CK=1.0;
      expkn=1.0;
    }
    // In G, all terms that does not contains either CHCK,... we must multiply by expCorr
    QtbRealType expCorr=exphn*expkn;
    // Each cosh or sinh has to be multiplied by either exp(he) or exp(ke)
    // he is always > ke, thus exp(he)>exp(ke).
    // In each term, CHCK, CHSK, ... there is a factor exp(he+ke)
    QtbRealType CHCK=CH*CK;
    QtbRealType SHSK=SH*SK;
    QtbRealType CHSK=CH*SK;
    QtbRealType SHCK=SH*CK;
    // CHCK is always real
    // Other may be imaginary or complex
    // Intermediate variable, avoid dupplicate operations
    QtbRealType gn = 2.0*k2 / (kb*kb);
    QtbRealType gn2 = gn*gn;
    QtbRealType adim1 = gn2-2.0*gn+1.0;
    QtbRealType adim2 = hn2*kn2;
    QtbRealType adim3 = gn2+adim1;
    QtbRealType adim4 = 1.0-gn;
    QtbRealType adim5 = gn2*adim2;
    QtbRealType CHCK1 = expCorr-CHCK;
    QtbRealType CHCK2 = 2.0*CHCK1;
    QtbRealType SHCK1 = gn*hn2*SHCK;
    c1=(QtbRealType)_model->rho(i)*w2/k;
    QtbRealType c2 = 1.0/c1;

    // Second order subdeterminants of propagator matrix
    // i means that the subdeterminant is imaginary
    QtbRealType G1212 = adim3*CHCK-(adim1+adim5)*SHSK-(adim3-1)*expCorr;
    QtbRealType G1213 = c2*(CHSK-hn2*SHCK);
    QtbRealType iG1214 = c2*((adim1-gn2)*CHCK1+(adim4-gn*adim2)*SHSK);
//#define iG1223 iG1214 --> R1223
    QtbRealType G1224 = c2*(kn2*CHSK-SHCK);
    QtbRealType G1234 = c2*c2*(CHCK2+(1+adim2)*SHSK);
    QtbRealType G1312 = c1*(gn2*kn2*CHSK-adim1*SHCK);
#define G1313 CHCK
    QtbRealType iG1314=adim4*SHCK+gn*kn2*CHSK;
//#define iG1323 iG1314 --> R1223
    QtbRealType G1324 = kn2*SHSK;
#define G1334 G1224
    QtbRealType iG1412 = c1*((adim1-adim4)*(adim4-gn)*CHCK1+(adim4*adim1-gn*adim5)*SHSK);
    QtbRealType iG1413 = SHCK1+adim4*CHSK;
    QtbRealType G1423 = CHCK-G1212;
#define G1414 expCorr+G1423
#define iG1424 iG1314
#define iG1434 iG1214
//#define iG2312 iG1412
//#define iG2313 iG1413
#define G2314 G1423
//#define G2323 G1414 --> R1223
//#define iG2324 iG1314
//#define iG2334 iG1214
    QtbRealType G2412= c1*(adim1*CHSK-gn*SHCK1);
    QtbRealType G2413 = hn2*SHSK;
#define iG2414 iG1413
//#define iG2423 iG1413 --> R1223
#define G2424 G1313
#define G2434 G1213
    QtbRealType G3412 = c1*c1*(gn2*adim1*CHCK2+(adim1*adim1+gn2*adim5)*SHSK);
#define G3413 G2412
#define iG3414 iG1412
//#define iG3423 iG1412 --> R1223
#define G3424 G1312
#define G3434 G1212

    /*
       Generic formula:
        R12cd = F1212*G12cd+F1213*G13cd+iF1214*iG14cd+iF1223*iG23cd+F1224*G24cd+F1234*G34cd
    */
    QtbRealType R1212 = F1212*G1212+fac*(F1213*G1312-2.0*iF1214*iG1412+F1224*G2412-F1234*G3412);
    QtbRealType R1213 = invFac*F1212*G1213+F1213*G1313+2.0*iF1214*iG1413-F1224*G2413+F1234*G3413;
    QtbRealType iR1214= invFac*F1212*iG1214+F1213*iG1314+iF1214*(G1414+G2314)-F1224*iG2414+F1234*iG3414;
    //     R1223 = invFac*F1212*iG1223+F1213*iG1323+iF1214*(G1423+G2323)-F1224*iG2423+F1234*iG3423;
    QtbRealType R1224 = invFac*F1212*G1224-F1213*G1324-2.0*iF1214*iG1424+F1224*G2424+F1234*G3424;
    QtbRealType R1234 = F1213*G1334-invFac*F1212*G1234-2.0*iF1214*iG1434+F1224*G2434+F1234*G3434;

    //  Normalize R before copying to F
    QtbRealType maxR=R1212;
    if (maxR<0) maxR=-maxR;
    tmpAbs=R1213;
    if (tmpAbs<0) tmpAbs=-tmpAbs;
    if (tmpAbs>maxR) maxR=tmpAbs;
    tmpAbs=iR1214;
    if (tmpAbs<0) tmpAbs=-tmpAbs;
    if (tmpAbs>maxR) maxR=tmpAbs;
    tmpAbs=R1224;
    if (tmpAbs<0) tmpAbs=-tmpAbs;
    if (tmpAbs>maxR) maxR=tmpAbs;
    tmpAbs=R1234;
    if (tmpAbs<0) tmpAbs=-tmpAbs;
    if (tmpAbs>maxR) maxR=tmpAbs;
    if (maxR>1.0e5) {
      maxR=1.0e5/maxR;
      F1212=R1212*maxR;
      F1213=R1213*maxR;
      iF1214=iR1214*maxR;
      F1224=R1224*maxR;
      F1234=R1234*maxR;
    }
    else {
      F1212=R1212;
      F1213=R1213;
      iF1214=iR1214;
      F1224=R1224;
      F1234=R1234;
    }
  }
  return F1212;
}

/*
kb, kn2, kn,
From Herrmann (srfdis.f), for water layer on top
      if(llw.ne.1) then
c-----
c     include water layer.
c-----
        xka = omega/dble(a(1))
        wvnop=wvno+xka
        wvnom=dabs(wvno-xka)
        ra=dsqrt(wvnop*wvnom)
        dpth=dble(d(1))
        rho1=dble(rho(1))
        p = ra*dpth
        beta = dble(b(1))
        znul = 1.0d-05
        call var(p,znul,ra,znul,wvno,xka,znul,dpth,w,cosp,exa)
        w0=-rho1*w
  dltar4 = cosp*e(1) + w0*e(2)
      else
  dltar4 = e(1)
      endif
*/

#endif // QTB_RAYLEIGH_H

