/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-09-17
**
***************************************************************************/

#include "qtbtiltlayeredmodel.h"
#include "qtblayeredmodel.h"
#include "qtbpoint2d.h"
#include "qtbpoint.h"
#include "qtbfile.h"
#include "qtbtiltnode.h"
#include "qtbtiltpath.h"
#include "qtbcoreapplication.h"
#include "qtbmemorychecker.h"

/*!
  \class QtbTiltLayeredModel qtbtiltlayeredmodel.h
  \brief A layered model with tilted interfaces

  Each layer has one unique slowness. The bottom interface geometry is defined by its left and right
  depths. Depths are defined at xLeft() and xRight().
*/

QtbTiltLayeredModel::QtbTiltLayeredModel()
{
  QTB_TRACE;
  initMembers();
}

QtbTiltLayeredModel::QtbTiltLayeredModel(int nLayers)
{
  QTB_TRACE;
  _layerCount=nLayers;
  initMembers();
  allocateData();
}

QtbTiltLayeredModel::~QtbTiltLayeredModel()
{
  QTB_TRACE;
  if (_cosangle) end();
  deleteData();
}

/*!
  Initialize internal members

  Internal use only.
*/
void QtbTiltLayeredModel::initMembers()
{
  _depthL=0;
  _depthR=0;
  _slow=0;
  _minX=0;
  _maxX=0;
  _cosangle=0;
  _sinangle=0;
  _shift=0;
  _h=0;
  _critl=0;
  _critr=0;
}

/*!
  Delete all data vectors

  Internal use only.
*/
void QtbTiltLayeredModel::deleteData()
{
  QTB_TRACE;
  delete [] _depthL;
  delete [] _depthR;
  delete [] _slow;
  delete [] _minX;
  delete [] _maxX;
  _depthL=0;
  _depthR=0;
  _slow=0;
  _minX=0;
  _maxX=0;
}

void QtbTiltLayeredModel::allocateData()
{
  QTB_TRACE;
  if (_layerCount>0) {
    _depthL=new double [_layerCount];
    _depthR=new double [_layerCount];
    _depthL[0] = 0.0;
    _depthR[0] = 0.0;
    _slow=new double [_layerCount];
    _minX=new double [_layerCount];
    _maxX=new double [_layerCount];
  }
}

/*!
  Convert \a ref horizontal layered model to a titl layered model using \a pitch profile.
  The layer pitch is a real number between -1 and 0. \a pitch profile must contain exactly
  the same number of layers than \a ref.

    \li -1  left thickness is forced to 0
    \li 0   means no pitch.
    \li 1   right thickness is forced to 0

*/
void QtbTiltLayeredModel::fromLayeredModel( const QtbLayeredModel& ref, const QVector<double>& pitch,
                                            QtbLayeredModel::SlownessType slowType )
{
  QTB_TRACE;
  if ( _layerCount != ref.layerCount() ) {
    _layerCount = ref.layerCount();
    deleteData();
    allocateData();
  }
  QTB_ASSERT(_layerCount==pitch.count());
  int n1 = _layerCount-1;
  for(int i = 0; i<n1;i++) {
    switch(slowType) {
    case QtbLayeredModel::P:
      _slow[i] = ref.slowP(i);
      break;
    case QtbLayeredModel::S:
      _slow[i] = ref.slowS(i);
      break;
    }
    _depthL[i+1] = _depthL[i] + (1.0+pitch[i])*ref.h(i);
    _depthR[i+1] = _depthR[i] + (1.0-pitch[i])*ref.h(i);
  }
  switch(slowType) {
  case QtbLayeredModel::P:
    _slow[n1] = ref.slowP(n1);
    break;
  case QtbLayeredModel::S:
    _slow[n1] = ref.slowS(n1);
    break;
  }
}

/*!
  Geometrical and physical parameters are supposed to be all initialized.

  Compute relative angles between layers. For each layer, the critical angle angle is computed.
  Incidence angle are then back propagated to the surface (refraction at angle lower than critical).
*/
void QtbTiltLayeredModel::begin()
{
  QTB_TRACE;
  _cosangle=new double [_layerCount];
  _sinangle=new double [_layerCount];
  _critl=new double *[_layerCount];
  _critr=new double *[_layerCount];
  _shift=new double [_layerCount];
  _h=new double [_layerCount];
  double * angle=new double [_layerCount];

  // Init computed parameter for first layer
  angle[0]=atan((_depthL[0]-_depthR[0]) / (_xR-_xL));
  _cosangle[0]=cos(angle[0]);
  _sinangle[0]=sin(angle[0]);
  _minX[0]=0.5*(_xR+_xL);
  _maxX[0]=0.5*(_xR+_xL);
  _critl[0]=0;
  _critr[0]=0;
  for (int i=1;i<_layerCount;i++) {
    angle[i]=atan((_depthL[i]-_depthR[i]) / (_xR-_xL));
    _cosangle[i]=cos(angle[i]);
    _sinangle[i]=sin(angle[i]);
    double tmp=(_depthL[i]-_depthL[i-1])/cos(angle[i]-angle[i-1]);
    _shift[i]=-tmp*_sinangle[i-1];
    _h[i]=tmp*_cosangle[i];
    _minX[i]=0.5*(_xR+_xL);
    _maxX[i]=0.5*(_xR+_xL);
    if (_slow[i]<_slow[i-1]) { // refraction at critical angle
      double inc0=asin(_slow[i]/_slow[i-1]);
      _critl[i]=new double [i];
      _critl[i][i-1]=inc0;
      int j;
      for (j=i-2;j>=0;j--) {
        double inc=_critl[i][j+1]-(angle[j+2]-angle[j+1]);
        double a=_slow[j+1]/_slow[j]*sin(inc);
        if (a<1.0) _critl[i][j]=asin(a);
        else {
          qtbApp << tr("Critical angle cannot be reached in layer %1 from LeftToRight").arg(j) << endl;
          delete [] _critl[i]; // this is not a good candidate for critical angle
          _critl[i]=0;
        }
      }
      _critr[i]=new double [i];
      _critr[i][i-1]=-inc0;
      for (j=i-2;j>=0;j--) {
        double inc=_critr[i][j+1]-(angle[j+2]-angle[j+1]);
        double a=_slow[j+1]/_slow[j]*sin(inc);
        if (a<1.0) _critr[i][j]=asin(a);
        else {
          qtbApp << tr("Critical angle cannot be reached in layer %1 from RightToLeft").arg(j) << endl;
          delete [] _critr[i]; // this is not a good candidate for critical angle
          _critr[i]=0;
        }
      }
    } else {
      _critl[i]=0;
      _critr[i]=0;
    }
  }
  delete [] angle;
}

void QtbTiltLayeredModel::end()
{
  QTB_TRACE;
  delete [] _cosangle;
  delete [] _sinangle;
  delete [] _shift;
  delete [] _h;
  _cosangle=0;
  _sinangle=0;
  _shift=0;
  _h=0;
  int i;
  for (i=1;i<_layerCount;i++) delete [] _critl[i];
  delete [] _critl;
  for (i=1;i<_layerCount;i++) delete [] _critr[i];
  delete [] _critr;
  _critl=0;
  _critr=0;
}

/*!
  Return time needed to propagate across layer \a iLayer if the target layer is \a iDeepestLayer.
  \a x is set the local coordinate of the ray at the bottom interface. Local coordinates
  are counted from the reference profile.
*/
double QtbTiltLayeredModel::propagate( Direction dir, int iLayer, int iDeepestLayer, double& x) const
{
  QTB_TRACE;
  double inc;
  QTB_SAFE_UNINITIALIZED(inc,0);
  switch(dir) {
  case LeftToRight:
    inc = _critl[iDeepestLayer][iLayer-1];
    break;
  case RightToLeft:
    inc = _critr[iDeepestLayer][iLayer-1];
    break;
  }
  double cosab, cosb, sinb, d, dt;
  d=sqrt(x*x+_h[iLayer]*_h[iLayer]);
  cosb=_h[iLayer]/d;
  sinb=x/d;
  cosab=_cosangle[iLayer]*cosb-_sinangle[iLayer]*sinb; // cos(a+b)
  double cosinc=cos(inc);
  double dcosinc=d/cosinc;
  // time to cross the layer
  dt=_slow[iLayer-1]*cosab*dcosinc;
  /* intersection of the ray with interface counted from the reference profile
     sin(a+b-i)=sin(a+b)cos(i)-cos(a+b)sin(i)
               =[sin(a)cos(b)+cos(a)sin(b)]cos(i)-cos(a+b)sin(i) */
  x=((_sinangle[iLayer]*cosb+_cosangle[iLayer]*sinb)*cosinc-cosab*sin(inc))*
    dcosinc-_shift[iLayer];
  return dt;
}

/*!
  Returns the paths for \a src and \a rec. No computation is performed here. Nodes are assumed to initialized
  with QtbTiltNode::init() which calculate the paths.
*/
bool QtbTiltLayeredModel::node2paths( const QtbTiltNode& src, const QtbTiltNode& rec,
                                      const QtbTiltPath *& srcPaths, const QtbTiltPath *& recPaths ) const
{
  double xSrc = src.x();
  double xRec = rec.x();
  if (xSrc<xRec) {
    srcPaths = src.right();
    recPaths = rec.left();
    return true;
  } else if (xSrc>xRec) {
    srcPaths = src.left();
    recPaths = rec.right();
    return true;
  } else {
    srcPaths = 0;
    recPaths = 0;
    return false;
  }
}

/*!
  Return travel time between two nodes \a src and \a rec. \a minTimeLayer is set to the deepest layer.
*/
double QtbTiltLayeredModel::travelTime( const QtbTiltNode& src, const QtbTiltNode& rec, int& minTimeLayer)
{
  QTB_TRACE;
  const QtbTiltPath * srcPaths, * recPaths;
  minTimeLayer=0;
  if (!node2paths(src, rec, srcPaths, recPaths)) return 0.0;
  return travelTime( srcPaths, recPaths, minTimeLayer );
}

/*!
  Return travel time between two paths \a srcPaths and \a recPaths. \a minTimeLayer is set to the deepest layer.
*/
double QtbTiltLayeredModel::travelTime( const QtbTiltPath *& srcPaths, const QtbTiltPath *& recPaths, int& minTimeLayer )
{
  QTB_TRACE;
  minTimeLayer=0;
  double t, newt, tupdown;
  t=_slow[0]*fabs(recPaths[0].abscissa(0)-srcPaths[0].abscissa(0));
  int i;
  for (i=1;i<_layerCount;i++) {
    newt=_slow[i]*fabs(recPaths[i-1].abscissa(i)-srcPaths[i-1].abscissa(i));
    tupdown=recPaths[i-1].time()+srcPaths[i-1].time();
    //printf("travelTime %lg %lg %lg %lg\n",newt,tupdown,newt+tupdown,t);
    if (newt+tupdown<t) {
      t=newt+tupdown;
      minTimeLayer=i;
    }
  }
  double x;
  int pathIndex=minTimeLayer-1;
  if (pathIndex<0) pathIndex=0;
  for (i=0;i<=minTimeLayer;i++) {
    x=srcPaths[pathIndex].abscissa(i);
    if (x<_minX[i]) _minX[i]=x;
    else if (x>_maxX[i]) _maxX[i]=x;
    x=recPaths[pathIndex].abscissa(i);
    if (x<_minX[i]) _minX[i]=x;
    else if (x>_maxX[i]) _maxX[i]=x;
  }
  return t;
}

/*!
  Return ray path between two nodes \a src and \a rec.
*/
QtbCurve<QtbPoint2D> QtbTiltLayeredModel::ray(const QtbTiltNode& src, const QtbTiltNode& rec)
{
  QtbCurve<QtbPoint2D> rayPoints;
  const QtbTiltPath * srcPaths, * recPaths;
  if (!node2paths(src, rec, srcPaths, recPaths)) return rayPoints;
  int endLayer;
  travelTime(srcPaths, recPaths, endLayer);
  double x;
  int i;
  int lastpray=endLayer*2+1;
  rayPoints.resize(lastpray+1);
  int pathIndex=endLayer-1;
  if (pathIndex<0) pathIndex=0;

  for (i=0;i<=endLayer;i++) {
    x=srcPaths[pathIndex].abscissa(i);
    rayPoints[i].setX( _xL+x*_cosangle[i] );
    rayPoints[i].setY( _depthL[i]-x*_sinangle[i] );
  }

  for (i=0;i<=endLayer;i++) {
    x=recPaths[pathIndex].abscissa(i);
    rayPoints[lastpray-i].setX( _xL+x*_cosangle[i] );
    rayPoints[lastpray-i].setY( _depthL[i]-x*_sinangle[i] );
  }

  return rayPoints;
}

bool QtbTiltLayeredModel::setValue( double& var, const QtbStringSection& field, int iLayer, const QString& varName, bool optional )
{
  QTB_TRACE;
  if (field.isValid()) {
    var = field.toDouble();
    if (var>0.0) {
      return true;
    } else {
      qtbApp << tr("Bad value for %1 at layer %2: %3").arg(varName).arg(iLayer).arg(var) << endl;
    }
  } else if (!optional) {
    qtbApp << tr("Missing value for %1 at layer %2").arg(varName).arg(iLayer) << endl;
  }
  return false;
}

QString QtbTiltLayeredModel::formatHelp()
{
  QTB_TRACE;
  return "  Line 1    <number of layers including half-space for first model>\n"
         "  Line 2    <x left depth profile> <x right depth profile>\n"
         "  Line 3    <left thickness (m)> <right thickness (m)> <V (m/s)>\n"
         "  ....\n"
         "  Line n    0 0 <V (m/s)>\n"
         "  Line n+1  <number of layers including half-space for second model>\n"
         "  ....";
}

/*!
  Loads model from a text stream

    Format:
      \verbatim
      layerCount
      xLeft xRight
      thicknessLeft thicknessRight V (layer 1)
      ...
      0 0 V (layer n, half-space)
      \endverbatim

  Returns false if an error occured during reading or if the read model is not consistent.
  It returns true even if the number of layers is null or cannot be read. This is the responsability
  of the caller to test if the number of layers is greater than 1 before doing anything else.
*/
bool QtbTiltLayeredModel::fromStream( QTextStream& s, QString * comments)
{
  QTB_TRACE;
  // read the number of layers
  QtbStringSection field, fields;
  const QChar * ptr;
  QString buf;
  buf = QtbFile::readLineNoComments( s, comments );
  fields.set(buf);
  fields.trimmed();
  ptr = 0;
  field = fields.nextField( ptr );
  if (field.isValid()) {
    _layerCount = field.toInt();
  } else {
    _layerCount = 0;
    return true;  // This is just a blank line, before doing anything, you must test if the number of layer is >0
  }
  if (_layerCount>0 && !s.atEnd()) {
    allocateData();
    buf = QtbFile::readLineNoComments( s, comments );
    fields.set(buf);
    fields.trimmed();
    ptr = 0;
    field = fields.nextField( ptr );
    if (field.isValid()) {
      _xL = field.toDouble();
    } else return false;
    field = fields.nextField( ptr );
    if (field.isValid()) {
      _xR = field.toDouble();
    } else return false;
    // read the data
    _depthL[0] = 0.0;
    _depthR[0] = 0.0;
    for (int i=0;i<_layerCount;i++) {
      if (s.atEnd()) {
        qtbApp << tr("Tilted layered model: reaching the end of file, %1 layers found, expecting %2")
                  .arg(i).arg(_layerCount) << endl;
        _layerCount = 0;
        return false;
      }
      buf = QtbFile::readLineNoComments( s, comments );
      fields.set(buf);
      fields.trimmed();
      ptr = 0;
      field = fields.nextField( ptr );
      if (i<_layerCount-1) {
        if (!setValue( _depthL[i+1], field, i, tr("left thickness"), false ) ) return false;
        _depthL[i+1] += _depthL[i];
      }
      field = fields.nextField( ptr );
      if (i<_layerCount-1) {
        if (!setValue( _depthR[i+1], field, i, tr("right thickness"), false ) ) return false;
        _depthR[i+1] += _depthR[i];
     }
      field = fields.nextField( ptr );
      if (!setValue( _slow[i], field, i, tr("V"), false ) ) return false;
      // Switch to slowness
      _slow[i]=1/_slow[i];
    }
    return true;
  } else {
    _layerCount = 0;
    return false;
  }
}

/*!
  Saves model to a text stream

    Format:
      \verbatim
      layerCount
      xLeft xRight
      thicknessLeft thicknessRight V (layer 1)
      ...
      0 0 V (layer n, half-space)
      \endverbatim
*/
void QtbTiltLayeredModel::toStream( QTextStream& s ) const
{
  QTB_TRACE;
  static const QString str3("%1 %2 %3\n");
  s << _layerCount << "\n"
    << _xL << " " << _xR << "\n";
  for (int i=0;i<_layerCount-1;i++)
    s << str3.arg( _depthL[i+1]-_depthL[i], 0, 'g', 20 )
             .arg( _depthR[i+1]-_depthR[i], 0, 'g', 20 )
             .arg( 1/_slow[i], 0, 'g', 20);
  s << str3.arg( 0.0, 0, 'g', 20 )
            .arg( 0.0, 0, 'g', 20)
            .arg( 1/_slow[_layerCount-1], 0, 'g', 20) << flush;
}
