/***************************************************************************
**
**  Copyright (C) 2002-2009 Marc Wathelet. All rights reserved.
**
**  This file is part of qtbwave.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  See http://www.geopsy.org for more information.
**
**  Created : 2004-09-17
**
***************************************************************************/

#ifndef QTB_TILTLAYEREDMODEL_H
#define QTB_TILTLAYEREDMODEL_H

#include <math.h>
#include <QDataStream>

#include "qtbcurve.h"
#include "qtblayeredmodel.h"
#include "qtbtranslations.h"
#include "qtbwavedllexport.h"

class QtbPoint;
class QtbPoint2D;
class QtbTiltNode;
class QtbTiltPath;

class QTBWAVE_EXPORT QtbTiltLayeredModel
{
  QTB_TRANSLATIONS("QtbTiltLayeredModel");
public:
  QtbTiltLayeredModel();
  QtbTiltLayeredModel(int nLayers);
  ~QtbTiltLayeredModel();

  int layerCount() const {return _layerCount;}
  double depthLeft( int iLayer ) const {return _depthL[iLayer];}
  double depthRight( int iLayer ) const {return _depthR[iLayer];}
  double slowness( int iLayer ) const {return _slow[iLayer];}

  void setXLeft( double x ) { _xL=x; }
  double xLeft() const { return _xL; }

  void setXRight( double x ) { _xR=x; }
  double xRight() const { return _xR; }

  void fromLayeredModel( const QtbLayeredModel& ref, const QVector<double>& pitch,
                         QtbLayeredModel::SlownessType slowType );

  enum Direction { LeftToRight, RightToLeft };

  void begin();
  inline bool isCriticalLayer( Direction dir, int iLayer ) const;
  double propagate( Direction dir, int iLayer, int iDeepestLayer, double& x ) const;
  void end();

  double travelTime(const QtbTiltNode& src, const QtbTiltNode& rec, int& minTimeLayer);
  QtbCurve<QtbPoint2D> ray(const QtbTiltNode& src, const QtbTiltNode& rec);

  static QString formatHelp();
  bool fromStream( QTextStream& s, QString * comments = 0 );
  void toStream( QTextStream& s ) const;
  void fromStream( QDataStream& s );
  void toStream( QDataStream& s ) const;
private:
  void initMembers();
  void allocateData();
  void deleteData();
  bool setValue( double& var, const QtbStringSection& field, int iLayer, const QString& varName, bool optional );
  double travelTime( const QtbTiltPath *& srcPaths, const QtbTiltPath *& recPaths, int& minTimeLayer);
  bool node2paths( const QtbTiltNode& src, const QtbTiltNode& rec,
                   const QtbTiltPath *& srcPaths, const QtbTiltPath *& recPaths ) const;

  // number of layers including the bottom half space
  int _layerCount;
  // pointer to depth vector (positive values) counted at profile Left
  double * _depthL;
  // pointer to depth vector (positive values) counted at profile Right
  double * _depthR;
  // pointer to slownesses vector
  double * _slow;
  double _xL, _xR;

  // Min and max x of rays along interfaces
  double * _minX;
  double * _maxX;

  // Cosine of the angle (to avoid multiple useless computations)
  double * _cosangle;
  // Sine of the angle (to avoid multiple useless computations)
  double * _sinangle;
  // ratio of slownesses at the interface
  double * _n;
  /* Abscissa to substract to abscissae measured from the intercept point
      to obtain the abcissa from the reference profile.
      The intercept is the projection of the intersection of the reference
      profile with the upper interface, made perpendiculary to the upper
      interface, and downto the current interface */
  double * _shift;
  /* Apparent thickness of the layer calculated prependiculary to upper interface
      and at reference profile */
  double * _h;
  /* For each layer a vector of relative angles may be allocated
    if the interface is a good candidate for refraction at critical angle */
  double ** _critl;
  double ** _critr;
};

inline bool QtbTiltLayeredModel::isCriticalLayer( Direction dir, int iLayer ) const
{
  QTB_TRACE;
  switch(dir) {
  case LeftToRight:
    return _critl[iLayer];
    break;
  default:
    return _critr[iLayer];
    break;
  }
}

#endif // QTB_TILTLAYEREDMODEL_H
